package main

import (
	"context"
	"fmt"
	"net"
	"os"
	"os/signal"
	"syscall"
	"time"

	"github.com/oklog/oklog/pkg/group"
	"github.com/opentracing/opentracing-go"
	account "gitlab.mashplace.com/mashplace/backend/account/api"
	"gitlab.mashplace.com/mashplace/backend/common/api/health"
	"gitlab.mashplace.com/mashplace/backend/common/generic/addr"
	"gitlab.mashplace.com/mashplace/backend/common/generic/grpc"
	logger "gitlab.mashplace.com/mashplace/backend/common/generic/log"
	"gitlab.mashplace.com/mashplace/backend/common/generic/metrics"
	"gitlab.mashplace.com/mashplace/backend/common/generic/tracer"
	"gitlab.mashplace.com/mashplace/backend/common/lib/httpclient"
	httpclientfactory "gitlab.mashplace.com/mashplace/backend/common/lib/httpclient/factory"
	httpclientutil "gitlab.mashplace.com/mashplace/backend/common/lib/httpclient/util"
	loggerfactory "gitlab.mashplace.com/mashplace/backend/common/lib/logger/factory"
	loggerservice "gitlab.mashplace.com/mashplace/backend/common/lib/logger/service"
	"gitlab.mashplace.com/mashplace/backend/common/lib/marshallunmarshall/jsoniterlocal"
	"gitlab.mashplace.com/mashplace/backend/common/lib/mongolocal"
	extras "gitlab.mashplace.com/mashplace/backend/extras/api"
	invitation "gitlab.mashplace.com/mashplace/backend/invitation/api"
	invoice "gitlab.mashplace.com/mashplace/backend/invoice/api"
	listing "gitlab.mashplace.com/mashplace/backend/listing/api"
	notification "gitlab.mashplace.com/mashplace/backend/notification/api"
	offer "gitlab.mashplace.com/mashplace/backend/offer/api"
	apirto "gitlab.mashplace.com/mashplace/backend/offer/api/rto"
	"gitlab.mashplace.com/mashplace/backend/offer/impl"
	factorydomainrtorepository "gitlab.mashplace.com/mashplace/backend/offer/internal/app/factory/domain/rto/repository"
	factoryservice "gitlab.mashplace.com/mashplace/backend/offer/internal/app/factory/service"
	factoryservicerto "gitlab.mashplace.com/mashplace/backend/offer/internal/app/factory/service/rto"
	factoryservicertodaemon "gitlab.mashplace.com/mashplace/backend/offer/internal/app/factory/service/rto/daemon"
	factorytransportgrpc "gitlab.mashplace.com/mashplace/backend/offer/internal/app/factory/transport/grpc"
	factoryusecaserto "gitlab.mashplace.com/mashplace/backend/offer/internal/app/factory/usecase/rto"
	servicertodaemon "gitlab.mashplace.com/mashplace/backend/offer/internal/app/service/rto/daemon"
	transportgrpc "gitlab.mashplace.com/mashplace/backend/offer/internal/app/transport/grpc"
	usecaserto "gitlab.mashplace.com/mashplace/backend/offer/internal/app/usecase/rto"
	"gitlab.mashplace.com/mashplace/backend/offer/internal/connections"
	_ "gitlab.mashplace.com/mashplace/backend/offer/internal/connections" // initialize connections
	"gitlab.mashplace.com/mashplace/backend/offer/internal/status"
	"gitlab.mashplace.com/mashplace/backend/offer/model"
	"gitlab.mashplace.com/mashplace/backend/offer/version"
	schedule "gitlab.mashplace.com/mashplace/backend/schedule/api"
	"go.mongodb.org/mongo-driver/mongo"
	"go.uber.org/dig"
)

var (
	container   *dig.Container
	log         loggerservice.Logger
	err         error
	ctx         context.Context
	cancel      context.CancelFunc
	serviceName string
)

func init() { //nolint
	container = dig.New()

	serviceName = "offer"

	ctx, cancel = context.WithCancel(context.Background())

	opentracingTracer := tracer.GlobalTracer()
	err = container.Provide(func() opentracing.Tracer {
		return opentracingTracer
	})
	if err != nil {
		panic(err)
	}

	model.InitRepo()

	mongolocal.Client = model.Client

	err = container.Provide(func() *mongo.Database {
		return model.DB
	})
	err = container.Provide(factorydomainrtorepository.NewRepository)
	if err != nil {
		panic(err)
	}
	err = container.Provide(factorydomainrtorepository.NewHistoryRepository)
	if err != nil {
		panic(err)
	}

	log = new(loggerfactory.StdoutFactory).CreateLogger()
	err = container.Provide(func() loggerservice.Logger {
		return log
	})
	if err != nil {
		panic(err)
	}

	err = container.Provide(func() extras.ExtrasServiceClient {
		return connections.ExtrasService
	})
	if err != nil {
		panic(err)
	}
	err = container.Provide(func() account.AccountServiceClient {
		return connections.AccountService
	})
	if err != nil {
		panic(err)
	}
	err = container.Provide(func() notification.NotificationServiceClient {
		return connections.NotificationService
	})
	if err != nil {
		panic(err)
	}
	err = container.Provide(func() listing.ListingServiceClient {
		return connections.ListingService
	})
	if err != nil {
		panic(err)
	}
	err = container.Provide(func() invoice.InvoiceServiceClient {
		return connections.InvoiceService
	})
	if err != nil {
		panic(err)
	}

	httpClient := new(httpclientfactory.StaticFactory).CreateHTTPClient(log, opentracingTracer,
		httpclientutil.FastHTTPClientType, 5, 5)
	err = container.Provide(func() httpclient.HTTPClient {
		return httpClient
	})
	if err != nil {
		panic(err)
	}

	err = container.Provide(jsoniterlocal.NewMarshalUnmarshal)
	if err != nil {
		panic(err)
	}

	err = container.Provide(factoryservice.NewAccountService)
	if err != nil {
		panic(err)
	}
	err = container.Provide(factoryservice.NewExtrasService)
	if err != nil {
		panic(err)
	}
	err = container.Provide(factoryservice.NewInvoiceService)
	if err != nil {
		panic(err)
	}
	err = container.Provide(factoryservice.NewListingService)
	if err != nil {
		panic(err)
	}

	err = container.Provide(factoryservicerto.NewCloseComService)
	if err != nil {
		panic(err)
	}
	err = container.Provide(factoryservicerto.NewNotificationService)
	if err != nil {
		panic(err)
	}
	err = container.Provide(factoryservicerto.NewSecurityService)
	if err != nil {
		panic(err)
	}

	err = container.Provide(factoryservicertodaemon.NewFinishService)
	if err != nil {
		panic(err)
	}
	err = container.Provide(factoryservicertodaemon.NewNotificationService)
	if err != nil {
		panic(err)
	}
	err = container.Provide(factoryservicertodaemon.NewUpdateLeadTenantUserService)
	if err != nil {
		panic(err)
	}

	err = container.Provide(factoryusecaserto.NewCreateUseCase)
	if err != nil {
		panic(err)
	}
	err = container.Provide(factoryusecaserto.NewAnswerFromTenantUseCase)
	if err != nil {
		panic(err)
	}
	err = container.Provide(factoryusecaserto.NewCreateByGuestUseCase)
	if err != nil {
		panic(err)
	}
	err = container.Provide(factoryusecaserto.NewCreateTenancyUseCase)
	if err != nil {
		panic(err)
	}
	err = container.Provide(factoryusecaserto.NewGetByLandlordIDUseCase)
	if err != nil {
		panic(err)
	}
	err = container.Provide(factoryusecaserto.NewGetByPropertyIDUseCase)
	if err != nil {
		panic(err)
	}
	err = container.Provide(factoryusecaserto.NewGetByPropertyIDsUseCase)
	if err != nil {
		panic(err)
	}
	err = container.Provide(factoryusecaserto.NewLandlordApprovedUseCase)
	if err != nil {
		panic(err)
	}
	err = container.Provide(factoryusecaserto.NewRenewalUseCase)
	if err != nil {
		panic(err)
	}
	err = container.Provide(factoryusecaserto.NewRentPaymentUseCase)
	if err != nil {
		panic(err)
	}
	err = container.Provide(factoryusecaserto.NewStandingOrderUseCase)
	if err != nil {
		panic(err)
	}
	err = container.Provide(factoryusecaserto.NewUpdateEndDateUseCase)
	if err != nil {
		panic(err)
	}
	err = container.Provide(factoryusecaserto.NewUpdateLandlordUseCase)
	if err != nil {
		panic(err)
	}
	err = container.Provide(factoryusecaserto.NewUpdateLeadTenantUserUseCase)
	if err != nil {
		panic(err)
	}
	err = container.Provide(factoryusecaserto.NewUpdateLeadTenantUseCase)
	if err != nil {
		panic(err)
	}
	err = container.Provide(factoryusecaserto.NewUpdateRentPlanUseCase)
	if err != nil {
		panic(err)
	}
	err = container.Provide(factoryusecaserto.NewUpdateStateToActiveUseCase)
	if err != nil {
		panic(err)
	}
	err = container.Provide(factoryusecaserto.NewUpdateStateToWaitingForPaymentUseCase)
	if err != nil {
		panic(err)
	}
	err = container.Provide(factoryusecaserto.NewUpdateStateToActiveRollingTenancyUseCase)
	if err != nil {
		panic(err)
	}
	err = container.Provide(factoryusecaserto.NewUpdateStateToCancelledUseCase)
	if err != nil {
		panic(err)
	}
	err = container.Provide(factoryusecaserto.NewUpdateStateToEndingUseCase)
	if err != nil {
		panic(err)
	}
	err = container.Provide(factoryusecaserto.NewUpdateStateToFinishedUseCase)
	if err != nil {
		panic(err)
	}
	err = container.Provide(factoryusecaserto.NewUpdateStateToWaitingForTenantConfirmationUseCase)
	if err != nil {
		panic(err)
	}

	err = container.Provide(factorytransportgrpc.NewRtoServer)
	if err != nil {
		panic(err)
	}
}

func main() {
	var transportGrpcRtoServer transportgrpc.RtoServer
	var rtoNotificationService servicertodaemon.NotificationService
	var updateStateToActiveUseCase usecaserto.UpdateStateToActiveUseCase
	var updateStateToWaitingForPaymentUseCase usecaserto.UpdateStateToWaitingForPaymentUseCase
	var updateStateToEndingUseCase usecaserto.UpdateStateToEndingUseCase
	var finishService servicertodaemon.FinishService
	var updateLeadTenantUserService servicertodaemon.UpdateLeadTenantUserService
	var standingOrderUseCase usecaserto.StandingOrderUseCase
	err = container.Invoke(func(rtoServer transportgrpc.RtoServer,
		rtoNotificationSrv servicertodaemon.NotificationService,
		UpdateStateToActiveUseCase usecaserto.UpdateStateToActiveUseCase,
		UpdateStateToWaitingForPaymentUseCase usecaserto.UpdateStateToWaitingForPaymentUseCase,
		UpdateStateToEndingUseCase usecaserto.UpdateStateToEndingUseCase,
		finishSrv servicertodaemon.FinishService,
		updateLeadTenantUserSrv servicertodaemon.UpdateLeadTenantUserService,
		StandingOrderUseCase usecaserto.StandingOrderUseCase) {
		transportGrpcRtoServer = rtoServer
		rtoNotificationService = rtoNotificationSrv
		updateStateToActiveUseCase = UpdateStateToActiveUseCase
		updateStateToWaitingForPaymentUseCase = UpdateStateToWaitingForPaymentUseCase
		updateStateToEndingUseCase = UpdateStateToEndingUseCase
		finishService = finishSrv
		updateLeadTenantUserService = updateLeadTenantUserSrv
		standingOrderUseCase = StandingOrderUseCase
	})
	if err != nil {
		panic(err)
	}

	status.UpdateStateToActiveUseCase = updateStateToActiveUseCase

	grpcListener, err := net.Listen("tcp", addr.Grpc)
	if err != nil {
		panic(fmt.Errorf("fatal error while init gRPC listener: %s", err))
	}

	defer func() {
		tracer.Close()

		_ = grpcListener.Close()
	}()

	var g group.Group
	{
		// This function just sits and waits for ctrl-C.
		cancelInterrupt := make(chan struct{})
		g.Add(func() error {
			c := make(chan os.Signal, 1)
			signal.Notify(c, syscall.SIGINT, syscall.SIGTERM)
			select {
			case sig := <-c:
				close(cancelInterrupt)
				cancel()

				return fmt.Errorf("received signal %s", sig)
			case <-cancelInterrupt:
				cancel()

				return nil
			}
		}, func(error) {
			close(cancelInterrupt)
			cancel()
		})
	}
	baseGrpcServer, err := grpc.NewServer(serviceName)
	if err != nil {
		panic(err)
	}
	{
		server := impl.NewServer(ctx, updateStateToEndingUseCase, standingOrderUseCase,
			updateStateToWaitingForPaymentUseCase, updateStateToActiveUseCase)

		apirto.RegisterRtoServiceServer(baseGrpcServer, transportGrpcRtoServer)
		offer.RegisterOfferServiceServer(baseGrpcServer, server)
		invoice.RegisterInvoiceNotifyServiceServer(baseGrpcServer, server)
		schedule.RegisterReminderConsumerServiceServer(baseGrpcServer, server)
		invitation.RegisterInvitationNotifyServiceServer(baseGrpcServer, server)

		err = server.RegisterSchedules(ctx)
		if err != nil {
			panic(err)
		}

		health.RegisterHealthServiceServer(baseGrpcServer, version.NewHealthServer())

		logger.WithContext(ctx).Info("transport ", "grpc ", "address ", addr.Grpc, " msg ", "listening "+serviceName+" api")

		g.Add(func() error {
			return baseGrpcServer.Serve(grpcListener)
		}, func(err error) {
			cancel()
		})
	}

	g.Add(func() error {
		logger.WithContext(ctx).Info("metrics.ConfigureListenAndServe start")

		err = metrics.ConfigureListenAndServe(baseGrpcServer)
		if err == nil {
			for ctx.Err() == nil {
				time.Sleep(time.Second)
			}

			return nil
		}

		return err
	}, func(err error) {
		cancel()
	})
	g.Add(func() error {
		logger.WithContext(ctx).Info("rtoNotificationService.RemindAll start")

		return rtoNotificationService.RemindAll(ctx)
	}, func(err error) {
		cancel()
	})
	g.Add(func() error {
		logger.WithContext(ctx).Info("finishService.FinishAll start")

		return finishService.FinishAll(ctx)
	}, func(err error) {
		cancel()
	})
	g.Add(func() error {
		logger.WithContext(ctx).Info("updateLeadTenantUserService.UpdateAllLeadTenantUsers start")

		return updateLeadTenantUserService.UpdateAllLeadTenantUsers(ctx)
	}, func(err error) {
		cancel()
	})

	logger.WithContext(ctx).Info("exit", g.Run())
}
