package impl

import (
	"context"
	"time"

	account "gitlab.mashplace.com/mashplace/backend/account/api"
	"gitlab.mashplace.com/mashplace/backend/common/api/grpcctx"
	"gitlab.mashplace.com/mashplace/backend/common/generic/log"
	"gitlab.mashplace.com/mashplace/backend/common/generic/util"
	extras "gitlab.mashplace.com/mashplace/backend/extras/api"
	invoice "gitlab.mashplace.com/mashplace/backend/invoice/api"
	messaging "gitlab.mashplace.com/mashplace/backend/messaging/api"
	notification "gitlab.mashplace.com/mashplace/backend/notification/api"
	offer "gitlab.mashplace.com/mashplace/backend/offer/api"
	request "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto/resource/request/usecase"
	"gitlab.mashplace.com/mashplace/backend/offer/internal/connections"
	"gitlab.mashplace.com/mashplace/backend/offer/internal/invoices"
	"gitlab.mashplace.com/mashplace/backend/offer/internal/status"
	"gitlab.mashplace.com/mashplace/backend/offer/model"
	"go.mongodb.org/mongo-driver/bson/primitive"
)

func (server *serverImpl) InvoiceChange(
	ctx context.Context,
	req *invoice.InvoiceChangeRequest,
) (resp *invoice.InvoiceChangeResponse, err error) {
	if req.GetInvoice().GetCorrelationType() != invoice.CorrelationTypeOffer {
		return &invoice.InvoiceChangeResponse{}, nil
	}

	inv := req.GetInvoice()
	pmt := req.GetPayment()

	offerID, err := primitive.ObjectIDFromHex(inv.GetCorrelationId())
	if err != nil {
		return nil, err
	}

	off, err := model.GetOfferByID(ctx, offerID)
	if err != nil {
		return nil, err
	}

	if off.Test {
		ctx = context.WithValue(ctx, grpcctx.CtxIsTestUser, true)
	}

	if inv.GetCorrelationTarget() == invoice.CorrelationTargetStandingOrder {
		err = server.paymentPeriodicSetup(ctx, pmt, inv, off)
		if err != nil {
			return nil, err
		}
	}

	if invoices.HoldingDepositPayment(pmt) {
		notifyHoldingDepositPaid(ctx, pmt, inv, off)
	}

	if pmt.Status == invoice.PaymentStatus_PAYMENT_SUCCESS {
		err = server.paymentSuccess(ctx, pmt, inv, off)
	}

	if err != nil {
		return nil, err
	}

	return &invoice.InvoiceChangeResponse{}, nil
}

func (server *serverImpl) paymentPeriodicSetup(
	ctx context.Context,
	pmt *invoice.Payment,
	inv *invoice.Invoice,
	off *model.OfferModel,
) (err error) {
	switch pmt.Status {
	case invoice.PaymentStatus_PAYMENT_PENDING, invoice.PaymentStatus_PAYMENT_CONFIRM,
		invoice.PaymentStatus_PAYMENT_SUCCESS:
	default:
		return nil
	}

	if off.PaidStatuses == nil {
		off.PaidStatuses = &model.PaidStatusesModel{}
	}

	if inv.GetSetup() && !off.PaidStatuses.StandingOrderSetup {
		off.PaidStatuses.StandingOrder = true
		off.PaidStatuses.StandingOrderSetup = true

		err = model.UpdateOffer(ctx, off)
		if err != nil {
			return err
		}
	}

	if int(pmt.Serial+1) == int(*off.Conditions.TermMonths) && pmt.Status == invoice.PaymentStatus_PAYMENT_SUCCESS {
		go func() {
			_, _ = server.updateStateToEndingUseCase.UpdateStateToEnding(context.Background(),
				&request.UpdateStateToEndingRequest{ID: off.ID.Hex()})
		}()
	}

	if off.State == offer.OfferState_RTO_STATE_WAITING_FOR_STANDING_ORDER_SETUP && inv.GetSetup() {
		err = server.setPropertyRentCollection(ctx, off)
		if err != nil {
			return err
		}
	}

	err = status.GenerateSchedule(ctx, off)
	if err != nil {
		return err
	}

	if off.State != offer.OfferState_RENTPAID &&
		off.State != offer.OfferState_RTO_STATE_WAITING_FOR_STANDING_ORDER_SETUP &&
		off.State != offer.OfferState_RTO_STATE_WAITING_FOR_PAYMENT {
		err = status.SetPropertyRented(ctx, off)
		if err != nil {
			return err
		}
	}

	return
}

func (server *serverImpl) setPropertyRentCollection(ctx context.Context, off *model.OfferModel) error {
	_, err := server.updateStateToWaitingForPaymentUseCase.UpdateStateToWaitingForPayment(ctx,
		&request.UpdateStateToWaitingForPaymentRequest{ID: off.ID.Hex()})
	if err != nil {
		return err
	}

	return nil
}

func (server *serverImpl) paymentSuccess(
	ctx context.Context,
	pmt *invoice.Payment,
	inv *invoice.Invoice,
	off *model.OfferModel,
) (err error) {
	offerLog := model.NewOfferLog(ctx, off, inv.GetUser())

	defer func() {
		if err != nil {
			return
		}

		offerLog.LogDiff()
	}()

	var tenants []string

	// handle referencing check payment
	for _, item := range inv.GetServices() {
		if item.GetService().GetType() != extras.ServiceType_REFERENCING_CHECK ||
			item.GetService().GetTargetType() != "user-reference" {
			continue
		}

		for _, t := range off.Tenants {
			if util.UnPtrString(t.EmailPrimary) == item.GetService().GetTargetId() {
				t.CheckNeeded = true

				break
			}
		}

		for _, g := range off.Guarantors {
			if util.UnPtrString(g.EmailPrimary) == item.GetService().GetTargetId() {
				g.CheckNeeded = true

				break
			}
		}

		tenants = append(tenants, item.GetService().GetTargetId())
	}

	if len(tenants) > 0 {
		if err = model.UpdateOffer(ctx, off); err != nil {
			return err
		}

		if err = server.applicantCheckPaymentPaid(ctx, off, tenants, inv.InvoiceId); err != nil {
			return err
		}
	}

	switch off.State {
	case offer.OfferState_HOLDING_DEPOSIT, offer.OfferState_AWAITING_CONFIRM:
		err = server.paymentSuccessHoldingConfirm(ctx, inv, off)
	case offer.OfferState_SIGNED, offer.OfferState_AWAITING_CONFIRM_RENT,
		offer.OfferState_AWAITING_TENANCY, offer.OfferState_RENTPAID,
		offer.OfferState_RTO_STATE_ACTIVE, offer.OfferState_RTO_STATE_WAITING_FOR_STANDING_ORDER_SETUP:
		err = server.paymentSuccessSignedConfirm(ctx, inv, off)
	case offer.OfferState_RTO_STATE_WAITING_FOR_PAYMENT:
		err = server.paymentSuccessRtoActive(ctx, off)
	case offer.OfferState_AWAITING_CONFIRM_EXTRAS:
		err = server.paymentSuccessConfirmExtras(ctx, inv, off)
	}

	if err != nil {
		log.WithContext(ctx).Error(err)
	}

	if _, err = connections.ExpensesInvoiceNotificationService.InvoiceChange(ctx, &invoice.InvoiceChangeRequest{
		Invoice: inv,
		Payment: pmt,
	}); err != nil {
		log.WithContext(ctx).Error(err)
	}

	return
}

func notifyHoldingDepositPaid(
	ctx context.Context,
	pmt *invoice.Payment,
	inv *invoice.Invoice,
	off *model.OfferModel,
) {
	if !invoices.HoldingDepositPayment(pmt) {
		return
	}

	logger := log.WithFields(log.Fields{
		"action":      "notifyHoldingDepositPayed",
		"payment-ref": pmt.Reference,
		"invoice":     inv.InvoiceId,
		"offer":       off.ID.Hex(),
	})

	notify := func(action notification.Action) bool {
		_, err := connections.NotificationService.SendNotification(ctx, &notification.SendNotificationRequest{
			Entity: notification.Entity_ENTITY_OFFER_HOLDING_DEPOSIT_PAYMENT,
			Action: action,
			Targets: []*notification.Target{
				{
					User: off.LandlordUser(),
					Role: notification.Role_ROLE_LANDLORD,
				},
				{
					User: off.TenantUser(),
					Role: notification.Role_ROLE_TENANT,
				},
			},
			Embeds: []*notification.EventEmbed{
				{
					EmbedType:  messaging.EmbedType_EMBED_TYPE_PROPERTY,
					EmbedValue: off.Property,
				},
				{
					EmbedType:  messaging.EmbedType_EMBED_TYPE_OFFER,
					EmbedValue: off.ID.Hex(),
				},
			},
		})
		if err != nil {
			logger.WithContext(ctx).WithError(err).Errorf("notification delivery failed for actio: %s", action.String())
			return false
		}
		return true
	}

	if off.Notifications == nil {
		off.Notifications = make(model.OfferNotifications)
	}

	upd := false
	switch pmt.GetStatus() {
	case invoice.PaymentStatus_PAYMENT_CONFIRM:
		if pmt.GetSystem().GetBank() != nil && !off.Notifications.Sent(model.NotificationDepositBankPaid) {
			if notify(notification.Action_ACTION_CREATE) {
				off.Notifications[model.NotificationDepositBankPaid] = time.Now()
				upd = true
			}
		}
	case invoice.PaymentStatus_PAYMENT_SUCCESS:
		switch {
		case pmt.GetSystem().GetTokenio() != nil:
			notify(notification.Action_ACTION_COMPLETE)
		case pmt.GetSystem().GetBank() != nil && !off.Notifications.Sent(model.NotificationDepositBankVerified):
			if notify(notification.Action_ACTION_VERIFY) {
				off.Notifications[model.NotificationDepositBankVerified] = time.Now()
				upd = true
			}
		}
	}

	if upd {
		err := model.UpdateOffer(ctx, off)
		if err != nil {
			log.WithContext(ctx).Error(err)
		}
	}
}

func (server *serverImpl) paymentSuccessConfirmExtras(
	ctx context.Context,
	inv *invoice.Invoice,
	old *model.OfferModel,
) error {
	user, err := connections.AccountService.Get(ctx, &account.GetRequest{
		Id:    inv.User,
		Admin: true,
	})
	if err != nil {
		log.WithContext(ctx).WithError(err).Error("get user account failed")

		return nil
	}

	_, err = connections.NotificationService.SendNotification(ctx, &notification.SendNotificationRequest{
		Entity: notification.Entity_ENTITY_OFFER_EXTRA_PAYMENT,
		Action: notification.Action_ACTION_COMPLETE,
		Targets: []*notification.Target{
			{
				User: user.GetAccount().GetIdValue(),
				Role: notification.Role_ROLE_LANDLORD,
			},
		},
		Embeds: []*notification.EventEmbed{
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_OFFER,
				EmbedValue: old.ID.Hex(),
			},
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_PROPERTY,
				EmbedValue: old.Property,
			},
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_FILE,
				EmbedValue: inv.GetRender(),
			},
		},
	})
	if err != nil {
		log.WithContext(ctx).WithError(err).Errorf("sending failed: %s", notification.EventType_EVENT_PAYMENT_PAID_EXTRA_LANDLORD)

		return nil
	}

	return nil
}

func (server *serverImpl) paymentSuccessSignedConfirm(
	ctx context.Context,
	inv *invoice.Invoice,
	old *model.OfferModel,
) error {
	if len(old.Tenants) == 0 && old.Tenants[0].User != inv.User {
		return nil
	}

	if old.PaidStatuses == nil {
		old.PaidStatuses = &model.PaidStatusesModel{}
	}

	if inv.Status == invoice.InvoiceStatus_INVOICE_FULFILLED &&
		inv.CorrelationTarget == invoice.CorrelationTargetTenantServices {
		old.PaidStatuses.TenantServices = true

		err := model.UpdateOffer(ctx, old)
		if err != nil {
			return err
		}

		status.TryExportClosecomMoveinStatus(ctx, old)
	}

	go func(ctx context.Context, offerID string) {
		ctxTimeout, cancel := context.WithTimeout(ctx, 20*time.Second)

		defer cancel()

		err := status.TrySetPropertyRentedCreateInsurances(
			ctxTimeout,
			offerID,
			true,
			false,
		)
		if err != nil {
			logger := log.WithFields(log.Fields{
				"offerID":                 offerID,
				"isImitationMode":         true,
				"isRollingTenancyProcess": false,
			})

			logger.WithContext(ctxTimeout).
				WithError(err).
				Errorln("paymentSuccessSignedConfirm -> TrySetPropertyRentedCreateInsurances")
		}
	}(ctx, old.ID.Hex())

	err := status.GenerateSchedule(ctx, old)
	if err != nil {
		return err
	}

	if old.State != offer.OfferState_RENTPAID &&
		old.State != offer.OfferState_RTO_STATE_WAITING_FOR_STANDING_ORDER_SETUP &&
		old.State != offer.OfferState_RTO_STATE_WAITING_FOR_PAYMENT &&
		old.State != offer.OfferState_RTO_STATE_ACTIVE {
		err = status.SetPropertyRented(ctx, old)
		if err != nil {
			return err
		}
	}

	return nil
}

func (server *serverImpl) paymentSuccessHoldingConfirm(
	ctx context.Context,
	inv *invoice.Invoice,
	old *model.OfferModel,
) (err error) {
	log.WithContext(ctx).Info("processReceipt offer-deposited start")

	logger := log.WithFields(log.Fields{
		"invoice": inv.InvoiceId,
		"offer":   old.ID.Hex(),
	})

	err = old.SetInvoiceAndUpdate(ctx)
	if err != nil {
		logger.WithContext(ctx).WithError(err).Error("updating offer with invoice")
	}

	err = status.SetOfferState(ctx, old, offer.OfferState_DEPOSIT_CONFIRMED)
	if err != nil {
		logger.WithContext(ctx).WithError(err).Error("set offer state DEPOSIT_CONFIRMED")
	}

	if old.DepositReplacement() {
		err = status.SetOfferState(ctx, old, offer.OfferState_AWAITING_APPLICANT_CHECK_PAYMENT)
		if err != nil {
			logger.WithContext(ctx).WithError(err).Error("set offer state AWAITING_APPLICANT_CHECK_PAYMENT")

			return err
		}
	}

	return nil
}

func (server *serverImpl) applicantCheckPaymentPaid(
	ctx context.Context,
	old *model.OfferModel,
	tenants []string,
	invoiceID string,
) error {
	if old.State == offer.OfferState_AWAITING_APPLICANT_CHECK_PAYMENT ||
		old.State == offer.OfferState_CHECKING {
		if old.State != offer.OfferState_CHECKING {
			err := status.SetOfferState(ctx, old, offer.OfferState_CHECKING)
			if err != nil {
				return err
			}
		}

		go func(old *model.OfferModel, tenants []string, invoiceID string) {
			err := server.executeChecks(context.Background(), old, tenants, invoiceID)
			if err != nil {
				log.WithError(err).Infof("executeChecks is failed for offerID:%s invoiceID:%s",
					old.ID.Hex(), invoiceID)
			}
		}(old, tenants, invoiceID)
	}

	return nil
}
func (server *serverImpl) paymentSuccessRtoActive(ctx context.Context, rto *model.OfferModel) error {
	_, err := server.updateStateToActiveUseCase.UpdateStateToActive(ctx,
		&request.UpdateStateToActiveRequest{ID: rto.ID.Hex()})

	return err
}
