package impl

import (
	"context"

	"gitlab.mashplace.com/mashplace/backend/common/api/generic/query"
	expenses "gitlab.mashplace.com/mashplace/backend/expenses/api"
	offer "gitlab.mashplace.com/mashplace/backend/offer/api"
	"gitlab.mashplace.com/mashplace/backend/offer/internal/connections"
	"gitlab.mashplace.com/mashplace/backend/offer/internal/status"
	"gitlab.mashplace.com/mashplace/backend/offer/model"
	"go.mongodb.org/mongo-driver/bson/primitive"
)

func (server *serverImpl) ScheduledPaymentList(
	ctx context.Context,
	req *offer.ScheduledPaymentListRequest,
) (*offer.ScheduledPaymentListResponse, error) {
	id, err := primitive.ObjectIDFromHex(req.GetOffer())
	if err != nil {
		return nil, err
	}

	var offerModel *model.OfferModel

	if req.GetUser() != "" {
		offerModel, err = model.GetOffer(ctx, id, req.GetUser())
	} else {
		offerModel, err = model.GetOfferByID(ctx, id)
	}
	if err != nil {
		return nil, err
	}

	switch offerModel.State {
	case offer.OfferState_SIGNED, offer.OfferState_AWAITING_CONFIRM_RENT, offer.OfferState_RENTPAID:
	default:
		return &offer.ScheduledPaymentListResponse{}, nil
	}

	params := req.GetParams()

	if params == nil {
		params = &query.QueryParams{}
	}

	params.Filter = make([]*query.QueryParamsFilter, 0)

	params.Filter = append(
		params.Filter,
		&query.QueryParamsFilter{
			Fields:   []string{"correlation_type"},
			Value:    []string{expenses.CorrelationType_CORRELATION_TYPE_OFFER.String()},
			Operator: query.QueryFilterOperator_EQUALS,
		},
		&query.QueryParamsFilter{
			Fields:   []string{"correlation_id"},
			Value:    []string{req.GetOffer()},
			Operator: query.QueryFilterOperator_EQUALS,
		},
		&query.QueryParamsFilter{
			Fields:   []string{"schedule_id"},
			Value:    []string{"0"},
			Operator: query.QueryFilterOperator_EQUALS,
		},
	)

	if req.GetUser() != "" {
		params.Filter = append(params.Filter, &query.QueryParamsFilter{
			Fields:   []string{"user_id"},
			Value:    []string{req.GetUser()},
			Operator: query.QueryFilterOperator_EQUALS,
		})
	}

	records, err := connections.ExpensesService.RecordFilterList(ctx, &expenses.RecordFilterListRequest{
		Params: params,
	})
	if err != nil {
		return nil, err
	}

	res := make([]*offer.ScheduledPayment, 0)

	recordsMap := make(map[string][]*expenses.Record)

	for _, r := range records.GetRecords() {
		sliceRecords, ok := recordsMap[r.Time.AsTime().Format("200601")]
		if !ok {
			sliceRecords = make([]*expenses.Record, 0)
		}

		sliceRecords = append(sliceRecords, r)

		recordsMap[r.Time.AsTime().Format("200601")] = sliceRecords
	}

	for _, v := range recordsMap {
		res = append(res, model.LegacyRenderAPISchedulePaymentModel(
			ctx,
			v,
			req.GetUser(),
			offerModel.Landlord.User,
			req.GetOffer()),
		)
	}

	return &offer.ScheduledPaymentListResponse{
		Items: res,
		Total: int32(len(res)),
	}, nil
}

func (server *serverImpl) FilterScheduledPayments(
	ctx context.Context,
	req *offer.FilterScheduledPaymentsRequest,
) (*offer.ScheduledPaymentListResponse, error) {
	params := req.GetParams()
	if params == nil {
		params = &query.QueryParams{}
	}

	params.Filter = transformLegacyScheduleFilters(params.GetFilter())
	params.Sort = transformLegacyScheduleSort(params.GetSort())

	records, err := connections.ExpensesService.RecordFilterList(ctx, &expenses.RecordFilterListRequest{
		Params: params,
	})
	if err != nil {
		return nil, err
	}

	var res []*offer.ScheduledPayment

	recordsMap := make(map[string][]*expenses.Record)

	var keys []string

	for _, r := range records.GetRecords() {
		k := r.Time.AsTime().Format("200601")

		_, ok := recordsMap[k]
		if !ok {
			keys = append(keys, k)
		}

		recordsMap[k] = append(recordsMap[k], r)
	}

	for _, k := range keys {
		res = append(res, model.LegacyRenderAPISchedulePaymentModel(ctx, recordsMap[k], "", "", ""))
	}

	return &offer.ScheduledPaymentListResponse{
		Items: res,
		Total: records.GetTotal(),
	}, nil
}

var transformLegacyScheduleFieldMapping = map[string]string{
	"offer":              "correlation_id",
	"payout.transaction": "value.transaction_id",
	"date":               "time",
}

func transformLegacyScheduleFieldMap(field string) string {
	transformed, ok := transformLegacyScheduleFieldMapping[field]
	if !ok {
		return field
	}

	return transformed
}

func transformLegacyScheduleSort(sort []*query.QueryParamsSort) (res []*query.QueryParamsSort) {
	for _, s := range sort {
		if s == nil {
			continue
		}

		s.Field = transformLegacyScheduleFieldMap(s.GetField())

		res = append(res, s)
	}

	return
}

func transformLegacyScheduleFilters(filter []*query.QueryParamsFilter) (res []*query.QueryParamsFilter) {
	var extra []*query.QueryParamsFilter

	for _, f := range filter {
		if f == nil {
			continue
		}

		for i, field := range f.GetFields() {
			if field == "offer" {
				extra = append(
					extra,
					&query.QueryParamsFilter{
						Fields:   []string{"correlation_type"},
						Value:    []string{expenses.CorrelationType_CORRELATION_TYPE_OFFER.String()},
						Operator: query.QueryFilterOperator_EQUALS,
					},
				)
			}

			f.Fields[i] = transformLegacyScheduleFieldMap(field)
		}

		res = append(res, f)
	}

	res = append(res, extra...)

	return
}

func (server *serverImpl) ScheduledPaymentAdvanceCreate(
	ctx context.Context,
	req *offer.ScheduledPaymentAdvanceCreateRequest,
) (*offer.ScheduledPaymentAdvanceCreateResponse, error) {
	return &offer.ScheduledPaymentAdvanceCreateResponse{}, nil
}

func (server *serverImpl) ScheduledPaymentAdvanceCancel(
	ctx context.Context,
	req *offer.ScheduledPaymentAdvanceCancelRequest,
) (*offer.ScheduledPaymentAdvanceCancelResponse, error) {
	return &offer.ScheduledPaymentAdvanceCancelResponse{}, nil
}

func (server *serverImpl) ScheduledPaymentSave(
	ctx context.Context,
	req *offer.ScheduledPaymentSaveRequest,
) (*offer.ScheduledPaymentSaveResponse, error) {
	return &offer.ScheduledPaymentSaveResponse{}, nil
}

func (server *serverImpl) ScheduledPaymentPay(
	ctx context.Context,
	req *offer.ScheduledPaymentPayRequest,
) (*offer.ScheduledPaymentPayResponse, error) {
	return nil, status.ErrInvalidState
}

func (server *serverImpl) FilterScheduledDeposites(
	ctx context.Context,
	req *offer.FilterScheduledDepositesRequest,
) (*offer.FilterScheduledDepositesResponse, error) {
	return nil, status.ErrInvalidState
}

func (server *serverImpl) ScheduledDepositUpdate(
	ctx context.Context,
	req *offer.ScheduledDepositUpdateRequest,
) (*offer.ScheduledDepositUpdateResponse, error) {
	return nil, status.ErrInvalidState
}
