package impl

import (
	"context"
	"time"

	"gitlab.mashplace.com/mashplace/backend/common/api/grpcctx"
	"gitlab.mashplace.com/mashplace/backend/common/generic/log"
	"gitlab.mashplace.com/mashplace/backend/common/generic/util"
	messaging "gitlab.mashplace.com/mashplace/backend/messaging/api"
	notification "gitlab.mashplace.com/mashplace/backend/notification/api"
	offer "gitlab.mashplace.com/mashplace/backend/offer/api"
	"gitlab.mashplace.com/mashplace/backend/offer/internal/connections"
	"gitlab.mashplace.com/mashplace/backend/offer/internal/schedule/oneoff"
	"gitlab.mashplace.com/mashplace/backend/offer/internal/status"
	"gitlab.mashplace.com/mashplace/backend/offer/model"
	"go.mongodb.org/mongo-driver/bson/primitive"
)

// Document rendering option for document service.
const (
	SignMargin           = 0
	SignLineHeight       = 54
	SignWidth            = 206
	SignHeight           = 30
	SignX                = 306
	SignPageAST          = 16
	SignPageASTNoDeposit = 17
	SignPageLandlord     = 29
	SignPageTenant       = 32
)

// nolint
func (server *serverImpl) SignatureCallbackHandler(
	ctx context.Context,
	req *offer.SignatureCallbackHandlerRequest,
) (*offer.SignatureCallbackHandlerResponse, error) {
	id, err := primitive.ObjectIDFromHex(req.SignatureRequest.Metadata.OfferId)
	if err != nil {
		return nil, err
	}
	m, err := model.GetOfferByID(ctx, id)
	if err != nil {
		return nil, err
	}

	if m.State != offer.OfferState_SIGNING {
		log.Infof("%s offer is not in SIGNING state, actual state is %s", m.ID.Hex(), m.State.String())

		return nil, nil
	}

	bgctx := grpcctx.ValuesBackgroundContext(ctx)

	waserr := req.SignatureRequest.IsDeclined || req.SignatureRequest.HasError

	for _, s := range req.SignatureRequest.Signatures {
		var signatureStatus *model.SignatureModel

		signatureStatus, err = TranslateSignatureStatus(s.StatusCode)
		if err != nil {
			log.WithError(err).Errorf("invalid signature %+v", s)
			waserr = true
			break
		}
		if s.SignerEmailAddress == util.UnPtrString(m.Landlord.EmailPrimary) {
			if len(m.Landlord.Signature) == 0 ||
				m.Landlord.Signature[len(m.Landlord.Signature)-1].State != signatureStatus.State {
				m.Landlord.Signature = append(m.Landlord.Signature, signatureStatus)
			}
			if signatureStatus.State == offer.SignStatusState_SIGNATURE_PROVIDED {
				go func() {
					notifySignatureProvided(
						bgctx,
						notification.Role_ROLE_LANDLORD,
						m,
					)

					_ = oneoff.ScheduleInvocation(
						bgctx,
						time.Now().UTC().Add(2*time.Hour),
						oneoff.InvocationTypeSignatureProvided,
						&oneoff.InvocationSignatureProvided{
							OfferID: m.ID.Hex(),
							Actor:   notification.Role_ROLE_LANDLORD,
						},
					)
				}()
			}
		} else {
			for _, t := range m.Tenants {
				if s.SignerEmailAddress == util.UnPtrString(t.EmailPrimary) {
					if len(t.Signature) == 0 ||
						t.Signature[len(t.Signature)-1].State != signatureStatus.State {
						t.Signature = append(t.Signature, signatureStatus)
					}
					if signatureStatus.State == offer.SignStatusState_SIGNATURE_PROVIDED {
						go func() {
							notifySignatureProvided(
								bgctx,
								notification.Role_ROLE_TENANT,
								m,
							)

							_ = oneoff.ScheduleInvocation(
								bgctx,
								time.Now().UTC().Add(2*time.Hour),
								oneoff.InvocationTypeSignatureProvided,
								&oneoff.InvocationSignatureProvided{
									OfferID: m.ID.Hex(),
									Actor:   notification.Role_ROLE_TENANT,
								},
							)
						}()
					}
					break
				}
			}

			for _, g := range m.Guarantors {
				if s.SignerEmailAddress == util.UnPtrString(g.EmailPrimary) {
					if len(g.Signature) == 0 ||
						g.Signature[len(g.Signature)-1].State != signatureStatus.State {
						g.Signature = append(g.Signature, signatureStatus)
					}

					break
				}
			}
		}
	}

	if waserr {
		err = status.SetOfferState(ctx, m, offer.OfferState_ERROR)
		if err != nil {
			return nil, err
		}

		err = server.clearActiveOffersFromListing(ctx, m.Property)
		if err != nil {
			log.WithError(err).WithField("property", m.Property).Error("VersionSetActiveOffer to empty" +
				" failed as a part of SignatureCallbackHandler grpc method")
		}

		return &offer.SignatureCallbackHandlerResponse{}, nil
	}

	// set SIGNATURE_PROVIDED status if it's not set
	if req.SignatureRequest.IsComplete {
		if len(m.Landlord.Signature) == 0 ||
			m.Landlord.Signature[len(m.Landlord.Signature)-1].State !=
				offer.SignStatusState_SIGNATURE_PROVIDED {
			m.Landlord.Signature = append(m.Landlord.Signature, &model.SignatureModel{
				State: offer.SignStatusState_SIGNATURE_PROVIDED,
				Time:  time.Now(),
			})
		}

		for _, t := range m.Tenants {
			if len(t.Signature) == 0 ||
				t.Signature[len(t.Signature)-1].State != offer.SignStatusState_SIGNATURE_PROVIDED {
				t.Signature = append(t.Signature, &model.SignatureModel{
					State: offer.SignStatusState_SIGNATURE_PROVIDED,
					Time:  time.Now(),
				})
			}
		}

		for _, g := range m.Guarantors {
			if len(g.Signature) == 0 ||
				g.Signature[len(g.Signature)-1].State != offer.SignStatusState_SIGNATURE_PROVIDED {
				g.Signature = append(g.Signature, &model.SignatureModel{
					State: offer.SignStatusState_SIGNATURE_PROVIDED,
					Time:  time.Now(),
				})
			}
		}
	}

	err = model.UpdateOffer(ctx, m)
	if err != nil {
		return nil, err
	}

	if req.SignatureRequest.IsComplete {
		err = status.SetOfferState(ctx, m, offer.OfferState_SIGNED)
		if err != nil {
			return nil, err
		}
	}

	return &offer.SignatureCallbackHandlerResponse{}, nil
}

func notifySignatureProvided(ctx context.Context, actor notification.Role, off *model.OfferModel) {
	if off == nil {
		log.Errorf("unable to send signature invitation notification: nil offer")
		return
	}

	var recipientID, actorID string

	var target notification.Role

	switch actor {
	case notification.Role_ROLE_LANDLORD:
		actorID = off.LandlordUser()
		recipientID = off.TenantUser()
		target = notification.Role_ROLE_TENANT
	case notification.Role_ROLE_TENANT:
		actorID = off.TenantUser()
		recipientID = off.LandlordUser()
		target = notification.Role_ROLE_LANDLORD
	}

	_, err := connections.NotificationService.SendNotification(ctx, &notification.SendNotificationRequest{
		Actor: &notification.Actor{
			User: actorID,
			Role: actor,
		},
		Entity: notification.Entity_ENTITY_OFFER_SIGNATURE_INVITATION,
		Action: notification.Action_ACTION_PENDING,
		Targets: []*notification.Target{
			{
				User: recipientID,
				Role: target,
			},
		},
		Embeds: []*notification.EventEmbed{
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_PROPERTY,
				EmbedValue: off.Property,
			},
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_OFFER,
				EmbedValue: off.ID.Hex(),
			},
		},
	})
	if err != nil {
		log.WithError(err).Errorf("failed to send signature invitation notification, offer %s", off.ID.Hex())
	}
}

func getSignPageTenant(isGuarantor bool) int32 {
	if isGuarantor {
		return SignPageTenant + 1
	}
	return SignPageTenant
}

func getSignPageLandlord(isGuarantor bool) int32 {
	if isGuarantor {
		return SignPageLandlord + 1
	}
	return SignPageLandlord
}

func getPageAST(isDeposit bool, isGuarantor bool) int32 {
	var page int32

	if isDeposit {
		page = SignPageAST
	} else {
		page = SignPageASTNoDeposit
	}

	if isGuarantor {
		page++
	}

	return page
}

// TranslateSignatureStatus creates new model based on signature status code value.
func TranslateSignatureStatus(code string) (*model.SignatureModel, error) {
	switch code {
	case "success", "signed":
		return &model.SignatureModel{
			State: offer.SignStatusState_SIGNATURE_PROVIDED,
			Time:  time.Now(),
		}, nil
	case "on_hold", "on_hold_by_requester", "awaiting_signature":
		return &model.SignatureModel{
			State: offer.SignStatusState_AWAITING_SIGNATURE,
			Time:  time.Now(),
		}, nil
	case "declined":
		return &model.SignatureModel{
			State: offer.SignStatusState_SIGNATURE_REFUSED,
			Time:  time.Now(),
		}, nil
	case "error_unknown", "error_file", "error_component_position", "error_text_tag":
		return nil, ErrSign
	default:
		return nil, ErrSign
	}
}
