package impl

import (
	"context"

	closecom "gitlab.mashplace.com/mashplace/backend/closecom/api"
	contract "gitlab.mashplace.com/mashplace/backend/contract/api"
	invitation "gitlab.mashplace.com/mashplace/backend/invitation/api"
	invoice "gitlab.mashplace.com/mashplace/backend/invoice/api"
	offer "gitlab.mashplace.com/mashplace/backend/offer/api"
	"gitlab.mashplace.com/mashplace/backend/offer/docs"
	"gitlab.mashplace.com/mashplace/backend/offer/internal/app/usecase/rto"
	"gitlab.mashplace.com/mashplace/backend/offer/internal/connections"
	"gitlab.mashplace.com/mashplace/backend/offer/model"
	schedule "gitlab.mashplace.com/mashplace/backend/schedule/api"
)

// Server describes Offer service server interface.
type Server interface {
	offer.OfferServiceServer
	invitation.InvitationNotifyServiceServer
	invoice.InvoiceNotifyServiceServer
	schedule.ReminderConsumerServiceServer
	RegisterSchedules(ctx context.Context) error
}

type serverImpl struct {
	offer.UnimplementedOfferServiceServer
	invoice.UnimplementedInvoiceNotifyServiceServer
	closecom.UnimplementedClosecomServiceServer
	updateStateToEndingUseCase            rto.UpdateStateToEndingUseCase
	standingOrderUseCase                  rto.StandingOrderUseCase
	updateStateToWaitingForPaymentUseCase rto.UpdateStateToWaitingForPaymentUseCase
	updateStateToActiveUseCase            rto.UpdateStateToActiveUseCase
}

// NewServer builds Offer server interface and runs background processing.
func NewServer(ctx context.Context,
	updateStateToEndingUseCase rto.UpdateStateToEndingUseCase,
	standingOrderUseCase rto.StandingOrderUseCase,
	updateStateToWaitingForPaymentUseCase rto.UpdateStateToWaitingForPaymentUseCase,
	updateStateToActiveUseCase rto.UpdateStateToActiveUseCase) Server {
	var err error

	// template names are constant now, no need to update them each time template is updated
	// template revision is updated automatically if bindata modification time is after current
	// template revision modification date.
	// see contract.TemplateID for details.

	model.TemplateASTID, err = contract.TemplateID(
		ctx,
		docs.Asset,
		docs.AssetInfo,
		connections.ContractService,
		"offer-contract-template",
		"template.html",
	)
	if err != nil {
		panic(err)
	}

	model.TemplateSOMFID, err = contract.TemplateID(
		ctx,
		docs.Asset,
		docs.AssetInfo,
		connections.ContractService,
		"offer-somf-template",
		"StandingOrderMandateForm.html",
	)
	if err != nil {
		panic(err)
	}

	srv := &serverImpl{
		updateStateToEndingUseCase:            updateStateToEndingUseCase,
		standingOrderUseCase:                  standingOrderUseCase,
		updateStateToWaitingForPaymentUseCase: updateStateToWaitingForPaymentUseCase,
		updateStateToActiveUseCase:            updateStateToActiveUseCase,
	}

	return srv
}
