package mongo

import (
	"context"

	api "gitlab.mashplace.com/mashplace/backend/offer/api/rto"
	"gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto"
	"gitlab.mashplace.com/mashplace/backend/offer/model"

	"github.com/pkg/errors"
	"go.mongodb.org/mongo-driver/bson"
	"go.mongodb.org/mongo-driver/bson/primitive"
	"go.mongodb.org/mongo-driver/mongo"
)

type repository struct {
	collection *mongo.Collection
}

func NewRepository(db *mongo.Database) rto.Repository {
	return &repository{
		collection: db.Collection(model.MongoOffersCollection),
	}
}

func (r *repository) Create(ctx context.Context, rto *rto.Rto) error {
	_, err := r.collection.InsertOne(ctx, rto)

	return err
}

func (r *repository) Update(ctx context.Context, rto *rto.Rto) (int64, error) {
	res, err := r.collection.ReplaceOne(ctx, bson.M{"_id": rto.ID}, rto)
	if err != nil {
		return 0, err
	}

	return res.ModifiedCount, nil
}

func (r *repository) MustUpdate(ctx context.Context, rto *rto.Rto) error {
	updatedCount, err := r.Update(ctx, rto)
	if err != nil {
		return err
	}
	if updatedCount == 0 {
		return errors.Wrap(errors.New("updated count = 0"), "repository MustUpdate updatedCount == 0")
	}

	return nil
}

func (r *repository) DeleteByID(ctx context.Context, id primitive.ObjectID) (int64, error) {
	res, err := r.collection.DeleteOne(ctx, bson.M{"_id": id})
	if err != nil {
		return 0, err
	}

	return res.DeletedCount, nil
}

func (r *repository) FindByID(ctx context.Context, id primitive.ObjectID) (*rto.Rto, error) {
	res := r.collection.FindOne(ctx, bson.M{"_id": id})
	if res.Err() != nil {
		return nil, res.Err()
	}
	rto := new(rto.Rto)
	err := res.Decode(rto)
	if err != nil {
		return nil, errors.Wrap(err, "repository FindByID res.Decode")
	}

	return rto, nil
}

func (r *repository) Drop(ctx context.Context) error {
	return r.collection.Drop(ctx)
}

func (r *repository) FindAllByFilter(ctx context.Context, filter *rto.Filter) ([]*rto.Rto, error) {
	andFilter := bson.A{
		bson.M{"state": bson.M{"$in": filter.States}},
		bson.M{filter.GetDateColumnName(): bson.M{"$gte": filter.DateFrom}},
		bson.M{filter.GetDateColumnName(): bson.M{"$lt": filter.DateTo}},
	}
	if filter.NotificationType != "" {
		andFilter = append(andFilter, bson.D{
			{"notifications." + string(filter.NotificationType), bson.D{ //nolint
				{"$exists", false}, //nolint
			}},
		})
	}

	return r.findAll(ctx, bson.M{"$and": andFilter})
}

func (r *repository) findAll(ctx context.Context, filter bson.M) ([]*rto.Rto, error) {
	cur, err := r.collection.Find(ctx, filter)
	if err != nil {
		return []*rto.Rto{}, errors.Wrap(err, "repository findAll collection.Find")
	}
	var rtos []*rto.Rto
	for cur.Next(ctx) {
		r := &rto.Rto{}
		err = cur.Decode(r)
		if err != nil {
			return []*rto.Rto{}, errors.Wrap(err, "repository findAll cur.Decode")
		}
		rtos = append(rtos, r)
	}

	return rtos, nil
}

func (r *repository) FindAllEmptyLeadTenantUsersByStates(ctx context.Context, states []api.RtoState) (
	[]*rto.Rto, error) {
	andFilter := bson.A{
		bson.M{"state": bson.M{"$in": states}},
		bson.M{"lead_tenant.user": bson.M{"$in": []interface{}{""}}},
	}

	return r.findAll(ctx, bson.M{"$and": andFilter})
}

func (r *repository) FindAllEmptyLeadTenantUsersByEmail(ctx context.Context, leadTenantEmail string) (
	[]*rto.Rto, error) {
	andFilter := bson.A{
		bson.M{"lead_tenant.email_primary": leadTenantEmail},
		bson.M{"lead_tenant.user": bson.M{"$in": []interface{}{""}}},
	}

	return r.findAll(ctx, bson.M{"$and": andFilter})
}

func (r *repository) FindAllByPropertyIDs(ctx context.Context, filter *rto.AllByPropertyIDsFilter) ([]*rto.Rto, error) {
	andFilter := bson.A{
		bson.M{"property": bson.M{"$in": filter.PropertyIDs}},
		bson.M{"state": bson.M{"$gte": filter.MinState}},
		bson.M{"state": bson.M{"$lte": filter.MaxState}},
	}

	return r.findAll(ctx, bson.M{"$and": andFilter})
}

func (r *repository) FindByPropertyID(ctx context.Context, propertyID string) (*rto.Rto, error) {
	res := r.collection.FindOne(ctx, bson.M{"property": propertyID})
	if res.Err() != nil {
		return nil, res.Err()
	}
	rto := new(rto.Rto)
	err := res.Decode(rto)
	if err != nil {
		return nil, errors.Wrap(err, "repository FindByPropertyID res.Decode")
	}

	return rto, nil
}

func (r *repository) FindAllByLandlordID(ctx context.Context, filter *rto.AllByLandlordIDFilter) ([]*rto.Rto, error) {
	andFilter := bson.A{
		bson.M{"landlord.user": bson.M{"$in": []interface{}{filter.LandlordID}}},
		bson.M{"state": bson.M{"$in": filter.States}},
	}

	return r.findAll(ctx, bson.M{"$and": andFilter})
}
