package rto

import (
	"context"
	"fmt"
	"os"
	"time"

	account "gitlab.mashplace.com/mashplace/backend/account/api"
	listing "gitlab.mashplace.com/mashplace/backend/listing/api"
	api "gitlab.mashplace.com/mashplace/backend/offer/api/rto"

	"go.mongodb.org/mongo-driver/bson/primitive"
	"google.golang.org/protobuf/types/known/timestamppb"
)

type Rto struct {
	UpdatedAt       time.Time          `bson:"updated_at"`
	CreatedAt       time.Time          `bson:"created_at"`
	Conditions      *Conditions        `bson:"conditions"`
	Landlord        *Landlord          `bson:"landlord"`
	PaidStatuses    *PaidStatuses      `bson:"paid_statuses"`
	Notifications   Notifications      `bson:"notifications"`
	LeadTenant      *Tenant            `bson:"lead_tenant"`
	PropertyIvr     string             `bson:"property_ivr"`
	PropertyAddress string             `bson:"property_address"`
	Property        string             `bson:"property"`
	PropertyListing string             `bson:"property_listing"`
	OpportunityID   string             `bson:"opportunity_id"`
	InvoiceID       string             `bson:"invoice_id"`
	Tenants         []*Tenant          `bson:"tenants"`
	RentPlan        api.RentPlan       `bson:"rent_plan"`
	State           api.RtoState       `bson:"state"`
	ID              primitive.ObjectID `bson:"_id"`
	IsRto           bool               `bson:"is_rto"`
}

type Repository interface {
	Create(ctx context.Context, rto *Rto) error

	Update(ctx context.Context, rto *Rto) (int64, error)
	MustUpdate(ctx context.Context, rto *Rto) error

	DeleteByID(ctx context.Context, id primitive.ObjectID) (int64, error)

	Drop(ctx context.Context) error

	FindByID(ctx context.Context, id primitive.ObjectID) (*Rto, error)
	FindByPropertyID(ctx context.Context, propertyID string) (*Rto, error)

	FindAllByFilter(ctx context.Context, filter *Filter) ([]*Rto, error)
	FindAllEmptyLeadTenantUsersByStates(ctx context.Context, states []api.RtoState) ([]*Rto, error)
	FindAllEmptyLeadTenantUsersByEmail(ctx context.Context, leadTenantEmail string) ([]*Rto, error)
	FindAllByPropertyIDs(ctx context.Context, filter *AllByPropertyIDsFilter) ([]*Rto, error)
	FindAllByLandlordID(ctx context.Context, filter *AllByLandlordIDFilter) ([]*Rto, error)
}

func (r *Rto) CreateApiResponse() *api.Rto {
	return &api.Rto{
		Id:              r.ID.Hex(),
		RentPlan:        r.RentPlan,
		State:           r.State,
		CreatedAt:       timestamppb.New(r.CreatedAt),
		UpdatedAt:       timestamppb.New(r.UpdatedAt),
		Landlord:        CreateApiLandlord(r.Landlord),
		Property:        r.Property,
		PropertyIvr:     r.PropertyIvr,
		PropertyListing: r.PropertyListing,
		PropertyAddress: r.PropertyAddress,
		InvoiceId:       r.InvoiceID,
		Conditions:      CreateApiConditions(r.Conditions),
		LeadTenant:      CreateApiTenant(r.LeadTenant),
	}
}

func (r *Rto) GetDate(dateColumnName DateColumnName) time.Time {
	switch dateColumnName {
	case DateColumnNameCreatedAt:
		return r.CreatedAt
	case DateColumnNameUpdatedAt:
		return r.UpdatedAt
	case DateColumnNameEndDate:
		if r.Conditions != nil {
			return r.Conditions.EndDate
		}

		return time.Time{}
	default:
		return time.Time{}
	}
}

func (r *Rto) GetLeadTenantID() string {
	if r.LeadTenant != nil {
		return r.LeadTenant.User
	}

	return ""
}

func (r *Rto) GetLeadTenantEmail() string {
	if r.LeadTenant != nil {
		return r.LeadTenant.EmailPrimary
	}

	return ""
}

func (r *Rto) GetLeadTenantPhone() string {
	if r.LeadTenant != nil {
		return r.LeadTenant.PhonePrimary
	}

	return ""
}

func (r *Rto) GetLeadTenantFirstName() string {
	if r.LeadTenant != nil {
		return r.LeadTenant.FirstName
	}

	return ""
}

func (r *Rto) GetLeadTenantLastName() string {
	if r.LeadTenant != nil {
		return r.LeadTenant.LastName
	}

	return ""
}

func (r *Rto) GetPrice() int {
	if r.Conditions != nil {
		return int(r.Conditions.Price)
	}

	return 0
}

func (r *Rto) GetMoveIn() time.Time {
	if r.Conditions != nil {
		return r.Conditions.MoveInDate
	}

	return time.Time{}
}

func (r *Rto) GetEndDate() time.Time {
	if r.Conditions != nil {
		return r.Conditions.EndDate
	}

	return time.Time{}
}

func (r *Rto) GetPropertyAdminLink() string {
	return fmt.Sprintf("https://%s/#/listing/%s/edit", os.Getenv("ADMIN_SITE_URL"), r.Property)
}

func (r *Rto) GetRtoAdminLink() string {
	return fmt.Sprintf("https://%s/#/offers/%s/edit", os.Getenv("ADMIN_SITE_URL"), r.ID.Hex())
}

func (r *Rto) GetLeadTenantForCloseCom() string {
	if r.GetLeadTenantID() != "" {
		return fmt.Sprintf("%s || https://%s/#/user/%s/view",
			r.LeadTenant.EmailPrimary, os.Getenv("ADMIN_SITE_URL"), r.LeadTenant.User)
	}

	return r.GetLeadTenantEmail()
}

func (r *Rto) GetLandlordID() string {
	if r.Landlord != nil {
		return r.Landlord.GetID()
	}

	return ""
}

func (r *Rto) GetLeadID() string {
	if r.Landlord != nil {
		return r.Landlord.LeadID
	}

	return ""
}

func (r *Rto) GetRentPlanForCloseCom() string {
	switch r.RentPlan {
	case api.RentPlan_RENT_PLAN_ESSENTIAL:
		return "2% - Essential"
	case api.RentPlan_RENT_PLAN_SECURED:
		return "5% - Secured"
	default:
		return "Free"
	}
}

func CreateRto(now time.Time,
	rentPlan api.RentPlan,
	landlordID string,
	landlordAccount *account.Account,
	listingVersion *listing.Version) *Rto {
	return &Rto{
		ID:              primitive.NewObjectID(),
		IsRto:           true,
		State:           api.RtoState_RTO_STATE_PENDING_SETUP,
		CreatedAt:       now,
		UpdatedAt:       now,
		RentPlan:        rentPlan,
		Landlord:        CreateLandlord(landlordID, landlordAccount),
		Property:        listingVersion.GetPropertyId(),
		PropertyIvr:     listingVersion.GetIvr(),
		PropertyListing: listingVersion.GetListingId(),
		PropertyAddress: listingVersion.GetLocation().FormatAddress(),
		PaidStatuses:    &PaidStatuses{},
	}
}
