package rto

import (
	"context"
	"time"

	"gitlab.mashplace.com/mashplace/backend/common/generic/log"

	"gitlab.mashplace.com/mashplace/backend/common/lib/constants"
	loggerservice "gitlab.mashplace.com/mashplace/backend/common/lib/logger/service"
	api "gitlab.mashplace.com/mashplace/backend/offer/api/rto"
	domain "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto"

	"github.com/davecgh/go-spew/spew"
	"go.mongodb.org/mongo-driver/bson/primitive"
)

type loggingMiddleware struct {
	next          domain.Repository
	interfaceName string
	logger        loggerservice.Logger
}

func NewLoggingMiddleware(s domain.Repository,
	interfaceName string, logger loggerservice.Logger) domain.Repository {
	return &loggingMiddleware{
		next:          s,
		interfaceName: interfaceName,
		logger:        logger,
	}
}

func (s *loggingMiddleware) Create(ctx context.Context, rto *domain.Rto) (err error) {
	defer func(begin time.Time) {
		log.WithContext(ctx).WithFields(log.Fields{
			"method":              s.interfaceName + "_" + "Create",
			"rto":                 spew.Sdump(rto),
			"took":                time.Since(begin),
			constants.LogErrorKey: err,
		}).Debug()
	}(time.Now())

	return s.next.Create(ctx, rto)
}

func (s *loggingMiddleware) Update(ctx context.Context, rto *domain.Rto) (updatedCount int64, err error) {
	defer func(begin time.Time) {
		log.WithContext(ctx).WithFields(log.Fields{
			"method":              s.interfaceName + "_" + "Update",
			"rto":                 spew.Sdump(rto),
			"updatedCount":        updatedCount,
			"took":                time.Since(begin),
			constants.LogErrorKey: err,
		}).Debug()
	}(time.Now())

	return s.next.Update(ctx, rto)
}

func (s *loggingMiddleware) MustUpdate(ctx context.Context, rto *domain.Rto) (err error) {
	defer func(begin time.Time) {
		log.WithContext(ctx).WithFields(log.Fields{
			"method":              s.interfaceName + "_" + "MustUpdate",
			"rto":                 spew.Sdump(rto),
			"took":                time.Since(begin),
			constants.LogErrorKey: err,
		}).Debug()
	}(time.Now())

	return s.next.MustUpdate(ctx, rto)
}

func (s *loggingMiddleware) DeleteByID(ctx context.Context, id primitive.ObjectID) (deletedCount int64, err error) {
	defer func(begin time.Time) {
		log.WithContext(ctx).WithFields(log.Fields{
			"method":              s.interfaceName + "_" + "DeleteByID",
			"deletedCount":        deletedCount,
			"took":                time.Since(begin),
			constants.LogErrorKey: err,
		}).Debug()
	}(time.Now())

	return s.next.DeleteByID(ctx, id)
}

func (s *loggingMiddleware) Drop(ctx context.Context) (err error) {
	defer func(begin time.Time) {
		log.WithContext(ctx).WithFields(log.Fields{
			"method":              s.interfaceName + "_" + "Drop",
			"took":                time.Since(begin),
			constants.LogErrorKey: err,
		}).Debug()
	}(time.Now())

	return s.next.Drop(ctx)
}

func (s *loggingMiddleware) FindByID(ctx context.Context, id primitive.ObjectID) (rto *domain.Rto, err error) {
	defer func(begin time.Time) {
		log.WithContext(ctx).WithFields(log.Fields{
			"method":              s.interfaceName + "_" + "FindByID",
			"rto":                 spew.Sdump(rto),
			"took":                time.Since(begin),
			constants.LogErrorKey: err,
		}).Debug()
	}(time.Now())

	return s.next.FindByID(ctx, id)
}

func (s *loggingMiddleware) FindByPropertyID(ctx context.Context, propertyID string) (rto *domain.Rto, err error) {
	defer func(begin time.Time) {
		log.WithContext(ctx).WithFields(log.Fields{
			"method":              s.interfaceName + "_" + "FindByPropertyID",
			"propertyID":          spew.Sdump(propertyID),
			"rto":                 spew.Sdump(rto),
			"took":                time.Since(begin),
			constants.LogErrorKey: err,
		}).Debug()
	}(time.Now())

	return s.next.FindByPropertyID(ctx, propertyID)
}

func (s *loggingMiddleware) FindAllByFilter(ctx context.Context, filter *domain.Filter) (
	rtos []*domain.Rto, err error) {
	defer func(begin time.Time) {
		log.WithContext(ctx).WithFields(log.Fields{
			"method":              s.interfaceName + "_" + "FindAllByFilter",
			"filter":              spew.Sdump(filter),
			"rtos":                spew.Sdump(rtos),
			"took":                time.Since(begin),
			constants.LogErrorKey: err,
		}).Debug()
	}(time.Now())

	return s.next.FindAllByFilter(ctx, filter)
}

func (s *loggingMiddleware) FindAllEmptyLeadTenantUsersByStates(ctx context.Context,
	states []api.RtoState) (rtos []*domain.Rto, err error) {
	defer func(begin time.Time) {
		log.WithContext(ctx).WithFields(log.Fields{
			"method":              s.interfaceName + "_" + "FindAllEmptyLeadTenantUsersByStates",
			"states":              spew.Sdump(states),
			"rtos":                spew.Sdump(rtos),
			"took":                time.Since(begin),
			constants.LogErrorKey: err,
		}).Debug()
	}(time.Now())

	return s.next.FindAllEmptyLeadTenantUsersByStates(ctx, states)
}

func (s *loggingMiddleware) FindAllEmptyLeadTenantUsersByEmail(ctx context.Context, leadTenantEmail string) (
	rtos []*domain.Rto, err error) {
	defer func(begin time.Time) {
		log.WithContext(ctx).WithFields(log.Fields{
			"method":              s.interfaceName + "_" + "FindAllEmptyLeadTenantUsersByEmail",
			"leadTenantEmail":     leadTenantEmail,
			"rtos":                spew.Sdump(rtos),
			"took":                time.Since(begin),
			constants.LogErrorKey: err,
		}).Debug()
	}(time.Now())

	return s.next.FindAllEmptyLeadTenantUsersByEmail(ctx, leadTenantEmail)
}

func (s *loggingMiddleware) FindAllByPropertyIDs(ctx context.Context,
	filter *domain.AllByPropertyIDsFilter) (rtos []*domain.Rto, err error) {
	defer func(begin time.Time) {
		log.WithContext(ctx).WithFields(log.Fields{
			"method":              s.interfaceName + "_" + "FindAllByPropertyIDs",
			"filter":              spew.Sdump(filter),
			"rtos":                spew.Sdump(rtos),
			"took":                time.Since(begin),
			constants.LogErrorKey: err,
		}).Debug()
	}(time.Now())

	return s.next.FindAllByPropertyIDs(ctx, filter)
}

func (s *loggingMiddleware) FindAllByLandlordID(ctx context.Context,
	filter *domain.AllByLandlordIDFilter) (rtos []*domain.Rto, err error) {
	defer func(begin time.Time) {
		log.WithContext(ctx).WithFields(log.Fields{
			"method":              s.interfaceName + "_" + "FindAllByLandlordID",
			"filter":              spew.Sdump(filter),
			"rtos":                spew.Sdump(rtos),
			"took":                time.Since(begin),
			constants.LogErrorKey: err,
		}).Debug()
	}(time.Now())

	return s.next.FindAllByLandlordID(ctx, filter)
}
