package account

import (
	"context"
	"time"

	"gitlab.mashplace.com/mashplace/backend/common/generic/log"

	account "gitlab.mashplace.com/mashplace/backend/account/api"
	"gitlab.mashplace.com/mashplace/backend/common/lib/constants"
	loggerservice "gitlab.mashplace.com/mashplace/backend/common/lib/logger/service"
	request "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto/resource/request/service"
	appservice "gitlab.mashplace.com/mashplace/backend/offer/internal/app/service"

	"github.com/davecgh/go-spew/spew"
)

type loggingMiddleware struct {
	next          appservice.AccountService
	interfaceName string
	logger        loggerservice.Logger
}

func NewLoggingMiddleware(s appservice.AccountService,
	interfaceName string, logger loggerservice.Logger) appservice.AccountService {
	return &loggingMiddleware{
		next:          s,
		interfaceName: interfaceName,
		logger:        logger,
	}
}

func (s *loggingMiddleware) IsVeriffValid(ctx context.Context, req *request.IsVeriffValidRequest) (err error) {
	defer func(begin time.Time) {
		log.WithContext(ctx).WithFields(log.Fields{
			"method":              s.interfaceName + "_" + "IsVeriffValid",
			"req":                 spew.Sdump(req),
			"took":                time.Since(begin),
			constants.LogErrorKey: err,
		}).Debug()
	}(time.Now())

	return s.next.IsVeriffValid(ctx, req)
}

func (s *loggingMiddleware) IsOwnershipVerified(ctx context.Context, req *request.IsOwnershipVerifiedRequest) (err error) {
	defer func(begin time.Time) {
		log.WithContext(ctx).WithFields(log.Fields{
			"method":              s.interfaceName + "_" + "IsOwnershipVerified",
			"req":                 spew.Sdump(req),
			"took":                time.Since(begin),
			constants.LogErrorKey: err,
		}).Debug()
	}(time.Now())

	return s.next.IsOwnershipVerified(ctx, req)
}

func (s *loggingMiddleware) GetAccountByID(ctx context.Context, id string) (resp *account.Account, err error) {
	defer func(begin time.Time) {
		log.WithContext(ctx).WithFields(log.Fields{
			"method":              s.interfaceName + "_" + "GetAccountByID",
			"id":                  id,
			"resp":                spew.Sdump(resp),
			"took":                time.Since(begin),
			constants.LogErrorKey: err,
		}).Debug()
	}(time.Now())

	return s.next.GetAccountByID(ctx, id)
}

func (s *loggingMiddleware) GetAccountByEmail(ctx context.Context, email string) (resp *account.Account, err error) {
	defer func(begin time.Time) {
		log.WithContext(ctx).WithFields(log.Fields{
			"method":              s.interfaceName + "_" + "GetAccountByEmail",
			"email":               email,
			"resp":                spew.Sdump(resp),
			"took":                time.Since(begin),
			constants.LogErrorKey: err,
		}).Debug()
	}(time.Now())

	return s.next.GetAccountByEmail(ctx, email)
}
