package listing

import (
	"context"
	"time"

	"gitlab.mashplace.com/mashplace/backend/common/generic/log"
	"gitlab.mashplace.com/mashplace/backend/common/lib/constants"
	loggerservice "gitlab.mashplace.com/mashplace/backend/common/lib/logger/service"
	listing "gitlab.mashplace.com/mashplace/backend/listing/api"
	request "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto/resource/request/service"
	appservice "gitlab.mashplace.com/mashplace/backend/offer/internal/app/service"

	"github.com/davecgh/go-spew/spew"
)

type loggingMiddleware struct {
	next          appservice.ListingService
	interfaceName string
	logger        loggerservice.Logger
}

func NewLoggingMiddleware(s appservice.ListingService,
	interfaceName string, logger loggerservice.Logger) appservice.ListingService {
	return &loggingMiddleware{
		next:          s,
		interfaceName: interfaceName,
		logger:        logger,
	}
}

func (s *loggingMiddleware) GetVersion(ctx context.Context, req *request.ListingGetVersionRequest) (
	resp *listing.Version, err error) {
	defer func(begin time.Time) {
		log.WithContext(ctx).WithFields(log.Fields{
			"method":              s.interfaceName + "_" + "GetVersion",
			"req":                 spew.Sdump(req),
			"resp":                spew.Sdump(resp),
			"took":                time.Since(begin),
			constants.LogErrorKey: err,
		}).Debug()
	}(time.Now())

	return s.next.GetVersion(ctx, req)
}

func (s *loggingMiddleware) VersionHasNotEmptyBankDetails(ctx context.Context, version *listing.Version) (ok bool) {
	defer func(begin time.Time) {
		log.WithContext(ctx).WithFields(log.Fields{
			"method":  s.interfaceName + "_" + "VersionHasNotEmptyBankDetails",
			"version": spew.Sdump(version),
			"ok":      ok,
			"took":    time.Since(begin),
		}).Debug()
	}(time.Now())

	return s.next.VersionHasNotEmptyBankDetails(ctx, version)
}

func (s *loggingMiddleware) UpdateTenantID(ctx context.Context, req *request.ListingUpdateTenantIDRequest) (err error) {
	defer func(begin time.Time) {
		log.WithContext(ctx).WithFields(log.Fields{
			"method":              s.interfaceName + "_" + "UpdateTenantID",
			"req":                 spew.Sdump(req),
			"took":                time.Since(begin),
			constants.LogErrorKey: err,
		}).Debug()
	}(time.Now())

	return s.next.UpdateTenantID(ctx, req)
}

func (s *loggingMiddleware) UpdateVersionStatus(ctx context.Context,
	req *request.ListingUpdateVersionStatusRequest) (err error) {
	defer func(begin time.Time) {
		log.WithContext(ctx).WithFields(log.Fields{
			"method":              s.interfaceName + "_" + "UpdateVersionStatus",
			"req":                 spew.Sdump(req),
			"took":                time.Since(begin),
			constants.LogErrorKey: err,
		}).Debug()
	}(time.Now())

	return s.next.UpdateVersionStatus(ctx, req)
}
