package rtoserver

import (
	"context"
	"time"

	"gitlab.mashplace.com/mashplace/backend/common/generic/log"

	"gitlab.mashplace.com/mashplace/backend/common/lib/constants"
	loggerservice "gitlab.mashplace.com/mashplace/backend/common/lib/logger/service"
	api "gitlab.mashplace.com/mashplace/backend/offer/api/rto"
	"gitlab.mashplace.com/mashplace/backend/offer/internal/app/transport/grpc"

	"github.com/davecgh/go-spew/spew"
)

type loggingMiddleware struct {
	api.UnimplementedRtoServiceServer
	next          grpc.RtoServer
	interfaceName string
	logger        loggerservice.Logger
}

func NewLoggingMiddleware(s grpc.RtoServer,
	interfaceName string, logger loggerservice.Logger) grpc.RtoServer {
	return &loggingMiddleware{
		next:          s,
		interfaceName: interfaceName,
		logger:        logger,
	}
}

func (s *loggingMiddleware) Create(ctx context.Context, req *api.CreateRequest) (resp *api.RtoResponse, err error) {
	defer func(begin time.Time) {
		log.WithContext(ctx).WithFields(log.Fields{
			"method":              s.interfaceName + "_" + "Create",
			"req":                 spew.Sdump(req),
			"resp":                spew.Sdump(resp),
			"took":                time.Since(begin),
			constants.LogErrorKey: err,
		}).Debug()
	}(time.Now())

	return s.next.Create(ctx, req)
}

func (s *loggingMiddleware) CreateByGuest(ctx context.Context, req *api.CreateByGuestRequest) (
	resp *api.RtoResponse, err error) {
	defer func(begin time.Time) {
		log.WithContext(ctx).WithFields(log.Fields{
			"method":              s.interfaceName + "_" + "CreateByGuest",
			"req":                 spew.Sdump(req),
			"resp":                spew.Sdump(resp),
			"took":                time.Since(begin),
			constants.LogErrorKey: err,
		}).Debug()
	}(time.Now())

	return s.next.CreateByGuest(ctx, req)
}

func (s *loggingMiddleware) UpdateLandlord(ctx context.Context, req *api.UpdateLandlordRequest) (
	resp *api.RtoResponse, err error) {
	defer func(begin time.Time) {
		log.WithContext(ctx).WithFields(log.Fields{
			"method":              s.interfaceName + "_" + "UpdateLandlord",
			"req":                 spew.Sdump(req),
			"resp":                spew.Sdump(resp),
			"took":                time.Since(begin),
			constants.LogErrorKey: err,
		}).Debug()
	}(time.Now())

	return s.next.UpdateLandlord(ctx, req)
}

func (s *loggingMiddleware) CreateTenancy(ctx context.Context, req *api.CreateTenancyRequest) (
	resp *api.RtoResponse, err error) {
	defer func(begin time.Time) {
		log.WithContext(ctx).WithFields(log.Fields{
			"method":              s.interfaceName + "_" + "CreateTenancy",
			"req":                 spew.Sdump(req),
			"resp":                spew.Sdump(resp),
			"took":                time.Since(begin),
			constants.LogErrorKey: err,
		}).Debug()
	}(time.Now())

	return s.next.CreateTenancy(ctx, req)
}

func (s *loggingMiddleware) UpdateRentPlan(ctx context.Context, req *api.UpdateRentPlanRequest) (
	resp *api.RtoResponse, err error) {
	defer func(begin time.Time) {
		log.WithContext(ctx).WithFields(log.Fields{
			"method":              s.interfaceName + "_" + "UpdateRentPlan",
			"req":                 spew.Sdump(req),
			"resp":                spew.Sdump(resp),
			"took":                time.Since(begin),
			constants.LogErrorKey: err,
		}).Debug()
	}(time.Now())

	return s.next.UpdateRentPlan(ctx, req)
}

func (s *loggingMiddleware) UpdateEndDate(ctx context.Context, req *api.UpdateEndDateRequest) (
	resp *api.RtoResponse, err error) {
	defer func(begin time.Time) {
		log.WithContext(ctx).WithFields(log.Fields{
			"method":              s.interfaceName + "_" + "UpdateEndDate",
			"req":                 spew.Sdump(req),
			"resp":                spew.Sdump(resp),
			"took":                time.Since(begin),
			constants.LogErrorKey: err,
		}).Debug()
	}(time.Now())

	return s.next.UpdateEndDate(ctx, req)
}

func (s *loggingMiddleware) Renewal(ctx context.Context, req *api.RenewalRequest) (resp *api.RtoResponse, err error) {
	defer func(begin time.Time) {
		log.WithContext(ctx).WithFields(log.Fields{
			"method":              s.interfaceName + "_" + "Renewal",
			"req":                 spew.Sdump(req),
			"resp":                spew.Sdump(resp),
			"took":                time.Since(begin),
			constants.LogErrorKey: err,
		}).Debug()
	}(time.Now())

	return s.next.Renewal(ctx, req)
}

func (s *loggingMiddleware) UpdateStateToActive(ctx context.Context,
	req *api.UpdateStateToActiveRequest) (resp *api.RtoResponse, err error) {
	defer func(begin time.Time) {
		log.WithContext(ctx).WithFields(log.Fields{
			"method":              s.interfaceName + "_" + "UpdateStateToActive",
			"req":                 spew.Sdump(req),
			"resp":                spew.Sdump(resp),
			"took":                time.Since(begin),
			constants.LogErrorKey: err,
		}).Debug()
	}(time.Now())

	return s.next.UpdateStateToActive(ctx, req)
}

func (s *loggingMiddleware) Cancel(ctx context.Context, req *api.CancelRequest) (resp *api.RtoResponse, err error) {
	defer func(begin time.Time) {
		log.WithContext(ctx).WithFields(log.Fields{
			"method":              s.interfaceName + "_" + "Cancel",
			"req":                 spew.Sdump(req),
			"resp":                spew.Sdump(resp),
			"took":                time.Since(begin),
			constants.LogErrorKey: err,
		}).Debug()
	}(time.Now())

	return s.next.Cancel(ctx, req)
}

func (s *loggingMiddleware) GetByLandlordID(ctx context.Context, req *api.GetByLandlordIDRequest) (
	resp *api.RtosResponse, err error) {
	defer func(begin time.Time) {
		log.WithContext(ctx).WithFields(log.Fields{
			"method":              s.interfaceName + "_" + "GetByLandlordID",
			"req":                 spew.Sdump(req),
			"resp":                spew.Sdump(resp),
			"took":                time.Since(begin),
			constants.LogErrorKey: err,
		}).Debug()
	}(time.Now())

	return s.next.GetByLandlordID(ctx, req)
}

func (s *loggingMiddleware) GetByPropertyID(ctx context.Context, req *api.GetByPropertyIDRequest) (
	resp *api.RtoResponse, err error) {
	defer func(begin time.Time) {
		log.WithContext(ctx).WithFields(log.Fields{
			"method":              s.interfaceName + "_" + "GetByPropertyID",
			"req":                 spew.Sdump(req),
			"resp":                spew.Sdump(resp),
			"took":                time.Since(begin),
			constants.LogErrorKey: err,
		}).Debug()
	}(time.Now())

	return s.next.GetByPropertyID(ctx, req)
}

func (s *loggingMiddleware) GetByPropertyIDs(ctx context.Context, req *api.GetByPropertyIDsRequest) (
	resp *api.RtosResponse, err error) {
	defer func(begin time.Time) {
		log.WithContext(ctx).WithFields(log.Fields{
			"method":              s.interfaceName + "_" + "GetByPropertyIDs",
			"req":                 spew.Sdump(req),
			"resp":                spew.Sdump(resp),
			"took":                time.Since(begin),
			constants.LogErrorKey: err,
		}).Debug()
	}(time.Now())

	return s.next.GetByPropertyIDs(ctx, req)
}

func (s *loggingMiddleware) AnswerFromTenant(ctx context.Context, req *api.AnswerFromTenantRequest) (
	resp *api.RtoResponse, err error) {
	defer func(begin time.Time) {
		log.WithContext(ctx).WithFields(log.Fields{
			"method":              s.interfaceName + "_" + "AnswerFromTenant",
			"req":                 spew.Sdump(req),
			"resp":                spew.Sdump(resp),
			"took":                time.Since(begin),
			constants.LogErrorKey: err,
		}).Debug()
	}(time.Now())

	return s.next.AnswerFromTenant(ctx, req)
}

func (s *loggingMiddleware) RentPayment(ctx context.Context, req *api.RentPaymentRequest) (
	resp *api.RtoResponse, err error) {
	defer func(begin time.Time) {
		log.WithContext(ctx).WithFields(log.Fields{
			"method":              s.interfaceName + "_" + "RentPayment",
			"req":                 spew.Sdump(req),
			"resp":                spew.Sdump(resp),
			"took":                time.Since(begin),
			constants.LogErrorKey: err,
		}).Debug()
	}(time.Now())

	return s.next.RentPayment(ctx, req)
}

func (s *loggingMiddleware) LandlordApproved(ctx context.Context, req *api.LandlordApprovedRequest) (
	resp *api.IdsResponse, err error) {
	defer func(begin time.Time) {
		log.WithContext(ctx).WithFields(log.Fields{
			"method":              s.interfaceName + "_" + "LandlordApproved",
			"req":                 spew.Sdump(req),
			"resp":                spew.Sdump(resp),
			"took":                time.Since(begin),
			constants.LogErrorKey: err,
		}).Debug()
	}(time.Now())

	return s.next.LandlordApproved(ctx, req)
}

func (s *loggingMiddleware) UpdateLeadTenant(ctx context.Context, req *api.UpdateLeadTenantRequest) (
	resp *api.UpdateLeadTenantResponse, err error) {
	defer func(begin time.Time) {
		log.WithContext(ctx).WithFields(log.Fields{
			"method":              s.interfaceName + "_" + "UpdateLeadTenant",
			"req":                 spew.Sdump(req),
			"resp":                spew.Sdump(resp),
			"took":                time.Since(begin),
			constants.LogErrorKey: err,
		}).Debug()
	}(time.Now())

	return s.next.UpdateLeadTenant(ctx, req)
}

func (s *loggingMiddleware) UpdateLeadTenantUser(ctx context.Context, req *api.UpdateLeadTenantUserRequest) (
	resp *api.RtoResponse, err error) {
	defer func(begin time.Time) {
		log.WithContext(ctx).WithFields(log.Fields{
			"method":              s.interfaceName + "_" + "UpdateLeadTenantUser",
			"req":                 spew.Sdump(req),
			"resp":                spew.Sdump(resp),
			"took":                time.Since(begin),
			constants.LogErrorKey: err,
		}).Debug()
	}(time.Now())

	return s.next.UpdateLeadTenantUser(ctx, req)
}
