package service

import (
	"context"

	account "gitlab.mashplace.com/mashplace/backend/account/api"
	request "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto/resource/request/service"

	"github.com/pkg/errors"
)

var VeriffStatusNotValidError = errors.New("veriff status not valid error")
var OwnershipStatusNotValidError = errors.New("ownership status not valid error")

const (
	IsVeriffValidNotValidRequestErrorText  = "accountService IsVeriffValid req.Validate"
	IsVeriffValidErrorText                 = "accountService IsVeriffValid client.IsVeriffValid"
	IsVeriffValidNotSuccessErrorText       = "accountService IsVeriffValid !isVeriffValidResponse.Success"
	IsOwnershipVerifiedNotSuccessErrorText = "accountService IsOwnershipVerified !IsOwnershipVerifiedResponse.Success"
)

type AccountService interface {
	IsVeriffValid(ctx context.Context, req *request.IsVeriffValidRequest) error
	IsOwnershipVerified(ctx context.Context, req *request.IsOwnershipVerifiedRequest) error

	GetAccountByID(ctx context.Context, id string) (*account.Account, error)
	GetAccountByEmail(ctx context.Context, email string) (*account.Account, error)
}

type accountService struct {
	client account.AccountServiceClient
}

func NewAccountService(accountServiceClient account.AccountServiceClient) AccountService {
	return &accountService{
		client: accountServiceClient,
	}
}

func (as *accountService) IsVeriffValid(ctx context.Context, req *request.IsVeriffValidRequest) error {
	err := req.Validate()
	if err != nil {
		return errors.Wrap(err, IsVeriffValidNotValidRequestErrorText)
	}

	isVeriffValidResponse, err := as.client.IsVeriffValid(ctx, &account.IsVeriffValidRequest{
		AccountId: req.LandlordID,
	})
	if err != nil {
		return errors.Wrap(err, IsVeriffValidErrorText)
	}
	if !isVeriffValidResponse.Success {
		return errors.Wrap(VeriffStatusNotValidError, IsVeriffValidNotSuccessErrorText)
	}

	return nil
}

func (as *accountService) IsOwnershipVerified(ctx context.Context, req *request.IsOwnershipVerifiedRequest) error {
	err := req.Validate()
	if err != nil {
		return errors.Wrap(err, IsVeriffValidNotValidRequestErrorText)
	}

	isVeriffValidResponse, err := as.client.OwnershipIsVerified(ctx, &account.OwnershipIsVerifiedRequest{
		PropertyId: req.PropertyID,
	})
	if err != nil {
		return errors.Wrap(err, IsVeriffValidErrorText)
	}
	if isVeriffValidResponse.Status != account.OwnershipStatus_OWNERSHIP_APPROVED {
		return errors.Wrap(OwnershipStatusNotValidError, IsOwnershipVerifiedNotSuccessErrorText)
	}

	return nil
}

func (as *accountService) GetAccountByID(ctx context.Context, id string) (*account.Account, error) {
	resp, err := as.client.Get(ctx, &account.GetRequest{
		Id: id,
	})
	if err != nil {
		return nil, errors.Wrap(err, "accountService GetAccountByID client.Get")
	}

	return resp.GetAccount(), nil
}

func (as *accountService) GetAccountByEmail(ctx context.Context, email string) (*account.Account, error) {
	resp, err := as.client.GetByEmail(ctx, &account.GetByEmailRequest{
		Email: email,
	})
	if err != nil {
		return nil, errors.Wrap(err, "accountService GetAccountByEmail client.GetByEmail")
	}

	return resp.GetAccount(), nil
}
