package service

import (
	"context"

	"gitlab.mashplace.com/mashplace/backend/common/api/generic"
	"gitlab.mashplace.com/mashplace/backend/common/api/grpcctx"
	listing "gitlab.mashplace.com/mashplace/backend/listing/api"
	request "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto/resource/request/service"

	"github.com/pkg/errors"
)

const PropertyNotFoundErrorText = "property not found"

type ListingService interface {
	GetVersion(ctx context.Context, req *request.ListingGetVersionRequest) (*listing.Version, error)
	VersionHasNotEmptyBankDetails(ctx context.Context, version *listing.Version) bool
	UpdateTenantID(ctx context.Context, req *request.ListingUpdateTenantIDRequest) error
	UpdateVersionStatus(ctx context.Context, req *request.ListingUpdateVersionStatusRequest) error
}

type listingService struct {
	client listing.ListingServiceClient
}

func NewListingService(listingServiceClient listing.ListingServiceClient) ListingService {
	return &listingService{
		client: listingServiceClient,
	}
}

func (ls *listingService) GetVersion(ctx context.Context, req *request.ListingGetVersionRequest) (
	*listing.Version, error) {
	propertyVersion, err := ls.client.VersionGet(grpcctx.WithServiceUser(ctx), &listing.VersionGetRequest{
		Spec: listing.VersionSpecPropertyID(req.LandlordID, req.PropertyID),
	})
	if err != nil {
		return nil, errors.Wrap(errors.New(PropertyNotFoundErrorText), "listingService client.VersionGet")
	}

	return propertyVersion.GetVersion(), nil
}

func (ls *listingService) VersionHasNotEmptyBankDetails(ctx context.Context, version *listing.Version) bool {
	if version.Rental != nil && version.Rental.BankDetails != nil {
		bankDetails := version.Rental.BankDetails

		return bankDetails.AccountName != "" && bankDetails.AccountNumber != "" && bankDetails.SortCode != ""
	}

	return false
}

func (ls *listingService) UpdateTenantID(ctx context.Context, req *request.ListingUpdateTenantIDRequest) error {
	err := req.Validate()
	if err != nil {
		return errors.Wrap(err, "listingService UpdateTenantID req.Validate")
	}

	version, err := ls.GetVersion(ctx, &request.ListingGetVersionRequest{
		PropertyID: req.PropertyID,
		LandlordID: req.LandlordID,
	})
	if err != nil {
		return errors.Wrap(err, "listingService UpdateTenantID GetVersion")
	}

	version.People.Tenants = &generic.StringArrayValue{Value: []string{req.LeadTenantID}}
	_, err = ls.client.VersionUpdate(grpcctx.WithServiceUser(ctx), &listing.VersionUpdateRequest{
		People: version.People,
		Spec:   listing.VersionSpecPropertyID(req.LandlordID, req.PropertyID),
	})
	if err != nil {
		return errors.Wrap(err, "listingService UpdateTenantID client.VersionUpdate")
	}

	return nil
}

func (ls *listingService) UpdateVersionStatus(ctx context.Context,
	req *request.ListingUpdateVersionStatusRequest) error {
	err := req.Validate()
	if err != nil {
		return errors.Wrap(err, "listingService UpdateVersionStatus req.Validate")
	}

	_, err = ls.client.ServiceVersionUpdate(ctx, &listing.ServiceVersionUpdateRequest{
		Spec:   listing.VersionSpecVersionID(grpcctx.ContextUserID(ctx), req.VersionID),
		Status: req.VersionStatus,
	})
	if err != nil {
		return errors.Wrap(err, "listingService UpdateVersionStatus ServiceVersionUpdate")
	}

	return nil
}
