package rto

import (
	"context"
	"encoding/base64"
	"fmt"
	"net/http"
	"os"

	"gitlab.mashplace.com/mashplace/backend/common/lib/constants"
	"gitlab.mashplace.com/mashplace/backend/common/lib/httpclient"
	"gitlab.mashplace.com/mashplace/backend/common/lib/httpclient/resource"
	"gitlab.mashplace.com/mashplace/backend/common/lib/marshallunmarshall"
	domain "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto"
	request "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto/resource/request/service/rto"
	response "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto/resource/response/service/rto"

	"github.com/pkg/errors"
)

const (
	CloseComProdApiKey = "api_5PfwDFFdK58N7NqcZr8zNM.6wYdIHdt1ZAn5D4kp1HaPi"
	CloseComDevApiKey  = "api_2hbydgRJlNztKN2E3bALHh.23gIYnc2T2DenJozn3Zhyf"

	CloseComApiBaseURL = "https://api.close.com/api/v1/"

	CloseComOpportunityURLPart = "opportunity/"
	CloseComLeadURLPart        = "lead/"
)

type CloseComService interface {
	SaveOpportunity(ctx context.Context, req *request.CloseComSaveOpportunityRequest) (
		*response.CloseComSaveOpportunityResponse, error)
	DeleteOpportunity(ctx context.Context, id string) error

	FindLead(ctx context.Context, req *request.CloseComFindLeadRequest) (string, error)
	GetLead(ctx context.Context, id string) (*response.CloseComGetLeadResponse, error)
	CreateLead(ctx context.Context, req *request.CloseComCreateLeadRequest) (string, error)
	DeleteLead(ctx context.Context, id string) error
}

type closeComService struct {
	httpClient         httpclient.HTTPClient
	marshallUnmarshall marshallunmarshall.MarshalUnmarshal
	isDev              bool
}

func NewCloseComService(httpClient httpclient.HTTPClient,
	marshallUnmarshall marshallunmarshall.MarshalUnmarshal) CloseComService {
	return &closeComService{
		httpClient:         httpClient,
		marshallUnmarshall: marshallUnmarshall,
		isDev:              os.Getenv("CLOSE_COM_API_ENVIRONMENT") != constants.EnvProd,
	}
}

func (ccs *closeComService) getDefaultHeaders() map[string]string {
	apiKey := CloseComProdApiKey
	if ccs.isDev {
		apiKey = CloseComDevApiKey
	}

	return map[string]string{
		"Authorization": "Basic " + base64.StdEncoding.EncodeToString([]byte(apiKey+":")),
	}
}

func (ccs *closeComService) SaveOpportunity(ctx context.Context,
	req *request.CloseComSaveOpportunityRequest) (*response.CloseComSaveOpportunityResponse, error) {
	err := req.Validate()
	if err != nil {
		return nil, errors.Wrap(err, "closeComService SaveOpportunity req.Validate")
	}

	var opportunity domain.CloseComOpportunity
	if ccs.isDev {
		opportunity = domain.NewCloseComOpportunityDev(req.Rto, req.RtoPreviousState, req.RtoPreviousStateDate)
	} else {
		opportunity = domain.NewCloseComOpportunityProd(req.Rto, req.RtoPreviousState, req.RtoPreviousStateDate)
	}

	resp := new(response.CloseComSaveOpportunityResponse)
	leadID := req.Rto.GetLeadID()
	if leadID == "" {
		leadID, err = ccs.FindLead(ctx, request.NewCloseComFindLeadRequest(req.Rto.GetLandlordID()))
		if err != nil {
			return nil, errors.Wrap(err, "closeComService SaveOpportunity FindLead")
		}

		if leadID == "" {
			leadID, err = ccs.CreateLead(ctx, &request.CloseComCreateLeadRequest{Landlord: req.Rto.Landlord})
			if err != nil {
				return nil, errors.Wrap(err, "closeComService SaveOpportunity CreateLead")
			}
		}

		resp.LeadID = leadID
	}
	opportunity.SetLeadID(leadID)

	httpMethod, url := http.MethodPost, CloseComOpportunityURLPart
	if req.Rto.OpportunityID != "" {
		httpMethod = http.MethodPut
		url += req.Rto.OpportunityID + "/"
	}

	body, err := ccs.saveEntity(ctx, opportunity, url, httpMethod)
	if err != nil {
		return nil, errors.Wrap(err, "closeComService SaveOpportunity saveEntity")
	}

	var closeComApiCreateOpportunityResponse response.CloseComApiSaveOpportunityResponse
	err = ccs.marshallUnmarshall.Unmarshal(body, &closeComApiCreateOpportunityResponse)
	if err != nil {
		return nil, errors.Wrap(err, "closeComService SaveOpportunity marshallUnmarshall.Unmarshal")
	}
	resp.OpportunityID = closeComApiCreateOpportunityResponse.Id

	return resp, nil
}

func (ccs *closeComService) FindLead(ctx context.Context, req *request.CloseComFindLeadRequest) (string, error) {
	reqJSON, err := ccs.marshallUnmarshall.Marshal(req)
	if err != nil {
		return "", errors.Wrap(err, "closeComService FindLead marshallUnmarshall.Marshal")
	}

	httpRequest, err := resource.NewHTTPRequest(CloseComApiBaseURL+"data/search/", http.MethodPost, "",
		reqJSON, ccs.getDefaultHeaders())
	if err != nil {
		return "", errors.Wrap(err, "closeComService FindLead resource.NewHTTPRequest")
	}

	resp := ccs.httpClient.Request(ctx, httpRequest)
	if resp.Error != nil {
		return "", errors.Wrap(resp.Error, "closeComService FindLead httpClient.Request")
	}

	if resp.StatusCode != http.StatusOK {
		return "", errors.Wrap(fmt.Errorf("bad status code: %d, error: %s", resp.StatusCode, string(resp.Body)),
			"closeComService FindLead resp.StatusCode != http.StatusOK")
	}

	var closeComFindLeadResponse response.CloseComFindLeadResponse
	err = ccs.marshallUnmarshall.Unmarshal(resp.Body, &closeComFindLeadResponse)
	if err != nil {
		return "", errors.Wrap(err, "closeComService FindLead marshallUnmarshall.Unmarshal")
	}

	if len(closeComFindLeadResponse.Data) > 0 {
		return closeComFindLeadResponse.Data[0].Id, nil
	}

	return "", nil
}

func (ccs *closeComService) GetLead(ctx context.Context, id string) (*response.CloseComGetLeadResponse, error) {
	httpRequest, err := resource.NewHTTPRequest(CloseComApiBaseURL+CloseComLeadURLPart+id+"/", http.MethodGet,
		"", nil, ccs.getDefaultHeaders())
	if err != nil {
		return nil, errors.Wrap(err, "closeComService GetLead resource.NewHTTPRequest")
	}

	resp := ccs.httpClient.Request(ctx, httpRequest)
	if resp.Error != nil {
		return nil, errors.Wrap(resp.Error, "closeComService GetLead httpClient.Request")
	}

	if resp.StatusCode != http.StatusOK {
		return nil, errors.Wrap(fmt.Errorf("bad status code: %d, error: %s", resp.StatusCode, string(resp.Body)),
			"closeComService GetLead resp.StatusCode != http.StatusOK")
	}

	var closeComGetLeadResponse response.CloseComGetLeadResponse
	err = ccs.marshallUnmarshall.Unmarshal(resp.Body, &closeComGetLeadResponse)
	if err != nil {
		return nil, errors.Wrap(err, "closeComService GetLead marshallUnmarshall.Unmarshal")
	}

	return &closeComGetLeadResponse, nil
}

func (ccs *closeComService) CreateLead(ctx context.Context, req *request.CloseComCreateLeadRequest) (string, error) {
	err := req.Validate()
	if err != nil {
		return "", errors.Wrap(err, "closeComService CreateLead req.Validate")
	}

	var closeComCreateLeadRequest interface{}
	landlord := req.Landlord
	if ccs.isDev {
		closeComCreateLeadRequest = &request.CloseComCreateLeadDevRequest{
			Name:      landlord.GetFullName(),
			AccountID: landlord.GetID(),
			AdminUser: landlord.GetAdminLink(),
		}
	} else {
		closeComCreateLeadRequest = &request.CloseComCreateLeadProdRequest{
			Name:      landlord.GetFullName(),
			AccountID: landlord.GetID(),
			AdminUser: landlord.GetAdminLink(),
		}
	}

	body, err := ccs.saveEntity(ctx, closeComCreateLeadRequest, CloseComLeadURLPart, http.MethodPost)
	if err != nil {
		return "", errors.Wrap(err, "closeComService CreateLead saveEntity")
	}

	var closeComCreateLeadResponse response.CloseComCreateLeadResponse
	err = ccs.marshallUnmarshall.Unmarshal(body, &closeComCreateLeadResponse)
	if err != nil {
		return "", errors.Wrap(err, "closeComService CreateLead marshallUnmarshall.Unmarshal")
	}

	return closeComCreateLeadResponse.Id, nil
}

func (ccs *closeComService) DeleteLead(ctx context.Context, id string) error {
	return ccs.deleteEntity(ctx, id, CloseComLeadURLPart)
}

func (ccs *closeComService) DeleteOpportunity(ctx context.Context, id string) error {
	return ccs.deleteEntity(ctx, id, CloseComOpportunityURLPart)
}

func (ccs *closeComService) saveEntity(ctx context.Context,
	req interface{}, urlPart, httpMethod string) ([]byte, error) {
	reqJSON, err := ccs.marshallUnmarshall.Marshal(req)
	if err != nil {
		return nil, errors.Wrap(err, "closeComService saveEntity marshallUnmarshall.Marshal")
	}

	httpRequest, err := resource.NewHTTPRequest(CloseComApiBaseURL+urlPart, httpMethod, "",
		reqJSON, ccs.getDefaultHeaders())
	if err != nil {
		return nil, errors.Wrap(err, "closeComService saveEntity resource.NewHTTPRequest")
	}

	resp := ccs.httpClient.Request(ctx, httpRequest)
	if resp.Error != nil {
		return nil, errors.Wrap(resp.Error, "closeComService saveEntity httpClient.Request")
	}

	if resp.StatusCode != http.StatusOK {
		return nil, errors.Wrap(fmt.Errorf("bad status code: %d, error: %s", resp.StatusCode, string(resp.Body)),
			"closeComService saveEntity resp.StatusCode != http.StatusOK")
	}

	return resp.Body, nil
}

func (ccs *closeComService) deleteEntity(ctx context.Context, id, urlPart string) error {
	httpRequest, err := resource.NewHTTPRequest(CloseComApiBaseURL+urlPart+id+"/",
		http.MethodDelete, "", nil, ccs.getDefaultHeaders())
	if err != nil {
		return errors.Wrap(err, "closeComService deleteEntity resource.NewHTTPRequest")
	}

	resp := ccs.httpClient.Request(ctx, httpRequest)
	if resp.Error != nil {
		return errors.Wrap(resp.Error, "closeComService deleteEntity httpClient.Request")
	}

	if resp.StatusCode != http.StatusOK {
		return errors.Wrap(fmt.Errorf("bad status code: %d, error: %s", resp.StatusCode, string(resp.Body)),
			"closeComService deleteEntity resp.StatusCode != http.StatusOK")
	}

	return nil
}
