package daemon

import (
	"context"
	"time"

	api "gitlab.mashplace.com/mashplace/backend/offer/api/rto"
	domain "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto"
	requestservice "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto/resource/request/service"
	request "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto/resource/request/usecase"
	"gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto/resource/response"
	"gitlab.mashplace.com/mashplace/backend/offer/internal/app/service"
	usecaserto "gitlab.mashplace.com/mashplace/backend/offer/internal/app/usecase/rto"

	"github.com/pkg/errors"
)

type UpdateLeadTenantUserService interface {
	UpdateAllLeadTenantUsers(ctx context.Context) error
	UpdateLeadTenantUsers(ctx context.Context) (*response.IDsResponse, error)
	UpdateLeadTenantUsersInListing(ctx context.Context) (*response.IDsResponse, error)
}

type updateLeadTenantUserService struct {
	repository                  domain.Repository
	updateLeadTenantUserUseCase usecaserto.UpdateLeadTenantUserUseCase
	listingService              service.ListingService
}

func NewUpdateLeadTenantUserService(repository domain.Repository,
	updateLeadTenantUserUseCase usecaserto.UpdateLeadTenantUserUseCase,
	listingService service.ListingService) UpdateLeadTenantUserService {
	return &updateLeadTenantUserService{
		repository:                  repository,
		updateLeadTenantUserUseCase: updateLeadTenantUserUseCase,
		listingService:              listingService,
	}
}

func (ultus *updateLeadTenantUserService) UpdateAllLeadTenantUsers(ctx context.Context) error {
	for {
		select {
		case <-ctx.Done():
			return ctx.Err()
		default:
			go func() {
				_, _ = ultus.UpdateLeadTenantUsers(ctx)
			}()
			go func() {
				_, _ = ultus.UpdateLeadTenantUsersInListing(ctx)
			}()
			time.Sleep(time.Hour)
		}
	}
}

func (ultus *updateLeadTenantUserService) UpdateLeadTenantUsers(ctx context.Context) (*response.IDsResponse, error) {
	states := []api.RtoState{
		api.RtoState_RTO_STATE_PENDING_SETUP,
		api.RtoState_RTO_STATE_WAITING_FOR_TENANT_CONFIRMATION,
		api.RtoState_RTO_STATE_WAITING_FOR_STANDING_ORDER_SETUP,
	}
	rtos, err := ultus.repository.FindAllEmptyLeadTenantUsersByStates(ctx, states)
	if err != nil {
		return nil, errors.Wrap(err, "rto daemon updateLeadTenantUserService "+
			"UpdateLeadTenantUsers repository.FindAllEmptyLeadTenantUsers")
	}
	var rtoIDs []string
	for _, rto := range rtos {
		resp, err := ultus.updateLeadTenantUserUseCase.UpdateLeadTenantUser(ctx,
			&request.UpdateLeadTenantUserRequest{
				RtoID:               rto.ID.Hex(),
				LeadTenantEmail:     rto.GetLeadTenantEmail(),
				LeadTenantPhone:     rto.GetLeadTenantPhone(),
				LeadTenantFirstName: rto.GetLeadTenantFirstName(),
				LeadTenantLastName:  rto.GetLeadTenantLastName(),
			})
		if err != nil {
			return nil, errors.Wrap(err, "rto daemon updateLeadTenantUserService "+
				"UpdateLeadTenantUsers updateLeadTenantUserUseCase.UpdateLeadTenantUser")
		}
		if resp.Rto != nil {
			rtoIDs = append(rtoIDs, rto.ID.Hex())
		}
	}

	return &response.IDsResponse{
		IDs: rtoIDs,
	}, nil
}

func (ultus *updateLeadTenantUserService) UpdateLeadTenantUsersInListing(
	ctx context.Context) (*response.IDsResponse, error) {
	states := []api.RtoState{
		api.RtoState_RTO_STATE_WAITING_FOR_TENANT_CONFIRMATION,
		api.RtoState_RTO_STATE_WAITING_FOR_STANDING_ORDER_SETUP,
	}
	rtos, err := ultus.repository.FindAllEmptyLeadTenantUsersByStates(ctx, states)
	if err != nil {
		return nil, errors.Wrap(err, "rto daemon updateLeadTenantUserService "+
			"UpdateLeadTenantUsersInListing repository.FindAllEmptyLeadTenantUsers")
	}
	var rtoIDs []string
	for _, rto := range rtos {
		if rto.GetLeadTenantID() != "" {
			err = ultus.listingService.UpdateTenantID(ctx, &requestservice.ListingUpdateTenantIDRequest{
				PropertyID:   rto.Property,
				LeadTenantID: rto.GetLeadTenantID(),
				LandlordID:   rto.GetLandlordID(),
			})
			if err != nil {
				return nil, errors.Wrap(err, "rto daemon updateLeadTenantUserService "+
					"UpdateLeadTenantUsersInListing listingService.UpdateTenantID")
			}
			rtoIDs = append(rtoIDs, rto.ID.Hex())
		}
	}

	return &response.IDsResponse{
		IDs: rtoIDs,
	}, nil
}
