package rto

import (
	"context"

	messaging "gitlab.mashplace.com/mashplace/backend/messaging/api"
	notification "gitlab.mashplace.com/mashplace/backend/notification/api"
	domain "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto"
	request "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto/resource/request/service/rto"

	"github.com/pkg/errors"
)

const IsRtoTrue = `{"is_rto":"TRUE"}`

type NotificationService interface {
	SendWaitingForStandingOrderSetup(ctx context.Context, req *request.SendWaitingForStandingOrderSetupRequest) error
	SendSetupComplete(ctx context.Context, req *request.SendSetupCompleteRequest) error
	SendActive(ctx context.Context, req *request.SendActiveRequest) error
}

type notificationService struct {
	client     notification.NotificationServiceClient
	repository domain.Repository
}

func NewNotificationService(notificationServiceClient notification.NotificationServiceClient,
	repository domain.Repository) NotificationService {
	return &notificationService{
		client:     notificationServiceClient,
		repository: repository,
	}
}

func (ns *notificationService) SendWaitingForStandingOrderSetup(ctx context.Context,
	req *request.SendWaitingForStandingOrderSetupRequest) error {
	err := req.Validate()
	if err != nil {
		return errors.Wrap(err, "rto notificationService SendWaitingForStandingOrderSetup req.Validate")
	}

	sendNotificationRequest := &notification.SendNotificationRequest{
		Entity: notification.Entity_ENTITY_OFFER_RTO,
		Action: notification.Action_ACTION_RTO_STATE_WAITING_FOR_STANDING_ORDER_SETUP,
		Targets: []*notification.Target{
			{
				User: req.LandlordID,
				Role: notification.Role_ROLE_LANDLORD,
			},
		},
		Embeds: []*notification.EventEmbed{
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_PROPERTY,
				EmbedValue: req.PropertyID,
			},
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_OFFER,
				EmbedValue: req.ID,
			},
		},
	}
	_, err = ns.client.SendNotification(ctx, sendNotificationRequest)
	if err != nil {
		return errors.Wrap(err, "rto notificationService SendWaitingForStandingOrderSetup "+
			"client.SendNotification for landlord")
	}

	if req.TenantID != "" || req.TenantEmail != "" {
		sendNotificationRequest.Targets = []*notification.Target{
			{
				User: req.TenantID,
				Contacts: &notification.Contacts{
					Email: req.TenantEmail,
				},
				Role: notification.Role_ROLE_TENANT,
			},
		}
		_, err = ns.client.SendNotification(ctx, sendNotificationRequest)
		if err != nil {
			return errors.Wrap(err, "rto notificationService SendWaitingForStandingOrderSetup "+
				"client.SendNotification for tenant")
		}
	}

	return nil
}

func (ns *notificationService) SendActive(ctx context.Context, req *request.SendActiveRequest) error {
	err := req.Validate()
	if err != nil {
		return errors.Wrap(err, "rto notificationService SendActive req.Validate")
	}

	_, err = ns.client.SendNotification(ctx, &notification.SendNotificationRequest{
		Entity: notification.Entity_ENTITY_OFFER_RTO,
		Action: req.GetNotificationAction(),
		Targets: []*notification.Target{
			{
				User: req.LandlordID,
				Role: notification.Role_ROLE_LANDLORD,
			},
		},
		Embeds: []*notification.EventEmbed{
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_PROPERTY,
				EmbedValue: req.PropertyID,
			},
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_OFFER,
				EmbedValue: req.OfferID,
			},
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_JSON,
				EmbedValue: IsRtoTrue,
			},
		},
	})
	if err != nil {
		return errors.Wrap(err, "rto notificationService SendActive "+
			"client.SendNotification for landlord")
	}

	return nil
}

func (ns *notificationService) SendSetupComplete(ctx context.Context,
	req *request.SendSetupCompleteRequest) error {
	err := req.Validate()
	if err != nil {
		return errors.Wrap(err, "rto notificationService SendSetupComplete req.Validate")
	}

	_, err = ns.client.SendNotification(ctx, &notification.SendNotificationRequest{
		Entity: notification.Entity_ENTITY_OFFER_RTO,
		Action: notification.Action_ACTION_RTO_SETUP_COMPLETE,
		Targets: []*notification.Target{
			{
				User: req.LandlordID,
				Role: notification.Role_ROLE_LANDLORD,
			},
		},
		Embeds: []*notification.EventEmbed{
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_PROPERTY,
				EmbedValue: req.PropertyID,
			},
		},
	})
	if err != nil {
		return errors.Wrap(err, "rto notificationService SendSetupComplete "+
			"client.SendNotification for landlord")
	}

	return nil
}
