package rto

import (
	"context"
	"fmt"

	domain "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto"
	requestservice "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto/resource/request/service"
	request "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto/resource/request/service/rto"
	"gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto/resource/response"
	"gitlab.mashplace.com/mashplace/backend/offer/internal/app/service"

	"github.com/pkg/errors"
	"go.mongodb.org/mongo-driver/bson/primitive"
)

const (
	CanUserReadRtoNotValidUserErrorText  = "rto securityService CanUserReadRto bad user"
	CanUserReadRtoNotValidStateErrorText = "rto securityService CanUserReadRto bad state"
)

type SecurityService interface {
	CanLandlordUpdateRto(ctx context.Context, req *request.CanLandlordUpdateRtoRequest) (*response.RtoResponse, error)
	CanUserReadRto(ctx context.Context, req *request.CanUserReadRtoRequest) error
	CanTenantUpdateRto(ctx context.Context, req *request.CanTenantUpdateRtoRequest) (*response.RtoResponse, error)
}

type securityService struct {
	repository     domain.Repository
	accountService service.AccountService
}

func NewSecurityService(repository domain.Repository,
	accountService service.AccountService) SecurityService {
	return &securityService{
		repository:     repository,
		accountService: accountService,
	}
}

func (ss *securityService) CanLandlordUpdateRto(ctx context.Context, req *request.CanLandlordUpdateRtoRequest) (
	*response.RtoResponse, error) {
	err := req.Validate()
	if err != nil {
		return nil, errors.Wrap(err, "rto securityService CanLandlordUpdateRto req.Validate")
	}

	if req.NeedCheckVeriff {
		err = ss.accountService.IsVeriffValid(ctx, &requestservice.IsVeriffValidRequest{
			LandlordID: req.LandlordID,
		})
		if err != nil {
			return nil, errors.Wrap(err, "rto securityService CanLandlordUpdateRto accountService.IsVeriffValid")
		}
	}

	rtoID, err := primitive.ObjectIDFromHex(req.ID)
	if err != nil {
		return nil, errors.Wrap(err, "rto securityService CanLandlordUpdateRto primitive.ObjectIDFromHex")
	}

	rto, err := ss.repository.FindByID(ctx, rtoID)
	if err != nil {
		return nil, errors.Wrap(err, "rto securityService CanLandlordUpdateRto repository.FindByID")
	}

	if rto.GetLandlordID() != req.LandlordID {
		return nil, errors.Wrap(fmt.Errorf("403 error rto.GetLandlordID() (%s) != req.LandlordID (%s)",
			rto.GetLandlordID(), req.LandlordID),
			"rto securityService CanLandlordUpdateRto rto.GetLandlordID() != req.LandlordID")
	}

	return &response.RtoResponse{
		Rto: rto,
	}, nil
}

func (ss *securityService) CanUserReadRto(ctx context.Context, req *request.CanUserReadRtoRequest) error {
	err := req.Validate()
	if err != nil {
		return errors.Wrap(err, "rto securityService CanUserReadRto req.Validate")
	}

	if req.Rto.GetLandlordID() == req.UserID || req.Rto.GetLeadTenantID() == req.UserID {
		return nil
	}

	return errors.Wrap(errors.New(CanUserReadRtoNotValidUserErrorText), "rto securityService CanUserReadRto")
}

func (ss *securityService) CanTenantUpdateRto(ctx context.Context, req *request.CanTenantUpdateRtoRequest) (
	*response.RtoResponse, error) {
	err := req.Validate()
	if err != nil {
		return nil, errors.Wrap(err, "rto securityService CanTenantUpdateRto req.Validate")
	}

	rtoID, err := primitive.ObjectIDFromHex(req.ID)
	if err != nil {
		return nil, errors.Wrap(err, "rto securityService CanTenantUpdateRto primitive.ObjectIDFromHex")
	}

	rto, err := ss.repository.FindByID(ctx, rtoID)
	if err != nil {
		return nil, errors.Wrap(err, "rto securityService CanTenantUpdateRto repository.FindByID")
	}

	if rto.GetLeadTenantID() != req.TenantID {
		return nil, errors.Wrap(fmt.Errorf("403 error rto.GetLeadTenantID() (%s) != req.TenantID (%s)",
			rto.GetLeadTenantID(), req.TenantID),
			"rto securityService CanTenantUpdateRto rto.GetLeadTenantID() != req.TenantID")
	}

	return &response.RtoResponse{
		Rto: rto,
	}, nil
}
