package grpc

import (
	"context"

	api "gitlab.mashplace.com/mashplace/backend/offer/api/rto"
	request "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto/resource/request/usecase"
	usecase "gitlab.mashplace.com/mashplace/backend/offer/internal/app/usecase/rto"
)

type RtoServer interface {
	api.RtoServiceServer
}

type rtoServer struct {
	api.UnimplementedRtoServiceServer
	createUseCase                         usecase.CreateUseCase
	createByGuestUseCase                  usecase.CreateByGuestUseCase
	updateLandlordUseCase                 usecase.UpdateLandlordUseCase
	createTenancyUseCase                  usecase.CreateTenancyUseCase
	updateRentPlanUseCase                 usecase.UpdateRentPlanUseCase
	updateEndDateUseCase                  usecase.UpdateEndDateUseCase
	renewalUseCase                        usecase.RenewalUseCase
	updateStateToActiveUseCase            usecase.UpdateStateToActiveUseCase
	updateStateToWaitingForPaymentUseCase usecase.UpdateStateToWaitingForPaymentUseCase
	updateStateToCancelledUseCase         usecase.UpdateStateToCancelledUseCase
	getByLandlordIDUseCase                usecase.GetByLandlordIDUseCase
	getByPropertyIDUseCase                usecase.GetByPropertyIDUseCase
	getByPropertyIDsUseCase               usecase.GetByPropertyIDsUseCase
	answerFromTenantUseCase               usecase.AnswerFromTenantUseCase
	landlordApprovedUseCase               usecase.LandlordApprovedUseCase
	rentPaymentUseCase                    usecase.RentPaymentUseCase
	updateLeadTenantUseCase               usecase.UpdateLeadTenantUseCase
	updateLeadTenantUserUseCase           usecase.UpdateLeadTenantUserUseCase
}

func NewRtoServer(createUseCase usecase.CreateUseCase,
	createByGuestUseCase usecase.CreateByGuestUseCase,
	updateLandlordUseCase usecase.UpdateLandlordUseCase,
	createTenancyUseCase usecase.CreateTenancyUseCase,
	updateRentPlanUseCase usecase.UpdateRentPlanUseCase,
	updateEndDateUseCase usecase.UpdateEndDateUseCase,
	renewalUseCase usecase.RenewalUseCase,
	updateStateToActiveUseCase usecase.UpdateStateToActiveUseCase,
	updateStateToWaitingForPaymentUseCase usecase.UpdateStateToWaitingForPaymentUseCase,
	updateStateToCancelledUseCase usecase.UpdateStateToCancelledUseCase,
	getByLandlordIDUseCase usecase.GetByLandlordIDUseCase,
	getByPropertyIDUseCase usecase.GetByPropertyIDUseCase,
	getByPropertyIDsUseCase usecase.GetByPropertyIDsUseCase,
	answerFromTenantUseCase usecase.AnswerFromTenantUseCase,
	landlordApprovedUseCase usecase.LandlordApprovedUseCase,
	rentPaymentUseCase usecase.RentPaymentUseCase,
	updateLeadTenantUseCase usecase.UpdateLeadTenantUseCase,
	updateLeadTenantUserUseCase usecase.UpdateLeadTenantUserUseCase) RtoServer {
	return &rtoServer{
		createUseCase:                         createUseCase,
		createByGuestUseCase:                  createByGuestUseCase,
		updateLandlordUseCase:                 updateLandlordUseCase,
		createTenancyUseCase:                  createTenancyUseCase,
		updateRentPlanUseCase:                 updateRentPlanUseCase,
		updateEndDateUseCase:                  updateEndDateUseCase,
		renewalUseCase:                        renewalUseCase,
		updateStateToActiveUseCase:            updateStateToActiveUseCase,
		updateStateToWaitingForPaymentUseCase: updateStateToWaitingForPaymentUseCase,
		updateStateToCancelledUseCase:         updateStateToCancelledUseCase,
		getByLandlordIDUseCase:                getByLandlordIDUseCase,
		getByPropertyIDUseCase:                getByPropertyIDUseCase,
		getByPropertyIDsUseCase:               getByPropertyIDsUseCase,
		answerFromTenantUseCase:               answerFromTenantUseCase,
		landlordApprovedUseCase:               landlordApprovedUseCase,
		rentPaymentUseCase:                    rentPaymentUseCase,
		updateLeadTenantUseCase:               updateLeadTenantUseCase,
		updateLeadTenantUserUseCase:           updateLeadTenantUserUseCase,
	}
}

func (rs *rtoServer) Create(ctx context.Context, req *api.CreateRequest) (*api.RtoResponse, error) {
	resp, err := rs.createUseCase.Create(ctx, &request.CreateRequest{
		LandlordID: req.GetLandlordId(),
		RentPlan:   req.GetRentPlan(),
		PropertyID: req.GetPropertyId(),
	})
	if err != nil {
		return nil, err
	}

	return &api.RtoResponse{
		Rto: resp.Rto.CreateApiResponse(),
	}, nil
}

func (rs *rtoServer) CreateByGuest(ctx context.Context, req *api.CreateByGuestRequest) (*api.RtoResponse, error) {
	resp, err := rs.createByGuestUseCase.CreateByGuest(ctx, &request.CreateByGuestRequest{
		RentPlan:   req.GetRentPlan(),
		PropertyID: req.GetPropertyId(),
	})
	if err != nil {
		return nil, err
	}

	return &api.RtoResponse{
		Rto: resp.Rto.CreateApiResponse(),
	}, nil
}

func (rs *rtoServer) UpdateLandlord(ctx context.Context, req *api.UpdateLandlordRequest) (*api.RtoResponse, error) {
	resp, err := rs.updateLandlordUseCase.UpdateLandlord(ctx, &request.UpdateLandlordRequest{
		ID:         req.GetId(),
		LandlordID: req.GetLandlordId(),
	})
	if err != nil {
		return nil, err
	}

	return &api.RtoResponse{
		Rto: resp.Rto.CreateApiResponse(),
	}, nil
}

func (rs *rtoServer) CreateTenancy(ctx context.Context, req *api.CreateTenancyRequest) (*api.RtoResponse, error) {
	resp, err := rs.createTenancyUseCase.CreateTenancy(ctx, &request.CreateTenancyRequest{
		ID:         req.GetId(),
		LandlordID: req.GetLandlordId(),
		PaymentData: &request.PaymentData{
			StartDate:                     req.GetPaymentData().GetStartDate().AsTime(),
			EndDate:                       req.GetPaymentData().GetEndDate().AsTime(),
			RentAmount:                    req.GetPaymentData().GetRentAmount(),
			PaymentDay:                    req.GetPaymentData().GetPaymentDay(),
			IsPaymentDayWasLastDayInMonth: req.GetPaymentData().GetIsPaymentDayWasLastDayInMonth(),
			BreakUpClauseDate:             req.GetPaymentData().GetBreakUpClauseDate().AsTime(),
		},
		LeadTenant: &request.User{
			FirstName: req.GetLeadTenant().GetFirstNameValue(),
			LastName:  req.GetLeadTenant().GetLastNameValue(),
			Email:     req.GetLeadTenant().GetEmail(),
			Phone:     req.GetLeadTenant().GetPhoneValue(),
		},
	})
	if err != nil {
		return nil, err
	}

	return &api.RtoResponse{
		Rto: resp.Rto.CreateApiResponse(),
	}, nil
}

func (rs *rtoServer) UpdateRentPlan(ctx context.Context, req *api.UpdateRentPlanRequest) (*api.RtoResponse, error) {
	resp, err := rs.updateRentPlanUseCase.UpdateRentPlan(ctx, &request.UpdateRentPlanRequest{
		ID:         req.GetId(),
		LandlordID: req.GetLandlordId(),
		RentPlan:   req.GetRentPlan(),
	})
	if err != nil {
		return nil, err
	}

	return &api.RtoResponse{
		Rto: resp.Rto.CreateApiResponse(),
	}, nil
}

func (rs *rtoServer) UpdateEndDate(ctx context.Context, req *api.UpdateEndDateRequest) (*api.RtoResponse, error) {
	resp, err := rs.updateEndDateUseCase.UpdateEndDate(ctx, &request.UpdateEndDateRequest{
		ID:         req.GetId(),
		LandlordID: req.GetLandlordId(),
		EndDate:    req.GetEndDate().AsTime(),
	})
	if err != nil {
		return nil, err
	}

	return &api.RtoResponse{
		Rto: resp.Rto.CreateApiResponse(),
	}, nil
}

func (rs *rtoServer) Renewal(ctx context.Context, req *api.RenewalRequest) (*api.RtoResponse, error) {
	resp, err := rs.renewalUseCase.Renewal(ctx, &request.RenewalRequest{
		ID:         req.GetId(),
		LandlordID: req.GetLandlordId(),
		EndDate:    req.GetEndDate().AsTime(),
	})
	if err != nil {
		return nil, err
	}

	return &api.RtoResponse{
		Rto: resp.Rto.CreateApiResponse(),
	}, nil
}
func (rs *rtoServer) UpdateStateToWaitingForPayment(ctx context.Context,
	req *api.UpdateStateToWaitingForPaymentRequest) (*api.RtoResponse, error) {
	resp, err := rs.updateStateToWaitingForPaymentUseCase.UpdateStateToWaitingForPayment(ctx,
		&request.UpdateStateToWaitingForPaymentRequest{
			ID: req.GetId(),
		})
	if err != nil {
		return nil, err
	}

	return &api.RtoResponse{
		Rto: resp.Rto.CreateApiResponse(),
	}, nil
}

func (rs *rtoServer) UpdateStateToActive(ctx context.Context,
	req *api.UpdateStateToActiveRequest) (*api.RtoResponse, error) {
	resp, err := rs.updateStateToActiveUseCase.UpdateStateToActive(ctx,
		&request.UpdateStateToActiveRequest{
			ID: req.GetId(),
		})
	if err != nil {
		return nil, err
	}

	return &api.RtoResponse{
		Rto: resp.Rto.CreateApiResponse(),
	}, nil
}

func (rs *rtoServer) Cancel(ctx context.Context, req *api.CancelRequest) (*api.RtoResponse, error) {
	resp, err := rs.updateStateToCancelledUseCase.UpdateStateToCancelled(ctx, &request.UpdateStateToCancelledRequest{
		ID: req.GetId(),
	})
	if err != nil {
		return nil, err
	}

	return &api.RtoResponse{
		Rto: resp.Rto.CreateApiResponse(),
	}, nil
}

func (rs *rtoServer) GetByLandlordID(ctx context.Context, req *api.GetByLandlordIDRequest) (
	*api.RtosResponse, error) {
	resp, err := rs.getByLandlordIDUseCase.GetByLandlordID(ctx, &request.GetByLandlordIDRequest{
		LandlordID: req.GetUserId(),
		States:     req.GetStates(),
	})
	if err != nil {
		return nil, err
	}

	var rtos []*api.Rto
	for _, rto := range resp.Rtos {
		rtos = append(rtos, rto.CreateApiResponse())
	}

	return &api.RtosResponse{
		Rtos: rtos,
	}, nil
}

func (rs *rtoServer) GetByPropertyID(ctx context.Context, req *api.GetByPropertyIDRequest) (*api.RtoResponse, error) {
	resp, err := rs.getByPropertyIDUseCase.GetByPropertyID(ctx, &request.GetByPropertyIDRequest{
		PropertyID: req.GetPropertyId(),
		UserID:     req.GetUserId(),
	})
	if err != nil {
		return nil, err
	}
	if resp == nil || resp.Rto == nil {
		return &api.RtoResponse{
			Rto: nil,
		}, nil
	}

	return &api.RtoResponse{
		Rto: resp.Rto.CreateApiResponse(),
	}, nil
}

func (rs *rtoServer) GetByPropertyIDs(ctx context.Context, req *api.GetByPropertyIDsRequest) (
	*api.RtosResponse, error) {
	resp, err := rs.getByPropertyIDsUseCase.GetByPropertyIDs(ctx, &request.GetByPropertyIDsRequest{
		PropertyIDs: req.GetPropertyIds(),
		UserID:      req.GetUserId(),
	})
	if err != nil {
		return nil, err
	}
	var rtos []*api.Rto
	for _, rto := range resp.Rtos {
		rtos = append(rtos, rto.CreateApiResponse())
	}

	return &api.RtosResponse{
		Rtos: rtos,
	}, nil
}

func (rs *rtoServer) AnswerFromTenant(ctx context.Context, req *api.AnswerFromTenantRequest) (*api.RtoResponse, error) {
	resp, err := rs.answerFromTenantUseCase.AnswerFromTenant(ctx, &request.AnswerFromTenantRequest{
		ID:       req.GetId(),
		TenantID: req.GetTenantId(),
		Accept:   req.GetAcceptValue(),
	})
	if err != nil {
		return nil, err
	}

	return &api.RtoResponse{
		Rto: resp.Rto.CreateApiResponse(),
	}, nil
}

func (rs *rtoServer) LandlordApproved(ctx context.Context, req *api.LandlordApprovedRequest) (
	*api.IdsResponse, error) {
	resp, err := rs.landlordApprovedUseCase.LandlordApproved(ctx, &request.LandlordApprovedRequest{
		LandlordID: req.GetLandlordId(),
	})
	if err != nil {
		return nil, err
	}

	return &api.IdsResponse{
		Ids: resp.IDs,
	}, nil
}

func (rs *rtoServer) RentPayment(ctx context.Context, req *api.RentPaymentRequest) (*api.RtoResponse, error) {
	resp, err := rs.rentPaymentUseCase.RentPayment(ctx, &request.RentPaymentRequest{
		ListingVersion: req.GetVersion(),
	})
	if err != nil {
		return nil, err
	}

	if resp.Rto != nil {
		return &api.RtoResponse{
			Rto: resp.Rto.CreateApiResponse(),
		}, nil
	}

	return &api.RtoResponse{}, nil
}

func (rs *rtoServer) UpdateLeadTenant(ctx context.Context, req *api.UpdateLeadTenantRequest) (
	*api.UpdateLeadTenantResponse, error) {
	resp, err := rs.updateLeadTenantUseCase.UpdateLeadTenant(ctx, &request.UpdateLeadTenantRequest{
		LeadTenantID:    req.GetLeadTenantId(),
		LeadTenantEmail: req.GetLeadTenantEmail(),
	})
	if err != nil {
		return nil, err
	}

	return &api.UpdateLeadTenantResponse{
		Ids:         resp.IDs,
		PropertyIds: resp.PropertyIDs,
	}, nil
}

func (rs *rtoServer) UpdateLeadTenantUser(ctx context.Context,
	req *api.UpdateLeadTenantUserRequest) (*api.RtoResponse, error) {
	resp, err := rs.updateLeadTenantUserUseCase.UpdateLeadTenantUser(ctx, &request.UpdateLeadTenantUserRequest{
		RtoID:               req.GetRtoId(),
		LeadTenantEmail:     req.GetLeadTenantEmail(),
		LeadTenantPhone:     req.GetLeadTenantPhone(),
		LeadTenantFirstName: req.GetLeadTenantFirstName(),
		LeadTenantLastName:  req.GetLeadTenantLastName(),
	})
	if err != nil {
		return nil, err
	}

	return &api.RtoResponse{
		Rto: resp.Rto.CreateApiResponse(),
	}, nil
}
