package rto

import (
	"context"
	"fmt"
	"time"

	"gitlab.mashplace.com/mashplace/backend/common/generic/log"
	"gitlab.mashplace.com/mashplace/backend/common/lib/mongolocal"
	extras "gitlab.mashplace.com/mashplace/backend/extras/api"
	api "gitlab.mashplace.com/mashplace/backend/offer/api/rto"
	domain "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto"
	requestservice "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto/resource/request/service"
	requestservrto "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto/resource/request/service/rto"
	request "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto/resource/request/usecase"
	"gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto/resource/response"
	appservice "gitlab.mashplace.com/mashplace/backend/offer/internal/app/service"
	service "gitlab.mashplace.com/mashplace/backend/offer/internal/app/service/rto"

	"github.com/pkg/errors"
)

const AnswerFromTenantNotValidStateErrorText = "bad rto state for AnswerFromTenant"

type AnswerFromTenantUseCase interface {
	AnswerFromTenant(ctx context.Context, req *request.AnswerFromTenantRequest) (*response.RtoResponse, error)
}

type answerFromTenantUseCase struct {
	repository          domain.Repository
	historyRepository   domain.HistoryRepository
	securityService     service.SecurityService
	invoiceService      appservice.InvoiceService
	extrasService       appservice.ExtrasService
	notificationService service.NotificationService
	closeComService     service.CloseComService
}

func NewAnswerFromTenantUseCase(repository domain.Repository,
	historyRepository domain.HistoryRepository,
	securityService service.SecurityService,
	invoiceService appservice.InvoiceService,
	extrasService appservice.ExtrasService,
	notificationService service.NotificationService,
	closeComService service.CloseComService) AnswerFromTenantUseCase {
	return &answerFromTenantUseCase{
		repository:          repository,
		historyRepository:   historyRepository,
		securityService:     securityService,
		invoiceService:      invoiceService,
		extrasService:       extrasService,
		notificationService: notificationService,
		closeComService:     closeComService,
	}
}

func (aftuc *answerFromTenantUseCase) AnswerFromTenant(ctx context.Context, req *request.AnswerFromTenantRequest) (
	*response.RtoResponse, error) {
	err := req.Validate()
	if err != nil {
		return nil, errors.Wrap(err, "rto answerFromTenantUseCase req.Validate")
	}

	resp, err := aftuc.securityService.CanTenantUpdateRto(ctx, &requestservrto.CanTenantUpdateRtoRequest{
		ID:       req.ID,
		TenantID: req.TenantID,
	})
	if err != nil {
		return nil, errors.Wrap(err, "rto answerFromTenantUseCase securityService.CanTenantUpdateRto")
	}

	rto := resp.Rto
	if rto.State != api.RtoState_RTO_STATE_WAITING_FOR_TENANT_CONFIRMATION {
		return nil, errors.Wrap(fmt.Errorf("rto state must be in state %s, not in state %s",
			api.RtoState_RTO_STATE_WAITING_FOR_TENANT_CONFIRMATION, rto.State),
			AnswerFromTenantNotValidStateErrorText)
	}

	if rto.GetLeadTenantID() == "" {
		rto.LeadTenant.User, rto.Tenants[0].User = req.TenantID, req.TenantID
	}

	closeComSaveOpportunityRequest := requestservrto.CreateCloseComSaveOpportunityRequest(rto)
	if req.Accept {
		rto.State = api.RtoState_RTO_STATE_WAITING_FOR_STANDING_ORDER_SETUP
		rto.PaidStatuses.StandingOrder = true

		servicePriceSnapshot, err := aftuc.extrasService.GetServicePriceSnapshot(ctx,
			&requestservice.GetServicePriceSnapshotRequest{
				Code:  extras.ProductTenantStandingOrder,
				Price: rto.Conditions.Price,
			})
		if err != nil {
			return nil, errors.Wrap(err, "rto answerFromTenantUseCase extrasService.GetServicePriceSnapshot")
		}

		invoiceID, err := aftuc.invoiceService.CreateStandingOrderInvoice(ctx,
			&requestservice.CreateStandingOrderInvoiceRequest{
				Rto:             rto,
				ServiceSnapshot: servicePriceSnapshot,
			})
		if err != nil {
			return nil, errors.Wrap(err, "rto answerFromTenantUseCase invoiceService.CreateStandingOrderInvoice")
		}

		rto.InvoiceID = invoiceID

		err = aftuc.notificationService.SendSetupComplete(ctx, &requestservrto.SendSetupCompleteRequest{
			LandlordID: rto.Landlord.User,
			PropertyID: rto.Property,
		})
		if err != nil {
			return nil, errors.Wrap(err, "rto answerFromTenantUseCase notificationService.SendSetupComplete")
		}
	} else {
		rto.State = api.RtoState_RTO_STATE_CANCELLED
	}

	now := time.Now()
	rto.UpdatedAt = now
	err = mongolocal.Transaction(ctx, func() error {
		err = aftuc.repository.MustUpdate(ctx, rto)
		if err != nil {
			return errors.Wrap(err, "rto answerFromTenantUseCase repository.MustUpdate")
		}

		err = aftuc.historyRepository.Create(ctx, domain.CreateHistory(domain.HistoryActionUpdate, now, rto))
		if err != nil {
			return errors.Wrap(err, "rto answerFromTenantUseCase historyRepository.Create")
		}

		return nil
	})
	if err != nil {
		return nil, errors.Wrap(err, "rto answerFromTenantUseCase mongolocal.Transaction")
	}

	go func() {
		ctx := context.Background()
		_, err = aftuc.closeComService.SaveOpportunity(ctx, closeComSaveOpportunityRequest)
		if err != nil {
			log.WithContext(ctx).WithError(err).Error("rto answerFromTenantUseCase closeComService.SaveOpportunity")
		}
	}()

	return &response.RtoResponse{
		Rto: rto,
	}, nil
}
