package rto_test

import (
	"strings"
	"testing"

	extras "gitlab.mashplace.com/mashplace/backend/extras/api"
	api "gitlab.mashplace.com/mashplace/backend/offer/api/rto"
	domain "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto"
	requestservice "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto/resource/request/service"
	requestservrto "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto/resource/request/service/rto"
	request "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto/resource/request/usecase"
	mockappservice "gitlab.mashplace.com/mashplace/backend/offer/internal/app/service/mock"
	service "gitlab.mashplace.com/mashplace/backend/offer/internal/app/service/rto"
	mockservicerto "gitlab.mashplace.com/mashplace/backend/offer/internal/app/service/rto/mock"
	"gitlab.mashplace.com/mashplace/backend/offer/internal/app/usecase/rto"
	"gitlab.mashplace.com/mashplace/backend/offer/test"

	"github.com/golang/mock/gomock"
	. "github.com/smartystreets/goconvey/convey"
	"go.mongodb.org/mongo-driver/bson/primitive"
)

func TestAnswerFromTenantUseCase(t *testing.T) {
	if test.CanRunTests() {
		Convey("AnswerFromTenantUseCase", t, func() {
			ctrl := gomock.NewController(t)
			defer ctrl.Finish()

			securityService := service.NewSecurityService(repository, nil)
			invoiceService := mockappservice.NewMockInvoiceService(ctrl)
			extrasService := mockappservice.NewMockExtrasService(ctrl)
			rtoNotificationService := mockservicerto.NewMockNotificationService(ctrl)
			useCase := rto.NewAnswerFromTenantUseCase(repository, historyRepository,
				securityService, invoiceService, extrasService, rtoNotificationService, closeComService)

			Convey("AnswerFromTenant error "+rto.AnswerFromTenantNotValidStateErrorText, func() {
				id := primitive.NewObjectID()
				tenantUserID, landlordID := "barligikimmagantuisiemeschertishlyuhi", primitive.NewObjectID().Hex()
				rtoModel := &domain.Rto{
					ID:    id,
					State: api.RtoState_RTO_STATE_PENDING_SETUP,
					LeadTenant: &domain.Tenant{
						User: tenantUserID,
					},
					Conditions: &domain.Conditions{
						Price: test.Chmolgani,
					},
					Landlord: &domain.Landlord{
						User: landlordID,
					},
				}
				err := repository.Create(ctx, rtoModel)
				So(err, ShouldBeNil)
				rtoID := id.Hex()

				saveOpportunityResponse, err := closeComService.SaveOpportunity(ctx,
					&requestservrto.CloseComSaveOpportunityRequest{
						RtoPreviousStateDate: rtoModel.UpdatedAt,
						Rto:                  rtoModel,
						RtoPreviousState:     rtoModel.State,
					})
				So(err, ShouldBeNil)
				So(saveOpportunityResponse.OpportunityID, ShouldNotBeEmpty)
				So(saveOpportunityResponse.LeadID, ShouldNotBeEmpty)

				rtoModel.OpportunityID = saveOpportunityResponse.OpportunityID
				rtoModel.Landlord.LeadID = saveOpportunityResponse.LeadID
				So(repository.MustUpdate(ctx, rtoModel), ShouldBeNil)

				req := &request.AnswerFromTenantRequest{
					ID:       rtoID,
					TenantID: tenantUserID,
				}
				res, err := useCase.AnswerFromTenant(ctx, req)
				So(res, ShouldBeNil)
				So(err, ShouldBeError)
				So(strings.Contains(err.Error(), rto.AnswerFromTenantNotValidStateErrorText), ShouldBeTrue)

				rtoModel.State = api.RtoState_RTO_STATE_WAITING_FOR_TENANT_CONFIRMATION
				err = repository.MustUpdate(ctx, rtoModel)
				So(err, ShouldBeNil)

				Convey("AnswerFromTenant success decline", func() {
					res, err := useCase.AnswerFromTenant(ctx, req)
					So(err, ShouldBeNil)
					So(res, ShouldNotBeNil)
					So(res.Rto.State, ShouldEqual, api.RtoState_RTO_STATE_CANCELLED)

					rtoModel.State = api.RtoState_RTO_STATE_WAITING_FOR_TENANT_CONFIRMATION
					err = repository.MustUpdate(ctx, rtoModel)
					So(err, ShouldBeNil)

					Convey("AnswerFromTenant success accept", func() {
						extrasService.EXPECT().GetServicePriceSnapshot(ctx,
							&requestservice.GetServicePriceSnapshotRequest{
								Code:  extras.ProductTenantStandingOrder,
								Price: test.Chmolgani,
							}).Return(&extras.ServiceSnapshot{}, nil)
						invoiceService.EXPECT().CreateStandingOrderInvoice(ctx, gomock.Any()).Return("", nil)
						rtoNotificationService.EXPECT().SendSetupComplete(ctx, gomock.Any()).Return(nil)
						req.Accept = true
						res, err := useCase.AnswerFromTenant(ctx, req)
						So(err, ShouldBeNil)
						So(res, ShouldNotBeNil)
						So(res.Rto.State, ShouldEqual, api.RtoState_RTO_STATE_WAITING_FOR_STANDING_ORDER_SETUP)

						err = closeComService.DeleteOpportunity(ctx, saveOpportunityResponse.OpportunityID)
						So(err, ShouldBeNil)

						err = closeComService.DeleteLead(ctx, saveOpportunityResponse.LeadID)
						So(err, ShouldBeNil)

						test.DeleteRto(ctx, t, repository, historyRepository, rtoModel.ID, 2)
					})
				})
			})
		})
	}
}
