package rto

import (
	"context"
	"time"

	"gitlab.mashplace.com/mashplace/backend/common/generic/log"

	"gitlab.mashplace.com/mashplace/backend/common/lib/mongolocal"
	domain "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto"
	servicerequest "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto/resource/request/service"
	requestservrto "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto/resource/request/service/rto"
	request "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto/resource/request/usecase"
	"gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto/resource/response"
	appservice "gitlab.mashplace.com/mashplace/backend/offer/internal/app/service"
	service "gitlab.mashplace.com/mashplace/backend/offer/internal/app/service/rto"

	"github.com/pkg/errors"
	"go.mongodb.org/mongo-driver/bson/primitive"
	"go.mongodb.org/mongo-driver/mongo"
)

const WithThisPropertyIDRtoAlreadyExistsErrorText = "with this property id rto already exists"

type CreateUseCase interface {
	Create(ctx context.Context, req *request.CreateRequest) (*response.RtoResponse, error)
}

type createUseCase struct {
	repository        domain.Repository
	historyRepository domain.HistoryRepository
	accountService    appservice.AccountService
	listingService    appservice.ListingService
	closeComService   service.CloseComService
}

func NewCreateUseCase(repository domain.Repository,
	historyRepository domain.HistoryRepository,
	accountService appservice.AccountService,
	listingService appservice.ListingService,
	closeComService service.CloseComService) CreateUseCase {
	return &createUseCase{
		repository:        repository,
		historyRepository: historyRepository,
		accountService:    accountService,
		listingService:    listingService,
		closeComService:   closeComService,
	}
}

func (cuc *createUseCase) Create(ctx context.Context, req *request.CreateRequest) (
	*response.RtoResponse, error) {
	err := req.Validate()
	if err != nil {
		return nil, errors.Wrap(err, "rto createUseCase req.Validate")
	}

	listingVersion, err := cuc.listingService.GetVersion(ctx, &servicerequest.ListingGetVersionRequest{
		PropertyID: req.PropertyID,
		LandlordID: req.LandlordID,
	})
	if err != nil {
		return nil, errors.Wrap(err, "rto createUseCase listingService.GetVersionByPropertyID")
	}

	rto, err := cuc.repository.FindByPropertyID(ctx, req.PropertyID)
	if err != nil && !errors.Is(err, mongo.ErrNoDocuments) {
		return nil, errors.Wrap(err, "rto createUseCase repository.FindByPropertyID")
	}
	if rto != nil {
		return nil, errors.Wrap(errors.New(WithThisPropertyIDRtoAlreadyExistsErrorText),
			"rto createUseCase rto != nil")
	}

	landlord, err := cuc.accountService.GetAccountByID(ctx, req.LandlordID)
	if err != nil {
		return nil, errors.Wrap(err, "rto createUseCase accountService.GetAccountByID")
	}

	now := time.Now()
	rto = domain.CreateRto(now, req.RentPlan, req.LandlordID, landlord, listingVersion)

	err = mongolocal.Transaction(ctx, func() error {
		err = cuc.repository.Create(ctx, rto)
		if err != nil {
			return errors.Wrap(err, "rto createUseCase repository.Create")
		}

		err = cuc.historyRepository.Create(ctx, domain.CreateHistory(domain.HistoryActionCreate, now, rto))
		if err != nil {
			return errors.Wrap(err, "rto createUseCase historyRepository.Create")
		}

		return nil
	})
	if err != nil {
		return nil, errors.Wrap(err, "rto createUseCase mongolocal.Transaction")
	}

	go func() {
		ctx := context.Background()
		saveOpportunityResponse, err := cuc.closeComService.SaveOpportunity(ctx,
			&requestservrto.CloseComSaveOpportunityRequest{Rto: rto})
		if err != nil {
			log.WithError(err).Error("rto createUseCase closeComService.SaveOpportunity")
			return
		}

		err = cuc.saveOpportunityID(ctx, rto.ID.Hex(),
			saveOpportunityResponse.OpportunityID, saveOpportunityResponse.LeadID)
		if err != nil {
			log.WithError(err).Error("rto createUseCase saveOpportunityID")
			return
		}
	}()

	return &response.RtoResponse{
		Rto: rto,
	}, nil
}

func (cuc *createUseCase) saveOpportunityID(ctx context.Context, offID string, oppID string, leadID string) error {
	return mongolocal.Transaction(ctx, func() error {
		rtoID, err := primitive.ObjectIDFromHex(offID)
		if err != nil {
			return errors.Wrap(err, "rto createUseCase saveOpportunityID primitive.ObjectIDFromHex")
		}

		rto, err := cuc.repository.FindByID(ctx, rtoID)
		if err != nil {
			return errors.Wrap(err, "rto createUseCase saveOpportunityID repository.FindByID")
		}

		rto.OpportunityID = oppID
		rto.Landlord.LeadID = leadID
		err = cuc.repository.MustUpdate(ctx, rto)
		if err != nil {
			return errors.Wrap(err, "rto createUseCase saveOpportunityID repository.MustUpdate")
		}

		err = cuc.historyRepository.Create(ctx, domain.CreateHistory(domain.HistoryActionCreate, time.Now(), rto))
		if err != nil {
			return errors.Wrap(err, "rto createUseCase saveOpportunityID historyRepository.Create")
		}

		return nil
	})
}
