package rto

import (
	"context"
	"strings"
	"time"

	"gitlab.mashplace.com/mashplace/backend/common/lib/mongolocal"
	domain "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto"
	servicerequest "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto/resource/request/service"
	request "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto/resource/request/usecase"
	"gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto/resource/response"
	appservice "gitlab.mashplace.com/mashplace/backend/offer/internal/app/service"

	"github.com/pkg/errors"
	"go.mongodb.org/mongo-driver/mongo"
)

type CreateByGuestUseCase interface {
	CreateByGuest(ctx context.Context, req *request.CreateByGuestRequest) (*response.RtoResponse, error)
}

type createByGuestUseCase struct {
	repository        domain.Repository
	historyRepository domain.HistoryRepository
	listingService    appservice.ListingService
}

func NewCreateByGuestUseCase(repository domain.Repository,
	historyRepository domain.HistoryRepository,
	listingService appservice.ListingService) CreateByGuestUseCase {
	return &createByGuestUseCase{
		repository:        repository,
		historyRepository: historyRepository,
		listingService:    listingService,
	}
}

func (cbguc *createByGuestUseCase) CreateByGuest(ctx context.Context, req *request.CreateByGuestRequest) (
	*response.RtoResponse, error) {
	err := req.Validate()
	if err != nil {
		return nil, errors.Wrap(err, "rto createByGuestUseCase req.Validate")
	}

	listingVersion, err := cbguc.listingService.GetVersion(ctx, &servicerequest.ListingGetVersionRequest{
		PropertyID: req.PropertyID,
	})
	if err != nil {
		return nil, errors.Wrap(err, "rto createByGuestUseCase listingService.GetVersionByPropertyID")
	}

	rto, err := cbguc.repository.FindByPropertyID(ctx, req.PropertyID)
	if err != nil && !strings.Contains(err.Error(), mongo.ErrNoDocuments.Error()) {
		return nil, errors.Wrap(err, "rto createByGuestUseCase repository.FindByPropertyID")
	}
	if rto != nil {
		return nil, errors.Wrap(errors.New(WithThisPropertyIDRtoAlreadyExistsErrorText),
			"rto createByGuestUseCase rto != nil")
	}

	now := time.Now()
	rto = domain.CreateRto(now, req.RentPlan, "", nil, listingVersion)
	err = mongolocal.Transaction(ctx, func() error {
		err = cbguc.repository.Create(ctx, rto)
		if err != nil {
			return errors.Wrap(err, "rto createByGuestUseCase repository.CreateByGuest")
		}

		err = cbguc.historyRepository.Create(ctx, domain.CreateHistory(domain.HistoryActionCreate, now, rto))
		if err != nil {
			return errors.Wrap(err, "rto createByGuestUseCase historyRepository.CreateByGuest")
		}

		return nil
	})
	if err != nil {
		return nil, errors.Wrap(err, "rto createByGuestUseCase mongolocal.Transaction")
	}

	return &response.RtoResponse{
		Rto: rto,
	}, nil
}
