package rto

import (
	"context"
	"strings"
	"time"

	"github.com/pkg/errors"
	"gitlab.mashplace.com/mashplace/backend/common/lib/mongolocal"
	domain "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto"
	requestservrto "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto/resource/request/service/rto"
	request "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto/resource/request/usecase"
	"gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto/resource/response"
	appservice "gitlab.mashplace.com/mashplace/backend/offer/internal/app/service"
	service "gitlab.mashplace.com/mashplace/backend/offer/internal/app/service/rto"
	"go.mongodb.org/mongo-driver/mongo"
)

type CreateTenancyUseCase interface {
	CreateTenancy(ctx context.Context, req *request.CreateTenancyRequest) (*response.RtoResponse, error)
}

type createTenancyUseCase struct {
	repository        domain.Repository
	historyRepository domain.HistoryRepository
	securityService   service.SecurityService
	accountService    appservice.AccountService
}

func NewCreateTenancyUseCase(repository domain.Repository,
	historyRepository domain.HistoryRepository,
	securityService service.SecurityService,
	accountService appservice.AccountService) CreateTenancyUseCase {
	return &createTenancyUseCase{
		repository:        repository,
		historyRepository: historyRepository,
		securityService:   securityService,
		accountService:    accountService,
	}
}

func (ctuc *createTenancyUseCase) CreateTenancy(ctx context.Context, req *request.CreateTenancyRequest) (
	*response.RtoResponse, error) {
	err := req.Validate()
	if err != nil {
		return nil, errors.Wrap(err, "rto createTenancyUseCase req.Validate")
	}

	var rto *domain.Rto
	err = mongolocal.Transaction(ctx, func() error {
		canLandlordUpdateRtoResponse, err := ctuc.securityService.CanLandlordUpdateRto(ctx,
			&requestservrto.CanLandlordUpdateRtoRequest{
				ID:         req.ID,
				LandlordID: req.LandlordID,
			})
		if err != nil {
			return errors.Wrap(err, "rto createTenancyUseCase securityService.CanLandlordUpdateRto")
		}

		rto = canLandlordUpdateRtoResponse.Rto

		tenant, err := ctuc.accountService.GetAccountByEmail(ctx, req.LeadTenant.Email)
		if err != nil && !strings.Contains(err.Error(), mongo.ErrNoDocuments.Error()) {
			return errors.Wrap(err, "rto createTenancyUseCase accountService.GetAccountByEmail")
		}

		rto.LeadTenant = req.GetLeadTenant()
		if tenant != nil {
			rto.LeadTenant.User = tenant.GetIdValue()
		}

		rto.Tenants = rto.LeadTenant.GetTenants()
		rto.Conditions = req.GetConditions()

		now := time.Now()
		rto.UpdatedAt = now

		err = ctuc.repository.MustUpdate(ctx, rto)
		if err != nil {
			return errors.Wrap(err, "rto createTenancyUseCase repository.MustUpdate")
		}

		err = ctuc.historyRepository.Create(ctx, domain.CreateHistory(domain.HistoryActionUpdate, now, rto))
		if err != nil {
			return errors.Wrap(err, "rto createTenancyUseCase historyRepository.Create")
		}

		return nil
	})
	if err != nil {
		return nil, errors.Wrap(err, "rto createTenancyUseCase mongolocal.Transaction")
	}

	return &response.RtoResponse{
		Rto: rto,
	}, nil
}
