package rto_test

import (
	"sort"
	"testing"

	api "gitlab.mashplace.com/mashplace/backend/offer/api/rto"
	domain "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto"
	request "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto/resource/request/usecase"
	service "gitlab.mashplace.com/mashplace/backend/offer/internal/app/service/rto"
	"gitlab.mashplace.com/mashplace/backend/offer/internal/app/usecase/rto"
	"gitlab.mashplace.com/mashplace/backend/offer/test"

	. "github.com/smartystreets/goconvey/convey"
	"go.mongodb.org/mongo-driver/bson/primitive"
)

func TestGetByPropertyIDsUseCase(t *testing.T) {
	if test.CanRunTests() {
		Convey("GetByPropertyIDsUseCase", t, func() {
			securityService := service.NewSecurityService(repository, nil)
			useCase := rto.NewGetByPropertyIDsUseCase(repository, securityService)

			Convey("GetByPropertyIDsUseCase empty", func() {
				propertyID, landlordID, leadTenantID := "chertivsektomnetuisiemes",
					"chmorikivsektomnetuisiemes", "barligikimmagantuisiemeschertichmoshiebannieolsin"
				req := &request.GetByPropertyIDsRequest{
					PropertyIDs: []string{propertyID},
					UserID:      landlordID,
				}
				resp, err := useCase.GetByPropertyIDs(ctx, req)
				So(err, ShouldBeNil)
				So(len(resp.Rtos), ShouldBeZeroValue)

				rtoIDObjectPendingSetup := createRtoModel(t, api.RtoState_RTO_STATE_PENDING_SETUP,
					propertyID, landlordID, leadTenantID)
				rtoIDObjectActive := createRtoModel(t, api.RtoState_RTO_STATE_ACTIVE,
					propertyID, landlordID, leadTenantID)
				rtoIDObjectCancelled := createRtoModel(t, api.RtoState_RTO_STATE_CANCELLED,
					propertyID, landlordID, leadTenantID)
				rtoIDObjects := []primitive.ObjectID{rtoIDObjectPendingSetup, rtoIDObjectActive, rtoIDObjectCancelled}

				rtoIDPendingSetup := rtoIDObjectPendingSetup.Hex()
				rtoIDActive := rtoIDObjectActive.Hex()
				rtoIDCancelled := rtoIDObjectCancelled.Hex()

				Convey("GetByPropertyIDsUseCase success", func() {
					resp, err = useCase.GetByPropertyIDs(ctx, req)
					So(err, ShouldBeNil)

					var respRtoIDs []string
					for _, rtoModel := range resp.Rtos {
						respRtoIDs = append(respRtoIDs, rtoModel.ID.Hex())
					}
					sort.Strings(respRtoIDs)

					var rtoIDs []string
					rtoIDs = append(rtoIDs, rtoIDPendingSetup, rtoIDActive, rtoIDCancelled)
					sort.Strings(rtoIDs)

					So(rtoIDs, ShouldResemble, respRtoIDs)
				})

				for _, rtoID := range rtoIDObjects {
					_, err = repository.DeleteByID(ctx, rtoID)
					So(err, ShouldBeNil)
				}
			})
		})
	}
}

func createRtoModel(t *testing.T, state api.RtoState, propertyID, landlordID, leadTenantID string) primitive.ObjectID {
	id := primitive.NewObjectID()
	rtoModel := &domain.Rto{
		ID:       id,
		Property: propertyID,
		Landlord: &domain.Landlord{
			User: landlordID,
		},
		State: state,
		LeadTenant: &domain.Tenant{
			User: leadTenantID,
		},
	}
	err := repository.Create(ctx, rtoModel)
	if err != nil {
		t.Error(err)
	}

	return id
}
