package rto

import (
	"context"

	api "gitlab.mashplace.com/mashplace/backend/offer/api/rto"
	domain "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto"
	servicerequest "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto/resource/request/service"
	request "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto/resource/request/usecase"
	"gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto/resource/response"
	appservice "gitlab.mashplace.com/mashplace/backend/offer/internal/app/service"

	"github.com/pkg/errors"
)

type LandlordApprovedUseCase interface {
	LandlordApproved(ctx context.Context, req *request.LandlordApprovedRequest) (*response.IDsResponse, error)
}

type landlordApprovedUseCase struct {
	repository                                       domain.Repository
	listingService                                   appservice.ListingService
	updateStateToWaitingForTenantConfirmationUseCase UpdateStateToWaitingForTenantConfirmationUseCase
}

func NewLandlordApprovedUseCase(repository domain.Repository,
	listingService appservice.ListingService,
	updateStateToWaitingForTenantConfirmationUseCase UpdateStateToWaitingForTenantConfirmationUseCase,
) LandlordApprovedUseCase {
	return &landlordApprovedUseCase{
		repository:     repository,
		listingService: listingService,
		updateStateToWaitingForTenantConfirmationUseCase: updateStateToWaitingForTenantConfirmationUseCase,
	}
}

func (lauc *landlordApprovedUseCase) LandlordApproved(ctx context.Context, req *request.LandlordApprovedRequest) (
	*response.IDsResponse, error) {
	err := req.Validate()
	if err != nil {
		return nil, errors.Wrap(err, "rto landlordApprovedUseCase req.Validate")
	}

	rtos, err := lauc.repository.FindAllByLandlordID(ctx, &domain.AllByLandlordIDFilter{
		LandlordID: req.LandlordID,
		States: []api.RtoState{
			api.RtoState_RTO_STATE_PENDING_SETUP,
		},
	})
	if err != nil {
		return nil, errors.Wrap(err, "rto landlordApprovedUseCase repository.FindAllByLandlordID")
	}

	var ids []string
	for _, rto := range rtos {
		listingVersion, err := lauc.listingService.GetVersion(ctx, &servicerequest.ListingGetVersionRequest{
			PropertyID: rto.Property,
			LandlordID: rto.GetLandlordID(),
		})
		if err != nil {
			return nil, errors.Wrap(err, "rto landlordApprovedUseCase listingService.GetVersionByPropertyID")
		}

		if lauc.listingService.VersionHasNotEmptyBankDetails(ctx, listingVersion) {
			_, err = lauc.updateStateToWaitingForTenantConfirmationUseCase.UpdateStateToWaitingForTenantConfirmation(
				ctx, &request.UpdateStateToWaitingForTenantConfirmationRequest{Rto: rto})
			if err != nil {
				return nil, errors.Wrap(err,
					"rto landlordApprovedUseCase updateStateToWaitingForTenantConfirmationUseCase")
			}
			ids = append(ids, rto.ID.Hex())
		}
	}

	return &response.IDsResponse{
		IDs: ids,
	}, nil
}
