package rto

import (
	"context"
	"fmt"
	"time"

	"gitlab.mashplace.com/mashplace/backend/common/lib/mongolocal"
	api "gitlab.mashplace.com/mashplace/backend/offer/api/rto"
	domain "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto"
	requestservrto "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto/resource/request/service/rto"
	request "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto/resource/request/usecase"
	"gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto/resource/response"
	service "gitlab.mashplace.com/mashplace/backend/offer/internal/app/service/rto"
	"gitlab.mashplace.com/mashplace/backend/offer/internal/pkg/util"

	"github.com/pkg/errors"
)

const (
	RenewalNotValidRtoEndDateErrorText = "bad rto end date for Renewal"
	RenewalNotValidStateErrorText      = "bad rto state for Renewal"
)

type RenewalUseCase interface {
	Renewal(ctx context.Context, req *request.RenewalRequest) (*response.RtoResponse, error)
}

type renewalUseCase struct {
	repository        domain.Repository
	historyRepository domain.HistoryRepository
	securityService   service.SecurityService
}

func NewRenewalUseCase(repository domain.Repository,
	historyRepository domain.HistoryRepository,
	securityService service.SecurityService) RenewalUseCase {
	return &renewalUseCase{
		repository:        repository,
		historyRepository: historyRepository,
		securityService:   securityService,
	}
}

func (ruc *renewalUseCase) Renewal(ctx context.Context, req *request.RenewalRequest) (*response.RtoResponse, error) {
	err := req.Validate()
	if err != nil {
		return nil, errors.Wrap(err, "rto renewalUseCase req.Validate")
	}

	canLandlordUpdateRtoResponse, err := ruc.securityService.CanLandlordUpdateRto(ctx,
		&requestservrto.CanLandlordUpdateRtoRequest{
			ID:              req.ID,
			LandlordID:      req.LandlordID,
			NeedCheckVeriff: true,
		})
	if err != nil {
		return nil, errors.Wrap(err, "rto renewalUseCase rs.securityService.CanLandlordUpdateRto")
	}
	rto := canLandlordUpdateRtoResponse.Rto

	if rto.State != api.RtoState_RTO_STATE_ACTIVE {
		return nil, errors.Wrap(fmt.Errorf("rto state must be in state %s, not in state %s",
			api.RtoState_RTO_STATE_ACTIVE, rto.State), RenewalNotValidStateErrorText)
	}

	if req.EndDate.IsZero() {
		rto.State = api.RtoState_RTO_STATE_ACTIVE_ROLLING_TENANCY
	} else {
		if req.EndDate.Before(rto.Conditions.MoveInDate) {
			return nil, errors.Wrap(fmt.Errorf("end date (%s) can not be early than start date (%s)",
				req.EndDate, rto.Conditions.MoveInDate), RenewalNotValidRtoEndDateErrorText)
		}
		rto.State = api.RtoState_RTO_STATE_FINISHED
		rto.Conditions.EndDate = req.EndDate
		rto.Conditions.TermMonths = util.GetTermMonths(rto.Conditions.MoveInDate, rto.Conditions.EndDate)
	}
	now := time.Now()
	rto.UpdatedAt = now

	err = mongolocal.Transaction(ctx, func() error {
		err = ruc.repository.MustUpdate(ctx, rto)
		if err != nil {
			return errors.Wrap(err, "rto renewalUseCase repository.MustUpdate")
		}

		err = ruc.historyRepository.Create(ctx, domain.CreateHistory(domain.HistoryActionUpdate, now, rto))
		if err != nil {
			return errors.Wrap(err, "rto renewalUseCase historyRepository.Create")
		}

		return nil
	})
	if err != nil {
		return nil, errors.Wrap(err, "rto renewalUseCase mongolocal.Transaction")
	}

	return &response.RtoResponse{
		Rto: rto,
	}, nil
}
