package rto_test

import (
	"strings"
	"testing"
	"time"

	api "gitlab.mashplace.com/mashplace/backend/offer/api/rto"
	requestservrto "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto/resource/request/service/rto"
	request "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto/resource/request/usecase"
	"gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto/resource/response"
	mockservice "gitlab.mashplace.com/mashplace/backend/offer/internal/app/service/rto/mock"
	"gitlab.mashplace.com/mashplace/backend/offer/internal/app/usecase/rto"
	"gitlab.mashplace.com/mashplace/backend/offer/test"

	"github.com/golang/mock/gomock"
	. "github.com/smartystreets/goconvey/convey"
)

func TestRenewalUseCase(t *testing.T) {
	if test.CanRunTests() {
		Convey("RenewalUseCase", t, func() {
			ctrl := gomock.NewController(t)
			defer ctrl.Finish()

			securityService := mockservice.NewMockSecurityService(ctrl)
			useCase := rto.NewRenewalUseCase(repository, historyRepository, securityService)

			Convey("Renewal error "+rto.RenewalNotValidStateErrorText, func() {
				rtoModel := test.CreateRto(ctx, t, repository, api.RtoState_RTO_STATE_PENDING_SETUP,
					test.KorshilerDogs, test.Chmolgani)
				rtoID, landlordID := rtoModel.ID.Hex(), rtoModel.Landlord.User

				securityService.EXPECT().CanLandlordUpdateRto(ctx, &requestservrto.CanLandlordUpdateRtoRequest{
					ID:              rtoID,
					LandlordID:      landlordID,
					NeedCheckVeriff: true,
				}).Return(&response.RtoResponse{Rto: rtoModel}, nil)
				rtoRenewalRequest := &request.RenewalRequest{
					ID:         rtoID,
					LandlordID: landlordID,
					EndDate:    rtoModel.CreatedAt.Add(-1 * time.Hour),
				}
				res, err := useCase.Renewal(ctx, rtoRenewalRequest)
				So(res, ShouldBeNil)
				So(err, ShouldBeError)
				So(strings.Contains(err.Error(), rto.RenewalNotValidStateErrorText), ShouldBeTrue)

				rtoModel.State = api.RtoState_RTO_STATE_ACTIVE
				err = repository.MustUpdate(ctx, rtoModel)
				So(err, ShouldBeNil)

				Convey("Renewal error "+rto.RenewalNotValidRtoEndDateErrorText, func() {
					securityService.EXPECT().CanLandlordUpdateRto(ctx, &requestservrto.CanLandlordUpdateRtoRequest{
						ID:              rtoID,
						LandlordID:      landlordID,
						NeedCheckVeriff: true,
					}).Return(&response.RtoResponse{Rto: rtoModel}, nil)
					res, err = useCase.Renewal(ctx, rtoRenewalRequest)
					So(res, ShouldBeNil)
					So(err, ShouldBeError)
					So(strings.Contains(err.Error(), rto.RenewalNotValidRtoEndDateErrorText), ShouldBeTrue)

					Convey("Renewal success "+string(api.RtoState_RTO_STATE_ACTIVE_ROLLING_TENANCY), func() {
						securityService.EXPECT().CanLandlordUpdateRto(ctx,
							&requestservrto.CanLandlordUpdateRtoRequest{
								ID:              rtoID,
								LandlordID:      landlordID,
								NeedCheckVeriff: true,
							}).Return(&response.RtoResponse{Rto: rtoModel}, nil)
						rtoRenewalRequest.EndDate = time.Time{}
						res, err = useCase.Renewal(ctx, rtoRenewalRequest)
						So(err, ShouldBeNil)
						So(res, ShouldNotBeNil)
						So(res.Rto.State, ShouldEqual, api.RtoState_RTO_STATE_ACTIVE_ROLLING_TENANCY)

						rtoModel.State = api.RtoState_RTO_STATE_ACTIVE
						err = repository.MustUpdate(ctx, rtoModel)
						So(err, ShouldBeNil)

						Convey("Renewal success "+string(api.RtoState_RTO_STATE_FINISHED), func() {
							securityService.EXPECT().CanLandlordUpdateRto(ctx,
								&requestservrto.CanLandlordUpdateRtoRequest{
									ID:              rtoID,
									LandlordID:      landlordID,
									NeedCheckVeriff: true,
								}).Return(&response.RtoResponse{Rto: rtoModel}, nil)
							rtoRenewalRequest.EndDate = rtoModel.CreatedAt.Add(2 * time.Hour)
							res, err = useCase.Renewal(ctx, rtoRenewalRequest)
							So(err, ShouldBeNil)
							So(res, ShouldNotBeNil)
							So(res.Rto.State, ShouldEqual, api.RtoState_RTO_STATE_FINISHED)

							test.DeleteRto(ctx, t, repository, historyRepository, rtoModel.ID, 2)
						})
					})
				})
			})
		})
	}
}
