package rto

import (
	"context"
	"fmt"

	"gitlab.mashplace.com/mashplace/backend/common/generic/log"

	"go.mongodb.org/mongo-driver/mongo"

	appservice "gitlab.mashplace.com/mashplace/backend/offer/internal/app/service"
	service "gitlab.mashplace.com/mashplace/backend/offer/internal/app/service/rto"

	api "gitlab.mashplace.com/mashplace/backend/offer/api/rto"
	domain "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto"
	requestservice "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto/resource/request/service"
	requestservrto "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto/resource/request/service/rto"
	request "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto/resource/request/usecase"
	"gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto/resource/response"

	"github.com/pkg/errors"
)

const (
	RentPaymentNotValidStateErrorText                     = "bad rto state for RentPayment"
	RentPaymentNotValidListingVersionBankDetailsErrorText = "rentPayment not valid listing version bank details"
)

type RentPaymentUseCase interface {
	RentPayment(ctx context.Context, req *request.RentPaymentRequest) (*response.RtoResponse, error)
}

type rentPaymentUseCase struct {
	repository                                       domain.Repository
	accountService                                   appservice.AccountService
	listingService                                   appservice.ListingService
	closeComService                                  service.CloseComService
	updateStateToWaitingForTenantConfirmationUseCase UpdateStateToWaitingForTenantConfirmationUseCase
}

func NewRentPaymentUseCase(repository domain.Repository,
	accountService appservice.AccountService,
	listingService appservice.ListingService,
	closeComService service.CloseComService,
	updateStateToWaitingForTenantConfirmationUseCase UpdateStateToWaitingForTenantConfirmationUseCase,
) RentPaymentUseCase {
	return &rentPaymentUseCase{
		repository:      repository,
		accountService:  accountService,
		listingService:  listingService,
		closeComService: closeComService,
		updateStateToWaitingForTenantConfirmationUseCase: updateStateToWaitingForTenantConfirmationUseCase,
	}
}

func (rpuc *rentPaymentUseCase) RentPayment(ctx context.Context, req *request.RentPaymentRequest) (
	*response.RtoResponse, error) {
	err := req.Validate()
	if err != nil {
		return nil, errors.Wrap(err, "rto rentPaymentUseCase req.Validate")
	}

	rto, err := rpuc.repository.FindByPropertyID(ctx, req.ListingVersion.PropertyId)
	if err != nil && !errors.Is(err, mongo.ErrNoDocuments) {
		return nil, errors.Wrap(err, "rto rentPaymentUseCase repository.FindAllByLandlordID")
	}

	if rto == nil || !rto.IsRto {
		return &response.RtoResponse{}, nil
	}

	if !rpuc.listingService.VersionHasNotEmptyBankDetails(ctx, req.ListingVersion) {
		return nil, errors.Wrap(errors.New(RentPaymentNotValidListingVersionBankDetailsErrorText),
			"rto rentPaymentUseCase listingService.VersionHasNotEmptyBankDetails")
	}

	if rto.State != api.RtoState_RTO_STATE_PENDING_SETUP {
		return nil, errors.Wrap(fmt.Errorf("rto state must be in state %s, not in state %s",
			api.RtoState_RTO_STATE_PENDING_SETUP, rto.State), RentPaymentNotValidStateErrorText)
	}

	err = rpuc.accountService.IsVeriffValid(ctx, &requestservice.IsVeriffValidRequest{LandlordID: rto.GetLandlordID()})
	if err != nil {
		return nil, errors.Wrap(err, "rto rentPaymentUseCase accountService.IsVeriffValid")
	}

	err = rpuc.accountService.IsOwnershipVerified(ctx, &requestservice.IsOwnershipVerifiedRequest{PropertyID: rto.Property})
	if err != nil {
		return nil, errors.Wrap(err, "rto rentPaymentUseCase accountService.IsOwnershipVerified")
	}

	_, err = rpuc.updateStateToWaitingForTenantConfirmationUseCase.UpdateStateToWaitingForTenantConfirmation(ctx,
		&request.UpdateStateToWaitingForTenantConfirmationRequest{Rto: rto})
	if err != nil {
		return nil, errors.Wrap(err, "rto rentPaymentUseCase updateStateToWaitingForTenantConfirmationUseCase")
	}

	go func() {
		ctx := context.Background()
		_, err = rpuc.closeComService.SaveOpportunity(ctx, &requestservrto.CloseComSaveOpportunityRequest{Rto: rto})
		if err != nil {
			log.WithContext(ctx).WithError(err).Error("rto RentPayment closeComService.SaveOpportunity")
		}
	}()

	return &response.RtoResponse{
		Rto: rto,
	}, nil
}
