package rto

import (
	"context"
	"fmt"
	"time"

	"gitlab.mashplace.com/mashplace/backend/common/lib/mongolocal"
	api "gitlab.mashplace.com/mashplace/backend/offer/api/rto"
	domain "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto"
	request "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto/resource/request/usecase"
	"gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto/resource/response"

	"github.com/pkg/errors"
)

const StandingOrderNotValidStateErrorText = "bad rto state for StandingOrder"

type StandingOrderUseCase interface {
	StandingOrder(ctx context.Context, req *request.StandingOrderRequest) (*response.RtoResponse, error)
}

type standingOrderUseCase struct {
	repository        domain.Repository
	historyRepository domain.HistoryRepository
}

func NewStandingOrderUseCase(repository domain.Repository,
	historyRepository domain.HistoryRepository,
) StandingOrderUseCase {
	return &standingOrderUseCase{
		repository:        repository,
		historyRepository: historyRepository,
	}
}

func (souc *standingOrderUseCase) StandingOrder(ctx context.Context, req *request.StandingOrderRequest) (
	*response.RtoResponse, error) {
	err := req.Validate()
	if err != nil {
		return nil, errors.Wrap(err, "rto standingOrderUseCase req.Validate")
	}

	rto, err := souc.repository.FindByID(ctx, req.ID)
	if err != nil {
		return nil, errors.Wrap(err, "rto standingOrderUseCase repository.FindByID")
	}

	if rto.State != api.RtoState_RTO_STATE_WAITING_FOR_STANDING_ORDER_SETUP {
		return nil, errors.Wrap(fmt.Errorf("rto state must be in state %s, not in state %s",
			api.RtoState_RTO_STATE_WAITING_FOR_STANDING_ORDER_SETUP, rto.State), StandingOrderNotValidStateErrorText)
	}

	if rto.PaidStatuses == nil {
		rto.PaidStatuses = &domain.PaidStatuses{}
	}
	rto.PaidStatuses.StandingOrder = true
	now := time.Now()
	rto.UpdatedAt = now

	err = mongolocal.Transaction(ctx, func() error {
		err = souc.repository.MustUpdate(ctx, rto)
		if err != nil {
			return errors.Wrap(err, "rto standingOrderUseCase repository.MustUpdate")
		}

		err = souc.historyRepository.Create(ctx, domain.CreateHistory(domain.HistoryActionUpdate, now, rto))
		if err != nil {
			return errors.Wrap(err, "rto standingOrderUseCase historyRepository.Create")
		}

		return nil
	})
	if err != nil {
		return nil, errors.Wrap(err, "rto standingOrderUseCase mongolocal.Transaction")
	}

	return &response.RtoResponse{
		Rto: rto,
	}, nil
}
