package rto

import (
	"context"
	"fmt"
	"time"

	"gitlab.mashplace.com/mashplace/backend/common/lib/mongolocal"
	api "gitlab.mashplace.com/mashplace/backend/offer/api/rto"
	domain "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto"
	requestservrto "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto/resource/request/service/rto"
	request "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto/resource/request/usecase"
	"gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto/resource/response"
	service "gitlab.mashplace.com/mashplace/backend/offer/internal/app/service/rto"
	"gitlab.mashplace.com/mashplace/backend/offer/internal/pkg/util"

	"github.com/pkg/errors"
)

const (
	UpdateEndDateNotValidEndDateErrorText = "bad rto end date for UpdateEndDate"
	UpdateEndDateNotValidStateErrorText   = "bad rto state for UpdateEndDate"
)

type UpdateEndDateUseCase interface {
	UpdateEndDate(ctx context.Context, req *request.UpdateEndDateRequest) (*response.RtoResponse, error)
}

type updateEndDateUseCase struct {
	repository        domain.Repository
	historyRepository domain.HistoryRepository
	securityService   service.SecurityService
	closeComService   service.CloseComService
}

func NewUpdateEndDateUseCase(repository domain.Repository,
	historyRepository domain.HistoryRepository,
	securityService service.SecurityService,
	closeComService service.CloseComService) UpdateEndDateUseCase {
	return &updateEndDateUseCase{
		repository:        repository,
		historyRepository: historyRepository,
		securityService:   securityService,
		closeComService:   closeComService,
	}
}

func (ueduc *updateEndDateUseCase) UpdateEndDate(ctx context.Context, req *request.UpdateEndDateRequest) (
	*response.RtoResponse, error) {
	err := req.Validate()
	if err != nil {
		return nil, errors.Wrap(err, "rto updateEndDateUseCase req.Validate")
	}

	canLandlordUpdateRtoResponse, err := ueduc.securityService.CanLandlordUpdateRto(ctx,
		&requestservrto.CanLandlordUpdateRtoRequest{
			ID:              req.ID,
			LandlordID:      req.LandlordID,
			NeedCheckVeriff: true,
		})
	if err != nil {
		return nil, errors.Wrap(err, "rto updateEndDateUseCase securityService.CanLandlordUpdateRto")
	}
	rto := canLandlordUpdateRtoResponse.Rto

	if rto.State != api.RtoState_RTO_STATE_ENDING {
		return nil, errors.Wrap(fmt.Errorf("rto state must be in state %s, not in state %s",
			api.RtoState_RTO_STATE_ENDING, rto.State), UpdateEndDateNotValidStateErrorText)
	}

	if req.EndDate.Before(rto.Conditions.MoveInDate) {
		return nil, errors.Wrap(fmt.Errorf("rto end date (%s) can not be early than start date (%s)",
			req.EndDate, rto.Conditions.MoveInDate), UpdateEndDateNotValidEndDateErrorText)
	}

	rto.Conditions.EndDate = req.EndDate
	rto.Conditions.TermMonths = util.GetTermMonths(rto.Conditions.MoveInDate, rto.Conditions.EndDate)
	rto.State = api.RtoState_RTO_STATE_FINISHED
	now := time.Now()
	rto.UpdatedAt = now

	err = mongolocal.Transaction(ctx, func() error {
		err = ueduc.repository.MustUpdate(ctx, rto)
		if err != nil {
			return errors.Wrap(err, "rto updateEndDateUseCase repository.MustUpdate")
		}

		err = ueduc.historyRepository.Create(ctx, domain.CreateHistory(domain.HistoryActionUpdate, now, rto))
		if err != nil {
			return errors.Wrap(err, "rto updateEndDateUseCase historyRepository.Create")
		}

		return nil
	})
	if err != nil {
		return nil, errors.Wrap(err, "rto updateEndDateUseCase mongolocal.Transaction")
	}

	closeComSaveOpportunityRequest := requestservrto.CreateCloseComSaveOpportunityRequest(rto)
	_, err = ueduc.closeComService.SaveOpportunity(ctx, closeComSaveOpportunityRequest)
	if err != nil {
		return nil, errors.Wrap(err, "rto updateEndDateUseCase closeComService.SaveOpportunity")
	}

	return &response.RtoResponse{
		Rto: rto,
	}, nil
}
