package rto

import (
	"context"
	"time"

	account "gitlab.mashplace.com/mashplace/backend/account/api"
	"gitlab.mashplace.com/mashplace/backend/offer/internal/connections"

	"gitlab.mashplace.com/mashplace/backend/common/lib/mongolocal"
	api "gitlab.mashplace.com/mashplace/backend/offer/api/rto"
	domain "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto"
	requestservice "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto/resource/request/service"
	request "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto/resource/request/usecase"
	response "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto/resource/response/usecase"
	"gitlab.mashplace.com/mashplace/backend/offer/internal/app/service"

	"github.com/pkg/errors"
)

type UpdateLeadTenantUseCase interface {
	UpdateLeadTenant(ctx context.Context, req *request.UpdateLeadTenantRequest) (
		*response.UpdateLeadTenantResponse, error)
}

type updateLeadTenantUseCase struct {
	repository        domain.Repository
	historyRepository domain.HistoryRepository
	listingService    service.ListingService
}

func NewUpdateLeadTenantUseCase(repository domain.Repository,
	historyRepository domain.HistoryRepository,
	listingService service.ListingService) UpdateLeadTenantUseCase {
	return &updateLeadTenantUseCase{
		repository:        repository,
		historyRepository: historyRepository,
		listingService:    listingService,
	}
}

func (ultuc *updateLeadTenantUseCase) UpdateLeadTenant(ctx context.Context,
	req *request.UpdateLeadTenantRequest) (*response.UpdateLeadTenantResponse, error) {
	err := req.Validate()
	if err != nil {
		return nil, errors.Wrap(err, "rto updateLeadTenantUseCase req.Validate")
	}

	rtos, err := ultuc.repository.FindAllEmptyLeadTenantUsersByEmail(ctx, req.LeadTenantEmail)
	if err != nil {
		return nil, errors.Wrap(err, "rto updateLeadTenantUseCase repository.FindAllEmptyLeadTenantUsersByEmail")
	}

	leadTenant, err := connections.AccountService.Get(ctx, &account.GetRequest{
		Id: req.LeadTenantID,
	})
	if err != nil {
		return nil, errors.Wrap(err, "rto updateLeadTenantUseCase connections.AccountService")
	}

	var ids, propertyIDs []string

	for _, rto := range rtos {
		now := time.Now()
		rto.UpdatedAt = now
		rto.LeadTenant.User = req.LeadTenantID
		rto.LeadTenant.FirstName = leadTenant.GetAccount().GetDetailsValue().GetFirstNameValue()
		rto.LeadTenant.LastName = leadTenant.GetAccount().GetDetailsValue().GetLastNameValue()

		for _, tenant := range rto.Tenants {
			if tenant.EmailPrimary == req.LeadTenantEmail {
				tenant.User = req.LeadTenantID
				tenant.FirstName = leadTenant.GetAccount().GetDetailsValue().GetFirstNameValue()
				tenant.LastName = leadTenant.GetAccount().GetDetailsValue().GetLastNameValue()
			}
		}

		err = mongolocal.Transaction(ctx, func() error {
			err = ultuc.repository.MustUpdate(ctx, rto)
			if err != nil {
				return errors.Wrap(err, "rto updateLeadTenantUseCase repository.MustUpdate")
			}

			err = ultuc.historyRepository.Create(ctx, domain.CreateHistory(domain.HistoryActionUpdate, now, rto))
			if err != nil {
				return errors.Wrap(err, "rto updateLeadTenantUseCase historyRepository.Create")
			}

			return nil
		})
		if err != nil {
			return nil, errors.Wrap(err, "rto updateLeadTenantUseCase mongolocal.Transaction")
		}

		ids = append(ids, rto.ID.Hex())

		if rto.State > api.RtoState_RTO_STATE_PENDING_SETUP {
			err = ultuc.listingService.UpdateTenantID(ctx, &requestservice.ListingUpdateTenantIDRequest{
				PropertyID:   rto.Property,
				LeadTenantID: rto.GetLeadTenantID(),
				LandlordID:   rto.GetLandlordID(),
			})
			if err != nil {
				return nil, errors.Wrap(err, "rto updateLeadTenantUseCase listingService.UpdateTenantID")
			}

			propertyIDs = append(propertyIDs, rto.Property)
		}
	}

	return &response.UpdateLeadTenantResponse{
		IDs:         ids,
		PropertyIDs: propertyIDs,
	}, nil
}
