package rto

import (
	"context"
	"time"

	"gitlab.mashplace.com/mashplace/backend/common/api/grpcerr"

	"go.mongodb.org/mongo-driver/bson/primitive"

	"gitlab.mashplace.com/mashplace/backend/common/lib/mongolocal"
	api "gitlab.mashplace.com/mashplace/backend/offer/api/rto"
	domain "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto"
	requestservice "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto/resource/request/service"
	request "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto/resource/request/usecase"
	"gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto/resource/response"
	"gitlab.mashplace.com/mashplace/backend/offer/internal/app/service"

	"github.com/pkg/errors"
)

const UpdateLeadTenantUserTenantNotFound = "rto updateLeadTenantUserUseCase tenant not found"

type UpdateLeadTenantUserUseCase interface {
	UpdateLeadTenantUser(ctx context.Context, req *request.UpdateLeadTenantUserRequest) (*response.RtoResponse, error)
}

type updateLeadTenantUserUseCase struct {
	repository        domain.Repository
	historyRepository domain.HistoryRepository
	accountService    service.AccountService
	listingService    service.ListingService
}

func NewUpdateLeadTenantUserUseCase(repository domain.Repository,
	historyRepository domain.HistoryRepository,
	accountService service.AccountService,
	listingService service.ListingService) UpdateLeadTenantUserUseCase {
	return &updateLeadTenantUserUseCase{
		repository:        repository,
		historyRepository: historyRepository,
		accountService:    accountService,
		listingService:    listingService,
	}
}

func (ultuuc *updateLeadTenantUserUseCase) UpdateLeadTenantUser(ctx context.Context,
	req *request.UpdateLeadTenantUserRequest) (*response.RtoResponse, error) {
	err := req.Validate()
	if err != nil {
		return nil, errors.Wrap(err, "rto updateLeadTenantUserUseCase req.Validate")
	}

	rtoId, err := primitive.ObjectIDFromHex(req.RtoID)
	if err != nil {
		return nil, errors.Wrap(err, "rto updateLeadTenantUserUseCase wrong rto id format")
	}

	rto, err := ultuuc.repository.FindByID(ctx, rtoId)
	if err != nil {
		return nil, errors.Wrap(err, "rto updateLeadTenantUserUseCase repository.FindByID")
	}

	var email string
	if req.LeadTenantEmail != "" {
		email = req.LeadTenantEmail
	} else {
		email = rto.GetLeadTenantEmail()
	}

	tenant, err := ultuuc.accountService.GetAccountByEmail(ctx, email)
	if err != nil && !errors.Is(err, grpcerr.NotFound) {
		return nil, errors.Wrap(err, "rto updateLeadTenantUserUseCase accountService.GetAccountByEmail")
	}

	var (
		tenantId        string
		tenantEmail     string
		tenantPhone     string
		tenantFirstName string
		tenantLastName  string
	)

	if tenant != nil && tenant.GetIdValue() != "" {
		tenantId = tenant.GetIdValue()
		tenantEmail = tenant.GetContactValue().GetEmailPrimaryValue()
		tenantPhone = tenant.GetContactValue().GetPhonePrimaryValue()
		tenantFirstName = tenant.GetDetailsValue().GetFirstNameValue()
		tenantLastName = tenant.GetDetailsValue().GetLastNameValue()
	} else {
		tenantId = ""
		tenantEmail = req.LeadTenantEmail
		tenantPhone = req.LeadTenantPhone
		tenantFirstName = req.LeadTenantFirstName
		tenantLastName = req.LeadTenantLastName
	}

	rto.LeadTenant.User = tenantId
	rto.LeadTenant.EmailPrimary = tenantEmail
	rto.LeadTenant.PhonePrimary = tenantPhone
	rto.LeadTenant.FirstName = tenantFirstName
	rto.LeadTenant.LastName = tenantLastName
	if len(rto.Tenants) > 0 {
		rto.Tenants[0].User = tenantId
		rto.Tenants[0].EmailPrimary = tenantEmail
		rto.Tenants[0].PhonePrimary = tenantPhone
		rto.Tenants[0].FirstName = tenantFirstName
		rto.Tenants[0].LastName = tenantLastName
	}

	now := time.Now()
	rto.UpdatedAt = now

	err = mongolocal.Transaction(ctx, func() error {
		err = ultuuc.repository.MustUpdate(ctx, rto)
		if err != nil {
			return errors.Wrap(err, "rto updateLeadTenantUserUseCase repository.MustUpdate")
		}

		err = ultuuc.historyRepository.Create(ctx, domain.CreateHistory(domain.HistoryActionUpdate, now, rto))
		if err != nil {
			return errors.Wrap(err, "rto updateLeadTenantUserUseCase historyRepository.Create")
		}

		return nil
	})
	if err != nil {
		return nil, errors.Wrap(err, "rto updateLeadTenantUserUseCase mongolocal.Transaction")
	}

	if rto.GetLeadTenantID() != "" && rto.State > api.RtoState_RTO_STATE_PENDING_SETUP {
		err = ultuuc.listingService.UpdateTenantID(ctx, &requestservice.ListingUpdateTenantIDRequest{
			PropertyID:   rto.Property,
			LeadTenantID: rto.GetLeadTenantID(),
			LandlordID:   rto.GetLandlordID(),
		})
		if err != nil {
			return nil, errors.Wrap(err, "rto updateLeadTenantUserUseCase listingService.UpdateTenantID")
		}
	}

	return &response.RtoResponse{
		Rto: rto,
	}, nil
}
