package rto

import (
	"context"
	"fmt"
	"time"

	"gitlab.mashplace.com/mashplace/backend/common/generic/log"
	"gitlab.mashplace.com/mashplace/backend/common/lib/mongolocal"
	api "gitlab.mashplace.com/mashplace/backend/offer/api/rto"
	domain "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto"
	requestservrto "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto/resource/request/service/rto"
	request "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto/resource/request/usecase"
	"gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto/resource/response"
	service "gitlab.mashplace.com/mashplace/backend/offer/internal/app/service/rto"

	"github.com/pkg/errors"
	"go.mongodb.org/mongo-driver/bson/primitive"
)

const UpdateStateToActiveNotValidStateErrorText = "bad rto state for Active"

type UpdateStateToActiveUseCase interface {
	UpdateStateToActive(ctx context.Context, req *request.UpdateStateToActiveRequest) (*response.RtoResponse, error)
}

type updateStateToActiveUseCase struct {
	repository          domain.Repository
	historyRepository   domain.HistoryRepository
	notificationService service.NotificationService
	closeComService     service.CloseComService
}

func NewUpdateStateToActiveUseCase(repository domain.Repository,
	historyRepository domain.HistoryRepository,
	notificationService service.NotificationService,
	closeComService service.CloseComService) UpdateStateToActiveUseCase {
	return &updateStateToActiveUseCase{
		repository:          repository,
		historyRepository:   historyRepository,
		notificationService: notificationService,
		closeComService:     closeComService,
	}
}

func (ustauc *updateStateToActiveUseCase) UpdateStateToActive(ctx context.Context,
	req *request.UpdateStateToActiveRequest) (*response.RtoResponse, error) {
	err := req.Validate()
	if err != nil {
		return nil, errors.Wrap(err, "rto updateStateToActiveUseCase req.Validate")
	}

	rtoID, err := primitive.ObjectIDFromHex(req.ID)
	if err != nil {
		return nil, errors.Wrap(err,
			"rto updateStateToActiveUseCase primitive.ObjectIDFromHex")
	}

	var (
		closeComSaveOpportunityRequest *requestservrto.CloseComSaveOpportunityRequest
		rto                            *domain.Rto
	)
	err = mongolocal.Transaction(ctx, func() error {
		rto, err = ustauc.repository.FindByID(ctx, rtoID)
		if err != nil {
			return errors.Wrap(err,
				"rto updateStateToActiveUseCase repository.FindByID")
		}

		if rto.State != api.RtoState_RTO_STATE_WAITING_FOR_PAYMENT {
			return errors.Wrap(fmt.Errorf("rto state must be in state %s, not in state %s",
				api.RtoState_RTO_STATE_WAITING_FOR_PAYMENT, rto.State),
				UpdateStateToActiveNotValidStateErrorText)
		}

		closeComSaveOpportunityRequest = requestservrto.CreateCloseComSaveOpportunityRequest(rto)
		rto.State = api.RtoState_RTO_STATE_ACTIVE
		now := time.Now()
		rto.UpdatedAt = now

		err = ustauc.repository.MustUpdate(ctx, rto)
		if err != nil {
			return errors.Wrap(err,
				"rto updateStateToActiveUseCase repository.MustUpdate")
		}

		err = ustauc.historyRepository.Create(ctx, domain.CreateHistory(domain.HistoryActionUpdate, now, rto))
		if err != nil {
			return errors.Wrap(err,
				"rto updateStateToActiveUseCase historyRepository.Create")
		}

		return nil
	})
	if err != nil {
		return nil, errors.Wrap(err,
			"rto updateStateToActiveUseCase mongolocal.Transaction")
	}

	err = ustauc.notificationService.SendActive(ctx,
		&requestservrto.SendActiveRequest{
			LandlordID: rto.Landlord.User,
			PropertyID: rto.Property,
			OfferID:    rto.ID.Hex(),
			RentPlan:   rto.RentPlan,
		})
	if err != nil {
		return nil, errors.Wrap(err, "rto updateStateToActiveUseCase notificationService.SendActive")
	}

	go func() {
		ctx := context.Background()
		_, err = ustauc.closeComService.SaveOpportunity(ctx, closeComSaveOpportunityRequest)
		if err != nil {
			log.WithContext(ctx).WithError(err).Error("rto updateStateToActiveUseCase closeComService.SaveOpportunity")
		}
	}()

	return &response.RtoResponse{
		Rto: rto,
	}, nil
}
