package rto

import (
	"context"
	"fmt"
	"time"

	"gitlab.mashplace.com/mashplace/backend/common/lib/mongolocal"
	api "gitlab.mashplace.com/mashplace/backend/offer/api/rto"
	domain "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto"
	requestservrto "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto/resource/request/service/rto"
	request "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto/resource/request/usecase"
	"gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto/resource/response"
	service "gitlab.mashplace.com/mashplace/backend/offer/internal/app/service/rto"

	"github.com/pkg/errors"
	"go.mongodb.org/mongo-driver/bson/primitive"
)

const UpdateStateToActiveRollingTenancyNotValidStateErrorText = "bad rto state for ActiveRollingTenancy"

// todo: not used, in renewal
type UpdateStateToActiveRollingTenancyUseCase interface {
	UpdateStateToActiveRollingTenancy(ctx context.Context, req *request.UpdateStateToActiveRollingTenancyRequest) (
		*response.RtoResponse, error)
}

type updateStateToActiveRollingTenancyUseCase struct {
	repository        domain.Repository
	historyRepository domain.HistoryRepository
	closeComService   service.CloseComService
}

func NewUpdateStateToActiveRollingTenancyUseCase(repository domain.Repository,
	historyRepository domain.HistoryRepository,
	closeComService service.CloseComService) UpdateStateToActiveRollingTenancyUseCase {
	return &updateStateToActiveRollingTenancyUseCase{
		repository:        repository,
		historyRepository: historyRepository,
		closeComService:   closeComService,
	}
}

func (ustartuc *updateStateToActiveRollingTenancyUseCase) UpdateStateToActiveRollingTenancy(ctx context.Context,
	req *request.UpdateStateToActiveRollingTenancyRequest) (*response.RtoResponse, error) {
	err := req.Validate()
	if err != nil {
		return nil, errors.Wrap(err, "rto updateStateToActiveRollingTenancyUseCase req.Validate")
	}

	rtoID, err := primitive.ObjectIDFromHex(req.ID)
	if err != nil {
		return nil, errors.Wrap(err,
			"rto updateStateToActiveRollingTenancyUseCase primitive.ObjectIDFromHex")
	}

	rto, err := ustartuc.repository.FindByID(ctx, rtoID)
	if err != nil {
		return nil, errors.Wrap(err,
			"rto updateStateToActiveRollingTenancyUseCase repository.FindByID")
	}

	if rto.State != api.RtoState_RTO_STATE_ACTIVE {
		return nil, errors.Wrap(fmt.Errorf("rto state must be in state %s, not in state %s",
			api.RtoState_RTO_STATE_ACTIVE, rto.State),
			UpdateStateToActiveRollingTenancyNotValidStateErrorText)
	}

	closeComSaveOpportunityRequest := requestservrto.CreateCloseComSaveOpportunityRequest(rto)
	rto.State = api.RtoState_RTO_STATE_ACTIVE_ROLLING_TENANCY
	now := time.Now()
	rto.UpdatedAt = now

	err = mongolocal.Transaction(ctx, func() error {
		err = ustartuc.repository.MustUpdate(ctx, rto)
		if err != nil {
			return errors.Wrap(err,
				"rto updateStateToActiveRollingTenancyUseCase repository.MustUpdate")
		}

		err = ustartuc.historyRepository.Create(ctx, domain.CreateHistory(domain.HistoryActionUpdate, now, rto))
		if err != nil {
			return errors.Wrap(err,
				"rto updateStateToActiveRollingTenancyUseCase historyRepository.Create")
		}

		return nil
	})
	if err != nil {
		return nil, errors.Wrap(err,
			"rto updateStateToActiveRollingTenancyUseCase mongolocal.Transaction")
	}

	_, err = ustartuc.closeComService.SaveOpportunity(ctx, closeComSaveOpportunityRequest)
	if err != nil {
		return nil, errors.Wrap(err,
			"rto updateStateToActiveRollingTenancyUseCase closeComService.SaveOpportunity")
	}

	return &response.RtoResponse{
		Rto: rto,
	}, nil
}
