package rto

import (
	"context"
	"fmt"
	"time"

	"gitlab.mashplace.com/mashplace/backend/common/generic/log"

	listing "gitlab.mashplace.com/mashplace/backend/listing/api"

	requestservice "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto/resource/request/service"

	"gitlab.mashplace.com/mashplace/backend/common/lib/mongolocal"
	api "gitlab.mashplace.com/mashplace/backend/offer/api/rto"
	domain "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto"
	requestservrto "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto/resource/request/service/rto"
	request "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto/resource/request/usecase"
	"gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto/resource/response"
	appservice "gitlab.mashplace.com/mashplace/backend/offer/internal/app/service"
	service "gitlab.mashplace.com/mashplace/backend/offer/internal/app/service/rto"

	"github.com/pkg/errors"
	"go.mongodb.org/mongo-driver/bson/primitive"
)

const UpdateStateToWaitingForPaymentNotValidStateErrorText = "bad rto state for Waiting For Payment"

type UpdateStateToWaitingForPaymentUseCase interface {
	UpdateStateToWaitingForPayment(ctx context.Context, req *request.UpdateStateToWaitingForPaymentRequest) (*response.RtoResponse, error)
}

type updateStateToWaitingForPaymentUseCase struct {
	repository          domain.Repository
	historyRepository   domain.HistoryRepository
	notificationService service.NotificationService
	closeComService     service.CloseComService
	listingService      appservice.ListingService
}

func NewUpdateStateToWaitingForPaymentUseCase(repository domain.Repository,
	historyRepository domain.HistoryRepository,
	notificationService service.NotificationService,
	closeComService service.CloseComService,
	listingService appservice.ListingService) UpdateStateToWaitingForPaymentUseCase {
	return &updateStateToWaitingForPaymentUseCase{
		repository:          repository,
		historyRepository:   historyRepository,
		notificationService: notificationService,
		closeComService:     closeComService,
		listingService:      listingService,
	}
}

func (ustauc *updateStateToWaitingForPaymentUseCase) UpdateStateToWaitingForPayment(ctx context.Context,
	req *request.UpdateStateToWaitingForPaymentRequest) (*response.RtoResponse, error) {
	err := req.Validate()
	if err != nil {
		return nil, errors.Wrap(err, "rto updateStateToWaitingForPaymentUseCase req.Validate")
	}

	rtoID, err := primitive.ObjectIDFromHex(req.ID)
	if err != nil {
		return nil, errors.Wrap(err,
			"rto updateStateToWaitingForPaymentUseCase primitive.ObjectIDFromHex")
	}

	var (
		closeComSaveOpportunityRequest *requestservrto.CloseComSaveOpportunityRequest
		rto                            *domain.Rto
	)
	err = mongolocal.Transaction(ctx, func() error {
		rto, err = ustauc.repository.FindByID(ctx, rtoID)
		if err != nil {
			return errors.Wrap(err,
				"rto updateStateToWaitingForPaymentUseCase repository.FindByID")
		}

		if rto.State != api.RtoState_RTO_STATE_WAITING_FOR_STANDING_ORDER_SETUP {
			return errors.Wrap(fmt.Errorf("rto state must be in state %s, not in state %s",
				api.RtoState_RTO_STATE_WAITING_FOR_STANDING_ORDER_SETUP, rto.State),
				UpdateStateToWaitingForPaymentNotValidStateErrorText)
		}

		closeComSaveOpportunityRequest = requestservrto.CreateCloseComSaveOpportunityRequest(rto)
		rto.State = api.RtoState_RTO_STATE_WAITING_FOR_PAYMENT
		now := time.Now()
		rto.UpdatedAt = now

		err = ustauc.repository.MustUpdate(ctx, rto)
		if err != nil {
			return errors.Wrap(err,
				"rto updateStateToWaitingForPaymentUseCase repository.MustUpdate")
		}

		err = ustauc.historyRepository.Create(ctx, domain.CreateHistory(domain.HistoryActionUpdate, now, rto))
		if err != nil {
			return errors.Wrap(err,
				"rto updateStateToWaitingForPaymentUseCase historyRepository.Create")
		}

		err = ustauc.listingService.UpdateTenantID(ctx, &requestservice.ListingUpdateTenantIDRequest{
			PropertyID:   rto.Property,
			LeadTenantID: rto.GetLeadTenantID(),
			LandlordID:   rto.GetLandlordID(),
		})
		if err != nil {
			return errors.Wrap(err, "rto updateStateToWaitingForPaymentUseCase listingService.UpdateTenantID")
		}

		err := ustauc.listingService.UpdateVersionStatus(ctx, &requestservice.ListingUpdateVersionStatusRequest{
			VersionID:     rto.PropertyListing,
			VersionStatus: listing.VersionStatus_VERSION_RENT_COLLECTION,
		})
		if err != nil {
			return errors.Wrap(err, "rto updateStateToWaitingForPaymentUseCase listingService.UpdateVersionStatus")
		}

		return nil
	})
	if err != nil {
		return nil, errors.Wrap(err,
			"rto updateStateToWaitingForPaymentUseCase mongolocal.Transaction")
	}

	go func() {
		ctx := context.Background()
		_, err = ustauc.closeComService.SaveOpportunity(ctx, closeComSaveOpportunityRequest)
		if err != nil {
			log.WithContext(ctx).WithError(err).Error("rto updateStateToActiveUseCase closeComService.SaveOpportunity")
		}
	}()

	return &response.RtoResponse{
		Rto: rto,
	}, nil
}
