package rto

import (
	"context"
	"fmt"
	"strings"
	"time"

	"gitlab.mashplace.com/mashplace/backend/common/generic/log"

	"gitlab.mashplace.com/mashplace/backend/common/lib/mongolocal"
	api "gitlab.mashplace.com/mashplace/backend/offer/api/rto"
	domain "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto"
	requestservice "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto/resource/request/service"
	requestservrto "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto/resource/request/service/rto"
	request "gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto/resource/request/usecase"
	"gitlab.mashplace.com/mashplace/backend/offer/internal/app/domain/rto/resource/response"
	appservice "gitlab.mashplace.com/mashplace/backend/offer/internal/app/service"
	service "gitlab.mashplace.com/mashplace/backend/offer/internal/app/service/rto"

	"github.com/pkg/errors"
	"go.mongodb.org/mongo-driver/mongo"
)

const UpdateStateToWaitingForTenantConfirmationNotValidStateErrorText = "bad rto state for WaitingForTenantConfirmation"

type UpdateStateToWaitingForTenantConfirmationUseCase interface {
	UpdateStateToWaitingForTenantConfirmation(ctx context.Context,
		req *request.UpdateStateToWaitingForTenantConfirmationRequest) (*response.RtoResponse, error)
}

type updateStateToWaitingForTenantConfirmationUseCase struct {
	repository                  domain.Repository
	historyRepository           domain.HistoryRepository
	updateLeadTenantUserUseCase UpdateLeadTenantUserUseCase
	notificationService         service.NotificationService
	closeComService             service.CloseComService
	listingService              appservice.ListingService
}

func NewUpdateStateToWaitingForTenantConfirmationUseCase(repository domain.Repository,
	historyRepository domain.HistoryRepository,
	updateLeadTenantUserUseCase UpdateLeadTenantUserUseCase,
	notificationService service.NotificationService,
	closeComService service.CloseComService,
	listingService appservice.ListingService) UpdateStateToWaitingForTenantConfirmationUseCase {
	return &updateStateToWaitingForTenantConfirmationUseCase{
		repository:                  repository,
		historyRepository:           historyRepository,
		updateLeadTenantUserUseCase: updateLeadTenantUserUseCase,
		notificationService:         notificationService,
		closeComService:             closeComService,
		listingService:              listingService,
	}
}

func (ustwftcuc *updateStateToWaitingForTenantConfirmationUseCase) UpdateStateToWaitingForTenantConfirmation(
	ctx context.Context, req *request.UpdateStateToWaitingForTenantConfirmationRequest) (*response.RtoResponse, error) {
	err := req.Validate()
	if err != nil {
		return nil, errors.Wrap(err, "rto updateStateToWaitingForTenantConfirmationUseCase req.Validate")
	}

	rto := req.Rto
	if rto.State != api.RtoState_RTO_STATE_PENDING_SETUP {
		return nil, errors.Wrap(fmt.Errorf("rto state must be in state %s, not in state %s",
			api.RtoState_RTO_STATE_PENDING_SETUP, rto.State),
			UpdateStateToWaitingForTenantConfirmationNotValidStateErrorText)
	}

	closeComSaveOpportunityRequest := requestservrto.CreateCloseComSaveOpportunityRequest(rto)
	rto.State = api.RtoState_RTO_STATE_WAITING_FOR_TENANT_CONFIRMATION
	now := time.Now()
	rto.UpdatedAt = now

	err = mongolocal.Transaction(ctx, func() error {
		err = ustwftcuc.repository.MustUpdate(ctx, rto)
		if err != nil {
			return errors.Wrap(err,
				"rto updateStateToWaitingForTenantConfirmationUseCase repository.MustUpdate")
		}

		err = ustwftcuc.historyRepository.Create(ctx, domain.CreateHistory(domain.HistoryActionUpdate, now, rto))
		if err != nil {
			return errors.Wrap(err,
				"rto updateStateToWaitingForTenantConfirmationUseCase historyRepository.Create")
		}

		return nil
	})
	if err != nil {
		return nil, errors.Wrap(err,
			"rto updateStateToWaitingForTenantConfirmationUseCase mongolocal.Transaction")
	}

	leadTenantID := rto.GetLeadTenantID()
	if leadTenantID == "" {
		resp, err := ustwftcuc.updateLeadTenantUserUseCase.UpdateLeadTenantUser(ctx,
			&request.UpdateLeadTenantUserRequest{
				RtoID:               rto.ID.Hex(),
				LeadTenantEmail:     rto.GetLeadTenantEmail(),
				LeadTenantFirstName: rto.GetLeadTenantFirstName(),
				LeadTenantLastName:  rto.GetLeadTenantLastName(),
			})

		if err == nil {
			leadTenantID = resp.Rto.GetLeadTenantID()
		} else if !strings.Contains(err.Error(), mongo.ErrNoDocuments.Error()) {
			return nil, errors.Wrap(err,
				"rto updateStateToWaitingForTenantConfirmationUseCase updateLeadTenantUserUseCase")
		}
	} else {
		err = ustwftcuc.listingService.UpdateTenantID(ctx, &requestservice.ListingUpdateTenantIDRequest{
			PropertyID:   rto.Property,
			LeadTenantID: rto.GetLeadTenantID(),
			LandlordID:   rto.GetLandlordID(),
		})
		if err != nil {
			return nil, errors.Wrap(err, "rto updateStateToWaitingForTenantConfirmationUseCase "+
				"listingService.UpdateTenantID")
		}
	}

	err = ustwftcuc.notificationService.SendWaitingForStandingOrderSetup(ctx,
		&requestservrto.SendWaitingForStandingOrderSetupRequest{
			ID:          rto.ID.Hex(),
			LandlordID:  rto.Landlord.User,
			TenantID:    leadTenantID,
			TenantEmail: rto.GetLeadTenantEmail(),
			PropertyID:  rto.Property,
		})
	if err != nil {
		return nil, errors.Wrap(err, "rto updateStateToWaitingForTenantConfirmationUseCase "+
			"notificationService.SendWaitingForStandingOrderSetup")
	}

	go func() {
		ctx := context.Background()
		_, err = ustwftcuc.closeComService.SaveOpportunity(ctx, closeComSaveOpportunityRequest)
		if err != nil {
			log.WithContext(ctx).WithError(err).Error("rto updateStateToWaitingForTenantConfirmationUseCase closeComService.SaveOpportunity")
		}
	}()

	return &response.RtoResponse{
		Rto: rto,
	}, nil
}
