package mixpanel

import (
	"context"
	"fmt"
	"strings"
	"time"

	"gitlab.mashplace.com/mashplace/backend/common/api/grpcctx"

	"gitlab.mashplace.com/mashplace/backend/common/generic/util"

	mixpanel "gitlab.mashplace.com/mashplace/backend/mixpanel/api"
	offer "gitlab.mashplace.com/mashplace/backend/offer/api"
	"gitlab.mashplace.com/mashplace/backend/offer/internal/connections"
	"gitlab.mashplace.com/mashplace/backend/offer/model"
	property "gitlab.mashplace.com/mashplace/backend/property/api"
)

func paymentFrequencyString(frequency int32) string {
	switch frequency {
	case 1:
		return "Monthly"
	case 3:
		return "Quartely"
	case 6:
		return "Twice a year"
	case 12:
		return "Yearly"
	default:
		return fmt.Sprintf("Once in %d months", frequency)
	}
}

func depositTypeString(dt offer.DepositType) string {
	switch dt {
	case offer.DepositType_UNKNOWN_DEPOSIT_TYPE:
		return "Unspecified"
	case offer.DepositType_DEPOSIT_REPLACEMENT:
		return "Deposit replacement"
	default:
		return "Traditional"
	}
}

func employmentStatusString(st offer.EmploymentStatus) string {
	switch st {
	case offer.EmploymentStatus_EMPLOYED:
		return "Employed"
	case offer.EmploymentStatus_RETIRED:
		return "Retired"
	case offer.EmploymentStatus_STUDENT:
		return "Student"
	case offer.EmploymentStatus_UNEMPLOYED:
		return "Unemployed"
	default:
		return "Unspecified"
	}
}

func guarantorDetailsString(guarantors []*model.GuarantorOfferModel) string {
	var parts []string

	for _, g := range guarantors {
		parts = append(parts, fmt.Sprintf(
			"First name: %s\nLast name: %s\nAddress: %s\nEmail:%s",
			util.UnPtrString(g.FirstName),
			util.UnPtrString(g.LastName),
			util.UnPtrString(g.BillingAddress),
			util.UnPtrString(g.EmailPrimary),
		))
	}

	return strings.Join(parts, ", ")
}

// TrackLandlordPropertyRented helps to track PropertyRented mixpanel event for LANDLORD.
func TrackLandlordPropertyRented(ctx context.Context, data *model.OfferModel) error {
	switch {
	case data == nil:
		return fmt.Errorf("offer data model is nil")
	case data.Landlord == nil:
		return fmt.Errorf("offer landlord data is nil")
	case data.Conditions == nil:
		return fmt.Errorf("offer conditions model is nil")
	}

	prop, err := connections.PropertyService.Get(grpcctx.WithServiceUser(ctx), &property.GetRequest{
		Id:     data.Property,
		Verify: false,
	})
	if err != nil {
		return err
	}

	instant := prop.GetProperty().GetListing().GetStartedListingAt().AsTime()
	hours := int(time.Since(instant).Hours())
	condpb := data.Conditions.RenderOfferConditionsModel()
	_, err = connections.MixpanelService.TrackEvent(ctx, &mixpanel.TrackRequest{
		UserId: data.Landlord.User,
		Email:  util.UnPtrString(data.Landlord.EmailPrimary),
		Event: &mixpanel.TrackRequest_PropertyRented{
			PropertyRented: &mixpanel.Event_PropertyRented{
				PropertyId:              data.Property,
				MonthlyRent:             condpb.GetPrice().GetValue(),
				PaymentFrequency:        paymentFrequencyString(condpb.GetPaymentFrequency().GetValue()),
				DepositReplacement:      data.DepositReplacement(),
				DaysSinceFirstPublished: int32(hours / 24),
			},
		},
	})

	return err
}

// TrackTenantFirstRentPaymentCompleted helps to track FirstRentPaymentCompleted mixpanel event for TENANT.
func TrackTenantFirstRentPaymentCompleted(ctx context.Context, data *model.OfferModel) error {
	switch {
	case data == nil:
		return fmt.Errorf("offer data model is nil")
	case len(data.Tenants) == 0 || data.Tenants[0] == nil:
		return fmt.Errorf("offer tenants list is empty")
	case data.Conditions == nil:
		return fmt.Errorf("offer conditions model is nil")
	}

	condpb := data.Conditions.RenderOfferConditionsModel()
	_, err := connections.MixpanelService.TrackEvent(ctx, &mixpanel.TrackRequest{
		UserId: data.Tenants[0].User,
		Email:  util.UnPtrString(data.Tenants[0].EmailPrimary),
		Event: &mixpanel.TrackRequest_FirstRentPaymentCompleted{
			FirstRentPaymentCompleted: &mixpanel.Event_FirstRentPaymentCompleted{
				PropertyId:  data.Property,
				OfferId:     data.ID.Hex(),
				DepositType: depositTypeString(condpb.GetDepositType()),
			},
		},
	})

	return err
}

// TrackTenantOfferSubmitted helps to track OfferSubmitted mixpanel event for TENANT.
func TrackTenantOfferSubmitted(ctx context.Context, data *model.OfferModel) error {
	switch {
	case data == nil:
		return fmt.Errorf("offer data model is nil")
	case len(data.Tenants) == 0 || data.Tenants[0] == nil:
		return fmt.Errorf("offer tenants list is empty")
	case data.Conditions == nil:
		return fmt.Errorf("offer conditions model is nil")
	}

	var (
		details       []string
		adults, teens int32
		annualIncome  float64
	)

	for _, t := range data.Tenants {
		pb := t.RenderTenantOfferModel()
		details = append(details, fmt.Sprintf(
			"First name: %s, Last name: %s, Email: %s",
			pb.GetFirstName().GetValue(),
			pb.GetLastName().GetValue(),
			pb.GetEmailPrimary(),
		))

		if pb.GetAge().GetValue() >= 18 {
			adults++
		} else {
			teens++
		}

		annualIncome += pb.GetAnnualIncome().GetValue()
	}

	condpb := data.Conditions.RenderOfferConditionsModel()
	tenantpb := data.Tenants[0].RenderTenantOfferModel()
	_, err := connections.MixpanelService.TrackEvent(ctx, &mixpanel.TrackRequest{
		UserId: data.Tenants[0].User,
		Email:  util.UnPtrString(data.Tenants[0].EmailPrimary),
		Event: &mixpanel.TrackRequest_OfferSubmitted{
			OfferSubmitted: &mixpanel.Event_OfferSubmitted{
				OfferId:                data.ID.Hex(),
				PropertyId:             data.Property,
				MoveinDate:             condpb.GetMoveinDate(),
				MonthlyRent:            condpb.GetPrice().GetValue(),
				PaymentFrequency:       paymentFrequencyString(condpb.GetPaymentFrequency().GetValue()),
				DepositType:            depositTypeString(condpb.GetDepositType()),
				TenancyLength:          condpb.GetTermMonths().GetValue(),
				BreakClauseLength:      condpb.GetBreakClause().GetValue(),
				TenantCurrentAddress:   tenantpb.GetBillingAddress().GetValue(),
				TenantAge:              tenantpb.GetAge().GetValue(),
				TenantEmploymentStatus: employmentStatusString(tenantpb.GetEmploymentStatus()),
				OtherTenantsDetails:    strings.Join(details, "\n"),
				GuarantorDetails:       guarantorDetailsString(data.Guarantors),
				AdultTenantsNum:        adults,
				TeenTenantsNum:         teens,
				TenantsHavePets:        condpb.GetPets().GetValue(),
				FinalPitch:             condpb.GetTenantsPitch().GetValue(),
				TotalAnnualIncome:      annualIncome,
			},
		},
	})

	return err
}

// TrackLandlordOfferReceived helps to track OfferReceived mixpanel event for LANDLORD.
func TrackLandlordOfferReceived(ctx context.Context, data *model.OfferModel) error {
	switch {
	case data == nil:
		return fmt.Errorf("offer data model is nil")
	case data.Landlord == nil:
		return fmt.Errorf("offer landlord data is nil")
	case data.Conditions == nil:
		return fmt.Errorf("offer conditions model is nil")
	case len(data.Tenants) == 0:
		return fmt.Errorf("offer tenants list is empty")
	}

	prop, err := connections.PropertyService.Get(grpcctx.WithServiceUser(ctx), &property.GetRequest{
		Id:     data.Property,
		Verify: false,
	})
	if err != nil {
		return err
	}

	var (
		adults       int32
		annualIncome float64
	)

	for _, t := range data.Tenants {
		pb := t.RenderTenantOfferModel()

		if pb.GetAge().GetValue() >= 18 {
			adults++
		}

		annualIncome += pb.GetAnnualIncome().GetValue()
	}

	condpb := data.Conditions.RenderOfferConditionsModel()

	_, err = connections.MixpanelService.TrackEvent(ctx, &mixpanel.TrackRequest{
		UserId: data.Landlord.User,
		Email:  util.UnPtrString(data.Landlord.EmailPrimary),
		Event: &mixpanel.TrackRequest_OfferReceived{
			OfferReceived: &mixpanel.Event_OfferReceived{
				OfferId:           data.ID.Hex(),
				PropertyId:        data.Property,
				MonthlyRent:       condpb.GetPrice().GetValue(),
				ListedPrice:       prop.GetProperty().GetRental().GetRentalPrice().GetValue(),
				PaymentFrequency:  paymentFrequencyString(condpb.GetPaymentFrequency().GetValue()),
				DepositType:       depositTypeString(condpb.GetDepositType()),
				AdultTenantsNum:   adults,
				TotalAnnualIncome: annualIncome,
				GuarantorAdded:    len(data.Guarantors) > 0,
			},
		},
	})

	return err
}

// TrackOfferAccepted helps to track OfferAccepted mixpanel event for LANDLORD.
func TrackOfferAccepted(ctx context.Context, userID, email string, data *model.OfferModel) error {
	switch {
	case data == nil:
		return fmt.Errorf("offer data model is nil")
	case data.Landlord == nil:
		return fmt.Errorf("offer landlord data is nil")
	case userID != data.Landlord.User:
		return fmt.Errorf("user %q is not landlord", userID)
	case data.Conditions == nil:
		return fmt.Errorf("offer conditions model is nil")
	case len(data.Tenants) == 0:
		return fmt.Errorf("offer tenants list is empty")
	}

	prop, err := connections.PropertyService.Get(grpcctx.WithServiceUser(ctx), &property.GetRequest{
		Id:     data.Property,
		Verify: false,
	})
	if err != nil {
		return err
	}

	var (
		adults       int32
		annualIncome float64
	)

	for _, t := range data.Tenants {
		pb := t.RenderTenantOfferModel()

		if pb.GetAge().GetValue() >= 18 {
			adults++
		}

		annualIncome += pb.GetAnnualIncome().GetValue()
	}

	condpb := data.Conditions.RenderOfferConditionsModel()

	_, err = connections.MixpanelService.TrackEvent(ctx, &mixpanel.TrackRequest{
		UserId: userID,
		Email:  email,
		Event: &mixpanel.TrackRequest_OfferAccepted{
			OfferAccepted: &mixpanel.Event_OfferAccepted{
				OfferId:           data.ID.Hex(),
				PropertyId:        data.Property,
				MonthlyRent:       condpb.GetPrice().GetValue(),
				ListedPrice:       prop.GetProperty().GetRental().GetRentalPrice().GetValue(),
				PaymentFrequency:  paymentFrequencyString(condpb.GetPaymentFrequency().GetValue()),
				DepositType:       depositTypeString(condpb.GetDepositType()),
				AdultTenantsNum:   adults,
				TotalAnnualIncome: annualIncome,
				GuarantorAdded:    len(data.Guarantors) > 0,
			},
		},
	})

	return err
}

// TrackOfferUpdated helps to track OfferUpdated mixpanel event for TENANT.
func TrackOfferUpdated(ctx context.Context, userID, email string, data *model.OfferModel) error {
	switch {
	case data == nil:
		return fmt.Errorf("offer data model is nil")
	case len(data.Tenants) == 0 || data.Tenants[0] == nil:
		return fmt.Errorf("offer tenants list is empty")
	case userID != data.Tenants[0].User:
		return fmt.Errorf("user %q is not a tennant", userID)
	case data.Conditions == nil:
		return fmt.Errorf("offer conditions model is nil")
	}

	prop, err := connections.PropertyService.Get(grpcctx.WithServiceUser(ctx), &property.GetRequest{
		Id:     data.Property,
		Verify: false,
	})
	if err != nil {
		return err
	}

	condpb := data.Conditions.RenderOfferConditionsModel()
	_, err = connections.MixpanelService.TrackEvent(ctx, &mixpanel.TrackRequest{
		UserId: userID,
		Email:  email,
		Event: &mixpanel.TrackRequest_OfferUpdated{
			OfferUpdated: &mixpanel.Event_OfferUpdated{
				OfferId:     data.ID.Hex(),
				PropertyId:  data.Property,
				MonthlyRent: condpb.GetPrice().GetValue(),
				ListedPrice: prop.GetProperty().GetRental().GetRentalPrice().GetValue(),
			},
		},
	})

	return err
}

// TrackOfferDeclined helps to track OfferDeclined mixpanel event for LANDLORD.
func TrackOfferDeclined(ctx context.Context, userID, email string, data *model.OfferModel) error {
	switch {
	case data == nil:
		return fmt.Errorf("offer data model is nil")
	case data.Landlord == nil:
		return fmt.Errorf("offer landlord data is nil")
	case userID != data.Landlord.User:
		return fmt.Errorf("user %q is not landlord", userID)
	case data.Conditions == nil:
		return fmt.Errorf("offer conditions model is nil")
	}

	prop, err := connections.PropertyService.Get(grpcctx.WithServiceUser(ctx), &property.GetRequest{
		Id:     data.Property,
		Verify: false,
	})
	if err != nil {
		return err
	}

	condpb := data.Conditions.RenderOfferConditionsModel()
	_, err = connections.MixpanelService.TrackEvent(ctx, &mixpanel.TrackRequest{
		UserId: userID,
		Email:  email,
		Event: &mixpanel.TrackRequest_OfferDeclined{
			OfferDeclined: &mixpanel.Event_OfferDeclined{
				OfferId:     data.ID.Hex(),
				PropertyId:  data.Property,
				MonthlyRent: condpb.GetPrice().GetValue(),
				ListedPrice: prop.GetProperty().GetRental().GetRentalPrice().GetValue(),
			},
		},
	})

	return err
}
