package hourly

import (
	"context"
	"fmt"
	"time"

	"gitlab.mashplace.com/mashplace/backend/common/api/grpcctx"

	messaging "gitlab.mashplace.com/mashplace/backend/messaging/api"
	notification "gitlab.mashplace.com/mashplace/backend/notification/api"
	offer "gitlab.mashplace.com/mashplace/backend/offer/api"
	"gitlab.mashplace.com/mashplace/backend/offer/internal/connections"
	"gitlab.mashplace.com/mashplace/backend/offer/model"
	"go.mongodb.org/mongo-driver/bson"
	"google.golang.org/protobuf/types/known/durationpb"
)

func consumeReminderHourlyNotificationsNotResponded3hrs(ctx context.Context) error {
	now := time.Now()
	models, err := model.ListNotNotified(
		ctx,
		[]offer.OfferState{offer.OfferState_ACTIVE},
		[]model.ScheduleNotificationType{model.ScheduleNotificationNotResponded3hrs},
		now,
		now.Add(-3*time.Hour),
	)
	if err != nil {
		return err
	}

	for _, m := range models {
		s := m.ScheduleNotifications[model.ScheduleNotificationNotResponded3hrs]
		if s == nil || s.IsNotified {
			continue
		}

		lctx := ctx

		if m.Test {
			lctx = context.WithValue(lctx, grpcctx.CtxIsTestUser, true)
		}

		_, err = connections.NotificationService.SendNotification(lctx, &notification.SendNotificationRequest{
			Entity: notification.Entity_ENTITY_OFFER,
			Action: notification.Action_ACTION_PENDING,
			Targets: []*notification.Target{
				{
					User: m.LandlordUser(),
					Role: notification.Role_ROLE_LANDLORD,
				},
			},
			Duration: durationpb.New(time.Hour * 3),
			Embeds: []*notification.EventEmbed{
				{
					EmbedType:  messaging.EmbedType_EMBED_TYPE_OFFER,
					EmbedValue: m.ID.Hex(),
				},
				{
					EmbedType:  messaging.EmbedType_EMBED_TYPE_PROPERTY,
					EmbedValue: m.Property,
				},
			},
		})
		if err != nil {
			continue
		}

		key := fmt.Sprintf("schedule_notifications.%d.is_notified", model.ScheduleNotificationNotResponded3hrs)

		_, _ = model.Offers.UpdateOne(lctx, bson.M{"_id": m.ID}, bson.M{"$set": bson.M{key: true}})
	}

	return nil
}

func consumeReminderHourlyNotificationsWeekLeftToMovein(ctx context.Context) (err error) {
	now := time.Now()
	models, err := model.ListNotNotified(
		ctx,
		[]offer.OfferState{offer.OfferState_ACTIVE},
		[]model.ScheduleNotificationType{model.ScheduleNotificationWeekLeftToMovein},
		now,
		now.Add(time.Hour*24*7),
	)
	if err != nil {
		return err
	}

	for _, m := range models {
		s := m.ScheduleNotifications[model.ScheduleNotificationWeekLeftToMovein]
		if s == nil || s.IsNotified {
			continue
		}

		lctx := ctx

		if m.Test {
			lctx = context.WithValue(lctx, grpcctx.CtxIsTestUser, true)
		}

		_, err = connections.NotificationService.SendNotification(lctx, &notification.SendNotificationRequest{
			Entity: notification.Entity_ENTITY_OFFER,
			Action: notification.Action_ACTION_REMIND,
			Targets: []*notification.Target{
				{
					User: m.LandlordUser(),
					Role: notification.Role_ROLE_LANDLORD,
				},
				{
					User: m.TenantUser(),
					Role: notification.Role_ROLE_TENANT,
				},
			},
			Duration: durationpb.New(time.Hour * 24 * 7),
			Embeds: []*notification.EventEmbed{
				{
					EmbedType:  messaging.EmbedType_EMBED_TYPE_OFFER,
					EmbedValue: m.ID.Hex(),
				},
				{
					EmbedType:  messaging.EmbedType_EMBED_TYPE_PROPERTY,
					EmbedValue: m.Property,
				},
			},
		})
		if err != nil {
			continue
		}

		key := fmt.Sprintf("schedule_notifications.%d.is_notified", model.ScheduleNotificationWeekLeftToMovein)

		_, _ = model.Offers.UpdateOne(lctx, bson.M{"_id": m.ID}, bson.M{"$set": bson.M{key: true}})
	}

	return nil
}
