// Package oneoff provides handlers for scheduled one-off operations
package oneoff

import (
	"context"
	"encoding/json"
	"time"

	"gitlab.mashplace.com/mashplace/backend/common/generic/log"
	"gitlab.mashplace.com/mashplace/backend/offer/internal/connections"
	schedule "gitlab.mashplace.com/mashplace/backend/schedule/api"
	"google.golang.org/protobuf/types/known/timestamppb"
)

// InvocationType identifies type of the oneoff operations
type InvocationType string

// Known oneoff types
const (
	InvocationTypeSignatureProvided   InvocationType = "signature-provided"
	InvocationTypeCheckHoldingDepsoit InvocationType = "check-holding-deposit"
)

// Invocation wrapper
type Invocation struct {
	Type InvocationType  `json:"type,omitempty"`
	Data json.RawMessage `json:"data,omitempty"`
}

// ScheduleInvocation schedules a oneoff invocation
func ScheduleInvocation(ctx context.Context, t time.Time, typ InvocationType, data interface{}) error {
	bs, err := json.Marshal(data)
	if err != nil {
		return err
	}

	bs, err = json.Marshal(&Invocation{
		Type: typ,
		Data: bs,
	})
	if err != nil {
		return err
	}

	_, err = connections.ScheduleService.ReminderSchedule(ctx, &schedule.Reminder_ScheduleRequest{
		ConsumerUrl: connections.SelfRemote,
		Instant:     timestamppb.New(t),
		Reminder: &schedule.Reminder{
			Data: bs,
		},
		Retries: 1,
	})

	return err
}

// ConsumeReminderInvocation handles triggered invocation
func ConsumeReminderInvocation(
	ctx context.Context,
	req *schedule.Reminder_ConsumeRequest,
) {
	fields := log.Fields{
		"invocation_id": req.GetInvocationId(),
		"task_id":       req.GetTaskId(),
	}

	var inv Invocation

	err := json.Unmarshal(req.GetReminder().GetData(), &inv)
	if err != nil {
		log.WithContext(ctx).WithError(err).WithFields(fields).Error("unmarshalling invocation")

		return
	}

	fields["type"] = inv.Type

	switch inv.Type {
	case InvocationTypeSignatureProvided:
		err = consumeReminderInvocationSignatureProvided(ctx, &inv)
	case InvocationTypeCheckHoldingDepsoit:
		err = consumeReminderInvocationCheckHoldingDeposit(ctx, &inv)
	}

	if err != nil {
		log.WithContext(ctx).WithError(err).WithFields(fields).Error("processing invocation")
	}
}
