package oneoff

import (
	"context"
	"encoding/json"
	"time"

	"gitlab.mashplace.com/mashplace/backend/common/api/grpcctx"

	messaging "gitlab.mashplace.com/mashplace/backend/messaging/api"
	notification "gitlab.mashplace.com/mashplace/backend/notification/api"
	offer "gitlab.mashplace.com/mashplace/backend/offer/api"
	"gitlab.mashplace.com/mashplace/backend/offer/internal/connections"
	"gitlab.mashplace.com/mashplace/backend/offer/model"
	"go.mongodb.org/mongo-driver/bson/primitive"
	"google.golang.org/protobuf/types/known/durationpb"
)

// InvocationSignatureProvided holds data for signature check invocation
type InvocationSignatureProvided struct {
	OfferID string            `json:"offer_id"`
	Actor   notification.Role `json:"role"`
}

func consumeReminderInvocationSignatureProvided(
	ctx context.Context,
	inv *Invocation,
) (err error) {
	var signcheck InvocationSignatureProvided

	err = json.Unmarshal(inv.Data, &signcheck)
	if err != nil {
		return
	}

	id, err := primitive.ObjectIDFromHex(signcheck.OfferID)
	if err != nil {
		return
	}

	off, err := model.GetOfferByID(ctx, id)
	if err != nil {
		return
	}

	if off.Test {
		ctx = context.WithValue(ctx, grpcctx.CtxIsTestUser, true)
	}

	var list []*model.SignatureModel

	var target notification.Role

	var actorID, targetID string

	switch signcheck.Actor {
	case notification.Role_ROLE_TENANT:
		list = off.LandlordModel().SignatureList()
		actorID = off.TenantUser()
		targetID = off.LandlordUser()
		target = notification.Role_ROLE_LANDLORD
	case notification.Role_ROLE_LANDLORD:
		list = off.TenantModel().SignatureList()
		actorID = off.LandlordUser()
		targetID = off.TenantUser()
		target = notification.Role_ROLE_TENANT
	}

	for _, item := range list {
		if item.State == offer.SignStatusState_SIGNATURE_PROVIDED {
			return
		}
	}

	_, err = connections.NotificationService.SendNotification(ctx, &notification.SendNotificationRequest{
		Actor: &notification.Actor{
			User: actorID,
			Role: signcheck.Actor,
		},
		Entity: notification.Entity_ENTITY_OFFER_SIGNATURE_INVITATION,
		Action: notification.Action_ACTION_PENDING,
		Targets: []*notification.Target{
			{
				User: targetID,
				Role: target,
			},
		},
		Duration: durationpb.New(time.Hour * 2),
		Embeds: []*notification.EventEmbed{
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_PROPERTY,
				EmbedValue: off.Property,
			},
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_OFFER,
				EmbedValue: off.ID.Hex(),
			},
		},
	})

	return err
}
