package status

import (
	"context"
	"errors"
	"time"

	account "gitlab.mashplace.com/mashplace/backend/account/api"
	closecom "gitlab.mashplace.com/mashplace/backend/closecom/api"
	"gitlab.mashplace.com/mashplace/backend/common/api/grpcctx"
	"gitlab.mashplace.com/mashplace/backend/common/api/grpcerr"
	"gitlab.mashplace.com/mashplace/backend/common/generic/bind"
	"gitlab.mashplace.com/mashplace/backend/common/generic/log"
	"gitlab.mashplace.com/mashplace/backend/common/generic/util"
	document "gitlab.mashplace.com/mashplace/backend/document/api"
	extras "gitlab.mashplace.com/mashplace/backend/extras/api"
	invoice "gitlab.mashplace.com/mashplace/backend/invoice/api"
	listing "gitlab.mashplace.com/mashplace/backend/listing/api"
	messaging "gitlab.mashplace.com/mashplace/backend/messaging/api"
	notification "gitlab.mashplace.com/mashplace/backend/notification/api"
	offer "gitlab.mashplace.com/mashplace/backend/offer/api"
	usecaserto "gitlab.mashplace.com/mashplace/backend/offer/internal/app/usecase/rto"
	"gitlab.mashplace.com/mashplace/backend/offer/internal/connections"
	"gitlab.mashplace.com/mashplace/backend/offer/internal/schedule/oneoff"
	"gitlab.mashplace.com/mashplace/backend/offer/internal/synchronization"
	"gitlab.mashplace.com/mashplace/backend/offer/model"
	property "gitlab.mashplace.com/mashplace/backend/property/api"
	viewing "gitlab.mashplace.com/mashplace/backend/viewing/api"
	"go.mongodb.org/mongo-driver/mongo"

	"go.mongodb.org/mongo-driver/bson"
	"go.mongodb.org/mongo-driver/bson/primitive"
)

// BPMTimeToLive TTL for BPM events
const BPMTimeToLive = time.Hour

// errors
var (
	ErrInvalidState        = grpcerr.New(grpcerr.InvalidState, "invalid state")
	ErrNoAccess            = grpcerr.New(grpcerr.AuthError, "no access")
	ErrPaidDepositNotFound = grpcerr.New(grpcerr.InvalidData, "paid deposit not found")
)

var UpdateStateToActiveUseCase usecaserto.UpdateStateToActiveUseCase

// SetOfferState handles offer state transition
func SetOfferState(ctx context.Context, m *model.OfferModel, state offer.OfferState) (err error) {
	if m.Test {
		ctx = context.WithValue(ctx, grpcctx.CtxIsTestUser, true)
	}

	ctx = grpcctx.ValuesBackgroundContext(ctx)

	m.StateChangedAt = time.Now()

	switch state {
	case offer.OfferState_ACTIVE:
		err = setOfferStateActive(ctx, m)
	case offer.OfferState_ACCEPTED:
		err = setOfferStateAccepted(ctx, m)
	case offer.OfferState_DENIED:
		err = setOfferStateDenied(ctx, m)
	case offer.OfferState_CANCELED:
		err = setOfferStateCanceled(ctx, m)
	case offer.OfferState_HOLDING_DEPOSIT:
		err = setOfferStateHoldingDeposit(ctx, m)
	case offer.OfferState_AWAITING_CONFIRM:
		err = setOfferStateConfirmingHoldingDeposit(ctx, m)
	case offer.OfferState_DEPOSIT_CONFIRMED:
		err = setOfferStateDepositConfirmed(ctx, m)
	case offer.OfferState_AWAITING_APPLICANT_CHECK_PAYMENT:
		err = setOfferStateApplicantCheckPayment(ctx, m)
	case offer.OfferState_CHECKING:
		err = setOfferStateChecking(ctx, m)
	case offer.OfferState_CONTRACT_REVIEW_AND_CLAUSES:
		err = setOfferStateContractReviewAndClauses(ctx, m)
	case offer.OfferState_SIGNING:
		err = setOfferStateSigning(ctx, m)
	case offer.OfferState_SIGNED:
		if m.AwaitingTermination == "" || TenancyTerminated(ctx, m.AwaitingTermination) {
			m.State = offer.OfferState_SIGNED

			err = setOfferStateSigned(ctx, m)

			if m.State == offer.OfferState_RENTPAID && err == nil {
				state = offer.OfferState_RENTPAID
			}
		} else {
			m.State = offer.OfferState_AWAITING_TENANCY
			id, er := primitive.ObjectIDFromHex(m.AwaitingTermination)

			if er != nil {
				return er
			}

			oldoff, erro := model.GetOfferByID(ctx, id)

			if erro != nil {
				return erro
			}

			if m.OriginalOffer == oldoff.ID.Hex() && oldoff.RentPlan != offer.RentPlan_RENT_PLAN_LEGACY {
				m.RentPlan = oldoff.RentPlan
			}

			state = offer.OfferState_AWAITING_TENANCY
			err = setOfferStateSigned(ctx, m)
		}
	case offer.OfferState_RENTPAID:
		err = setOfferStateRentpaid(ctx, m)
	case offer.OfferState_UNKNOWN_OFFER_STATE:
		err = ErrInvalidState
	}

	if err != nil {
		return err
	}

	m.State = state

	m.Sync.Reset()

	log.WithContext(ctx).Infof("Set %s offer state to %s \n", m.ID.Hex(), m.State.String())

	err = model.UpdateOffer(ctx, m)
	if err != nil {
		return err
	}

	synchronization.ServiceOfferSync(ctx, m)

	return nil
}

// TenancyTerminated checks if related offer is already terminated
func TenancyTerminated(ctx context.Context, oid string) bool {
	id, err := primitive.ObjectIDFromHex(oid)
	if err != nil {
		return false
	}

	oldoff, err := model.GetOfferByID(ctx, id)
	if err != nil {
		return false
	}

	if oldoff.Conditions == nil {
		return false
	}

	months := util.UnPtrInt32(oldoff.Conditions.TermMonths)

	return time.Since(util.UnPtrTime(oldoff.Conditions.MoveinDate).AddDate(0, int(months), 0)) > 0
}

func setOfferStateActive(ctx context.Context, m *model.OfferModel) error {
	go func() {
		ctx := context.Background()
		logger := log.WithContext(ctx).WithFields(log.Fields{
			"offer": m.ID.Hex(),
		})

		_, err := connections.NotificationService.SendNotification(
			ctx,
			&notification.SendNotificationRequest{
				Entity: notification.Entity_ENTITY_OFFER,
				Action: notification.Action_ACTION_CREATE,
				Targets: []*notification.Target{
					{
						User: m.LandlordUser(),
						Role: notification.Role_ROLE_LANDLORD,
					},
					{
						User: m.TenantUser(),
						Role: notification.Role_ROLE_TENANT,
					},
				},
				Embeds: []*notification.EventEmbed{
					{
						EmbedType:  messaging.EmbedType_EMBED_TYPE_PROPERTY,
						EmbedValue: m.Property,
					},
					{
						EmbedType:  messaging.EmbedType_EMBED_TYPE_OFFER,
						EmbedValue: m.ID.Hex(),
					},
					{
						EmbedType:  messaging.EmbedType_EMBED_TYPE_USER,
						EmbedValue: m.TenantUser(),
					},
				},
			},
		)
		if err != nil {
			logger.WithContext(ctx).WithError(err).Error("sending notification to landlord failed")
		}
	}()

	return nil
}

func setOfferStateAccepted(ctx context.Context, m *model.OfferModel) error {
	m.AcceptedAt = time.Now().UTC()

	ExportClosecomOpportunity(ctx, m, closecom.OfferStatus_OFFER_STATUS_AWAITING_HOLDING_DEPOSIT)

	return nil
}

func revertOriginalOffer(ctx context.Context, m *model.OfferModel) (err error) {
	if m.OriginalOffer == "" {
		return nil
	}

	origid, err := primitive.ObjectIDFromHex(m.OriginalOffer)
	if err != nil {
		return err
	}

	_, err = model.Offers.UpdateOne(
		ctx,
		bson.M{
			"_id":   origid,
			"state": offer.OfferState_AWAITING_RENEW,
		},
		bson.M{
			"$set": bson.M{
				"state": offer.OfferState_RENTPAID,
			},
		},
	)

	return err
}

func setOfferStateDenied(ctx context.Context, m *model.OfferModel) (err error) {
	prop, err := connections.PropertyService.Get(grpcctx.WithServiceUser(ctx), &property.GetRequest{Id: m.Property})
	if err != nil {
		return err
	}

	if m.OpportunityID != "" {
		ExportClosecomOpportunity(ctx, m, closecom.OfferStatus_OFFER_STATUS_CANCELLED)
	}

	switch {
	case m.State == offer.OfferState_HOLDING_DEPOSIT:
		err = setOfferStateDeniedAtHoldingDeposit(ctx, m)
	case (m.State == offer.OfferState_ACTIVE || m.State == offer.OfferState_ACCEPTED ||
		m.State == offer.OfferState_PENDING) &&
		prop.GetProperty().GetStatus() == property.PropertyStatus_OFFERING:
		err = setOfferStateDeniedAtActiveOffering(ctx, m)
	}

	if err != nil {
		return err
	}

	_, _ = connections.ViewingService.ViewingStatus(
		grpcctx.WithServiceUser(ctx),
		&viewing.ViewingStatusRequest{
			Spec: &viewing.ViewingSpec{
				Select: &viewing.ViewingSpecSelect{
					ViewingId: m.Viewing,
				},
			},
			Transition: viewing.ViewingState_VIEWING_STATE_COMPLETED,
		},
	)

	m.DeniedAt = time.Now().UTC()

	// Cancel pending holding deposit invoice
	if m.State == offer.OfferState_HOLDING_DEPOSIT || m.State == offer.OfferState_AWAITING_CONFIRM {
		_, err := connections.InvoiceService.InvoiceCancel(ctx, &invoice.InvoiceCancelRequest{
			CorrelationType:   invoice.CorrelationTypeOffer,
			CorrelationId:     m.ID.Hex(),
			CorrelationTarget: invoice.CorrelationTargetHoldingDeposit,
			Type:              invoice.InvoiceType_INVOICE_TYPE_PAYMENT,
			User:              m.Tenants[0].User,
		})
		if err != nil && !errors.Is(err, grpcerr.NotFound) {
			return err
		}
	}

	err = revertOriginalOffer(ctx, m)
	if err != nil {
		return err
	}

	return nil
}

func setOfferStateDeniedAtHoldingDeposit(ctx context.Context, m *model.OfferModel) (err error) {
	invs, err := connections.InvoiceService.InvoiceList(ctx, &invoice.InvoiceListRequest{
		CorrelationType: invoice.CorrelationTypeOffer,
		CorrelationId:   m.ID.Hex(),
		User:            m.Tenants[0].User,
		Statuses: []invoice.InvoiceStatus{
			invoice.InvoiceStatus_INVOICE_OPEN,
		},
	})
	if err != nil ||
		len(invs.GetInvoices()) == 0 ||
		time.Now().Unix()-invs.GetInvoices()[0].GetCreatedAt().Seconds < 86380 {
		return err
	}

	_, err = connections.NotificationService.SendNotification(ctx, &notification.SendNotificationRequest{
		Entity: notification.Entity_ENTITY_OFFER_HOLDING_DEPOSIT_PAYMENT,
		Action: notification.Action_ACTION_EXPIRE,
		Targets: []*notification.Target{
			{
				User: m.LandlordUser(),
				Role: notification.Role_ROLE_LANDLORD,
			},
			{
				User: m.TenantUser(),
				Role: notification.Role_ROLE_TENANT,
			},
		},
		Embeds: []*notification.EventEmbed{
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_PROPERTY,
				EmbedValue: m.Property,
			},
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_OFFER,
				EmbedValue: m.ID.Hex(),
			},
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_USER,
				EmbedValue: m.TenantUser(),
			},
		},
	})
	if err != nil {
		log.WithContext(ctx).WithError(err).Error("sending offer holding deposit not paid to tenant")
	}

	ExportClosecomOpportunity(ctx, m, closecom.OfferStatus_OFFER_STATUS_CANCELLED)

	return nil
}

func setOfferStateDeniedAtActiveOffering(ctx context.Context, m *model.OfferModel) (err error) {
	_, err = connections.NotificationService.SendNotification(ctx, &notification.SendNotificationRequest{
		Entity: notification.Entity_ENTITY_OFFER,
		Action: notification.Action_ACTION_REJECT,
		Targets: []*notification.Target{
			{
				User: m.TenantUser(),
				Role: notification.Role_ROLE_TENANT,
			},
		},
		Embeds: []*notification.EventEmbed{
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_PROPERTY,
				EmbedValue: m.Property,
			},
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_OFFER,
				EmbedValue: m.ID.Hex(),
			},
		},
	})
	if err != nil {
		log.WithContext(ctx).WithError(err).Error("sending offer landlord accepted other tenant to tenant")
	}

	ExportClosecomOpportunity(ctx, m, closecom.OfferStatus_OFFER_STATUS_CANCELLED)

	return nil
}

func setOfferStateCanceled(ctx context.Context, m *model.OfferModel) error {
	_, _ = connections.ViewingService.ViewingStatus(
		grpcctx.WithServiceUser(ctx),
		&viewing.ViewingStatusRequest{
			Spec: &viewing.ViewingSpec{
				Select: &viewing.ViewingSpecSelect{
					ViewingId: m.Viewing,
				},
			},
			Transition: viewing.ViewingState_VIEWING_STATE_COMPLETED,
		},
	)

	if m.OpportunityID != "" {
		ExportClosecomOpportunity(ctx, m, closecom.OfferStatus_OFFER_STATUS_CANCELLED)
	}

	m.CancelledAt = time.Now().UTC()

	// Cancel pending holding deposit invoice
	if m.State == offer.OfferState_HOLDING_DEPOSIT || m.State == offer.OfferState_AWAITING_CONFIRM {
		_, err := connections.InvoiceService.InvoiceCancel(ctx, &invoice.InvoiceCancelRequest{
			CorrelationType:   invoice.CorrelationTypeOffer,
			CorrelationId:     m.ID.Hex(),
			CorrelationTarget: invoice.CorrelationTargetHoldingDeposit,
			Type:              invoice.InvoiceType_INVOICE_TYPE_PAYMENT,
			User:              m.Tenants[0].User,
		})
		if err != nil && !errors.Is(err, grpcerr.NotFound) {
			return err
		}
	}

	err := revertOriginalOffer(ctx, m)
	if err != nil {
		return err
	}

	return nil
}

// OfferCreateHoldingDepositInvoice creates holding deposit invoice
func OfferCreateHoldingDepositInvoice(ctx context.Context, m *model.OfferModel) error {
	price := util.UnPtrFloat64(m.Conditions.Price) * 12.0 / 52.0

	service, err := extras.GetServicePriceSnapshot(
		ctx,
		connections.ExtrasService,
		extras.ProductTenantHoldingDeposit,
		price,
	)
	if err != nil {
		return err
	}

	_, err = connections.InvoiceService.InvoiceCreate(ctx, &invoice.InvoiceCreateRequest{
		CorrelationType:   invoice.CorrelationTypeOffer,
		CorrelationId:     m.ID.Hex(),
		CorrelationTarget: invoice.CorrelationTargetHoldingDeposit,
		Property:          m.Property,
		Ivr:               m.PropertyIvr,
		Listing:           m.PropertyListing,
		User:              m.Tenants[0].User,
		Services: []*extras.ServiceSnapshot{
			service,
		},
		NotifyCallback: connections.SelfRemote,
	})

	return err
}

func setOfferStateHoldingDeposit(ctx context.Context, m *model.OfferModel) error {
	err := OfferCreateHoldingDepositInvoice(ctx, m)
	if err != nil {
		return err
	}

	scheduleReminder := func(expiry time.Duration) {
		err = oneoff.ScheduleInvocation(
			ctx,
			time.Now().Add(expiry),
			oneoff.InvocationTypeCheckHoldingDepsoit,
			&oneoff.InvocationCheckHoldingDeposit{
				OfferID: m.ID.Hex(),
				Expiry:  expiry,
			},
		)
		if err != nil {
			log.WithFields(log.Fields{
				"offer_id": m.ID.Hex(),
				"expiry":   expiry.String(),
			}).WithError(err).
				Errorf("scheduling failed")
		}
	}

	scheduleReminder(12 * time.Hour)

	_, err = connections.NotificationService.SendNotification(ctx, &notification.SendNotificationRequest{
		Entity: notification.Entity_ENTITY_OFFER,
		Action: notification.Action_ACTION_ACCEPT,
		Targets: []*notification.Target{
			{
				User: m.LandlordUser(),
				Role: notification.Role_ROLE_LANDLORD,
			},
		},
		Embeds: []*notification.EventEmbed{
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_PROPERTY,
				EmbedValue: m.Property,
			},
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_OFFER,
				EmbedValue: m.ID.Hex(),
			},
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_USER,
				EmbedValue: m.TenantUser(),
			},
		},
	})
	if err != nil {
		log.WithContext(ctx).WithError(err).Error("sending offer accept notification")
	}

	_, err = connections.NotificationService.SendNotification(ctx, &notification.SendNotificationRequest{
		Entity: notification.Entity_ENTITY_OFFER,
		Action: notification.Action_ACTION_ACCEPT,
		Targets: []*notification.Target{
			{
				User: m.TenantUser(),
				Role: notification.Role_ROLE_TENANT,
			},
		},
		Embeds: []*notification.EventEmbed{
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_PROPERTY,
				EmbedValue: m.Property,
			},
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_OFFER,
				EmbedValue: m.ID.Hex(),
			},
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_USER,
				EmbedValue: m.LandlordUser(),
			},
		},
	})
	if err != nil {
		log.WithContext(ctx).WithError(err).Error("sending offer accept notification")
	}

	return nil
}

func setOfferStateConfirmingHoldingDeposit(ctx context.Context, m *model.OfferModel) error {
	ExportClosecomOpportunity(ctx, m, closecom.OfferStatus_OFFER_STATUS_CONFIRMING_HOLDING_DEPOSIT)

	return nil
}

func setOfferStateDepositConfirmed(ctx context.Context, m *model.OfferModel) error {
	err := GenerateDraftContract(ctx, m)
	if err != nil {
		return err
	}

	// reject all non-accepted active offers on this property
	log.WithContext(ctx).Info("ListActiveOffersByProperty property id :", m.Property)

	activeOffers, err := model.ListActiveOffersByProperty(ctx, m.Property)
	if err != nil {
		log.WithContext(ctx).WithError(err).Error("ListActiveOffersByProperty error")
	}

	if activeOffers != nil {
		for _, activeOffer := range activeOffers {
			log.WithContext(ctx).Info("activeOffer id = ", activeOffer.ID.Hex())

			if activeOffer.ID.Hex() == m.ID.Hex() {
				continue
			}

			newOfferState := offer.OfferState_DENIED
			if activeOffer.State == offer.OfferState_PENDING {
				newOfferState = offer.OfferState_CANCELED
			}

			_ = SetOfferState(ctx, activeOffer, newOfferState)
		}
	} else {
		log.WithContext(ctx).Infof("No active offers by property: %s", m.Property)
	}

	ExportClosecomOpportunity(ctx, m, closecom.OfferStatus_OFFER_STATUS_REFERENCE_PAYMENTS)

	return nil
}

func setOfferStateApplicantCheckPayment(ctx context.Context, m *model.OfferModel) error {
	ExportClosecomOpportunity(ctx, m, closecom.OfferStatus_OFFER_STATUS_REFERENCE_PAYMENTS)

	return nil
}

func setOfferStateChecking(ctx context.Context, m *model.OfferModel) error {
	ExportClosecomOpportunity(ctx, m, closecom.OfferStatus_OFFER_STATUS_AWAITING_REFERENCE)

	return nil
}

func setOfferStateSigned(ctx context.Context, m *model.OfferModel) error {
	_, err := connections.NotificationService.SendNotification(ctx, &notification.SendNotificationRequest{
		Entity: notification.Entity_ENTITY_OFFER_SIGNATURE,
		Action: notification.Action_ACTION_COMPLETE,
		Targets: []*notification.Target{
			{
				User: m.LandlordUser(),
				Role: notification.Role_ROLE_LANDLORD,
			},
		},
		Embeds: []*notification.EventEmbed{
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_PROPERTY,
				EmbedValue: m.Property,
			},
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_OFFER,
				EmbedValue: m.ID.Hex(),
			},
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_USER,
				EmbedValue: m.TenantUser(),
			},
		},
	})
	if err != nil {
		log.WithContext(ctx).WithError(err).Error("sending signing complete to landlord")
	}

	_, err = connections.NotificationService.SendNotification(ctx, &notification.SendNotificationRequest{
		Entity: notification.Entity_ENTITY_OFFER_SIGNATURE,
		Action: notification.Action_ACTION_COMPLETE,
		Targets: []*notification.Target{
			{
				User: m.TenantUser(),
				Role: notification.Role_ROLE_TENANT,
			},
		},
		Embeds: []*notification.EventEmbed{
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_PROPERTY,
				EmbedValue: m.Property,
			},
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_OFFER,
				EmbedValue: m.ID.Hex(),
			},
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_USER,
				EmbedValue: m.LandlordUser(),
			},
		},
	})
	if err != nil {
		log.WithContext(ctx).WithError(err).Error("sending signing complete to tenant")
	}

	_, err = connections.DocumentService.UpdateStatus(ctx, &document.UpdateStatusRequest{
		FileId: m.Contract,
		State:  document.DocumentState_ACTIVE,
	})
	if err != nil {
		log.WithContext(ctx).WithError(err).WithField("contract", m.Contract).Error("updating document state to active")
	}

	err = model.UpdateTenancyModel(ctx, m)
	if err != nil {
		log.WithContext(ctx).WithError(err).Errorf("model.UpdateTenancyModel model: %+v", *m)
	}

	_, err = connections.ListingService.VersionSetStatus(ctx, &listing.VersionSetStatusRequest{
		PropertyId: m.Property,
		Status:     listing.VersionStatus_VERSION_LET,
	})
	if err != nil {
		log.WithContext(ctx).WithError(err).WithField("property", m.Property).Error("VersionSetStatus to LET failed")
	}

	err = TryCreateRentInvoices(ctx, m, false)
	if err != nil {
		log.WithContext(ctx).Error(err)
	}

	return nil
}

// TryExportClosecomMoveinStatus exports offer opportunity as MOVING_IN once
// - [if DRP present] DRP setup is initiated
// - [if non-zero/legacy rent plan selected] first month payment is confirmed
func TryExportClosecomMoveinStatus(ctx context.Context, m *model.OfferModel) {
	if (m.State == offer.OfferState_SIGNED || m.State == offer.OfferState_RENTPAID) &&
		(m.RentPlan == offer.RentPlan_RENT_PLAN_MANUAL || m.RentPlan == offer.RentPlan_RENT_PLAN_LEGACY ||
			m.PaidStatuses.TenantServices) {
		ExportClosecomOpportunity(ctx, m, closecom.OfferStatus_OFFER_STATUS_MOVING_IN)
	}
}

func setOfferStateRentpaid(ctx context.Context, m *model.OfferModel) error {
	TryExportClosecomMoveinStatus(ctx, m)

	return nil
}

func setOfferStateSigning(ctx context.Context, m *model.OfferModel) error {
	_, err := connections.NotificationService.SendNotification(ctx, &notification.SendNotificationRequest{
		Entity: notification.Entity_ENTITY_OFFER_SIGNATURE,
		Action: notification.Action_ACTION_PENDING,
		Targets: []*notification.Target{
			{
				User: m.LandlordUser(),
				Role: notification.Role_ROLE_LANDLORD,
			},
		},
		Embeds: []*notification.EventEmbed{
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_PROPERTY,
				EmbedValue: m.Property,
			},
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_OFFER,
				EmbedValue: m.ID.Hex(),
			},
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_USER,
				EmbedValue: m.Tenants[0].User,
			},
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_DAYS,
				EmbedValue: "7",
			},
		},
	})
	if err != nil {
		log.WithContext(ctx).WithError(err).Error("sending signing ready to manager")
	}

	f, _ := m.GetFileByType(ctx, model.OfferDocsTypeAst, m.TenantUser())

	_, err = connections.NotificationService.SendNotification(ctx, &notification.SendNotificationRequest{
		Entity: notification.Entity_ENTITY_OFFER_SIGNATURE,
		Action: notification.Action_ACTION_PENDING,
		Targets: []*notification.Target{
			{
				User: m.TenantUser(),
				Role: notification.Role_ROLE_TENANT,
			},
		},
		Embeds: []*notification.EventEmbed{
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_PROPERTY,
				EmbedValue: m.Property,
			},
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_OFFER,
				EmbedValue: m.ID.Hex(),
			},
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_USER,
				EmbedValue: m.Landlord.User,
			},
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_DAYS,
				EmbedValue: "7",
			},
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_FILE,
				EmbedValue: f,
			},
		},
	})
	if err != nil {
		log.WithContext(ctx).WithError(err).Error("sending signing is ready to tenant")
	}

	ExportClosecomOpportunity(ctx, m, closecom.OfferStatus_OFFER_STATUS_CONTRACT_SIGNING)

	return nil
}

func setOfferStateContractReviewAndClauses(ctx context.Context, m *model.OfferModel) error {
	err := GenerateDraftContract(ctx, m)
	if err != nil {
		return err
	}

	ExportClosecomOpportunity(ctx, m, closecom.OfferStatus_OFFER_STATUS_CONTRACT_REVIEW)

	return nil
}

// GenerateDraftContract generates an updated draft contract
func GenerateDraftContract(ctx context.Context, m *model.OfferModel) error {
	prop, err := connections.PropertyService.Get(grpcctx.WithServiceUser(ctx), &property.GetRequest{
		Id:   m.Property,
		User: m.LandlordUser(),
	})
	if err != nil {
		return err
	}

	landlord, err := connections.AccountService.Get(ctx, &account.GetRequest{
		Id: m.Landlord.User,
	})
	if err != nil {
		return err
	}

	ownershipResponse, err := connections.AccountService.GetOwnershipVerifications(
		ctx,
		&account.GetOwnershipVerificationRequest{
			PropertyId: &account.GetOwnershipVerificationRequest_PropertyIdValue{
				PropertyIdValue: m.Property,
			},
		})
	if err != nil && !errors.Is(err, mongo.ErrNoDocuments) {
		return err
	}

	companyName := ""

	if ownershipResponse != nil && len(ownershipResponse.GetOwnerships()) > 0 {
		companyName = ownershipResponse.GetOwnerships()[0].CompanyName
	}

	cs := struct {
		Landlord    *account.Account   `json:"landlord"`
		Offer       *offer.Offer       `json:"offer"`
		Property    *property.Property `json:"property"`
		Now         string             `json:"now"`
		CompanyName string             `json:"company_name"`
	}{
		Now:         time.Now().Format(time.RFC1123),
		Landlord:    landlord.GetAccount(),
		Offer:       m.RenderOfferModel(ctx, ""),
		Property:    prop.GetProperty(),
		CompanyName: companyName,
	}

	bs, err := bind.ToJSON(cs, &bind.Config{
		GetOmitEmpty: false,
		SetOmitEmpty: false,
	})
	if err != nil {
		return err
	}

	fileID, err := CreateContract(ctx, bs, true)
	if err != nil {
		return err
	}

	m.DraftContract = fileID
	m.DraftContractCreatedAt = time.Now()

	return nil
}
