package status

import (
	"context"
	"errors"
	"fmt"
	"math"
	"strings"
	"time"

	closecom "gitlab.mashplace.com/mashplace/backend/closecom/api"
	"gitlab.mashplace.com/mashplace/backend/common/api/grpcerr"
	"gitlab.mashplace.com/mashplace/backend/common/generic/util"
	extras "gitlab.mashplace.com/mashplace/backend/extras/api"
	invoice "gitlab.mashplace.com/mashplace/backend/invoice/api"
	offer "gitlab.mashplace.com/mashplace/backend/offer/api"
	"gitlab.mashplace.com/mashplace/backend/offer/internal/connections"
	"gitlab.mashplace.com/mashplace/backend/offer/model"
	"google.golang.org/protobuf/types/known/timestamppb"
)

// TryCreateRentInvoices will attempt to create invoices for signed or
// for awaiting tenancy offer if it has rent plan selected
func TryCreateRentInvoices(ctx context.Context, m *model.OfferModel, isImitationMode bool) error {
	if m.State != offer.OfferState_SIGNED &&
		m.State != offer.OfferState_AWAITING_TENANCY &&
		m.State != offer.OfferState_RENTPAID && m.State != offer.OfferState_RTO_STATE_ACTIVE {
		if isImitationMode {
			return errors.New("available only for signed/awaiting tenancy/rent paid states")
		}

		return nil
	}

	if m.RentPlan == 0 || (m.RentPlan == offer.RentPlan_RENT_PLAN_LEGACY && !isImitationMode) {
		if isImitationMode {
			return errors.New("rent plan is not selected")
		}

		return nil
	}

	if m.SignedInvoicesGenerated {
		if isImitationMode {
			return errors.New("invoices already generated")
		}

		return nil
	}

	return CreateRentInvoices(ctx, m, isImitationMode)
}

// CreateRentInvoices creates invoices for signed offer with rent plan selected
func CreateRentInvoices(ctx context.Context, m *model.OfferModel, isImitationMode bool) error {
	if isImitationMode {
		setDefaultValuesInOldOffersForRentInvoices(m)
	}

	if m.Conditions == nil ||
		len(m.Tenants) == 0 ||
		m.Conditions.PaymentFrequency == nil ||
		m.Conditions.Price == nil ||
		m.Conditions.TermMonths == nil {
		return fmt.Errorf(
			"%w: empty condition payment frequency/price/term months/tenants",
			grpcerr.InvalidState,
		)
	}

	if m.RentPlan == offer.RentPlan_RENT_PLAN_MANUAL {
		m.SignedInvoicesGenerated = true

		err := model.UpdateOffer(ctx, m)
		if err != nil {
			return err
		}

		return SetPropertyRented(ctx, m)
	}

	var errs []string

	err := createFirstMonthPaymentInvoice(ctx, m, isImitationMode)
	if err != nil {
		errs = append(errs, "generating first month rent: "+err.Error())
	}

	// create standing order request invoice
	err = createStandingOrderInvoice(ctx, m, isImitationMode)
	if err != nil {
		errs = append(errs, "generating standing order: "+err.Error())
	}

	if len(errs) > 0 {
		return errors.New(strings.Join(errs, "; "))
	}

	ExportClosecomOpportunity(ctx, m, closecom.OfferStatus_OFFER_STATUS_AWAITING_PAYMENTS)

	m.SignedInvoicesGenerated = true

	return nil
}

func setDefaultValuesInOldOffersForRentInvoices(m *model.OfferModel) {
	if m.Conditions == nil {
		m.Conditions = &model.OfferConditionsModel{}
	}

	if m.Conditions.MoveinDate == nil {
		m.Conditions.MoveinDate = util.PtrTime(time.Now())
	}

	if m.Conditions.TermMonths == nil {
		m.Conditions.TermMonths = util.PtrInt32(12)
	}

	if m.Conditions.PaymentFrequency == nil {
		m.Conditions.PaymentFrequency = util.PtrInt32(1)
	}

	if m.Conditions.FirstRentPaymentDate == nil {
		m.Conditions.FirstRentPaymentDate = m.Conditions.MoveinDate
	}

	if m.Conditions.RentPaymentDay == nil {
		m.Conditions.RentPaymentDay = util.PtrInt32(int32(offer.PaymentDayClamp(m.Conditions.MoveinDate.Day())))
	}

	if m.Conditions.BreakClause == nil {
		m.Conditions.BreakClause = util.PtrInt32(0)
	}

	if m.Conditions.DepositType == offer.DepositType_UNKNOWN_DEPOSIT_TYPE {
		m.Conditions.DepositType = offer.DepositType_DEPOSIT_TRADITIONAL_5WEEK
	}
}

func firstMonthPayment(
	ctx context.Context,
	off *model.OfferModel,
) (services []*extras.ServiceSnapshot, err error) {
	if off.Conditions.MoveinDate == nil || util.UnPtrInt32(off.Conditions.RentPaymentDay) == 0 {
		err = fmt.Errorf(
			"%w: empty condition movein date/rent payment day",
			grpcerr.InvalidState,
		)

		return
	}

	services = make([]*extras.ServiceSnapshot, 0)

	condPrice := util.UnPtrFloat64(off.Conditions.Price)
	invoiceResp, err := connections.InvoiceService.InvoiceList(ctx, &invoice.InvoiceListRequest{
		CorrelationType:   invoice.CorrelationTypeOffer,
		CorrelationId:     off.ID.Hex(),
		CorrelationTarget: invoice.CorrelationTargetHoldingDeposit,
		Type:              invoice.InvoiceType_INVOICE_TYPE_PAYMENT,
		Property:          off.Property,
	})
	if err != nil || len(invoiceResp.Invoices) == 0 {
		err = ErrPaidDepositNotFound

		return
	}

	var holding float64 = 0

	for _, inv := range invoiceResp.Invoices {
		if inv.GetStatus() == invoice.InvoiceStatus_INVOICE_FULFILLED {
			holding = inv.GetAmount()
		}
	}

	firstMonth := condPrice - holding

	if *off.Conditions.PaymentFrequency > *off.Conditions.TermMonths {
		if *off.Conditions.TermMonths > 1 {
			firstMonth = condPrice * (float64(*off.Conditions.TermMonths) - 1)
		}
	} else {
		if *off.Conditions.PaymentFrequency > 1 {
			firstMonth += condPrice * (float64(*off.Conditions.PaymentFrequency) - 1)
		}
	}

	firstMonthRentService, err := extras.GetServicePriceSnapshot(
		ctx,
		connections.ExtrasService,
		extras.ProductTenantFirstMonthRent,
		firstMonth,
	)
	if err != nil {
		return nil, err
	}

	services = append(services, firstMonthRentService)

	var selectedDepositType extras.ProductCode

	var depositAmount float64

	if off.Conditions.DepositType == offer.DepositType_DEPOSIT_TRADITIONAL_5WEEK {
		selectedDepositType = extras.ProductTenantDeposit

		depositAmount = extras.CalculateTraditionalDepositAmount(condPrice)
	} else {
		selectedDepositType = extras.ProductTenantDepositReplacement

		depositAmount = extras.CalculateDepositReplacementAmount(condPrice)
	}

	depositService, err := extras.GetServicePriceSnapshot(
		ctx,
		connections.ExtrasService,
		selectedDepositType,
		depositAmount,
	)
	if err != nil {
		return nil, err
	}

	services = append(services, depositService)

	return
}

func createHoldingDepositPayoutInvoice(
	ctx context.Context,
	old *model.OfferModel,
) error {
	inv, err := connections.InvoiceService.InvoiceList(ctx, &invoice.InvoiceListRequest{
		CorrelationType:   invoice.CorrelationTypeOffer,
		CorrelationId:     old.ID.Hex(),
		CorrelationTarget: invoice.CorrelationTargetHoldingDeposit,
		Type:              invoice.InvoiceType_INVOICE_TYPE_PAYOUT,
		Statuses: []invoice.InvoiceStatus{
			invoice.InvoiceStatus_INVOICE_OPEN,
			invoice.InvoiceStatus_INVOICE_FULFILLED,
		},
	})
	if err != nil {
		return err
	}

	if len(inv.GetInvoices()) > 0 {
		return nil
	}

	invoiceResp, err := connections.InvoiceService.InvoiceGet(ctx, &invoice.InvoiceGetRequest{
		CorrelationType:   invoice.CorrelationTypeOffer,
		CorrelationId:     old.ID.Hex(),
		CorrelationTarget: invoice.CorrelationTargetHoldingDeposit,
		Type:              invoice.InvoiceType_INVOICE_TYPE_PAYMENT,
		Property:          old.Property,
	})
	if err != nil || invoiceResp.GetInvoice() == nil {
		return nil
	}

	var svcs []*extras.ServiceSnapshot

	for _, s := range invoiceResp.GetInvoice().GetServices() {
		svcs = append(svcs, s.GetService())
	}

	_, err = connections.InvoiceService.InvoiceCreate(ctx, &invoice.InvoiceCreateRequest{
		CorrelationType:   invoice.CorrelationTypeOffer,
		CorrelationId:     old.ID.Hex(),
		CorrelationTarget: invoice.CorrelationTargetHoldingDeposit,
		Type:              invoice.InvoiceType_INVOICE_TYPE_PAYOUT,
		Property:          old.Property,
		Ivr:               old.PropertyIvr,
		Listing:           old.PropertyListing,
		User:              old.TenantUser(),
		Services:          svcs,
		NotifyCallback:    connections.SelfRemote,
	})
	if err != nil {
		return err
	}

	return nil
}

// createFirstMonthPaymentInvoice creates tenant first month payment invoice after signing
func createFirstMonthPaymentInvoice(
	ctx context.Context,
	offerModel *model.OfferModel,
	isImitationMode bool,
) error {
	if offerModel.OriginalOffer != "" {
		return nil
	}

	services, err := firstMonthPayment(ctx, offerModel)
	if err != nil {
		return err
	}

	// check invoice existence
	inv, err := connections.InvoiceService.InvoiceList(ctx, &invoice.InvoiceListRequest{
		CorrelationType:   invoice.CorrelationTypeOffer,
		CorrelationId:     offerModel.ID.Hex(),
		CorrelationTarget: invoice.CorrelationTargetTenantServices,
		Statuses: []invoice.InvoiceStatus{
			invoice.InvoiceStatus_INVOICE_OPEN,
			invoice.InvoiceStatus_INVOICE_FULFILLED,
		},
	})
	if err != nil {
		return err
	}

	if len(inv.GetInvoices()) > 0 {
		if isImitationMode {
			_, err = connections.InvoiceService.GenerateInvoicePayments(ctx, &invoice.GenerateInvoicePaymentsRequest{
				InvoiceId: inv.GetInvoices()[0].GetInvoiceId(),
			})
			if err != nil {
				return err
			}
		}

		return nil
	}

	resp, err := connections.InvoiceService.InvoiceCreate(ctx, &invoice.InvoiceCreateRequest{
		CorrelationType:   invoice.CorrelationTypeOffer,
		CorrelationId:     offerModel.ID.Hex(),
		CorrelationTarget: invoice.CorrelationTargetTenantServices,
		Property:          offerModel.Property,
		Ivr:               offerModel.PropertyIvr,
		Listing:           offerModel.PropertyListing,
		User:              offerModel.TenantUser(),
		Services:          services,
		NotifyCallback:    connections.SelfRemote,
	})
	if err != nil {
		return err
	}

	if isImitationMode {
		_, err = connections.InvoiceService.GenerateInvoicePayments(ctx, &invoice.GenerateInvoicePaymentsRequest{
			InvoiceId: resp.GetInvoice().GetInvoiceId(),
		})
		if err != nil {
			return err
		}
	}

	return nil
}

func createStandingOrderInvoice(ctx context.Context, m *model.OfferModel, isImitationMode bool) error {
	freq := *m.Conditions.PaymentFrequency
	offerPrice := *m.Conditions.Price

	var frequency invoice.PeriodicPaymentFrequency
	if m.IsRto {
		frequency = invoice.PeriodicPaymentFrequency(freq)
	} else {
		frequency = offer.DataPeriodicPaymentFrequency[freq]
	}

	priceMonth := math.Round(offerPrice*100) / 100
	pricePayment := priceMonth * float64(freq)

	var lastPayMonths int32

	termMonths := *m.Conditions.TermMonths

	if termMonths%freq == 0 {
		lastPayMonths = termMonths - freq
	} else {
		lastPayMonths = termMonths - (termMonths % freq) - freq
	}

	movein := *m.Conditions.MoveinDate
	day := offer.PaymentDayClamp(int(util.UnPtrInt32(m.Conditions.RentPaymentDay)))
	startAt := timestamppb.New(movein.AddDate(0, 1, -movein.Day()+day))
	endAt := timestamppb.New(movein.AddDate(0, int(lastPayMonths)+1, 0))

	if isImitationMode {
		start := time.Date(movein.Year(), movein.Month()+1, day, 0, 0, 0, 0, time.UTC)
		end := movein.AddDate(0, int(lastPayMonths)+1, 0)

		startAt = timestamppb.New(start)
		endAt = timestamppb.New(end)
	}

	service, err := extras.GetServicePriceSnapshot(
		ctx,
		connections.ExtrasService,
		extras.ProductTenantStandingOrder,
		pricePayment,
	)
	if err != nil {
		return err
	}

	// check invoice existence
	inv, err := connections.InvoiceService.InvoiceList(ctx, &invoice.InvoiceListRequest{
		CorrelationType:   invoice.CorrelationTypeOffer,
		CorrelationId:     m.ID.Hex(),
		CorrelationTarget: invoice.CorrelationTargetStandingOrder,
		Statuses: []invoice.InvoiceStatus{
			invoice.InvoiceStatus_INVOICE_OPEN,
			invoice.InvoiceStatus_INVOICE_FULFILLED,
		},
	})
	if err != nil {
		return err
	}

	if len(inv.GetInvoices()) > 0 {
		if isImitationMode {
			_, err = connections.InvoiceService.GenerateInvoicePayments(ctx, &invoice.GenerateInvoicePaymentsRequest{
				InvoiceId: inv.GetInvoices()[0].GetInvoiceId(),
				Periodic: &invoice.GenerateInvoicePaymentsPeriodic{
					StartAt:   startAt,
					EndAt:     endAt,
					Frequency: frequency,
				},
			})
			if err != nil {
				return err
			}
		}

		return nil
	}

	resp, err := connections.InvoiceService.InvoiceCreate(ctx, &invoice.InvoiceCreateRequest{
		CorrelationType:   invoice.CorrelationTypeOffer,
		CorrelationId:     m.ID.Hex(),
		CorrelationTarget: invoice.CorrelationTargetStandingOrder,
		Periodic:          true,
		Property:          m.Property,
		Ivr:               m.PropertyIvr,
		Listing:           m.PropertyListing,
		User:              m.Tenants[0].User,
		Services: []*extras.ServiceSnapshot{
			service,
		},
		NotifyCallback: connections.SelfRemote,
	})

	if isImitationMode {
		_, err = connections.InvoiceService.GenerateInvoicePayments(ctx, &invoice.GenerateInvoicePaymentsRequest{
			InvoiceId: resp.GetInvoice().GetInvoiceId(),
			Periodic: &invoice.GenerateInvoicePaymentsPeriodic{
				StartAt:   startAt,
				EndAt:     endAt,
				Frequency: frequency,
			},
		})
		if err != nil {
			return err
		}
	}

	return err
}
