package status

import (
	"context"
	"errors"
	"fmt"
	"strings"
	"sync"
	"time"

	"gitlab.mashplace.com/mashplace/backend/common/api/generic/query"
	"gitlab.mashplace.com/mashplace/backend/common/api/grpcctx"
	"gitlab.mashplace.com/mashplace/backend/common/generic/log"
	"gitlab.mashplace.com/mashplace/backend/common/generic/util"
	"gitlab.mashplace.com/mashplace/backend/common/generic/wrap"
	"gitlab.mashplace.com/mashplace/backend/common/lib/mongolocal"
	document "gitlab.mashplace.com/mashplace/backend/document/api"
	expenses "gitlab.mashplace.com/mashplace/backend/expenses/api"
	extras "gitlab.mashplace.com/mashplace/backend/extras/api"
	invoice "gitlab.mashplace.com/mashplace/backend/invoice/api"
	messaging "gitlab.mashplace.com/mashplace/backend/messaging/api"
	notification "gitlab.mashplace.com/mashplace/backend/notification/api"
	offer "gitlab.mashplace.com/mashplace/backend/offer/api"
	"gitlab.mashplace.com/mashplace/backend/offer/internal/connections"
	"gitlab.mashplace.com/mashplace/backend/offer/internal/mixpanel"
	"gitlab.mashplace.com/mashplace/backend/offer/model"
	property "gitlab.mashplace.com/mashplace/backend/property/api"
	signature "gitlab.mashplace.com/mashplace/backend/signature/api"
	"go.mongodb.org/mongo-driver/bson/primitive"
	"google.golang.org/protobuf/types/known/timestamppb"
	"google.golang.org/protobuf/types/known/wrapperspb"
)

// FunctionCreateInsurancesLockState contains a called flag, which displays that TrySetPropertyRentedCreateInsurances is
// called or not by offerID
type FunctionCreateInsurancesLockState struct {
	mu     sync.Mutex
	called bool
}

// functionCreateInsurancesStates contains a map of used TrySetPropertyRentedCreateInsurances function by offerID.
var functionCreateInsurancesStates = struct {
	mu         sync.Mutex
	lockStates map[string]*FunctionCreateInsurancesLockState
}{lockStates: make(map[string]*FunctionCreateInsurancesLockState)}

func trySetPropertyRentedTenancyUpdate(ctx context.Context, old *model.OfferModel) {
	tn, err := model.UpdateTenancy(ctx, old)
	if err != nil {
		log.WithContext(ctx).WithError(err).Error("updating tenancy")
	}

	movein := util.UnPtrTime(old.Conditions.MoveinDate)

	var tenants []*property.Tenant

	for _, t := range old.Tenants {
		var risk property.TenantRisk

		switch t.Risk {
		case offer.TenantRisk_LOW_RISK:
			risk = property.TenantRisk_LOW_RISK
		case offer.TenantRisk_MEDIUM_RISK:
			risk = property.TenantRisk_MEDIUM_RISK
		case offer.TenantRisk_HIGH_RISK:
			risk = property.TenantRisk_HIGH_RISK
		}

		tenants = append(tenants, &property.Tenant{
			User:     wrapperspb.String(t.User),
			Contract: wrapperspb.String(old.Contract),
			Active:   wrapperspb.Bool(true),
			Risk:     risk,
			TenantContact: &property.TenantContact{
				FirstName:    wrap.StringValue(t.FirstName),
				LastName:     wrap.StringValue(t.LastName),
				EmailPrimary: wrap.StringValue(t.EmailPrimary),
				PhonePrimary: wrap.StringValue(t.PhonePrimary),
			},
		})
	}

	var guarantors []*property.Guarantor

	for _, g := range old.Guarantors {
		guarantors = append(guarantors, &property.Guarantor{
			FirstName:    wrap.StringValue(g.FirstName),
			LastName:     wrap.StringValue(g.LastName),
			EmailPrimary: wrap.StringValue(g.EmailPrimary),
			PhonePrimary: wrap.StringValue(g.PhonePrimary),
		})
	}

	_, err = connections.PropertyService.ServiceRented(ctx, &property.RentedRequest{
		Id:         old.Property,
		Offer:      old.ID.Hex(),
		Landlord:   old.LandlordUser(),
		ListingId:  old.PropertyListing,
		Tenants:    &property.TenantArrayValue{Value: tenants},
		Guarantors: &property.GuarantorArrayValue{Value: guarantors},
		TermStart:  timestamppb.New(movein),
		TermEnd:    timestamppb.New(movein.AddDate(0, int(util.UnPtrInt32(old.Conditions.TermMonths)), 0)),
		TenancyId:  tn.GetId(),
	})
	if err != nil {
		log.WithContext(ctx).WithError(err).Error("setting property to rented")
	}

	if err = mixpanel.TrackLandlordPropertyRented(ctx, old); err != nil {
		log.WithContext(ctx).WithError(err).Error("failed to track PropertyRented mixpanel event")
	}

	if err = mixpanel.TrackTenantFirstRentPaymentCompleted(ctx, old); err != nil {
		log.WithContext(ctx).WithError(err).Error("failed to track FirstRentpaymentCompleted mixpanel event")
	}
}

func getFunctionCreateInsurancesLockState(offerID *string) *FunctionCreateInsurancesLockState {
	functionCreateInsurancesStates.mu.Lock()

	defer functionCreateInsurancesStates.mu.Unlock()

	if functionCreateInsurancesStates.lockStates[*offerID] == nil {
		functionCreateInsurancesStates.lockStates[*offerID] = &FunctionCreateInsurancesLockState{}
	}

	return functionCreateInsurancesStates.lockStates[*offerID]
}

func deleteFunctionCreateInsurancesLockState(offerID *string) {
	functionCreateInsurancesStates.mu.Lock()

	defer functionCreateInsurancesStates.mu.Unlock()

	delete(functionCreateInsurancesStates.lockStates, *offerID)
}

// TrySetPropertyRentedCreateInsurances for offer with essential/secured rent collection plan
// nolint
func TrySetPropertyRentedCreateInsurances(
	ctx context.Context,
	offerID string,
	isImitationMode bool,
	isRollingTenancyProcess bool,
) error {
	logger := log.WithFields(log.Fields{
		"offerID":                 offerID,
		"isImitationMode":         isImitationMode,
		"isRollingTenancyProcess": isRollingTenancyProcess,
	})

	logger.Infoln("TrySetPropertyRentedCreateInsurances started")

	lockState := getFunctionCreateInsurancesLockState(&offerID)

	lockState.mu.Lock()

	defer func(ptrOfferID *string, lockState *FunctionCreateInsurancesLockState) {
		lockState.mu.Unlock()

		deleteFunctionCreateInsurancesLockState(ptrOfferID)
	}(&offerID, lockState)

	if lockState.called {
		return errors.New("rent insurances already generated")
	}

	objID, err := primitive.ObjectIDFromHex(offerID)
	if err != nil {
		return err
	}

	old, err := model.GetOfferByID(ctx, objID)
	if err != nil {
		return err
	}

	if old.RentInsurancesGenerated && !isRollingTenancyProcess {
		return errors.New("rent insurances already generated")
	}

	startDate := util.UnPtrTime(old.Conditions.MoveinDate)

	if isImitationMode {
		nowDate := time.Now().Truncate(24 * time.Hour)
		dateIn30days := nowDate.AddDate(0, 0, 30).Truncate(24 * time.Hour)
		startDate = startDate.Truncate(24 * time.Hour)

		if startDate.Before(nowDate) || dateIn30days.Before(startDate) {
			startDate = time.Now().Truncate(24 * time.Hour)
		}
	}

	if old.RentPlan == offer.RentPlan_RENT_PLAN_MANUAL {
		_, err := connections.NotificationService.SendNotification(ctx, &notification.SendNotificationRequest{
			Entity: notification.Entity_ENTITY_OFFER_RENT_COLLECTION_MANUAL,
			Action: notification.Action_ACTION_COMPLETE,
			Targets: []*notification.Target{
				{
					User: old.TenantUser(),
					Role: notification.Role_ROLE_TENANT,
				},
				{
					User: old.LandlordUser(),
					Role: notification.Role_ROLE_LANDLORD,
				},
			},
			Embeds: []*notification.EventEmbed{
				{
					EmbedType:  messaging.EmbedType_EMBED_TYPE_USER,
					EmbedValue: old.LandlordUser(),
				},
				{
					EmbedType:  messaging.EmbedType_EMBED_TYPE_PROPERTY,
					EmbedValue: old.Property,
				},
				{
					EmbedType:  messaging.EmbedType_EMBED_TYPE_OFFER,
					EmbedValue: old.ID.Hex(),
				},
			},
		})
		if err != nil {
			logger.WithContext(ctx).WithError(err).Errorf("sending failed: nanual rent for tenant")
		}
	}

	if old.RentPlan != offer.RentPlan_RENT_PLAN_ESSENTIAL && old.RentPlan != offer.RentPlan_RENT_PLAN_SECURED {
		return errors.New("available rent plan to generate insurances is not selected")
	}

	var codes []extras.ProductCode

	if isRollingTenancyProcess {
		var err error

		codes, err = checkRentPlanInsurancesExpirationDate(ctx, old)
		if err != nil {
			return err
		}

		if len(codes) == 0 {
			return nil
		}
	}

	products, err := generateProductsRentPlanInsurances(ctx, old, startDate, codes)
	if err != nil {
		return err
	}

	order, err := connections.InvoiceService.OrderSave(ctx, &invoice.OrderSaveRequest{
		Userid:            old.LandlordUser(),
		Products:          products,
		Clientid:          "",
		OrderType:         invoice.OrderType_REGULAR_ORDER,
		OrderSource:       invoice.OrderSource_RENT_COLLECTION,
		Periodic:          false,
		CorrelationType:   invoice.CorrelationTypeOffer,
		CorrelationId:     old.ID.Hex(),
		CorrelationTarget: invoice.CorrelationTargetLandlordServices,
	})
	if err != nil {
		logger.WithContext(ctx).WithError(err).Error("rented insurances: saving order")

		return err
	}

	commit, err := connections.InvoiceService.OrderCommit(ctx, &invoice.OrderCommitRequest{
		Userid:    old.LandlordUser(),
		Id:        order.GetOrder().GetId(),
		OrderType: invoice.OrderType_REGULAR_ORDER,
	})
	if err != nil {
		logger.WithContext(ctx).WithError(err).Error("rented insurances: committing order")

		return err
	}

	_, err = connections.InvoiceService.ServicePaymentCreate(ctx, &invoice.ServicePaymentCreateRequest{
		InvoiceId: commit.GetInvoice().GetInvoiceId(),
		Status:    invoice.PaymentStatus_PAYMENT_SUCCESS,
	})
	if err != nil {
		logger.WithContext(ctx).WithError(err).Error("rented insurances: creating payment")

		return err
	}

	err = model.OfferInsurancesGenerated(ctx, old.ID)
	if err != nil {
		logger.WithContext(ctx).WithError(err).Error("rented insurances: offer insurances generated save")

		return err
	}

	lockState.called = true

	logger.Infoln("TrySetPropertyRentedCreateInsurances ended")

	return nil
}

func checkRentPlanInsurancesExpirationDate(
	ctx context.Context,
	old *model.OfferModel,
) ([]extras.ProductCode, error) {
	codesToFilter := listInsuranceTypeCodesByRentPlan(old.RentPlan)

	docResp, err := connections.DocumentService.ServiceDocumentFilterList(ctx, &document.ServiceDocumentFilterListRequest{
		Params: &query.QueryParams{
			Filter: []*query.QueryParamsFilter{
				{
					Fields:   []string{"user"},
					Value:    []string{old.LandlordUser()},
					Operator: query.QueryFilterOperator_EQUALS,
				},
				{
					Fields:   []string{"property"},
					Value:    []string{old.Property},
					Operator: query.QueryFilterOperator_EQUALS,
				},
				{
					Fields:   []string{"type_code"},
					Value:    codesToFilter,
					Operator: query.QueryFilterOperator_ANY,
				},
				{
					Fields:   []string{"expiration"},
					Value:    []string{time.Now().Format(time.RFC3339)},
					Operator: query.QueryFilterOperator_MORE,
				},
			},
		},
	},
	)
	if err != nil {
		return nil, err
	}

	mp := make(map[string]struct{}, 0)

	for _, c := range codesToFilter {
		mp[c] = struct{}{}
	}

	for _, d := range docResp.GetDocuments() {
		delete(mp, d.TypeCode)
	}

	result := make([]extras.ProductCode, 0)

	for k, _ := range mp {
		switch document.TypeCode(k) {
		case document.TypeLegalExpensesInsurance:
			result = append(result, extras.ProductLandlordLegalExpensesInsurance)
		case document.TypeHomeEmergencyInsurance:
			result = append(result, extras.ProductLandlordHomeEmergencyInsurance)
		case document.TypeRentGuaranteeInsurance:
			result = append(result, extras.ProductLandlordRentGuaranteeInsurance)
		}
	}

	return result, nil
}

func generateProductsRentPlanInsurances(
	ctx context.Context,
	old *model.OfferModel,
	startDate time.Time,
	codes []extras.ProductCode,
) ([]*invoice.Product, error) {
	var rentplan invoice.RentPlan

	switch old.RentPlan {
	case offer.RentPlan_RENT_PLAN_LEGACY:
		rentplan = invoice.RentPlan_RENT_PLAN_LEGACY
	case offer.RentPlan_RENT_PLAN_MANUAL:
		rentplan = invoice.RentPlan_RENT_PLAN_MANUAL
	case offer.RentPlan_RENT_PLAN_ESSENTIAL:
		rentplan = invoice.RentPlan_RENT_PLAN_ESSENTIAL
	case offer.RentPlan_RENT_PLAN_SECURED:
		rentplan = invoice.RentPlan_RENT_PLAN_SECURED
	default:
		rentplan = invoice.RentPlan_RENT_PLAN_UNKNOWN
	}

	details := &invoice.ProductDetails{
		Time: &invoice.ProductDetailsTime{
			Start: timestamppb.New(startDate),
			End:   timestamppb.New(startDate.AddDate(1, 0, 0)),
		},
		Property: &invoice.ProductDetailsProperty{
			Id:      wrapperspb.String(old.Property),
			Listing: wrapperspb.String(old.PropertyListing),
		},
		Contract: &invoice.ProductDetailsContract{
			Id: wrapperspb.String(old.ID.Hex()),
			Conditions: &invoice.RentConditions{
				Rent:     wrapperspb.Double(util.UnPtrFloat64(old.Conditions.Price)),
				RentPlan: rentplan,
			},
			Tenant: wrapperspb.String(old.TenantUser()),
		},
	}

	var products []*invoice.Product

	if len(codes) == 0 {
		codes = listProductCodesByRentPlan(old.RentPlan)
	}

	for _, c := range codes {
		serv, err := extras.GetServiceSnapshot(
			ctx,
			connections.ExtrasService,
			c,
		)
		if err != nil {
			log.WithContext(ctx).WithError(err).Errorf("rented insurances: getting %s service", c.String())

			return nil, err
		}

		products = append(products, &invoice.Product{
			ServiceId: serv.GetServiceId(),
			Service: &extras.ServiceSnapshotStatus{
				Service:   serv,
				UpdatedAt: timestamppb.Now(),
				Status:    extras.ServiceStatus_ACTIVE,
			},
			Details: details,
		})
	}

	return products, nil
}

func listProductCodesByRentPlan(rentPlan offer.RentPlan) []extras.ProductCode {
	switch rentPlan {
	case offer.RentPlan_RENT_PLAN_ESSENTIAL:
		return []extras.ProductCode{
			extras.ProductLandlordLegalExpensesInsurance,
		}
	case offer.RentPlan_RENT_PLAN_SECURED:
		return []extras.ProductCode{
			extras.ProductLandlordRentGuaranteeInsurance,
			extras.ProductLandlordHomeEmergencyInsurance,
		}
	}

	return nil
}

func listInsuranceTypeCodesByRentPlan(rentPlan offer.RentPlan) []string {
	switch rentPlan {
	case offer.RentPlan_RENT_PLAN_ESSENTIAL:
		return []string{document.TypeLegalExpensesInsurance}
	case offer.RentPlan_RENT_PLAN_SECURED:
		return []string{document.TypeHomeEmergencyInsurance, document.TypeRentGuaranteeInsurance}
	}

	return nil
}

// SetPropertyRented attempts to set offer in rented status, if all setup payments are completed
func SetPropertyRented(ctx context.Context, old *model.OfferModel) error {
	if old.State == offer.OfferState_RENTPAID {
		return nil
	}

	if (old.RentPlan > offer.RentPlan_RENT_PLAN_MANUAL || old.RentPlan == offer.RentPlan_RENT_PLAN_LEGACY) &&
		!(old.PaidStatuses.TenantServices || old.PaidStatuses.StandingOrderSetup) {
		return nil
	}

	err := SetOfferState(ctx, old, offer.OfferState_RENTPAID)
	if err != nil {
		return err
	}

	if old.RentPlan == offer.RentPlan_RENT_PLAN_MANUAL {
		err = createHoldingDepositPayoutInvoice(ctx, old)
		if err != nil {
			return err
		}
	}

	bgctx := grpcctx.WithServiceUser(grpcctx.ValuesBackgroundContext(ctx))

	go trySetPropertyRentedTenancyUpdate(bgctx, old)

	err = GenerateSchedule(ctx, old)
	if err != nil {
		return err
	}

	offerRentPaymentEmbeds := []*notification.EventEmbed{
		{
			EmbedType:  messaging.EmbedType_EMBED_TYPE_USER,
			EmbedValue: old.Tenants[0].User,
		},
		{
			EmbedType:  messaging.EmbedType_EMBED_TYPE_PROPERTY,
			EmbedValue: old.Property,
		},
	}

	// TODO GetPDF 2 time (for tenant and for landlord) + retry while error
	resp, err := connections.SignatureService.SigningSignedDocument(ctx, &signature.SigningSignedDocumentRequest{
		SignatureRequestId: old.SignatureRequestID,
	})
	if err != nil {
		log.Println("Error getPDF: ", err)
	} else {
		offerRentPaymentEmbeds = append(offerRentPaymentEmbeds, &notification.EventEmbed{
			EmbedType:  messaging.EmbedType_EMBED_TYPE_FILE,
			EmbedValue: resp.GetSignedDocumentId(),
		})
	}

	_, err = connections.NotificationService.SendNotification(ctx, &notification.SendNotificationRequest{
		Entity: notification.Entity_ENTITY_OFFER_RENT_PAYMENT,
		Action: notification.Action_ACTION_COMPLETE,
		Targets: []*notification.Target{
			{
				User: old.LandlordUser(),
				Role: notification.Role_ROLE_LANDLORD,
			},
		},
		Embeds: offerRentPaymentEmbeds,
	})
	if err != nil {
		log.WithContext(ctx).WithError(err).Errorf("sending failed: offer rent paid for landlord")
	}

	fileType := model.OfferDocsTypeSomf
	fileModel := &model.OfferFileModel{
		FileType:    fileType,
		Version:     fileType.GetVersion(),
		Mime:        fileType.GetMime(),
		Description: fileType.GetDescription(),
		Name:        fileType.GetName(),
	}

	err = fileModel.RenderSOMF(ctx, old, old.TenantUser())
	if err != nil {
		log.Println("Error render")
	}

	_, err = connections.NotificationService.SendNotification(ctx, &notification.SendNotificationRequest{
		Entity: notification.Entity_ENTITY_OFFER_RENT_PAYMENT,
		Action: notification.Action_ACTION_COMPLETE,
		Targets: []*notification.Target{
			{
				User: old.TenantUser(),
				Role: notification.Role_ROLE_TENANT,
			},
		},
		Embeds: offerRentPaymentEmbeds,
	})
	if err != nil {
		log.WithContext(ctx).WithError(err).Errorf("sending failed: offer rent paid for tenant")
	}

	_, err = connections.NotificationService.SendNotification(
		ctx,
		&notification.SendNotificationRequest{
			Entity: notification.Entity_ENTITY_OFFER,
			Action: notification.Action_ACTION_COMPLETE,
			Targets: []*notification.Target{
				{
					User: old.TenantUser(),
					Role: notification.Role_ROLE_TENANT,
				},
				{
					User: old.LandlordUser(),
					Role: notification.Role_ROLE_LANDLORD,
				},
			},
			Embeds: []*notification.EventEmbed{
				{
					EmbedType:  messaging.EmbedType_EMBED_TYPE_PROPERTY,
					EmbedValue: old.Property,
				},
				{
					EmbedType:  messaging.EmbedType_EMBED_TYPE_OFFER,
					EmbedValue: old.ID.Hex(),
				},
			},
		},
	)
	if err != nil {
		log.WithContext(ctx).WithError(err).Errorf("sending failed: %s", notification.EventType_EVENT_PAYMENT_READY_MOVEIN_TENANT)
	}

	if old.Conditions.DepositType == offer.DepositType_DEPOSIT_REPLACEMENT {
		_, _ = RenderCSV(ctx, old, true)
	}

	return nil
}

// GenerateSchedule attempts to generate a payment schedule for signed and set up offer
func GenerateSchedule(ctx context.Context, old *model.OfferModel) error {
	if old.ScheduleGenerated {
		return nil
	}

	isRto := old.IsRto
	isOfferInAllowedState := old.State == offer.OfferState_RENTPAID ||
		old.State == offer.OfferState_AWAITING_TENANCY ||
		old.State == offer.OfferState_RTO_STATE_WAITING_FOR_STANDING_ORDER_SETUP ||
		old.State == offer.OfferState_RTO_STATE_WAITING_FOR_PAYMENT

	if (!isRto && !old.PaidStatuses.TenantServices) ||
		!old.PaidStatuses.StandingOrderSetup ||
		!isOfferInAllowedState {
		return nil
	}

	_, err := connections.ExpensesService.GenerateRentRecords(ctx, &expenses.GenerateRentRecordsRequest{
		OfferId:                  old.ID.Hex(),
		Property:                 old.Property,
		Ivr:                      old.PropertyIvr,
		ListingId:                old.PropertyListing,
		Landlord:                 old.LandlordUser(),
		Tenant:                   old.TenantUser(),
		Manager:                  old.Manager,
		Viewing:                  old.Viewing,
		Rent:                     util.UnPtrFloat64(old.Conditions.Price),
		BreakClause:              util.UnPtrInt32(old.Conditions.BreakClause),
		RentPaymentDay:           util.UnPtrInt32(old.Conditions.RentPaymentDay),
		TermMonths:               util.UnPtrInt32(old.Conditions.TermMonths),
		MoveinDate:               timestamppb.New(util.UnPtrTime(old.Conditions.MoveinDate)),
		FirstRentPaymentDate:     timestamppb.New(util.UnPtrTime(old.Conditions.FirstRentPaymentDate)),
		RentCollectionFeePercent: old.RentPlan.Percent(),
		Legacy:                   old.RentPlan == offer.RentPlan_RENT_PLAN_LEGACY,
	})
	if err != nil {
		return err
	}

	old.ScheduleGenerated = true
	err = markOfferScheduleGenerated(ctx, old.ID.Hex())
	if err != nil {
		return err
	}

	return err
}

func markOfferScheduleGenerated(ctx context.Context, offID string) error {
	return mongolocal.Transaction(ctx, func() error {
		off, err := model.GetOfferByHexID(ctx, offID)
		if err != nil {
			return err
		}

		off.ScheduleGenerated = true
		off.UpdatedAt = time.Now()

		err = model.UpdateOffer(ctx, off)

		return err
	})
}

// RenderCSV renders offer details as CSV file
func RenderCSV(ctx context.Context, old *model.OfferModel, notify bool) (fileID string, err error) {
	prop, err := connections.PropertyService.Get(grpcctx.WithServiceUser(ctx), &property.GetRequest{
		Id: old.Property,
	})
	if err != nil {
		return
	}

	header := `"Agent name", "Landlord name", "Risk Address", "Duration of tenancy", "Start date of tenancy",` +
		` "End date of tenancy", "LPG Premium", "Annual Rent", "Tenant Name", "Policy Type"`
	agentName := fmt.Sprintf("%s", "Mashroom Ltd")
	landlordName := fmt.Sprintf(
		"%s %s",
		util.UnPtrString(old.Landlord.FirstName),
		util.UnPtrString(old.Landlord.LastName),
	)
	riskAddress := fmt.Sprintf(
		"%s %s %s %s %s",
		prop.GetProperty().GetLocation().GetBuilding().GetValue(),
		prop.GetProperty().GetLocation().GetFlat().GetValue(),
		prop.GetProperty().GetLocation().GetStreet().GetValue(),
		prop.GetProperty().GetLocation().GetCity().GetValue(),
		prop.GetProperty().GetLocation().GetPostcode().GetValue(),
	)
	rentTerm := util.UnPtrInt32(old.Conditions.TermMonths)
	moveInDate := old.Conditions.MoveinDate
	condPrice := util.UnPtrFloat64(old.Conditions.Price)
	moveOutDate := moveInDate.AddDate(0, int(rentTerm), 0)
	lpgPremium := (condPrice * 12.0) * 0.0019
	annualRent := condPrice * 12.0
	tenantName := fmt.Sprintf(
		"%s %s",
		util.UnPtrString(old.Tenants[0].FirstName),
		util.UnPtrString(old.Tenants[0].LastName),
	)
	body := fmt.Sprintf(
		"%s, %s, %s, %d, %s, %s, %.2f, %.2f, %s, %s",
		agentName,
		landlordName,
		strings.ReplaceAll(riskAddress, "  ", " "),
		rentTerm,
		moveInDate.Format("2006-01-02"),
		moveOutDate.Format("2006-01-02"),
		lpgPremium,
		annualRent,
		tenantName,
		"New Business",
	)
	text := header + "\n" + body

	now := time.Now()
	name := fmt.Sprintf(
		"New-Deposit-Replacement-Product-order-%04d-%02d-%02d.%s",
		now.Year(),
		now.Month(),
		now.Day(),
		"csv",
	)

	resp, err := document.UploadFile(
		ctx,
		connections.DocumentService,
		name,
		"text/csv",
		&document.FileAttribute{
			Data: map[string]string{
				document.FileAttributeGenerated.String(): "true",
			},
		},
		strings.NewReader(text),
	)
	if err != nil {
		return
	}

	if notify {
		_, _ = connections.NotificationService.SendNotification(ctx, &notification.SendNotificationRequest{
			Entity: notification.Entity_ENTITY_OFFER_DEPOSIT_REPLACEMENT,
			Action: notification.Action_ACTION_CREATE,
			Targets: []*notification.Target{
				{
					Role: notification.Role_ROLE_ADMIN,
					Contacts: &notification.Contacts{
						Email: "orders@mashroom.com",
					},
				},
			},
			Embeds: []*notification.EventEmbed{
				{
					EmbedType:  messaging.EmbedType_EMBED_TYPE_FILE,
					EmbedValue: resp.GetId(),
				},
			},
		})
	}

	return resp.GetId(), nil
}
