package model

import (
	"bytes"
	"context"
	"errors"
	"fmt"
	"strconv"
	"strings"
	"time"

	account "gitlab.mashplace.com/mashplace/backend/account/api"
	cardpayment "gitlab.mashplace.com/mashplace/backend/cardpayment/api"
	q "gitlab.mashplace.com/mashplace/backend/common/api/generic/query"
	"gitlab.mashplace.com/mashplace/backend/common/api/grpcctx"
	"gitlab.mashplace.com/mashplace/backend/common/api/grpcerr"
	"gitlab.mashplace.com/mashplace/backend/common/generic/bind"
	"gitlab.mashplace.com/mashplace/backend/common/generic/log"
	"gitlab.mashplace.com/mashplace/backend/common/generic/mongoquery"
	"gitlab.mashplace.com/mashplace/backend/common/generic/unwrap"
	"gitlab.mashplace.com/mashplace/backend/common/generic/util"
	"gitlab.mashplace.com/mashplace/backend/common/generic/wrap"
	contract "gitlab.mashplace.com/mashplace/backend/contract/api"
	document "gitlab.mashplace.com/mashplace/backend/document/api"
	extras "gitlab.mashplace.com/mashplace/backend/extras/api"
	identity "gitlab.mashplace.com/mashplace/backend/identity/api"
	invoice "gitlab.mashplace.com/mashplace/backend/invoice/api"
	listing "gitlab.mashplace.com/mashplace/backend/listing/api"
	messaging "gitlab.mashplace.com/mashplace/backend/messaging/api"
	notification "gitlab.mashplace.com/mashplace/backend/notification/api"
	offer "gitlab.mashplace.com/mashplace/backend/offer/api"
	"gitlab.mashplace.com/mashplace/backend/offer/ctxmgr"
	"gitlab.mashplace.com/mashplace/backend/offer/docs"
	"gitlab.mashplace.com/mashplace/backend/offer/internal/connections"
	property "gitlab.mashplace.com/mashplace/backend/property/api"
	tenancy "gitlab.mashplace.com/mashplace/backend/tenancy/api"
	"go.mongodb.org/mongo-driver/bson"
	"go.mongodb.org/mongo-driver/bson/primitive"
	"go.mongodb.org/mongo-driver/mongo"
	"go.mongodb.org/mongo-driver/mongo/options"
	"google.golang.org/protobuf/types/known/timestamppb"
	"google.golang.org/protobuf/types/known/wrapperspb"
)

// offer contract template IDs
var (
	TemplateASTID  string
	TemplateSOMFID string
)

// Offer conditions rent date
const (
	// CreateAction for history table
	CreateAction string = "create"
	// UpdateAction for history table
	UpdateAction string = "update"
)

// GuarantorOfferModel represents data model for offer guarantor
type GuarantorOfferModel struct {
	PhonePrimary         *string           `bson:"phone_primary"`
	RentprofileCreatedAt *time.Time        `bson:"rentprofile_created_at"`
	BillingAddress       *string           `bson:"billing_address"`
	FirstName            *string           `bson:"first_name"`
	LastName             *string           `bson:"last_name"`
	EmailPrimary         *string           `bson:"email_primary"`
	RentprofileStatus    *string           `bson:"rentprofile_status"`
	RiskURL              *string           `bson:"risk_url"`
	RiskDocument         *string           `bson:"risk_document"`
	ReferencingInvoiceID *string           `bson:"referencing_invoice_id"`
	User                 string            `bson:"user"`
	Check                string            `bson:"check"`
	Signature            []*SignatureModel `bson:"signature"`
	Risk                 offer.TenantRisk  `bson:"risk"`
	CheckNeeded          bool              `bson:"check_needed"`
}

// ActiveCheck returns true if guarantor is currently under reference checking
func (model *GuarantorOfferModel) ActiveCheck(drp bool) bool {
	return (drp || model.CheckNeeded) && model.Risk <= offer.TenantRisk_PENDING_RISK
}

// SelectedCheck returns true if guarantor were selected for reference checking
func (model *GuarantorOfferModel) SelectedCheck(drp bool) bool {
	return drp || model.CheckNeeded || model.Risk > 0
}

func retriveSignature(list []*SignatureModel, state offer.SignStatusState) *SignatureModel {
	for _, s := range list {
		if s != nil && s.State == state {
			return s
		}
	}

	return nil
}

// RetriveSignature retrives guarantor signature with specified state.
// Method is safe to use with nil receiver.
func (model *GuarantorOfferModel) RetriveSignature(state offer.SignStatusState) *SignatureModel {
	if model == nil {
		return nil
	}

	return retriveSignature(model.Signature, state)
}

// TenantOfferModel represents data model for offer tenant
type TenantOfferModel struct {
	Paying               *float64               `bson:"paying"`
	RiskURL              *string                `bson:"risk_url"`
	AnnualIncome         *float64               `bson:"annual_income"`
	FirstName            *string                `bson:"first_name"`
	LastName             *string                `bson:"last_name"`
	EmailPrimary         *string                `bson:"email_primary"`
	PhonePrimary         *string                `bson:"phone_primary"`
	KidsNo               *int32                 `bson:"kids_no"`
	BillingAddress       *string                `bson:"billing_address"`
	RiskDocument         *string                `bson:"risk_document"`
	ReferencingInvoiceID *string                `bson:"referencing_invoice_id"`
	Age                  *int32                 `bson:"age"`
	RentprofileStatus    *string                `bson:"rentprofile_status"`
	RentprofileCreatedAt *time.Time             `bson:"rentprofile_created_at"`
	User                 string                 `bson:"user"`
	Check                string                 `bson:"check"`
	Signature            []*SignatureModel      `bson:"signature"`
	Risk                 offer.TenantRisk       `bson:"risk"`
	EmploymentStatus     offer.EmploymentStatus `bson:"employment_status"`
	CheckNotified        bool                   `bson:"check_notified"`
	CheckNeeded          bool                   `bson:"check_needed"`
}

// ActiveCheck returns true if tenant is currently under reference checking
func (model *TenantOfferModel) ActiveCheck(drp bool) bool {
	return (drp || model.CheckNeeded) && model.Risk <= offer.TenantRisk_PENDING_RISK
}

// SelectedCheck returns true if tenant were selected for reference checking
func (model *TenantOfferModel) SelectedCheck(drp bool) bool {
	return drp || model.CheckNeeded || model.Risk > 0
}

// RetriveSignature retrives tenant signature with specified state.
// Method is safe to use with nil receiver.
func (model *TenantOfferModel) RetriveSignature(state offer.SignStatusState) *SignatureModel {
	if model == nil {
		return nil
	}

	return retriveSignature(model.Signature, state)
}

// OfferConditionsModel represents data model for offer conditions.
type OfferConditionsModel struct {
	Comments                      *string           `bson:"comments"`
	TermMonths                    *int32            `bson:"term_months"`
	MoveinDate                    *time.Time        `bson:"movein_date"`
	EndDate                       time.Time         `bson:"end_date"`
	Pets                          *bool             `bson:"pets"`
	PetsDescription               *string           `bson:"pets_description"`
	Price                         *float64          `bson:"price"`
	TenantsPitch                  *string           `bson:"tenants_pitch"`
	PaymentFrequency              *int32            `bson:"payment_frequency"`
	BreakClause                   *int32            `bson:"break_clause"`
	BreakUpClauseDate             *time.Time        `bson:"break_up_clause_date"`
	FirstRentPaymentDate          *time.Time        `bson:"first_rent_payment_date"`
	RentPaymentDay                *int32            `bson:"rent_payment_day"`
	TerminateDate                 *time.Time        `bson:"terminate_date"`
	DepositType                   offer.DepositType `bson:"deposit_type"`
	IsPaymentDayWasLastDayInMonth bool              `bson:"is_payment_day_was_last_day_in_month"`
}

// SignatureModel represents offer signature data model.
type SignatureModel struct {
	Time  time.Time             `bson:"time"`
	State offer.SignStatusState `bson:"state"`
}

// LandlordModel represents landlord data model.
type LandlordModel struct {
	FirstName      *string           `bson:"first_name"`
	LastName       *string           `bson:"last_name"`
	EmailPrimary   *string           `bson:"email_primary"`
	PhonePrimary   *string           `bson:"phone_primary"`
	BillingAddress *string           `bson:"billing_address"`
	User           string            `bson:"user"`
	Signature      []*SignatureModel `bson:"signature"`
}

// RetriveSignature retrives landlord signature with specified state.
// Method is safe to use with nil receiver.
func (model *LandlordModel) RetriveSignature(state offer.SignStatusState) *SignatureModel {
	if model == nil {
		return nil
	}

	return retriveSignature(model.Signature, state)
}

// OfferDetailsModel represents data model of offer details.
type OfferDetailsModel struct {
	MovinAlone     *bool `bson:"movin_alone"`
	AdultFlatmates *bool `bson:"adult_flatmates"`
}

// AdminApproveModel represents data model for approve by admin event.
type AdminApproveModel struct {
	ApprovedAt time.Time `bson:"approved_at"`
	Admin      string    `bson:"admin"`
	Approved   bool      `bson:"approved"`
}

// OfferReadySigningModel represents data model for ready for signing
type OfferReadySigningModel struct {
	Landlord bool `bson:"landlord"`
	Tenant   bool `bson:"tenant"`
}

// OfferTimeNowString returns either time.Now or offer's test_now field
func OfferTimeNowString(ctx context.Context, id string) time.Time {
	oid, err := primitive.ObjectIDFromHex(id)
	if err != nil {
		return time.Now()
	}

	return OfferTimeNow(ctx, oid)
}

// OfferTimeNowInstance returns either time.Now or offer's test_now field
func OfferTimeNowInstance(ctx context.Context, off *OfferModel) time.Time {
	if off == nil || off.TestNow.IsZero() {
		return time.Now()
	}

	return off.TestNow
}

// OfferTimeNow returns either time.Now or offer's test_now field
func OfferTimeNow(ctx context.Context, id primitive.ObjectID) time.Time {
	var off struct {
		TestNow time.Time `bson:"test_now,omitempty"`
	}

	err := Offers.FindOne(ctx, bson.M{"_id": id}, options.FindOne().SetProjection(bson.M{
		"test_now": 1,
	})).Decode(&off)
	if err != nil || off.TestNow.IsZero() {
		return time.Now()
	}

	return off.TestNow
}

// OfferSync synchronization states
type OfferSync struct {
	Listing bool `bson:"listing"`
}

// Reset synchronization states
func (s *OfferSync) Reset() {
	s.Listing = false
}

// OfferModel represents offer data model.
type OfferModel struct {
	ContractCreatedAt       time.Time                                          `bson:"contract_created_at"`
	DraftContractCreatedAt  time.Time                                          `bson:"draft_contract_created_at"`
	CreatedAt               time.Time                                          `bson:"created_at"`
	UpdatedAt               time.Time                                          `bson:"updated_at"`
	UpdatedByTenantAt       time.Time                                          `bson:"updated_by_tenant_at"`
	AcceptedAt              time.Time                                          `bson:"accepted_at"`
	DeniedAt                time.Time                                          `bson:"denied_at"`
	CancelledAt             time.Time                                          `bson:"cancelled_at"`
	StateChangedAt          time.Time                                          `bson:"state_changed_at"`
	TestNow                 time.Time                                          `bson:"test_now,omitempty"`
	RentPlanSelectedAt      time.Time                                          `bson:"rent_plan_selected_at"`
	Notifications           OfferNotifications                                 `bson:"notifications"`
	ScheduleNotifications   map[ScheduleNotificationType]*ScheduleNotification `bson:"schedule_notifications"`
	PaidStatuses            *PaidStatusesModel                                 `bson:"paid_statuses"`
	ReadyForSigning         *OfferReadySigningModel                            `bson:"ready_for_signing"`
	Approve                 *AdminApproveModel                                 `bson:"approve"`
	Details                 *OfferDetailsModel                                 `bson:"details"`
	Landlord                *LandlordModel                                     `bson:"landlord"`
	Conditions              *OfferConditionsModel                              `bson:"conditions"`
	AwaitingTermination     string                                             `bson:"awaiting_termination"`
	Administrator           string                                             `bson:"administrator"`
	DraftContract           string                                             `bson:"draft_contract"`
	SignatureRequestID      string                                             `bson:"signature_request_id"`
	OriginalOffer           string                                             `bson:"original_offer"`
	PrevTenancy             string                                             `bson:"prev_tenancy"`
	Viewing                 string                                             `bson:"viewing"`
	OpportunityID           string                                             `bson:"opportunity_id"`
	PropertyAddress         string                                             `bson:"property_address"`
	PropertyIvr             string                                             `bson:"property_ivr"`
	WorkflowInstanceID      string                                             `bson:"workflow_instance_id"`
	Manager                 string                                             `bson:"manager"`
	PropertyListing         string                                             `bson:"property_listing"`
	Property                string                                             `bson:"property"`
	Contract                string                                             `bson:"conract"`
	Clause                  []*ClauseModel                                     `bson:"clause"`
	Landlords               []*LandlordModel                                   `bson:"landlords"`
	Log                     []*LogModel                                        `bson:"log"`
	Guarantors              []*GuarantorOfferModel                             `bson:"guarantors"`
	Tenants                 []*TenantOfferModel                                `bson:"tenants"`
	LeadTenant              *TenantOfferModel                                  `bson:"lead_tenant"`
	State                   offer.OfferState                                   `bson:"state"`
	RentPlan                offer.RentPlan                                     `bson:"rent_plan"`
	ID                      primitive.ObjectID                                 `bson:"_id"`
	SignedInvoicesGenerated bool                                               `bson:"signed_invoices_generated"`
	IsLocked                bool                                               `bson:"is_locked"`
	ScheduleGenerated       bool                                               `bson:"schedule_generated"`
	MoveoutScheduled        bool                                               `bson:"moveout_scheduled"`
	GuarantorRequired       bool                                               `bson:"guarantor_required"`
	OpportunityTerminated   bool                                               `bson:"opportunity_terminated"`
	RentInsurancesGenerated bool                                               `bson:"rent_insurances_generated,omitempty"`
	Sync                    OfferSync                                          `bson:"sync"`
	Test                    bool                                               `bson:"test"`
	Deleted                 bool                                               `bson:"deleted"`
	Renewal                 bool                                               `bson:"renewal"`
	IsRto                   bool                                               `bson:"is_rto"`
	InvoiceID               string                                             `bson:"invoice_id"`
}

// OfferHistoryModel represents offer data model.
type OfferHistoryModel struct {
	StateChangedAt          time.Time                                          `bson:"state_changed_at"`
	CreatedAt               time.Time                                          `bson:"created_at"`
	TestNow                 time.Time                                          `bson:"test_now,omitempty"`
	ContractCreatedAt       time.Time                                          `bson:"contract_created_at"`
	UpdatedByTenantAt       time.Time                                          `bson:"updated_by_tenant_at"`
	AcceptedAt              time.Time                                          `bson:"accepted_at"`
	DeniedAt                time.Time                                          `bson:"denied_at"`
	CancelledAt             time.Time                                          `bson:"cancelled_at"`
	UpdatedAt               time.Time                                          `bson:"updated_at"`
	DraftContractCreatedAt  time.Time                                          `bson:"draft_contract_created_at"`
	Landlord                *LandlordModel                                     `bson:"landlord"`
	ScheduleNotifications   map[ScheduleNotificationType]*ScheduleNotification `bson:"schedule_notifications"`
	PaidStatuses            *PaidStatusesModel                                 `bson:"paid_statuses"`
	ReadyForSigning         *OfferReadySigningModel                            `bson:"ready_for_signing"`
	Notifications           OfferNotifications                                 `bson:"notifications"`
	Approve                 *AdminApproveModel                                 `bson:"approve"`
	Details                 *OfferDetailsModel                                 `bson:"details"`
	Conditions              *OfferConditionsModel                              `bson:"conditions"`
	Property                string                                             `bson:"property"`
	PropertyAddress         string                                             `bson:"property_address"`
	DraftContract           string                                             `bson:"draft_contract"`
	SignatureRequestID      string                                             `bson:"signature_request_id"`
	OriginalOffer           string                                             `bson:"original_offer"`
	PrevTenancy             string                                             `bson:"prev_tenancy"`
	AwaitingTermination     string                                             `bson:"awaiting_termination"`
	OpportunityID           string                                             `bson:"opportunity_id"`
	OfferID                 string                                             `bson:"offer_id"`
	Action                  string                                             `bson:"action"`
	Viewing                 string                                             `bson:"viewing"`
	WorkflowInstanceID      string                                             `bson:"workflow_instance_id"`
	Administrator           string                                             `bson:"administrator"`
	Manager                 string                                             `bson:"manager"`
	PropertyIvr             string                                             `bson:"property_ivr"`
	PropertyListing         string                                             `bson:"property_listing"`
	Contract                string                                             `bson:"conract"`
	Landlords               []*LandlordModel                                   `bson:"landlords"`
	Clause                  []*ClauseModel                                     `bson:"clause"`
	Log                     []*LogModel                                        `bson:"log"`
	Guarantors              []*GuarantorOfferModel                             `bson:"guarantors"`
	Tenants                 []*TenantOfferModel                                `bson:"tenants"`
	State                   offer.OfferState                                   `bson:"state"`
	RentPlan                offer.RentPlan                                     `bson:"rent_plan"`
	ID                      primitive.ObjectID                                 `bson:"_id,omitempty"`
	IsLocked                bool                                               `bson:"is_locked"`
	SignedInvoicesGenerated bool                                               `bson:"signed_invoices_generated"`
	ScheduleGenerated       bool                                               `bson:"schedule_generated"`
	MoveoutScheduled        bool                                               `bson:"moveout_scheduled"`
	GuarantorRequired       bool                                               `bson:"guarantor_required"`
	OpportunityTerminated   bool                                               `bson:"opportunity_terminated"`
	Sync                    OfferSync                                          `bson:"sync"`
	Test                    bool                                               `bson:"test"`
	Deleted                 bool                                               `bson:"deleted"`
}

// DepositReplacement returns true if offer follows deposit replacement semantics
func (model *OfferModel) DepositReplacement() bool {
	return model != nil && model.Conditions != nil &&
		model.Conditions.DepositType == offer.DepositType_DEPOSIT_REPLACEMENT
}

// ActiveChecks returns true if there is any active referencing checks
func (model *OfferModel) ActiveChecks() bool {
	if model == nil {
		return false
	}

	drp := model.DepositReplacement()

	for _, t := range model.Tenants {
		if t.ActiveCheck(drp) {
			return true
		}
	}

	for _, g := range model.Guarantors {
		if g.ActiveCheck(drp) {
			return true
		}
	}

	return true
}

// ScheduleNotificationType identifies distinct notification type
type ScheduleNotificationType int32

// scheduled notification values, avoid using iota
const (
	ScheduleNotificationNotResponded3hrs ScheduleNotificationType = 1
	ScheduleNotificationWeekLeftToMovein ScheduleNotificationType = 2
)

// TODO: consider moving notification checks to notification service (i.e. 'once' parameter)

// NotificationType identifies distinct notification type
type NotificationType string

// notification values, avoid using iota
const (
	NotificationDepositBankPaid     NotificationType = "deposit_bank_paid"
	NotificationDepositBankVerified NotificationType = "deposit_bank_verified"
)

// OfferNotifications stores time for notifications sent
type OfferNotifications map[NotificationType]time.Time

// Sent checks if notification was already sent
func (model OfferNotifications) Sent(nType NotificationType) bool {
	notifiedAt, ok := model[nType]
	return ok && notifiedAt.IsZero()
}

// TenantUser returns main (first) tenant user ID from tenants list or empty string otherwise.
// Method is safe to use with nil receiver.
func (model *OfferModel) TenantUser() string {
	if t := model.TenantModel(); t != nil {
		return t.User
	}

	return ""
}

// TenantModel returns data model for the main (first) tenant from tenants list or nil.
// Method is safe to use with nil receiver.
func (model *OfferModel) TenantModel() *TenantOfferModel {
	if model == nil || len(model.Tenants) == 0 {
		return nil
	}

	return model.Tenants[0]
}

// TenantMatch finds first tenant for which match condition returns true or returns nil otherwise.
func (model *OfferModel) TenantMatch(match func(*TenantOfferModel) bool) *TenantOfferModel {
	if model == nil {
		return nil
	}

	for _, t := range model.Tenants {
		if t != nil && match(t) {
			return t
		}
	}

	return nil
}

// LandlordUser returns landlord user ID if landlord model was properly set or empty string otherwise.
func (model *OfferModel) LandlordUser() string {
	if model == nil || model.Landlord == nil {
		return ""
	}

	return model.Landlord.User
}

// LandlordModel returns landlord model or nil otherwise.
func (model *OfferModel) LandlordModel() *LandlordModel {
	if model == nil {
		return nil
	}

	return model.Landlord
}

// SignatureList returns landlord signature or nil slice
func (model *LandlordModel) SignatureList() []*SignatureModel {
	if model == nil {
		return nil
	}

	return model.Signature
}

// SignatureList returns tenant signature or nil slice
func (model *TenantOfferModel) SignatureList() []*SignatureModel {
	if model == nil {
		return nil
	}

	return model.Signature
}

// GuarantorMatch finds first guarantor for which match condition returns true or returns nil otherwise.
func (model *OfferModel) GuarantorMatch(match func(*GuarantorOfferModel) bool) *GuarantorOfferModel {
	if model == nil {
		return nil
	}

	for _, g := range model.Guarantors {
		if g != nil && match(g) {
			return g
		}
	}

	return nil
}

// ScheduleNotification ...
type ScheduleNotification struct {
	StartedAt  time.Time `bson:"started_at"`
	IsNotified bool      `bson:"is_notified"`
}

// PaidStatusesModel - ...
type PaidStatusesModel struct {
	// TenantServices flag indicates first rent paid or not
	TenantServices bool `bson:"tenant_services"`
	// StandingOrder flag indicates the setup of automatic funds write-off has started
	StandingOrder bool `bson:"standing_order"`
	// StandingOrderSetup flag indicates the setup of automatic funds write-off completed
	StandingOrderSetup bool `bson:"standing_order_setup"`
}

// ClauseModel represents offer clause data model.
type ClauseModel struct {
	CreatedAt time.Time         `bson:"created_at"`
	ID        string            `bson:"id"`
	Text      string            `bson:"text"`
	State     offer.ClauseState `bson:"state"`
}

// FieldChangelogModel represents a field values changes.
type FieldChangelogModel struct {
	Old   interface{} `bson:"old"`
	New   interface{} `bson:"new"`
	Field string      `bson:"field"`
}

// LogModel represents offer log data model.
type LogModel struct {
	Action string    `bson:"action"`
	Dt     time.Time `bson:"dt"`
	User   string    `bson:"user"`
}

const mimeApplicationPDF = "application/pdf"
const completeStatus = "Complete"

// ConvertRisk converts identity.TenantRisk into offer.TenantRisk value.
func ConvertRisk(risk identity.TenantRisk) offer.TenantRisk {
	tenantRisk := offer.TenantRisk_UNKNOWN_TENANT_RISK
	switch risk {
	// case identity.TenantRisk_UNKNOWN_TENANT_RISK:
	//	tenantRisk = offer.TenantRisk_UNKNOWN_TENANT_RISK
	case identity.TenantRisk_PENDING_RISK:
		tenantRisk = offer.TenantRisk_PENDING_RISK
	case identity.TenantRisk_LOW_RISK:
		tenantRisk = offer.TenantRisk_LOW_RISK
	case identity.TenantRisk_MEDIUM_RISK:
		tenantRisk = offer.TenantRisk_MEDIUM_RISK
	case identity.TenantRisk_HIGH_RISK:
		tenantRisk = offer.TenantRisk_HIGH_RISK
	}
	return tenantRisk
}

// RenderGuarantorOfferModel converts model.GuarantorOfferModel into gRPC API data.
func (model *GuarantorOfferModel) RenderGuarantorOfferModel() *offer.Guarantor {
	if model == nil {
		return nil
	}
	return &offer.Guarantor{
		User:                 wrapperspb.String(model.User),
		FirstName:            wrap.StringValue(model.FirstName),
		LastName:             wrap.StringValue(model.LastName),
		EmailPrimary:         wrap.StringValue(model.EmailPrimary),
		PhonePrimary:         wrap.StringValue(model.PhonePrimary),
		Risk:                 model.Risk,
		RiskProfile:          wrap.StringValue(model.RiskURL),
		RiskDocument:         wrap.StringValue(model.RiskDocument),
		BillingAddress:       wrap.StringValue(model.BillingAddress),
		Signature:            SignatureModelArray(model.Signature).RenderSignatureModel(),
		Check:                wrap.StringValue(&model.Check),
		RentprofileStatus:    wrap.StringValue(model.RentprofileStatus),
		RentprofileCreatedAt: wrap.TimestampValue(model.RentprofileCreatedAt),
	}
}

// RenderTenantOfferModel converts model.TenantOfferModel into gRPC API data.
func (model *TenantOfferModel) RenderTenantOfferModel() *offer.Tenant {
	return &offer.Tenant{
		User:                 model.User,
		FirstName:            wrap.StringValue(model.FirstName),
		LastName:             wrap.StringValue(model.LastName),
		EmailPrimary:         wrap.StringValue(model.EmailPrimary),
		PhonePrimary:         wrap.StringValue(model.PhonePrimary),
		Age:                  wrap.Int32Value(model.Age),
		EmploymentStatus:     model.EmploymentStatus,
		AnnualIncome:         wrap.DoubleValue(model.AnnualIncome),
		KidsNo:               wrap.Int32Value(model.KidsNo),
		Paying:               wrap.DoubleValue(model.Paying),
		Signature:            SignatureModelArray(model.Signature).RenderSignatureModel(),
		Risk:                 model.Risk,
		RiskProfile:          wrap.StringValue(model.RiskURL),
		RiskDocument:         wrap.StringValue(model.RiskDocument),
		BillingAddress:       wrap.StringValue(model.BillingAddress),
		Check:                wrap.StringValue(&model.Check),
		RentprofileStatus:    wrap.StringValue(model.RentprofileStatus),
		RentprofileCreatedAt: wrap.TimestampValue(model.RentprofileCreatedAt),
	}
}

// RenderLandlordModel converts model.LandlordModel into gRPC API data.
func (model *LandlordModel) RenderLandlordModel() *offer.Landlord {
	return &offer.Landlord{
		User:           model.User,
		FirstName:      wrap.StringValue(model.FirstName),
		LastName:       wrap.StringValue(model.LastName),
		Signature:      SignatureModelArray(model.Signature).RenderSignatureModel(),
		EmailPrimary:   wrap.StringValue(model.EmailPrimary),
		PhonePrimary:   wrap.StringValue(model.PhonePrimary),
		BillingAddress: wrap.StringValue(model.BillingAddress),
	}
}

// RenderOfferChangelog converts model.OfferChangelog into offer.OfferChangelog
func (model *OfferChangelog) RenderOfferChangelog() *offer.OfferChangelog {
	return &offer.OfferChangelog{
		Action:    model.Action,
		CreatedAt: wrap.TimestampValue(&model.CreatedAt),
		Old:       model.Old,
		New:       model.New,
		Initiator: model.Initiator,
	}
}

// RenderOfferConditionsModel converts model.OfferConditionsModel into gRPC API data.
func (model *OfferConditionsModel) RenderOfferConditionsModel() *offer.OfferConditions {
	return &offer.OfferConditions{
		Price:                wrap.DoubleValue(model.Price),
		DepositType:          model.DepositType,
		TermMonths:           wrap.Int32Value(model.TermMonths),
		MoveinDate:           wrap.TimestampValue(model.MoveinDate),
		Pets:                 wrap.BoolValue(model.Pets),
		PetsDescription:      wrap.StringValue(model.PetsDescription),
		Comments:             wrap.StringValue(model.Comments),
		TenantsPitch:         wrap.StringValue(model.TenantsPitch),
		PaymentFrequency:     wrap.Int32Value(model.PaymentFrequency),
		BreakClause:          wrap.Int32Value(model.BreakClause),
		FirstRentPaymentDate: wrap.TimestampValue(model.FirstRentPaymentDate),
		RentPaymentDay:       wrap.Int32Value(model.RentPaymentDay),
	}
}

// SignatureModelArray represents a list of model.SignatureModel.
type SignatureModelArray []*SignatureModel

// RenderSignatureModel converts model.SignatureModelArray into gRPC API data.
func (model SignatureModelArray) RenderSignatureModel() *offer.SignStatusArrayValue {
	res := &offer.SignStatusArrayValue{}
	for _, s := range model {
		res.Value = append(res.Value, &offer.SignStatus{
			State: s.State,
			Time:  timestamppb.New(s.Time),
		})
	}
	return res
}

// InProgress checks the model have in the progress logical state.
func (model *OfferModel) InProgress() bool {
	switch model.State {
	case offer.OfferState_DEPOSIT_CONFIRMED,
		offer.OfferState_CHECKING,
		offer.OfferState_AWAITING_FEES,
		offer.OfferState_AWAITING_CHECKS,
		offer.OfferState_AWAITING_DOCUMENTS,
		offer.OfferState_AWAITING_APPROVE,
		offer.OfferState_CONTRACT_PREPARATION,
		offer.OfferState_SIGNING,
		offer.OfferState_SIGNED,
		offer.OfferState_RENTPAID,
		offer.OfferState_AWAITING_APPLICANT_CHECK_PAYMENT,
		offer.OfferState_AWAITING_GUARANTOR,
		offer.OfferState_CONTRACT_REVIEW_AND_CLAUSES:
		return true
	}
	return false
}

// RenderScheduleNotification converts model.ScheduleNotification into gRPC API data
func (model *ScheduleNotification) RenderScheduleNotification() *offer.ScheduleNotification {
	return &offer.ScheduleNotification{
		StartedAt:  timestamppb.New(model.StartedAt),
		IsNotified: model.IsNotified,
	}
}

// RenderOfferModel converts model.OfferModel into gRPC API data.
func (model *OfferModel) RenderOfferModel(ctx context.Context, user string) *offer.Offer {
	var tenants []*offer.Tenant

	var guarantors []*offer.Guarantor

	for _, t := range model.Tenants {
		tenants = append(tenants, t.RenderTenantOfferModel())
	}

	for _, t := range model.Guarantors {
		guarantors = append(guarantors, t.RenderGuarantorOfferModel())
	}

	if model.Conditions == nil {
		model.Conditions = &OfferConditionsModel{}
	}

	if model.Landlord == nil {
		model.Landlord = &LandlordModel{}
	}

	if model.Details == nil {
		model.Details = &OfferDetailsModel{}
	}

	if model.Approve == nil {
		model.Approve = &AdminApproveModel{}
	}

	var logs []*offer.Log
	for _, l := range model.Log {
		logs = append(logs, &offer.Log{
			Action: l.Action,
			Dt:     timestamppb.New(l.Dt),
			User:   l.User,
		})
	}

	var changelogms []*OfferChangelog

	cursor, err := OfferLogs.Find(ctx, bson.M{"offer": model.ID.Hex()})
	if err != nil {
		log.Printf("OfferLogs.Find error: %s", err)
	} else {
		err = cursor.All(ctx, &changelogms)
		if err != nil {
			log.Printf("OfferLogs.Find error: %s", err)
		}
	}

	var changelogs []*offer.OfferChangelog

	for _, cms := range changelogms {
		changelogs = append(changelogs, cms.RenderOfferChangelog())
	}

	var clauses []*offer.Clause

	for _, c := range model.Clause {
		clauses = append(clauses, &offer.Clause{
			Id:        c.ID,
			Text:      c.Text,
			State:     c.State,
			CreatedAt: timestamppb.New(c.CreatedAt),
		})
	}

	var landlords = make([]*offer.Landlord, len(model.Landlords))
	for i, l := range model.Landlords {
		landlords[i] = l.RenderLandlordModel()
	}

	if model.ReadyForSigning == nil {
		model.ReadyForSigning = &OfferReadySigningModel{}
	}

	var scheduleNotifications = make(map[int32]*offer.ScheduleNotification)
	for eventType, scheduleNotification := range model.ScheduleNotifications {
		scheduleNotifications[int32(eventType)] = scheduleNotification.RenderScheduleNotification()
	}

	var notifications = make(map[string]*timestamppb.Timestamp)

	if model.Notifications == nil {
		model.Notifications = make(OfferNotifications)
	}

	for eventType, notifiedAt := range model.Notifications {
		if !notifiedAt.IsZero() {
			notifications[string(eventType)] = timestamppb.New(notifiedAt)
		}
	}

	if model.PaidStatuses == nil {
		model.PaidStatuses = &PaidStatusesModel{}
	}

	return &offer.Offer{
		Id:                     model.ID.Hex(),
		Property:               model.Property,
		PropertyListing:        model.PropertyListing,
		OriginalOffer:          model.OriginalOffer,
		Viewing:                model.Viewing,
		IsLocked:               model.IsLocked, // deprecate?
		Contract:               model.Contract,
		ContractCreatedAt:      timestamppb.New(model.ContractCreatedAt),
		DraftContract:          model.DraftContract,
		DraftContractCreatedAt: timestamppb.New(model.DraftContractCreatedAt),
		CreatedAtValue:         timestamppb.New(model.CreatedAt),
		UpdatedAtValue:         timestamppb.New(model.UpdatedAt),
		AcceptedAtValue:        timestamppb.New(model.AcceptedAt),
		DeniedAtValue:          timestamppb.New(model.DeniedAt),
		CancelledAtValue:       timestamppb.New(model.CancelledAt),
		RentPlanSelectedAt:     timestamppb.New(model.RentPlanSelectedAt),
		State:                  model.State,
		Tenants:                &offer.TenantArrayValue{Value: tenants},
		Guarantors:             &offer.GuarantorArrayValue{Value: guarantors},
		GuarantorRequired:      model.GuarantorRequired,
		Conditions:             model.Conditions.RenderOfferConditionsModel(),
		Landlord:               model.Landlord.RenderLandlordModel(),
		Landlords:              landlords,
		Details: &offer.OfferDetails{
			MoveinAlone:    wrap.BoolValue(model.Details.MovinAlone),
			AdultFlatmates: wrap.BoolValue(model.Details.AdultFlatmates),
		},
		Approve: &offer.Offer_ApproveValue{ApproveValue: &offer.OfferApprove{
			Admin:      model.Approve.Admin,
			Approved:   model.Approve.Approved,
			ApprovedAt: timestamppb.New(model.Approve.ApprovedAt),
		}},
		Deleted:     &offer.Offer_DeletedValue{DeletedValue: model.Deleted},
		Log:         logs,
		Test:        &offer.Offer_TestValue{TestValue: model.Test},
		Clause:      clauses,
		PropertyIvr: model.PropertyIvr,
		ReadyForSigning: &offer.OfferReadySigning{
			Landlord: wrapperspb.Bool(model.ReadyForSigning.Landlord),
			Tenant:   wrapperspb.Bool(model.ReadyForSigning.Tenant),
		},
		Changelog:             changelogs,
		WorkflowInstanceId:    model.WorkflowInstanceID,
		ScheduleNotifications: scheduleNotifications,
		Notifications:         notifications,
		Administrator:         model.Administrator,
		Manager:               model.Manager,
		ScheduleGenerated:     model.ScheduleGenerated,
		PaidStatuses: &offer.PaidStatuses{
			TenantsServices:    model.PaidStatuses.TenantServices,
			StandingOrder:      model.PaidStatuses.StandingOrder,
			StandingOrderSetup: model.PaidStatuses.StandingOrderSetup,
		},
		RentPlan:                model.RentPlan,
		RentInsurancesGenerated: model.RentInsurancesGenerated,
		IsRto:                   model.IsRto,
	}
}

// RenderInvoice converts invoice representation into gRPC API data.
func RenderInvoice(inv *invoice.Invoice) *offer.OfferPayment {
	var status offer.PaymentStatus

	switch inv.GetStatus() {
	case invoice.InvoiceStatus_INVOICE_FULFILLED:
		status = offer.PaymentStatus_PAYMENT_SUCCESS
	case invoice.InvoiceStatus_INVOICE_OPEN:
		status = offer.PaymentStatus_PAYMENT_PENDING
	case invoice.InvoiceStatus_INVOICE_CANCELED:
		status = offer.PaymentStatus_PAYMENT_FAILED
	}

	var typ offer.PaymentType

	var csrf string

	for _, p := range inv.GetPayments() {
		switch p.GetSystem().GetStripe().GetRequest().GetType() {
		case cardpayment.PaymentType_CARD:
			typ = offer.PaymentType_CARD
		case cardpayment.PaymentType_ACCOUNT:
			typ = offer.PaymentType_ACCOUNT
		case cardpayment.PaymentType_BANK_ACCOUNT:
			typ = offer.PaymentType_BANK_ACCOUNT
		case cardpayment.PaymentType_BITCOIN:
			typ = offer.PaymentType_BITCOIN
		case cardpayment.PaymentType_SOURCE:
			typ = offer.PaymentType_SOURCE
		}

		c := p.GetSystem().GetTokenio().GetResponse().GetCsrf()

		if c != "" {
			csrf = c
		}
	}

	var services []*extras.ServiceSnapshot

	for _, s := range inv.GetServices() {
		services = append(services, s.GetService())
	}

	return &offer.OfferPayment{
		ReceiptId:     inv.GetInvoiceId(),
		Amount:        inv.GetAmount(),
		Status:        status,
		Type:          typ,
		CreatedAt:     inv.GetCreatedAt(),
		UpdatedAt:     inv.GetUpdatedAt(),
		Services:      services,
		InvoiceNumber: inv.GetReference(),
		CsrfToken:     csrf,
	}
}

// UpdateAdminApprove changes model approve fieldset.
func (model *AdminApproveModel) UpdateAdminApprove(value bool, admin string) {
	if value && !model.Approved {
		model.Approved = true
		model.ApprovedAt = time.Now().UTC()
		model.Admin = admin
	} else if !value && model.Approved {
		model.Approved = false
		model.ApprovedAt = time.Unix(0, 0).UTC()
		model.Admin = ""
	}
}

// UpdateApproveOfferModel changes model.Approve fields values.
func (model *OfferModel) UpdateApproveOfferModel(approved bool, admin string) {
	if model.Approve == nil {
		model.Approve = &AdminApproveModel{}
	}
	model.Approve.UpdateAdminApprove(approved, admin)
}

// MatchTenant returns tenant matching userid/email from list
func MatchTenant(tenants []*TenantOfferModel, user, email string) *TenantOfferModel {
	for _, t := range tenants {
		switch {
		case email != "" && util.UnPtrString(t.EmailPrimary) == email,
			user != "" && t.User == user:
			return t
		}
	}

	return nil
}

// MatchGuarantor returns guarantor matching userid/email from list
func MatchGuarantor(guarantors []*GuarantorOfferModel, user, email string) *GuarantorOfferModel {
	for _, t := range guarantors {
		switch {
		case email != "" && util.UnPtrString(t.EmailPrimary) == email,
			user != "" && t.User == user:
			return t
		}
	}

	return nil
}

func (model *OfferModel) createTenant(ctx context.Context, user, email string) (*TenantOfferModel, error) {
	m := &TenantOfferModel{
		User: user,
	}

	mail := email
	m.EmailPrimary = &mail

	if model.OriginalOffer == "" {
		return m, nil
	}

	// If tenant which was checked on previous offering was removed from new one and then added again,
	// we should copy check and risk* fields of this tenant from previous offer.

	id, err := primitive.ObjectIDFromHex(model.OriginalOffer)
	if err != nil {
		return nil, err
	}

	origOff, err := GetOfferByID(ctx, id)
	if err != nil {
		return m, nil
	}

	ten := MatchTenant(origOff.Tenants, user, email)

	if ten != nil {
		m.Check = ten.Check
		m.RiskURL = ten.RiskURL
		m.RiskDocument = ten.RiskDocument
		m.Risk = ten.Risk
		m.RentprofileStatus = ten.RentprofileStatus
		m.RentprofileCreatedAt = ten.RentprofileCreatedAt
	}

	return m, nil
}

func (model *OfferModel) createGuarantor(ctx context.Context, user, email string) (*GuarantorOfferModel, error) {
	m := &GuarantorOfferModel{
		User: user,
	}

	mail := email
	m.EmailPrimary = &mail

	if model.OriginalOffer == "" {
		return m, nil
	}

	id, err := primitive.ObjectIDFromHex(model.OriginalOffer)
	if err != nil {
		return nil, err
	}

	origOff, err := GetOfferByID(ctx, id)
	if err != nil {
		return m, nil
	}

	gur := MatchGuarantor(origOff.Guarantors, user, email)

	if gur != nil {
		m.Check = gur.Check
		m.RiskURL = gur.RiskURL
		m.RiskDocument = gur.RiskDocument
		m.Risk = gur.Risk
		m.RentprofileStatus = gur.RentprofileStatus
		m.RentprofileCreatedAt = gur.RentprofileCreatedAt
	}

	return m, nil
}

// UpdateOfferGuarantorModel updates guarantor(s) in offer
func (model *OfferModel) UpdateOfferGuarantorModel(
	ctx context.Context,
	g *offer.Guarantor,
	create, uniqueCheck bool,
) (old *GuarantorOfferModel, err error) {
	old = MatchGuarantor(model.Guarantors, g.GetUser().GetValue(), g.GetEmailPrimary().GetValue())

	var newGuarantorObject bool

	if old == nil {
		newGuarantorObject = true

		if !create {
			return nil, ErrNoGuarantor
		}

		if uniqueCheck {
			err = model.uniqueEmailAddressCheck(g.GetEmailPrimary().GetValue())
			if err != nil {
				return nil, err
			}
		}

		old, err = model.createGuarantor(ctx, g.GetUser().GetValue(), g.GetEmailPrimary().GetValue())
		if err != nil {
			return
		}
	}

	oldemail := util.UnPtrString(old.EmailPrimary)

	if g.GetUser() != nil {
		old.User = g.GetUser().GetValue()
	}

	if g.GetFirstName() != nil {
		old.FirstName = unwrap.StringValue(g.GetFirstName())
	}

	if g.GetLastName() != nil {
		old.LastName = unwrap.StringValue(g.GetLastName())
	}

	if g.GetEmailPrimary() != nil {
		old.EmailPrimary = unwrap.StringValue(g.GetEmailPrimary())
	}

	if g.GetPhonePrimary() != nil {
		old.PhonePrimary = unwrap.StringValue(g.GetPhonePrimary())
	}

	if g.GetBillingAddress() != nil {
		old.BillingAddress = unwrap.StringValue(g.GetBillingAddress())
	}

	err = updateOfferGuarantorModelDetails(ctx, old)
	if err != nil {
		return nil, err
	}

	if oldemail != util.UnPtrString(old.EmailPrimary) {
		old.Check = ""
		old.RiskURL = util.PtrString("")
		old.RiskDocument = util.PtrString("")
		old.Risk = offer.TenantRisk_UNKNOWN_TENANT_RISK
		old.RentprofileStatus = nil
		old.RentprofileCreatedAt = nil
	}

	if newGuarantorObject {
		model.Guarantors = append(model.Guarantors, old)
	}

	return
}

// UpdateOfferTenantModel updates tenant(s) in offer
func (model *OfferModel) UpdateOfferTenantModel(
	ctx context.Context,
	t *offer.Tenant,
	create, uniqueCheck bool,
) (old *TenantOfferModel, err error) {
	old = MatchTenant(model.Tenants, t.GetUser(), t.GetEmailPrimary().GetValue())

	var newTenantObject bool

	if old == nil {
		newTenantObject = true

		if !create {
			return nil, ErrNoTenant
		}

		if uniqueCheck {
			err = model.uniqueEmailAddressCheck(t.GetEmailPrimary().GetValue())
			if err != nil {
				return nil, err
			}
		}

		old, err = model.createTenant(ctx, t.GetUser(), t.GetEmailPrimary().GetValue())
		if err != nil {
			return
		}
	}

	oldemail := util.UnPtrString(old.EmailPrimary)

	if t.GetBillingAddress() != nil {
		old.BillingAddress = unwrap.StringValue(t.GetBillingAddress())
	}

	if t.GetFirstName() != nil {
		old.FirstName = unwrap.StringValue(t.GetFirstName())
	}

	if t.GetLastName() != nil {
		old.LastName = unwrap.StringValue(t.GetLastName())
	}

	if t.GetEmailPrimary() != nil {
		old.EmailPrimary = unwrap.StringValue(t.GetEmailPrimary())
	}

	if t.GetPhonePrimary() != nil {
		old.PhonePrimary = unwrap.StringValue(t.GetPhonePrimary())
	}

	if t.GetAge() != nil {
		old.Age = unwrap.Int32Value(t.GetAge())
	}

	if t.GetEmploymentStatus() != offer.EmploymentStatus_UNKNOWN_EMPLOYMENT_STATUS {
		old.EmploymentStatus = t.GetEmploymentStatus()
	}

	if t.GetAnnualIncome() != nil {
		old.AnnualIncome = unwrap.DoubleValue(t.GetAnnualIncome())
	}

	if t.GetKidsNo() != nil {
		old.KidsNo = unwrap.Int32Value(t.GetKidsNo())
	}

	if t.GetPaying() != nil {
		old.Paying = unwrap.DoubleValue(t.GetPaying())
	}

	err = updateOfferTenantModelDetails(ctx, old)
	if err != nil {
		return nil, err
	}

	if oldemail != util.UnPtrString(old.EmailPrimary) {
		old.Check = ""
		old.RiskURL = util.PtrString("")
		old.RiskDocument = util.PtrString("")
		old.Risk = offer.TenantRisk_UNKNOWN_TENANT_RISK
		old.RentprofileStatus = nil
		old.RentprofileCreatedAt = nil
	}

	if newTenantObject {
		model.Tenants = append(model.Tenants, old)
	}

	return old, nil
}

func (model *OfferModel) uniqueEmailAddressCheck(email string) error {
	for _, oldg := range model.Guarantors {
		if util.UnPtrString(oldg.EmailPrimary) == email {
			return ErrNonUniqueTenantEmail
		}
	}

	for _, oldt := range model.Tenants {
		if util.UnPtrString(oldt.EmailPrimary) == email {
			return ErrNonUniqueTenantEmail
		}
	}

	return nil
}

func accountResolve(ctx context.Context, userid, email string) (*account.Account, error) {
	if userid != "" {
		acc, err := connections.AccountService.Get(ctx, &account.GetRequest{
			Id: userid,
		})
		if err != nil {
			return nil, err
		}

		return acc.GetAccount(), nil
	}

	if email != "" {
		acc, err := connections.AccountService.GetByEmail(ctx, &account.GetByEmailRequest{
			Email: email,
		})
		if err != nil {
			return nil, nil
		}

		return acc.GetAccount(), nil
	}

	return nil, nil
}

func updateOfferTenantModelDetails(ctx context.Context, old *TenantOfferModel) error {
	acc, err := accountResolve(ctx, old.User, util.UnPtrString(old.EmailPrimary))
	if err != nil || acc == nil {
		return err
	}

	if old.User == "" {
		old.User = acc.GetIdValue()
	}

	if util.UnPtrString(old.EmailPrimary) == "" {
		old.EmailPrimary = util.PtrString(acc.GetContactValue().GetEmailPrimaryValue())
	}

	if util.UnPtrString(old.PhonePrimary) == "" {
		old.PhonePrimary = util.PtrString(acc.GetContactValue().GetPhonePrimaryValue())
	}

	if util.UnPtrString(old.BillingAddress) == "" {
		old.BillingAddress = util.PtrString(GetBillingAddressFromLocation(acc.GetLocationValue()))
	}

	if util.UnPtrString(old.FirstName) == "" {
		old.FirstName = util.PtrString(acc.GetDetailsValue().GetFirstNameValue())
	}

	if util.UnPtrString(old.LastName) == "" {
		old.LastName = util.PtrString(acc.GetDetailsValue().GetLastNameValue())
	}

	if util.UnPtrInt32(old.Age) == 0 {
		t := acc.GetDetailsValue().GetBirthdayValue().AsTime()
		if !t.IsZero() && t.Unix() != 0 {
			old.Age = util.PtrInt32(int32(time.Since(t) / time.Hour * 24 * 365))
		}
	}

	return nil
}

func updateOfferGuarantorModelDetails(ctx context.Context, old *GuarantorOfferModel) error {
	acc, err := accountResolve(ctx, old.User, util.UnPtrString(old.EmailPrimary))
	if err != nil || acc == nil {
		return err
	}

	if old.User == "" {
		old.User = acc.GetIdValue()
	}

	if util.UnPtrString(old.EmailPrimary) == "" {
		old.EmailPrimary = util.PtrString(acc.GetContactValue().GetEmailPrimaryValue())
	}

	if util.UnPtrString(old.PhonePrimary) == "" {
		old.PhonePrimary = util.PtrString(acc.GetContactValue().GetPhonePrimaryValue())
	}

	if util.UnPtrString(old.BillingAddress) == "" {
		old.BillingAddress = util.PtrString(GetBillingAddressFromLocation(acc.GetLocationValue()))
	}

	if util.UnPtrString(old.FirstName) == "" {
		old.FirstName = util.PtrString(acc.GetDetailsValue().GetFirstNameValue())
	}

	if util.UnPtrString(old.LastName) == "" {
		old.LastName = util.PtrString(acc.GetDetailsValue().GetLastNameValue())
	}

	return nil
}

func updateOfferLandlordModelDetails(ctx context.Context, old *LandlordModel) error {
	acc, err := accountResolve(ctx, old.User, util.UnPtrString(old.EmailPrimary))
	if err != nil || acc == nil {
		return err
	}

	if old.User == "" {
		old.User = acc.GetIdValue()
	}

	if util.UnPtrString(old.EmailPrimary) == "" {
		old.EmailPrimary = util.PtrString(acc.GetContactValue().GetEmailPrimaryValue())
	}

	if util.UnPtrString(old.PhonePrimary) == "" {
		old.PhonePrimary = util.PtrString(acc.GetContactValue().GetPhonePrimaryValue())
	}

	if util.UnPtrString(old.BillingAddress) == "" {
		old.BillingAddress = util.PtrString(GetBillingAddressFromLocation(acc.GetLocationValue()))
	}

	if util.UnPtrString(old.FirstName) == "" {
		old.FirstName = util.PtrString(acc.GetDetailsValue().GetFirstNameValue())
	}

	if util.UnPtrString(old.LastName) == "" {
		old.LastName = util.PtrString(acc.GetDetailsValue().GetLastNameValue())
	}

	return nil
}

// UpdateTenancyModel synchronizes tenants data with property service.
func UpdateTenancyModel(ctx context.Context, of *OfferModel) error {
	if len(of.Tenants) > 0 {
		_, err := UpdateTenancy(ctx, of)
		if err != nil {
			return err
		}
	}
	return nil
}

// UpdateOfferModel updates model fields with specified gRPC API Offer data.
// FIXME context parameter is not used.
//
//nolint:gocognit // try to reduce high cognitive and cyclomatic complexty
func (model *OfferModel) UpdateOfferModel(ctx context.Context, req *offer.Offer) error {
	log.WithContext(ctx).Infof("offer model request = %v", *model)
	model.UpdatedAt = time.Now()

	if req.GetLandlord() != nil {
		if req.GetLandlord().GetBillingAddress() != nil {
			model.Landlord.BillingAddress = unwrap.StringValue(req.GetLandlord().GetBillingAddress())
		}
		if req.GetLandlord().GetFirstName() != nil {
			model.Landlord.FirstName = unwrap.StringValue(req.GetLandlord().GetFirstName())
		}
		if req.GetLandlord().GetLastName() != nil {
			model.Landlord.LastName = unwrap.StringValue(req.GetLandlord().GetLastName())
		}

		err := updateOfferLandlordModelDetails(ctx, model.Landlord)
		if err != nil {
			return err
		}
	}

	landlords := req.GetLandlords()
	err := model.updateOfferLandlords(ctx, landlords)
	if err != nil {
		return err
	}

	if req.GetTenants() != nil {
		uniq := make(map[string]bool)

		for _, t := range req.GetTenants().GetValue() {
			oldt, err := model.UpdateOfferTenantModel(ctx, t, true, true)
			if err != nil {
				return err
			}

			if oldt.EmailPrimary == nil {
				continue
			}

			if len(strings.TrimSpace(*oldt.EmailPrimary)) == 0 {
				return ErrInvalidTenantEmail
			}

			if _, ok := uniq[*oldt.EmailPrimary]; ok {
				return ErrNonUniqueTenantEmail
			}

			uniq[*oldt.EmailPrimary] = true
		}
	}

	if req.GetGuarantors() != nil {
		uniq := make(map[string]bool)

		for _, g := range req.GetGuarantors().GetValue() {
			oldg, err := model.UpdateOfferGuarantorModel(ctx, g, true, true)
			if err != nil {
				return err
			}

			if oldg.EmailPrimary == nil {
				continue
			}

			if len(strings.TrimSpace(*oldg.EmailPrimary)) == 0 {
				return ErrInvalidGuarantorEmail
			}

			if _, ok := uniq[*oldg.EmailPrimary]; ok {
				return ErrNonUniqueGuarantorEmail
			}

			uniq[*oldg.EmailPrimary] = true
		}
	}

	if err = model.updateConditions(req.GetConditions()); err != nil {
		return err
	}

	if model.Details == nil {
		model.Details = &OfferDetailsModel{}
	}

	if req.GetDetails() != nil {
		dets := req.GetDetails()
		old := model.Details
		if dets.GetMoveinAlone() != nil {
			old.MovinAlone = unwrap.BoolValue(dets.GetMoveinAlone())
		}
		if dets.GetAdultFlatmates() != nil {
			old.AdultFlatmates = unwrap.BoolValue(dets.GetAdultFlatmates())
		}
	}

	return nil
}

// UpdateLandlordModel updates landlord info
func (model *LandlordModel) UpdateLandlordModel(req *offer.Landlord) {
	if model == nil {
		return
	}

	if model.User != req.GetUser() {
		if model.User != "" {
			return
		}
		model.User = req.GetUser()
	}

	if req.GetEmailPrimary().GetValue() != "" {
		model.EmailPrimary = util.PtrString(req.GetEmailPrimary().GetValue())
	}

	if req.GetPhonePrimary().GetValue() != "" {
		model.PhonePrimary = util.PtrString(req.GetPhonePrimary().GetValue())
	}

	if req.GetBillingAddress().GetValue() != "" {
		model.BillingAddress = util.PtrString(req.GetBillingAddress().GetValue())
	}

	if req.GetFirstName().GetValue() != "" {
		model.FirstName = util.PtrString(req.GetFirstName().GetValue())
	}

	if req.GetLastName().GetValue() != "" {
		model.LastName = util.PtrString(req.GetLastName().GetValue())
	}
}

func (model *OfferModel) updateConditions(request *offer.OfferConditions) error {
	if request == nil {
		return nil
	}

	if model.Conditions == nil {
		model.Conditions = &OfferConditionsModel{}
	}

	old := model.Conditions

	if request.GetPrice() != nil {
		old.Price = unwrap.DoubleValue(request.GetPrice())
	}

	if request.GetDepositType() != offer.DepositType_UNKNOWN_DEPOSIT_TYPE {
		old.DepositType = request.GetDepositType()
	}

	if request.GetTermMonths() != nil {
		old.TermMonths = unwrap.Int32Value(request.GetTermMonths())
	}

	if request.GetMoveinDate() != nil {
		dayStart := time.Now().UTC().Truncate(24 * time.Hour)
		if request.GetMoveinDate().AsTime().Before(dayStart) {
			return fmt.Errorf(
				"%w: move-in date is in the past (%v/%v)",
				grpcerr.InvalidData,
				request.GetMoveinDate().AsTime(),
				dayStart,
			)
		}

		old.MoveinDate = unwrap.TimestampValue(request.GetMoveinDate())

		old.FirstRentPaymentDate = unwrap.TimestampValue(request.GetMoveinDate())

		rentPaymentDay := old.MoveinDate.Day()

		if rentPaymentDay < offer.LastRentPaymentDay {
			old.RentPaymentDay = util.PtrInt32(int32(rentPaymentDay))
		} else {
			old.RentPaymentDay = util.PtrInt32(int32(offer.LastRentPaymentDay))
		}
	}

	if request.GetPets() != nil {
		old.Pets = unwrap.BoolValue(request.GetPets())
	}

	if request.GetPetsDescription() != nil {
		old.PetsDescription = unwrap.StringValue(request.GetPetsDescription())
	}

	if request.GetComments() != nil {
		old.Comments = unwrap.StringValue(request.GetComments())
	}

	if request.GetTenantsPitch() != nil {
		old.TenantsPitch = unwrap.StringValue(request.GetTenantsPitch())
	}

	if request.GetPaymentFrequency() != nil {
		old.PaymentFrequency = unwrap.Int32Value(request.GetPaymentFrequency())
	}

	if request.GetBreakClause() != nil {
		old.BreakClause = unwrap.Int32Value(request.GetBreakClause())
	}

	if old.PaymentFrequency != nil && old.TermMonths != nil {
		if *old.TermMonths%*old.PaymentFrequency != 0 {
			return fmt.Errorf("%w: term in months is not divisible by payment frequency", grpcerr.InvalidData)
		}
	}

	return nil
}

// updateOfferLandlords updates only first in slice, others ignored
func (model *OfferModel) updateOfferLandlords(ctx context.Context, landlords []*offer.Landlord) error {
	if model.Landlord == nil || len(landlords) == 0 { // nothing to do
		return nil
	}

	if model.Landlords == nil {
		model.Landlords = make([]*LandlordModel, 0, 1)
	}

	landlord := &LandlordModel{
		User:           landlords[0].GetUser(),
		FirstName:      unwrap.StringValue(landlords[0].GetFirstName()),
		LastName:       unwrap.StringValue(landlords[0].GetLastName()),
		EmailPrimary:   unwrap.StringValue(landlords[0].GetEmailPrimary()),
		PhonePrimary:   unwrap.StringValue(landlords[0].GetPhonePrimary()),
		BillingAddress: unwrap.StringValue(landlords[0].GetBillingAddress()),
	}

	if util.UnPtrString(landlord.FirstName) != "" && util.UnPtrString(landlord.LastName) != "" &&
		util.UnPtrString(landlord.FirstName) == util.UnPtrString(model.Landlord.FirstName) &&
		util.UnPtrString(landlord.LastName) == util.UnPtrString(model.Landlord.LastName) {
		return ErrLandlordAlreadyExsistByNames
	}

	if util.UnPtrString(landlord.EmailPrimary) != "" &&
		util.UnPtrString(landlord.EmailPrimary) == util.UnPtrString(model.Landlord.EmailPrimary) {
		return ErrLandlordAlreadyExsistByEmail
	}

	if util.UnPtrString(landlord.PhonePrimary) != "" &&
		util.UnPtrString(landlord.PhonePrimary) == util.UnPtrString(model.Landlord.PhonePrimary) {
		return ErrLandlordAlreadyExsistByPhone
	}

	if len(model.Landlords) > 0 {
		model.Landlords[0] = landlord
		return nil
	}

	model.Landlords = append(model.Landlords, landlord)

	err := updateOfferLandlordModelDetails(ctx, landlord)
	if err != nil {
		return err
	}

	return nil
}

// UpdateTenancy synchronize tenancy data with property service.
func UpdateTenancy(ctx context.Context, of *OfferModel) (*tenancy.Tenancy, error) {
	var tenants []*tenancy.Tenant

	var tenantIDs []string

	for _, t := range of.Tenants {
		var risk tenancy.TenantRisk

		switch t.Risk {
		case offer.TenantRisk_LOW_RISK:
			risk = tenancy.TenantRisk_LOW_RISK
		case offer.TenantRisk_MEDIUM_RISK:
			risk = tenancy.TenantRisk_MEDIUM_RISK
		case offer.TenantRisk_HIGH_RISK:
			risk = tenancy.TenantRisk_HIGH_RISK
		}

		tenants = append(tenants, &tenancy.Tenant{
			User:     t.User,
			Contract: of.Contract,
			Active:   true,
			Risk:     risk,
			TenantContact: &tenancy.TenantContact{
				FirstName:    util.UnPtrString(t.FirstName),
				LastName:     util.UnPtrString(t.LastName),
				EmailPrimary: util.UnPtrString(t.EmailPrimary),
				PhonePrimary: util.UnPtrString(t.PhonePrimary),
			},
		})

		tenantIDs = append(tenantIDs, t.User)
	}

	var guarantors []*tenancy.Guarantor

	for _, g := range of.Guarantors {
		guarantors = append(guarantors, &tenancy.Guarantor{
			FirstName:    util.UnPtrString(g.FirstName),
			LastName:     util.UnPtrString(g.LastName),
			EmailPrimary: util.UnPtrString(g.EmailPrimary),
			PhonePrimary: util.UnPtrString(g.PhonePrimary),
		})
	}

	lresp, err := connections.ListingService.VersionGet(grpcctx.WithServiceUser(ctx), &listing.VersionGetRequest{
		Spec: listing.VersionSpecListingID(
			"",
			of.PropertyListing,
			listing.VersionStatus_VERSION_VERIFIED,
			listing.VersionStatus_VERSION_LET,
			listing.VersionStatus_VERSION_WITHDRAWN,
		),
	})
	if err != nil {
		return nil, err
	}

	location := lresp.GetVersion().GetLocation()

	t, err := connections.TenancyService.TenancySave(grpcctx.WithServiceUser(ctx), &tenancy.TenancySaveRequest{
		Ivr:                of.PropertyIvr,
		PropertyId:         of.Property,
		ListingId:          of.PropertyListing,
		OfferId:            of.ID.Hex(),
		Landlord:           of.LandlordUser(),
		Tenants:            tenants,
		Guarantors:         guarantors,
		TermMonths:         util.UnPtrInt32(of.Conditions.TermMonths),
		BreakClauseMonths:  util.UnPtrInt32(of.Conditions.BreakClause),
		MonthlyPayment:     util.UnPtrFloat64(of.Conditions.Price),
		DepositReplacement: of.DepositReplacement(),
		StartDate:          timestamppb.New(util.UnPtrTime(of.Conditions.MoveinDate)),
		Location: &tenancy.TenancyLocation{
			Postcode: location.GetPostcode().GetValue(),
			Address1: location.GetAddress1().GetValue(),
			Address2: location.GetAddress2().GetValue(),
			Address3: location.GetAddress3().GetValue(),
			Address4: location.GetAddress4().GetValue(),
			Flat:     location.GetFlat().GetValue(),
			Building: location.GetBuilding().GetValue(),
			Street:   location.GetStreet().GetValue(),
			City:     location.GetCity().GetValue(),
			Gps: &tenancy.GpsCoordinates{
				Longitude: location.GetGps().GetLongitude(),
				Latitude:  location.GetGps().GetLatitude(),
			},
			Country:              location.GetCountry().GetValue(),
			PostTown:             location.GetPostTown().GetValue(),
			County:               location.GetCounty().GetValue(),
			AdministrativeCounty: location.GetAdministrativeCounty().GetValue(),
			PostalCounty:         location.GetPostalCounty().GetValue(),
			TraditionalCounty:    location.GetTraditionalCounty().GetValue(),
			District:             location.GetDistrict().GetValue(),
			Address:              location.GetAddress().GetValue(),
			GeoRegion:            location.GetGeoRegion().GetValue(),
			GeoCounty:            location.GetGeoCounty().GetValue(),
			GeoTown:              location.GetGeoTown().GetValue(),
		},
	})
	if err != nil {
		return nil, err
	}

	_, err = connections.ListingService.ServiceVersionTenancyBegin(
		ctx,
		&listing.ServiceVersionTenancyBeginRequest{
			Spec: listing.VersionSpecListingID(
				"",
				of.PropertyListing,
				listing.VersionStatus_VERSION_VERIFIED,
			),
			TenancyId: t.GetTenancy().GetId(),
			Tenants:   tenantIDs,
		},
	)
	if err != nil {
		return nil, err
	}

	return t.GetTenancy(), nil
}

// SetInvoiceAndUpdate updates offer property calling Rightmove and Zoopla services if required
// and refreshes model state.
func (model *OfferModel) SetInvoiceAndUpdate(ctx context.Context) error {
	if model.State == offer.OfferState_HOLDING_DEPOSIT {
		ver, err := connections.ListingService.VersionGet(grpcctx.WithServiceUser(ctx), &listing.VersionGetRequest{
			Spec: listing.VersionSpecListingID("", model.PropertyListing),
		})
		if err != nil {
			return err
		}

		_, err = connections.ListingService.VersionPublish(
			grpcctx.WithServiceUser(ctx),
			&listing.VersionPublishRequest{
				VersionId: ver.GetVersion().GetId(),
				LetAgreed: wrapperspb.Bool(true),
			},
		)

		if errors.Is(err, grpcerr.ProviderError) {
			log.WithContext(ctx).WithError(err).Error("Temporary error updating publication status")

			return nil
		}

		return err
	}

	return nil
}

// RenewTenancy resets tenancy data of the model.
func (model *OfferModel) RenewTenancy() {
	originalOffer := model.ID.Hex()
	model.ID = primitive.NewObjectID()
	model.CreatedAt = time.Now()
	model.UpdatedAt = time.Now()
	model.DraftContract = ""
	model.DraftContractCreatedAt = time.Time{}
	model.Contract = ""
	model.ContractCreatedAt = time.Time{}
	if model.Approve != nil {
		model.Approve = nil
	}
	model.OriginalOffer = originalOffer
	model.State = offer.OfferState_PENDING
	if model.Conditions != nil {
		model.Conditions.MoveinDate.AddDate(0, int(util.UnPtrInt32(model.Conditions.TermMonths)), 0)
	}

	model.Sync.Reset()
}

// RemoveOffer removes offer from database.
// This action is unrecoverable and is intended to drop new offer on its creation process if Zeebe is unavailable.
func RemoveOffer(ctx context.Context, id primitive.ObjectID) error {
	_, err := Offers.DeleteOne(ctx, bson.M{"_id": id})

	return err
}

// DeleteOffer marks offer as deleted and saves it into the database.
func DeleteOffer(ctx context.Context, off *OfferModel, user string) error {
	off.Deleted = true
	off.Log = append(off.Log, &LogModel{
		Action: "delete",
		Dt:     time.Now(),
		User:   user,
	})
	err := UpdateOffer(ctx, off)

	return err
}

// RestoreOffer resets deleted flag and replaces offer into the database.
func RestoreOffer(ctx context.Context, off *OfferModel, user string) error {
	off.Deleted = false
	off.Log = append(off.Log, &LogModel{
		Action: "restore",
		Dt:     time.Now(),
		User:   user,
	})
	err := UpdateOffer(ctx, off)

	return err
}

// CreateHistoryModel add new entity into the history database.
func CreateHistoryModel(offer *OfferModel, action string) *OfferHistoryModel {
	return &OfferHistoryModel{
		ContractCreatedAt:       offer.ContractCreatedAt,
		DraftContractCreatedAt:  offer.DraftContractCreatedAt,
		CreatedAt:               offer.ContractCreatedAt,
		UpdatedAt:               offer.ContractCreatedAt,
		UpdatedByTenantAt:       offer.ContractCreatedAt,
		AcceptedAt:              offer.ContractCreatedAt,
		DeniedAt:                offer.ContractCreatedAt,
		CancelledAt:             offer.ContractCreatedAt,
		StateChangedAt:          offer.ContractCreatedAt,
		Tenants:                 offer.Tenants,
		Guarantors:              offer.Guarantors,
		Log:                     offer.Log,
		Clause:                  offer.Clause,
		Administrator:           offer.Administrator,
		Property:                offer.Property,
		PropertyListing:         offer.PropertyListing,
		PropertyIvr:             offer.PropertyIvr,
		PropertyAddress:         offer.PropertyAddress,
		Viewing:                 offer.Viewing,
		Contract:                offer.Contract,
		DraftContract:           offer.DraftContract,
		SignatureRequestID:      offer.SignatureRequestID,
		OriginalOffer:           offer.OriginalOffer,
		PrevTenancy:             offer.PrevTenancy,
		AwaitingTermination:     offer.AwaitingTermination,
		OpportunityID:           offer.OpportunityID,
		OfferID:                 offer.ID.Hex(),
		Action:                  action,
		Conditions:              offer.Conditions,
		Landlord:                offer.Landlord,
		Landlords:               offer.Landlords,
		Manager:                 offer.Manager,
		Details:                 offer.Details,
		Approve:                 offer.Approve,
		State:                   offer.State,
		ReadyForSigning:         offer.ReadyForSigning,
		PaidStatuses:            offer.PaidStatuses,
		WorkflowInstanceID:      offer.WorkflowInstanceID,
		ScheduleNotifications:   offer.ScheduleNotifications,
		Notifications:           offer.Notifications,
		Deleted:                 offer.Deleted,
		Test:                    offer.Test,
		IsLocked:                offer.IsLocked,
		SignedInvoicesGenerated: offer.SignedInvoicesGenerated,
		ScheduleGenerated:       offer.ScheduleGenerated,
		MoveoutScheduled:        offer.MoveoutScheduled,
		GuarantorRequired:       offer.GuarantorRequired,
		OpportunityTerminated:   offer.OpportunityTerminated,
		Sync:                    offer.Sync,
		RentPlan:                offer.RentPlan,
		TestNow:                 offer.TestNow,
	}
}

// CreateOffer inserts offer into the database.
func CreateOffer(ctx context.Context, offer *OfferModel) error {
	_, err := Offers.InsertOne(ctx, offer)
	if err != nil {
		return err
	}

	offer.CreatedAt = time.Now()
	offer.UpdatedAt = time.Now()

	offerHistory := CreateHistoryModel(offer, CreateAction)

	_, err = OffersHistory.InsertOne(ctx, offerHistory)

	return err
}

type updateOfferInternal struct {
	notificationUpdate bool
}

// WithForceNotificationUpdate forces check notification update
func WithForceNotificationUpdate() UpdateOption {
	return func(internal *updateOfferInternal) {
		internal.notificationUpdate = true
	}
}

// UpdateOption provides functional API for UpdateOffer
type UpdateOption func(*updateOfferInternal)

// UpdateOffer replaces existent offer into the database with specified one.
func UpdateOffer(ctx context.Context, model *OfferModel, opts ...UpdateOption) error {
	var internal updateOfferInternal

	for _, o := range opts {
		o(&internal)
	}

	if internal.notificationUpdate {
		needToSendVerifyNotification := true

		for _, t := range model.Tenants {
			if t.Check != "" && util.UnPtrString(t.RentprofileStatus) != completeStatus {
				needToSendVerifyNotification = false
				break
			}
		}

		for _, t := range model.Guarantors {
			if t.Check != "" && util.UnPtrString(t.RentprofileStatus) != completeStatus {
				needToSendVerifyNotification = false
				break
			}
		}

		if needToSendVerifyNotification {
			SendVerifyNotification(ctx, model)
		}
	}

	_, err := Offers.ReplaceOne(ctx, bson.M{"_id": model.ID}, model)
	if err != nil {
		return err
	}

	model.UpdatedAt = time.Now()

	offerHistory := CreateHistoryModel(model, UpdateAction)

	_, err = OffersHistory.InsertOne(ctx, offerHistory)

	return err
}

func addTestCondition(ctx context.Context, q bson.M) bson.M {
	if ctxmgr.IsTestUser(ctx) {
		return q
	}

	return bson.M{
		"$and": []bson.M{
			q,
			bson.M{
				"$or": []bson.M{
					bson.M{"test": bson.M{"$exists": false}},
					bson.M{"test": false},
				},
			},
		},
	}
}

// GetOfferByID finds offer by primary ID.
func GetOfferByID(ctx context.Context, id primitive.ObjectID) (*OfferModel, error) {
	var result = &OfferModel{}

	q := addTestCondition(ctx, bson.M{"_id": id})

	err := Offers.FindOne(ctx, q).Decode(result)
	return result, err
}

// GetOfferByHexID finds offer by primary ID specified as hex string.
func GetOfferByHexID(ctx context.Context, id string) (*OfferModel, error) {
	objID, err := primitive.ObjectIDFromHex(id)
	if err != nil {
		return nil, err
	}

	return GetOfferByID(ctx, objID)
}

// GetOffer finds offer by primary ID if specified user is among landlord or tenant.
func GetOffer(ctx context.Context, id primitive.ObjectID, user string) (*OfferModel, error) {
	var result = &OfferModel{}

	err := Offers.FindOne(ctx, bson.M{"$and": bson.A{
		bson.M{"_id": id},
		bson.M{"$or": bson.A{
			bson.M{"tenants.user": user},
			bson.M{"landlord.user": user},
			bson.M{"manager": user},
		}},
	}}).Decode(result)
	if err != nil {
		return nil, err
	}

	if result.Test && !ctxmgr.IsTestUser(ctx) {
		return nil, mongo.ErrNoDocuments
	}
	return result, err
}

// GetCurrentOffersByProperty finds all current offers for specified property ID.
func GetCurrentOffersByProperty(ctx context.Context, property string) ([]*OfferModel, error) {
	return findOffers(ctx,
		bson.M{"$and": bson.A{
			bson.M{"property": property},
			bson.M{"deleted": bson.M{"$not": bson.M{"$eq": true}}},
			bson.M{"state": bson.M{"$not": bson.M{"$in": bson.A{
				offer.OfferState_DENIED,
				offer.OfferState_CANCELED,
				offer.OfferState_TERMINATED,
				offer.OfferState_EXPIRED,
				offer.OfferState_RENTPAID,
				offer.OfferState_EARLY_TERMINATION,
				offer.OfferState_FINISHED,
			}}}},
		}},
	)
}

// GetOffersByProperty finds all offers for specified property ID.
func GetOffersByProperty(ctx context.Context, property string) ([]*OfferModel, error) {
	return findOffers(ctx, bson.M{"property": property})
}

// GetOffersByPrevTenancy finds all offers for specified prev_tenancy.
func GetOffersByPrevTenancy(ctx context.Context, tenancy string) ([]*OfferModel, error) {
	return findOffers(ctx, bson.M{"prev_tenancy": tenancy, "state": bson.M{"$not": bson.M{"$in": bson.A{
		offer.OfferState_UNKNOWN_OFFER_STATE,
		offer.OfferState_DENIED,
		offer.OfferState_CANCELED,
		offer.OfferState_TERMINATED,
		offer.OfferState_EXPIRED,
		offer.OfferState_RENTPAID,
		offer.OfferState_EARLY_TERMINATION,
		offer.OfferState_FINISHED,
	}}}})
}

// GetOffersAwaitingTermination finds all offers awaiting termination of another offer
func GetOffersAwaitingTermination(ctx context.Context) ([]*OfferModel, error) {
	return findOffers(ctx, bson.M{"$and": bson.A{
		bson.M{"state": offer.OfferState_AWAITING_TENANCY},
		bson.M{"awaiting_termination": bson.M{"$ne": ""}},
		bson.M{"awaiting_termination": bson.M{"$ne": nil}},
		bson.M{"awaiting_termination": bson.M{"$exists": true}},
	}})
}

// ListNotNotified finds all not notified offers for specified notification event types and by time range
func ListNotNotified(
	ctx context.Context,
	states []offer.OfferState,
	eventTypes []ScheduleNotificationType,
	from time.Time,
	to time.Time,
) ([]*OfferModel, error) {
	if len(eventTypes) == 0 {
		return nil, errors.New("missing event types")
	}

	orFilter := bson.A{}
	for _, eventType := range eventTypes {
		eventTypeKey := strconv.FormatInt(int64(eventType), 10)
		eventFilter := bson.M{
			"schedule_notifications." + eventTypeKey + ".is_notified": false,
		}

		startScheduleRangeFilter := bson.M{}
		if !from.IsZero() {
			startScheduleRangeFilter["$gte"] = from
		}

		if !to.IsZero() {
			startScheduleRangeFilter["$lt"] = to
		}

		if len(startScheduleRangeFilter) != 0 {
			eventFilter["schedule_notifications."+eventTypeKey+".started_at"] = startScheduleRangeFilter
		}

		orFilter = append(orFilter, eventFilter)
	}

	filter := bson.M{
		"state": bson.M{"$in": states},
		"$or":   orFilter,
	}

	return findOffers(ctx, filter)
}

func findOffers(ctx context.Context, filter bson.M) ([]*OfferModel, error) {
	// filter = addTestCondition(ctx, filter)
	cur, err := Offers.Find(ctx, filter)
	if err != nil {
		return []*OfferModel{}, err
	}
	var models []*OfferModel
	for cur.Next(ctx) {
		r := &OfferModel{}
		err = cur.Decode(r)
		if err != nil {
			return []*OfferModel{}, err
		}

		if r.Test && !ctxmgr.IsTestUser(ctx) {
			continue
		}
		models = append(models, r)
	}
	return models, nil
}

// could be like this but it has somw issues
// func RemoveOffersByProperty(ctx context.Context, propertyID string, user string, optFilter ...bson.M) error {
// 	filter := bson.M{
// 		"$and": bson.A{
// 			bson.M{"property": propertyID},
// 			bson.M{"deleted": bson.M{"$not": bson.M{"$eq": true}}},
// 		},
// 	}
//
// 	if len(optFilter) == 1 {
// 		filter["$and"] = append(filter["$and"].(bson.A), optFilter[0])
// 	}
//
// 	update := bson.M{
// 		"$set": bson.M{
// 			"deleted": true,
// 		},
// 		"$push": bson.M{ // this will trigger the error if there's no log field with array in document already
// 			"log": bson.M{
// 				"action": "delete",
// 				"dt":     time.Now().UTC(),
// 				"user":   user,
// 			},
// 		},
// 	}
//
// 	_, err := Offers.UpdateMany(ctx, filter, update)
// 	if err != nil {
// 		return err
// 	}
//
// 	return nil
// }

// RemoveOffersByProperty marks all offers for specified property as deleted.
func RemoveOffersByProperty(ctx context.Context, propertyID string, user string) error {
	offers, err := GetOffersByProperty(ctx, propertyID)
	if err != nil {
		return err
	}

	for _, of := range offers {
		if of.Deleted == true {
			continue
		}
		err := DeleteOffer(ctx, of, user)
		if err != nil {
			return err
		}
	}

	return nil
}

// DeleteFinishOffersByProperty marks all offers for specified property except active as deleted.
// marks active offer for specified property as finished.
func DeleteFinishOffersByProperty(
	ctx context.Context,
	propertyID string,
	user string,
) (offers []*OfferModel, err error) {
	offers, err = GetCurrentOffersByProperty(ctx, propertyID)
	if err != nil {
		return
	}

	now := time.Now().UTC()
	for _, of := range offers {
		switch of.State {
		case offer.OfferState_RENTPAID:
			of.State = offer.OfferState_FINISHED
			of.Sync.Reset()
			of.UpdatedAt = now
			err = UpdateOffer(ctx, of)
			if err != nil {
				return
			}
		default:
			err = DeleteOffer(ctx, of, user)
			if err != nil {
				return
			}
		}
	}

	return
}

// ListActiveOffersByProperty finds all offers with ACTIVE or PENDING state for specified property.
func ListActiveOffersByProperty(ctx context.Context, propertyID string) ([]*OfferModel, error) {
	states := []offer.OfferState{
		offer.OfferState_ACTIVE,
		offer.OfferState_PENDING,
	}

	filter := bson.M{}
	filter["property"] = propertyID
	// filter = addTestCondition(ctx, filter)

	cur, err := Offers.Find(ctx, filter)
	if err != nil {
		return nil, err
	}

	var models []*OfferModel
	for cur.Next(ctx) {
		r := &OfferModel{}
		err = cur.Decode(r)
		if err != nil {
			return nil, err
		}

		if r.Test && !ctxmgr.IsTestUser(ctx) {
			continue
		}
		for _, s := range states {
			if s == r.State {
				models = append(models, r)
				break
			}
		}
	}
	return models, nil
}

type offerPropertyField struct {
	Property string `bson:"property"`
}

// ListRelatedOutgoingProperties returns relate property IDs
func ListRelatedOutgoingProperties(ctx context.Context, userid string) (res []string, err error) {
	cur, err := Offers.Find(ctx, bson.M{
		"tenants.user": userid,
	})
	if err != nil {
		return nil, err
	}

	var f offerPropertyField

	m := make(map[string]bool)

	for cur.Next(ctx) {
		err = cur.Decode(&f)
		if err != nil {
			return nil, err
		}

		m[f.Property] = true
	}

	for p := range m {
		res = append(res, p)
	}

	return
}

// ListOpenOffersByProperty finds all "open" offers for specified property and tenant user.
func ListOpenOffersByProperty(ctx context.Context, propertyID, tenant string) ([]*OfferModel, error) {
	filter := bson.M{
		"property": propertyID,
		"state": bson.M{
			"$not": bson.M{
				"$in": []int32{
					int32(offer.OfferState_UNKNOWN_OFFER_STATE),
					int32(offer.OfferState_CANCELED),
					int32(offer.OfferState_DENIED),
					int32(offer.OfferState_SIGNED),
					int32(offer.OfferState_ERROR),
					int32(offer.OfferState_EXPIRED),
					int32(offer.OfferState_FINISHED),
				},
			},
		},
	}
	if tenant != "" {
		filter["tenants.user"] = tenant
	}

	// filter = addTestCondition(ctx, filter)

	cur, err := Offers.Find(ctx, filter)
	if err != nil {
		return nil, err
	}
	var models []*OfferModel
	for cur.Next(ctx) {
		r := &OfferModel{}
		err = cur.Decode(r)
		if err != nil {
			return nil, err
		}

		if r.Test && !ctxmgr.IsTestUser(ctx) {
			continue
		}

		// models = append(models, r)
	}
	return models, nil
}

// ListByPropertyLandlord finds all not PENDING and not CANCELLED offers for specified property.
func ListByPropertyLandlord(
	ctx context.Context,
	propertyID, listingID string,
	states []offer.OfferState,
) ([]*OfferModel, error) {
	var models []*OfferModel
	filter := bson.M{
		"property": propertyID,
		"state": bson.M{
			"$not": bson.M{
				"$in": []int32{
					int32(offer.OfferState_PENDING),
					int32(offer.OfferState_CANCELED),
					int32(offer.OfferState_FINISHED),
				},
			},
		},
	}

	if listingID != "" {
		filter["property_listing"] = listingID
	}

	if len(states) > 0 {
		filter["state"] = bson.M{
			"$in": states,
		}
	}

	// filter = addTestCondition(ctx, filter)

	cur, err := Offers.Find(ctx, filter)
	if err != nil {
		return nil, err
	}

	for cur.Next(ctx) {
		r := &OfferModel{}
		err = cur.Decode(r)
		if err != nil {
			return nil, err
		}

		if r.Test && !ctxmgr.IsTestUser(ctx) {
			continue
		}

		models = append(models, r)
	}
	return models, nil
}

// ListNotSignedByLandlord finds all active not signed offers for specified landlord.
func ListNotSignedByLandlord(ctx context.Context, landlordID string) ([]*OfferModel, error) {
	var models []*OfferModel
	filter := bson.M{
		"landlord.user": landlordID,
		"state": bson.M{
			"$not": bson.M{
				"$in": []int32{
					int32(offer.OfferState_SIGNED),
					int32(offer.OfferState_AWAITING_CONFIRM_RENT),
					int32(offer.OfferState_RENTPAID),
					int32(offer.OfferState_EARLY_TERMINATION),
					int32(offer.OfferState_FINISHED),
					int32(offer.OfferState_TERMINATED),
				},
			},
		},
	}

	cur, err := Offers.Find(ctx, filter)
	if err != nil {
		return nil, err
	}

	for cur.Next(ctx) {
		r := &OfferModel{}
		err = cur.Decode(r)
		if err != nil {
			return nil, err
		}

		models = append(models, r)
	}
	return models, nil
}

// ListByPropertyTenant finds all offers for specified property and tenant.
func ListByPropertyTenant(ctx context.Context, propertyID, tenant, listingID string) ([]*OfferModel, error) {
	var models []*OfferModel
	filter := bson.M{"property": propertyID, "tenants.user": tenant}

	if listingID != "" {
		filter["property_listing"] = listingID
	}

	// filter = addTestCondition(ctx, filter)

	cur, err := Offers.Find(ctx, filter)
	if err != nil {
		return nil, err
	}

	for cur.Next(ctx) {
		r := &OfferModel{}
		err = cur.Decode(r)
		if err != nil {
			return nil, err
		}

		if r.Test && !ctxmgr.IsTestUser(ctx) {
			continue
		}

		models = append(models, r)
	}
	return models, nil
}

// ListByProperty finds all offers for specified property.
func ListByProperty(ctx context.Context, propertyID, listingID string) ([]*OfferModel, error) {
	var models []*OfferModel
	filter := bson.M{"property": propertyID}

	if listingID != "" {
		filter["property_listing"] = listingID
	}

	// filter = addTestCondition(ctx, filter)

	cur, err := Offers.Find(ctx, filter)
	if err != nil {
		return nil, err
	}

	for cur.Next(ctx) {
		r := &OfferModel{}
		err = cur.Decode(r)
		if err != nil {
			return nil, err
		}

		if r.Test && !ctxmgr.IsTestUser(ctx) {
			continue
		}

		models = append(models, r)
	}
	return models, nil
}

// ListIncoming searches offers for specified property or if property ID is empty
// obtains all properties for specified user from property service
// and then find corresponding offers.
func ListIncoming(
	ctx context.Context,
	user, propertyID, listingID string,
	states []offer.OfferState,
) ([]*OfferModel, error) {
	var models []*OfferModel
	filter := bson.M{}
	if propertyID != "" {
		filter["property"] = propertyID
	} else {
		props, err := connections.PropertyService.List(
			grpcctx.WithServiceUser(ctx),
			&property.PropertyListRequest{
				User: wrap.StringValue(&user),
				Roles: &property.UserRoleArrayValue{
					Roles: []property.UserRole{
						property.UserRole_USER_ROLE_LANDLORD,
					},
				},
			},
		)
		if err != nil {
			return nil, err
		}
		var ids []string
		for _, p := range props.Properties {
			ids = append(ids, p.GetId())
		}
		if len(ids) == 0 {
			return models, nil
		}
		filter["property"] = bson.M{"$in": ids}
	}

	if listingID != "" {
		filter["property_listing"] = listingID
	}

	// filter = addTestCondition(ctx, filter)

	cur, err := Offers.Find(ctx, filter)
	if err != nil {
		return nil, err
	}

	for cur.Next(ctx) {
		r := &OfferModel{}
		err = cur.Decode(r)
		if err != nil {
			return nil, err
		}

		if r.Test && !ctxmgr.IsTestUser(ctx) {
			continue
		}

		for _, s := range states {
			if s == r.State {
				models = append(models, r)
				break
			}
		}
	}
	return models, nil
}

// ListOutgoing finds all offers for specified user and property, if it ID is not empty,
// considering required states.
func ListOutgoing(ctx context.Context, user, propertyID string, states []offer.OfferState) ([]*OfferModel, error) {
	filter := bson.M{"tenants.user": user}
	if propertyID != "" {
		filter["property"] = propertyID
	}
	// filter = addTestCondition(ctx, filter)

	cur, err := Offers.Find(ctx, filter)
	if err != nil {
		return nil, err
	}
	var models []*OfferModel
	for cur.Next(ctx) {
		r := &OfferModel{}
		err = cur.Decode(r)
		if err != nil {
			return nil, err
		}

		if r.Test && !ctxmgr.IsTestUser(ctx) {
			continue
		}

		for _, s := range states {
			if s == r.State {
				models = append(models, r)
				break
			}
		}
	}
	return models, nil
}

// FilterOffers finds all offers by specified query parameters.
func FilterOffers(ctx context.Context, params *q.QueryParams) ([]*OfferModel, int64, error) {
	var offers []*OfferModel

	// ////////check on property.location.postcode
	var queryFilters []*q.QueryParamsFilter
	for _, filter := range params.GetFilter() {
		for i, f := range filter.GetFields() {
			if f == "property.location.postcode" {
				properties, err := getPropertiesByPostcode(ctx, filter.Value, filter.Operator)
				if err != nil {
					return nil, 0, err
				}

				if len(properties) == 0 {
					return offers, 0, nil
				}

				queryFilters = append(queryFilters, &q.QueryParamsFilter{
					Fields:   []string{"property"},
					Value:    properties,
					Operator: q.QueryFilterOperator_ONEOF,
				})

				filter.Fields = append(filter.Fields[:i], filter.Fields[i+1:]...)
			}
		}

		if len(filter.Fields) > 0 {
			queryFilters = append(queryFilters, filter)
		}
	}

	params.Filter = queryFilters

	// ///////////////////////

	query, opt, err := mongoquery.BuildFilter(OfferModel{}, params, nil)
	if err != nil {
		return nil, 0, err
	}

	// query = addTestCondition(ctx, query)
	cur, err := Offers.Find(ctx, query, opt)
	if err != nil {
		return nil, 0, err
	}

	for cur.Next(ctx) {
		r := &OfferModel{}
		err = cur.Decode(r)
		if err != nil {
			return nil, 0, err
		}

		if r.Test && !ctxmgr.IsTestUser(ctx) {
			continue
		}

		offers = append(offers, r)
	}
	total, err := Offers.CountDocuments(ctx, query)
	return offers, total, err
}

func getPropertiesByPostcode(ctx context.Context, postcodes []string, op q.QueryFilterOperator) ([]string, error) {
	filter := &q.QueryParamsFilter{
		Fields:   []string{"location.postcode"},
		Value:    postcodes,
		Operator: op,
	}

	var filters []*q.QueryParamsFilter
	filters = append(filters, filter)
	// var params []*q.QueryParams
	//
	// params = append(params, &q.QueryParams{
	// 	Filter: filters,
	// })

	resp, err := connections.PropertyService.ListFilter(
		grpcctx.WithServiceUser(ctx),
		&property.PropertyListFilterRequest{
			Params: &q.QueryParams{Filter: filters},
		},
	)
	if err != nil {
		return nil, err
	}

	var res []string
	if resp.Total == 0 {
		return res, nil
	}

	for _, p := range resp.Properties {
		res = append(res, p.GetId())
	}

	return res, nil
}

// OfferFileType enumrates types of offer files
type OfferFileType int32

// Offer document type
const (
	OfferDocsTypeRightToRentLandlord OfferFileType = 0
	OfferDocsTypeHowToRentTenant     OfferFileType = 1
	OfferDocsTypeAst                 OfferFileType = 2 // Contract
	OfferDocsTypeSomf                OfferFileType = 3
)

// OfferFileModel represents data model for offer file.
type OfferFileModel struct {
	FileID      string             `bson:"file"`
	DocumentID  string             `bson:"document"`
	Mime        string             `bson:"mime"`
	Description string             `bson:"description"`
	Name        string             `bson:"name"`
	FileType    OfferFileType      `bson:"doc_type"`
	Version     int32              `bson:"version"`
	ID          primitive.ObjectID `bson:"_id"`
}

// GetVersion returns currently supported version of offer file.
func (fileType OfferFileType) GetVersion() int32 {
	switch fileType {
	case OfferDocsTypeRightToRentLandlord:
		return 1
	case OfferDocsTypeHowToRentTenant:
		return 1
	case OfferDocsTypeAst:
		return 1
	default:
		return 1
	}
}

// GetMime returns MIME specification of offer file.
func (fileType OfferFileType) GetMime() string {
	switch fileType {
	case OfferDocsTypeRightToRentLandlord:
		return mimeApplicationPDF
	case OfferDocsTypeHowToRentTenant:
		return mimeApplicationPDF
	case OfferDocsTypeAst:
		return mimeApplicationPDF
	default:
		return mimeApplicationPDF
	}
}

// GetDescription returns default offer file description.
func (fileType OfferFileType) GetDescription() string {
	switch fileType {
	case OfferDocsTypeRightToRentLandlord:
		return ""
	case OfferDocsTypeHowToRentTenant:
		return ""
	case OfferDocsTypeAst:
		return ""
	default:
		return ""
	}
}

// GetName returns default name of offer file.
func (fileType OfferFileType) GetName() string {
	switch fileType {
	case OfferDocsTypeRightToRentLandlord:
		return "Right-to-Rent-Guidance.pdf"
	case OfferDocsTypeHowToRentTenant:
		return "How_to_Rent.pdf"
	case OfferDocsTypeAst:
		return "AST-DRAFT.pdf"
	case OfferDocsTypeSomf:
		return "StandingOrderMandateForm.pdf"
	default:
		return ""
	}
}

// GetFileBytes returns offer file content.
func (fileType OfferFileType) GetFileBytes(offr *OfferModel) ([]byte, error) {
	buf, err := docs.Asset(fileType.GetName())
	if err != nil {
		return nil, err
	}
	return buf, nil
}

// GetFilesLandlord - is ...
func GetFilesLandlord(ctx context.Context, offr *OfferModel) ([]string, error) {
	result := make([]string, 0)

	f, err := offr.GetFileByType(ctx, OfferDocsTypeRightToRentLandlord, offr.LandlordUser())
	if err != nil {
		return nil, err
	}
	result = append(result, f)

	f, err = offr.GetFileByType(ctx, OfferDocsTypeAst, offr.LandlordUser())
	if err != nil {
		return nil, err
	}
	result = append(result, f)

	return result, nil
}

// GetFilesTenant - is ...
func GetFilesTenant(ctx context.Context, offr *OfferModel) ([]string, error) {
	result := make([]string, 0)

	f, err := offr.GetFileByType(ctx, OfferDocsTypeHowToRentTenant, offr.TenantUser())
	if err != nil {
		return nil, err
	}
	result = append(result, f)

	f, err = offr.GetFileByType(ctx, OfferDocsTypeAst, offr.TenantUser())
	if err != nil {
		return nil, err
	}
	result = append(result, f)

	files, err := GetPropertyDocs(ctx, offr)
	if err != nil {
		return nil, err
	}
	result = append(result, files...)

	return result, nil
}

// GetFileByType - is ...
func (model *OfferModel) GetFileByType(ctx context.Context, fileType OfferFileType, user string) (string, error) {
	fileModel := &OfferFileModel{
		FileType:    fileType,
		Version:     fileType.GetVersion(),
		Mime:        fileType.GetMime(),
		Description: fileType.GetDescription(),
		Name:        fileType.GetName(),
	}

	err := fileModel.UpdateFileModelByVersion(ctx, model, user)
	if err != nil {
		return "", err
	}

	return fileModel.FileID, nil
}

// UpdateFileModelByVersion - is ...
func (fileModel *OfferFileModel) UpdateFileModelByVersion(ctx context.Context, offr *OfferModel, user string) error {
	count, err := Documents.CountDocuments(ctx, bson.M{"$and": bson.A{
		bson.M{"version": fileModel.Version},
		bson.M{"doc_type": fileModel.FileType},
	}})
	if err != nil {
		return err
	}

	if count <= 0 { // not found
		switch fileModel.FileType {
		case OfferDocsTypeAst:
			err := fileModel.RenderAST(ctx, offr, user)
			if err != nil {
				return err
			}
		case OfferDocsTypeSomf:
			err := fileModel.RenderSOMF(ctx, offr, user)
			if err != nil {
				return err
			}
		default:
			err = fileModel.UploadFile(ctx, offr, user)
			if err != nil {
				return err
			}
			fileModel.ID = primitive.NewObjectID()
			_, err = Documents.InsertOne(ctx, fileModel)
			if err != nil {
				return err
			}
		}
	} else { // found
		var result = &OfferFileModel{}
		err := Documents.FindOne(ctx, bson.M{"$and": bson.A{
			bson.M{"version": fileModel.Version},
			bson.M{"doc_type": fileModel.FileType},
		}}).Decode(result)
		if err != nil {
			return err
		}
		fileModel.ID = result.ID
		fileModel.Mime = result.Mime
		fileModel.DocumentID = result.DocumentID
		fileModel.FileID = result.FileID
		fileModel.Description = result.Description
	}

	return nil
}

// UploadFile - is ...
func (fileModel *OfferFileModel) UploadFile(ctx context.Context, offr *OfferModel, user string) error {
	fbuf, err := fileModel.FileType.GetFileBytes(offr)
	if err != nil {
		return err
	}

	resp, err := document.UploadFile(
		ctx,
		connections.DocumentService,
		fileModel.Name,
		fileModel.Mime,
		&document.FileAttribute{
			User: user,
			Data: map[string]string{
				document.FileAttributeGenerated.String(): "true",
			},
		},
		bytes.NewReader(fbuf),
	)
	if err != nil {
		return err
	}

	fileModel.FileID = resp.GetId()

	return nil
}

// RenderAST - is ...
func (fileModel *OfferFileModel) RenderAST(ctx context.Context, offr *OfferModel, user string) error {
	m, err := GetOfferByID(ctx, offr.ID)
	if err != nil {
		return err
	}
	prop, err := connections.PropertyService.Get(grpcctx.WithServiceUser(ctx), &property.GetRequest{
		Id: m.Property,
	})
	if err != nil {
		return err
	}
	landlord, err := connections.AccountService.Get(ctx, &account.GetRequest{
		Id: prop.GetProperty().GetPeople().GetLandlord().GetValue(),
	})
	if err != nil {
		return err
	}
	cs := struct {
		Landlord *account.Account   `json:"landlord"`
		Offer    *offer.Offer       `json:"offer"`
		Property *property.Property `json:"property"`
		Now      string             `json:"now"`
	}{
		Now:      time.Now().Format(time.RFC1123),
		Landlord: landlord.GetAccount(),
		Offer:    m.RenderOfferModel(ctx, ""),
		Property: prop.GetProperty(),
	}

	bs, err := bind.ToJSON(cs, &bind.Config{
		GetOmitEmpty: false,
		SetOmitEmpty: false,
	})
	if err != nil {
		return err
	}

	resp, err := connections.ContractService.Render(ctx, &contract.RenderRequest{
		Id:         TemplateASTID,
		Format:     contract.ContractFormat_PDF,
		ValuesJson: bs,
		User:       user,
	})
	if err != nil {
		return err
	}

	fileModel.FileID = resp.GetFileid()
	return nil
}

// RenderSOMF - is ...
func (fileModel *OfferFileModel) RenderSOMF(ctx context.Context, offr *OfferModel, user string) error {
	m, err := GetOfferByID(ctx, offr.ID)
	if err != nil {
		return err
	}

	prop, err := connections.PropertyService.Get(grpcctx.WithServiceUser(ctx), &property.GetRequest{
		Id: m.Property,
	})
	if err != nil {
		return err
	}

	landlord, err := connections.AccountService.Get(ctx, &account.GetRequest{
		Id: prop.GetProperty().GetPeople().GetLandlord().GetValue(),
	})
	if err != nil {
		return err
	}

	cs := struct {
		Landlord *account.Account   `json:"landlord"`
		Offer    *offer.Offer       `json:"offer"`
		Property *property.Property `json:"property"`
		Now      string             `json:"now"`
	}{
		Now:      time.Now().Format(time.RFC1123),
		Landlord: landlord.GetAccount(),
		Offer:    m.RenderOfferModel(ctx, ""),
		Property: prop.GetProperty(),
	}

	bs, err := bind.ToJSON(cs, &bind.Config{
		GetOmitEmpty: false,
		SetOmitEmpty: false,
	})
	if err != nil {
		return err
	}

	now := time.Now()
	name := fmt.Sprintf(
		"StandingOrderMandateForm-%04d-%02d-%02d.pdf",
		now.Year(),
		now.Month(),
		now.Day(),
	)
	desc := fmt.Sprintf(
		"StandingOrderMandateForm as of %04d-%02d-%02d, rendered in pdf",
		now.Year(),
		now.Month(),
		now.Day(),
	)

	resp, err := connections.ContractService.Render(ctx, &contract.RenderRequest{
		Id:          TemplateSOMFID,
		Format:      contract.ContractFormat_PDF,
		Name:        name,
		Description: desc,
		ValuesJson:  bs,
		User:        user,
	})
	if err != nil {
		return err
	}

	fileModel.FileID = resp.GetFileid()

	return nil
}

// GetPropertyDocs examines document service for availability of property docs and returns their ID.
func GetPropertyDocs(ctx context.Context, offr *OfferModel) ([]string, error) {
	resp, err := connections.DocumentService.List(ctx, &document.ListRequest{
		Property: offr.Property,
	})
	if err != nil {
		return nil, err
	}

	result := make([]string, 0)
	if resp.GetDocuments() != nil {
		for _, doc := range resp.GetDocuments() {
			if doc.GetTypeCode() == document.TypeCertificate &&
				doc.GetState() == document.DocumentState_ACTIVE {
				result = append(result, doc.GetFile().GetId())
			}
		}
	}
	return result, nil
}

func performSendVerifyNotification(ctx context.Context, off *OfferModel, tenants []string) {
	embedProp := []*notification.EventEmbed{
		{
			EmbedType:  messaging.EmbedType_EMBED_TYPE_PROPERTY,
			EmbedValue: off.Property,
		},
	}

	embedCertificateFiles := make([]*notification.EventEmbed, 0)
	documents, _ := connections.DocumentService.List(ctx, &document.ListRequest{
		Property: off.Property,
	})
	if documents != nil {
		for _, docs := range documents.GetDocuments() {
			if docs.GetTypeCode() == document.TypeCertificate &&
				(docs.GetCode() == "landlord_epc_certificate" ||
					docs.GetCode() == "landlord_gas_certificate") &&
				docs.GetFile().GetId() != "" {
				embedCertificateFiles = append(embedCertificateFiles, &notification.EventEmbed{
					EmbedType:  messaging.EmbedType_EMBED_TYPE_FILE,
					EmbedValue: docs.GetFile().GetId(),
				})
			}
		}
	}

	files, err := GetFilesTenant(ctx, off)
	if err != nil {
		return
	}
	embedFiles := make([]*notification.EventEmbed, 0)
	for _, file := range files {
		embedFiles = append(embedFiles, &notification.EventEmbed{
			EmbedType:  messaging.EmbedType_EMBED_TYPE_FILE,
			EmbedValue: file,
		})
	}

	if len(embedCertificateFiles) > 0 {
		for _, f := range embedCertificateFiles {
			embedFiles = append(embedFiles, f)
		}
	}

	var targets []*notification.Target

	for _, tenant := range tenants {
		targets = append(targets, &notification.Target{
			User: tenant,
			Role: notification.Role_ROLE_TENANT,
		})
	}

	_, _ = connections.NotificationService.SendNotification(ctx, &notification.SendNotificationRequest{
		Entity:  notification.Entity_ENTITY_OFFER_REFERENCE_CHECKS,
		Action:  notification.Action_ACTION_COMPLETE,
		Targets: targets,
		Embeds:  append(embedProp, embedFiles...),
	})

	// landlord
	files, err = GetFilesLandlord(ctx, off)
	if err != nil {
		return
	}
	embedFiles = make([]*notification.EventEmbed, 0)
	for _, file := range files {
		embedFiles = append(embedFiles, &notification.EventEmbed{
			EmbedType:  messaging.EmbedType_EMBED_TYPE_FILE,
			EmbedValue: file,
		})
	}

	_, _ = connections.NotificationService.SendNotification(ctx, &notification.SendNotificationRequest{
		Entity: notification.Entity_ENTITY_OFFER_REFERENCE_CHECKS,
		Action: notification.Action_ACTION_COMPLETE,
		Targets: []*notification.Target{
			{
				User:     off.LandlordUser(),
				Role:     notification.Role_ROLE_LANDLORD,
				Contacts: nil,
			},
		},
		Embeds: append(embedProp, embedFiles...),
	})
}

// SendVerifyNotification notifies landlord and tenant for succesful offer verification.
func SendVerifyNotification(ctx context.Context, off *OfferModel) {
	var tenants []string

	for _, tenant := range off.Tenants {
		if tenant.CheckNotified {
			continue
		}

		tenant.CheckNotified = true

		tenants = append(tenants, tenant.User)
	}

	if len(tenants) == 0 {
		return
	}

	bgctx := grpcctx.ValuesBackgroundContext(ctx)

	go performSendVerifyNotification(bgctx, off, tenants)
}

// OfferTerminate sets offer termination date
func OfferTerminate(
	ctx context.Context,
	id primitive.ObjectID,
	date time.Time,
	state offer.OfferState,
) error {
	_, err := Offers.UpdateOne(
		ctx,
		bson.M{
			"_id": id,
		},
		bson.M{
			"$set": bson.M{
				"conditions.terminate_date": date,
				"state":                     state,
			},
		},
	)

	return err
}

// OfferCompleteCheck find offers eligible for terminated or finished state and updates state accordingly
func OfferCompleteCheck(
	ctx context.Context,
) error {
	now := time.Now()

	_, err := Offers.UpdateMany(
		ctx,
		bson.M{
			"$and": bson.A{
				bson.M{
					"conditions.terminate_date": bson.M{"$ne": nil, "$exists": true},
					"state":                     offer.OfferState_RENTPAID,
				},
				bson.M{
					"$expr": bson.M{
						"$lt": bson.A{
							"$conditions.terminate_date",
							now,
						},
					},
				},
			},
		},
		bson.M{
			"$set": bson.M{
				"state": offer.OfferState_TERMINATED,
			},
		},
	)
	if err != nil {
		return err
	}

	_, err = Offers.UpdateMany(
		ctx,
		bson.M{
			"$and": bson.A{
				bson.M{
					"conditions.movein_date": bson.M{"$ne": nil, "$exists": true},
					"conditions.term_months": bson.M{"$ne": nil, "$exists": true},
					"state":                  offer.OfferState_RENTPAID,
				},
				bson.M{
					"$expr": bson.M{
						"$lt": bson.A{
							bson.M{
								"$let": bson.M{
									"vars": bson.M{
										"foo": bson.M{
											"$mergeObjects": bson.A{
												bson.M{
													"$dateToParts": bson.M{
														"date": "$conditions.movein_date",
													},
												},
												bson.M{
													"month": bson.M{
														"$add": bson.A{
															"$conditions.term_months",
															bson.M{"$month": "$conditions.movein_date"},
														},
													},
												},
											},
										},
									},
									"in": bson.M{
										"$dateFromParts": bson.M{
											"year":        "$$foo.year",
											"month":       "$$foo.month",
											"day":         "$$foo.day",
											"hour":        "$$foo.hour",
											"minute":      "$$foo.minute",
											"second":      "$$foo.second",
											"millisecond": "$$foo.millisecond",
										},
									},
								},
							},
							now,
						},
					},
				},
			},
		},
		bson.M{
			"$set": bson.M{
				"state": offer.OfferState_FINISHED,
			},
		},
	)

	return err
}

// OfferInsurancesGenerated set rent_insurances_generated flag to true if rent insurances generated
func OfferInsurancesGenerated(
	ctx context.Context,
	id primitive.ObjectID,
) error {
	_, err := Offers.UpdateOne(
		ctx,
		bson.M{
			"_id": id,
		},
		bson.M{
			"$set": bson.M{
				"rent_insurances_generated": true,
			},
		},
	)

	return err
}

// GetBillingAddressFromLocation returns user billing address rendered from user location
func GetBillingAddressFromLocation(accountLocation *account.AccountLocation) string {
	result := ""

	if accountLocation.GetAddress1Value() != "" {
		result += accountLocation.GetAddress1Value() + " "
	}
	if accountLocation.GetAddress2Value() != "" {
		result += accountLocation.GetAddress2Value() + " "
	}
	if accountLocation.GetCityValue() != "" {
		result += accountLocation.GetCityValue() + " "
	}
	if accountLocation.GetCountryValue() != "" {
		result += accountLocation.GetCountryValue() + " "
	}
	if accountLocation.GetPostcodeValue() != "" {
		result += accountLocation.GetPostcodeValue() + " "
	}

	return result
}

// ListingOldNewMigrate updates listing ID
func ListingOldNewMigrate(ctx context.Context, oldl, newl string) error {
	_, err := Offers.UpdateMany(
		ctx,
		bson.M{"property_listing": oldl},
		bson.M{"$set": bson.M{"property_listing": newl}},
	)

	return err
}

// OfferPropertyIDs returns distinct property IDs user is a submitting tenant of
func OfferPropertyIDs(ctx context.Context, user string) (propertyIDs []string, err error) {
	vals, err := Offers.Distinct(ctx, "property", bson.M{"tenants.user": user})
	if err != nil {
		if err == mongo.ErrNoDocuments {
			return nil, nil
		}

		return nil, err
	}

	for _, v := range vals {
		s, ok := v.(string)

		if ok {
			propertyIDs = append(propertyIDs, s)
		}
	}

	return
}

// OfferCount returns count of offers
func OfferCount(ctx context.Context, params *q.QueryParams) (int64, error) {
	query, _, err := mongoquery.BuildFilter(OfferModel{}, params, nil)
	if err != nil {
		return 0, err
	}

	return Offers.CountDocuments(ctx, query)
}

// UserReferencingCheckResult represents values to update tenant's/guarantor's rentprofile check's data
type UserReferencingCheckResult struct {
	UserKindValue        UserKind
	UserArraysIndex      int
	RiskDocument         *string
	RentprofileStatus    *string
	RiskURL              *string
	RentprofileCreatedAt *time.Time
	Risk                 offer.TenantRisk
	OfferID              primitive.ObjectID
	CheckNeeded          bool
}

// UserKind tenants/guarantors
type UserKind string

// user kind values
const (
	UserKindTenants    UserKind = "tenants"
	UserKindGuarantors UserKind = "guarantors"
)

// UpdateUsersReferencingCheckResult updates tenant's/guarantor's rentprofile check's data
func UpdateUsersReferencingCheckResult(
	ctx context.Context,
	referencingCheckData *UserReferencingCheckResult,
) error {
	setPart := fmt.Sprintf("%s.%d.", referencingCheckData.UserKindValue, referencingCheckData.UserArraysIndex)

	_, err := Offers.UpdateOne(
		ctx,
		bson.M{
			"_id": referencingCheckData.OfferID,
		},
		bson.M{
			"$set": bson.M{
				setPart + "risk":                   referencingCheckData.Risk,
				setPart + "risk_document":          referencingCheckData.RiskDocument,
				setPart + "rentprofile_status":     referencingCheckData.RentprofileStatus,
				setPart + "risk_url":               referencingCheckData.RiskURL,
				setPart + "rentprofile_created_at": referencingCheckData.RentprofileCreatedAt,
				setPart + "check_needed":           referencingCheckData.CheckNeeded,
			},
		},
	)
	if err != nil {
		return err
	}

	return nil
}
