package model

import (
	"context"
	"fmt"
	"time"

	"gitlab.mashplace.com/mashplace/backend/common/generic/log"
	"gitlab.mashplace.com/mashplace/backend/common/generic/util"
	offer "gitlab.mashplace.com/mashplace/backend/offer/api"
	"go.mongodb.org/mongo-driver/bson/primitive"
)

// OfferLogger interface, use it when diff between initial state and changes should be stored
type OfferLogger interface {
	SetInitialState(ofr *OfferModel) error
	LogDiff(ofr *OfferModel, action string, user string) error
}

// SensitiveOfferFields contains fields that should be logged via OfferLogger
type SensitiveOfferFields struct {
	Conditions             OfferConditionsModel `bson:"conditions"`
	Landlord               LandlordModel        `bson:"landlord"`
	Approve                AdminApproveModel    `bson:"approve"`
	DraftContractCreatedAt time.Time            `bson:"draft_contract_created_at"`
	CreatedAt              time.Time            `bson:"created_at"`
	UpdatedAt              time.Time            `bson:"updated_at"`
	AcceptedAt             time.Time            `bson:"accepted_at"`
	DeniedAt               time.Time            `bson:"denied_at"`
	CancelledAt            time.Time            `bson:"cancelled_at"`
	ContractCreatedAt      time.Time            `bson:"contract_created_at"`
	Details                OfferDetailsModel    `bson:"details"`
	PrevTenancy            string               `bson:"prev_tenancy"`
	Contract               string               `bson:"conract"`
	DraftContract          string               `bson:"draft_contract"`
	SignatureRequestID     string               `bson:"signature_request_id"`
	Viewing                string               `bson:"viewing"`
	PropertyIvr            string               `bson:"property_ivr"`
	Property               string               `bson:"property"`
	OriginalOffer          string               `bson:"original_offer"`
	Tenants                []TenantOfferModel   `bson:"tenants"`
	Clause                 []ClauseModel        `bson:"clause"`
	State                  offer.OfferState     `bson:"state"`
	ID                     primitive.ObjectID   `bson:"_id"`
	Deleted                bool                 `bson:"deleted"`
	Test                   bool                 `bson:"test"`
	IsLocked               bool                 `bson:"is_locked"`
	ScheduleGenerated      bool                 `bson:"schedule_generated"`
	MoveoutScheduled       bool                 `bson:"moveout_scheduled"`
}

// OfferChangelog represents changes in Offer record.
type OfferChangelog struct {
	CreatedAt time.Time          `bson:"created_at"`
	Offer     string             `bson:"offer"`
	Initiator string             `bson:"initiator"` // User or system if null
	Old       string             `bson:"old"`
	New       string             `bson:"new"`
	Action    offer.OfferAction  `bson:"action"`
	ID        primitive.ObjectID `bson:"_id,omitempty"`
}

// OfferLog is a base logging struct
type OfferLog struct {
	ctx          context.Context
	InitialState *SensitiveOfferFields
	Instance     *OfferModel
	User         string
	Action       offer.OfferAction
}

// NewOfferLog returns new offer logger using provided offer as reference offer state and user as changes initiator
func NewOfferLog(ctx context.Context, ofr *OfferModel, user string) *OfferLog {
	o := &OfferLog{
		User:     user,
		Instance: ofr,
		ctx:      ctx,
	}

	o.InitialState = &SensitiveOfferFields{
		ContractCreatedAt:      ofr.ContractCreatedAt,
		DraftContractCreatedAt: ofr.DraftContractCreatedAt,
		CreatedAt:              ofr.CreatedAt,
		UpdatedAt:              ofr.UpdatedAt,
		AcceptedAt:             ofr.AcceptedAt,
		DeniedAt:               ofr.DeniedAt,
		CancelledAt:            ofr.CancelledAt,
		Property:               ofr.Property,
		PropertyIvr:            ofr.PropertyIvr,
		Viewing:                ofr.Viewing,
		Contract:               ofr.Contract,
		DraftContract:          ofr.DraftContract,
		SignatureRequestID:     ofr.SignatureRequestID,
		State:                  ofr.State,
		ID:                     ofr.ID,
		Deleted:                ofr.Deleted,
		Test:                   ofr.Test,
		IsLocked:               ofr.IsLocked, // deprecate?
		ScheduleGenerated:      ofr.ScheduleGenerated,
		MoveoutScheduled:       ofr.MoveoutScheduled,
		OriginalOffer:          ofr.OriginalOffer,
		PrevTenancy:            ofr.PrevTenancy,
	}

	o.InitialState.Tenants = make([]TenantOfferModel, 0, len(ofr.Tenants))
	o.InitialState.Clause = make([]ClauseModel, 0, len(ofr.Clause))
	for _, t := range ofr.Tenants {
		if t != nil {
			o.InitialState.Tenants = append(o.InitialState.Tenants, *t)
		}
	}
	for _, c := range ofr.Clause {
		if c != nil {
			o.InitialState.Clause = append(o.InitialState.Clause, *c)
		}
	}

	if ofr.Conditions != nil {
		o.InitialState.Conditions = *ofr.Conditions
	}
	if ofr.Landlord != nil {
		o.InitialState.Landlord = *ofr.Landlord
	}
	if ofr.Details != nil {
		o.InitialState.Details = *ofr.Details
	}
	if ofr.Approve != nil {
		o.InitialState.Approve = *ofr.Approve
	}

	return o
}

// Return changed fields
func (o *OfferLog) getLogs() (fieldsToLog []*OfferChangelog) {
	fieldsToLog = append(fieldsToLog, o.logLandlordName()...)

	fieldsToLog = append(fieldsToLog, o.logNewTenants()...)

	fieldsToLog = append(fieldsToLog, o.logPaymentFrequency()...)

	fieldsToLog = append(fieldsToLog, o.logFirstRentPaymentDate()...)

	fieldsToLog = append(fieldsToLog, o.logMoveInDate()...)

	return
}

func (o *OfferLog) logNewTenants() (fieldsToLog []*OfferChangelog) {
	if len(o.InitialState.Tenants) < len(o.Instance.Tenants) {
		fieldsToLog = append(fieldsToLog, &OfferChangelog{
			Action: offer.OfferAction_TENANT_ADD,
			Old: fmt.Sprintf("%s %s",
				util.UnPtrString(o.InitialState.Landlord.FirstName),
				util.UnPtrString(o.InitialState.Landlord.LastName),
			),
			New: fmt.Sprintf("%s %s",
				util.UnPtrString(o.Instance.Landlord.FirstName),
				util.UnPtrString(o.Instance.Landlord.LastName),
			),
		})
	}

	return
}

func (o *OfferLog) logPaymentFrequency() (fieldsToLog []*OfferChangelog) {
	initPayFreq := util.UnPtrInt32(o.InitialState.Conditions.PaymentFrequency)
	curPayFreq := util.UnPtrInt32(o.Instance.Conditions.PaymentFrequency)

	if initPayFreq != curPayFreq {
		fieldsToLog = append(fieldsToLog, &OfferChangelog{
			Action: offer.OfferAction_PAYMENT_FREQ_CHANGED,
			Old:    fmt.Sprint(initPayFreq),
			New:    fmt.Sprint(curPayFreq),
		})
	}

	return
}

func (o *OfferLog) logFirstRentPaymentDate() (fieldsToLog []*OfferChangelog) {
	initDate := UnPtrTime(o.InitialState.Conditions.FirstRentPaymentDate)
	curDate := UnPtrTime(o.Instance.Conditions.FirstRentPaymentDate)

	if initDate != curDate {
		fieldsToLog = append(fieldsToLog, &OfferChangelog{
			Action: offer.OfferAction_FIRST_PAYMENT_DATE_CHANGED,
			Old:    initDate.Format(time.RFC3339),
			New:    curDate.Format(time.RFC3339),
		})
	}

	return fieldsToLog
}

func (o *OfferLog) logMoveInDate() (fieldsToLog []*OfferChangelog) {
	initDate := UnPtrTime(o.InitialState.Conditions.MoveinDate)
	curDate := UnPtrTime(o.Instance.Conditions.MoveinDate)

	if initDate != curDate {
		fieldsToLog = append(fieldsToLog, &OfferChangelog{
			Action: offer.OfferAction_MOVE_IN_DATE_CHANGED,
			Old:    initDate.Format(time.RFC3339),
			New:    curDate.Format(time.RFC3339),
		})
	}

	return
}

func (o *OfferLog) logLandlordName() (fieldsToLog []*OfferChangelog) {
	initFirstName := util.UnPtrString(o.InitialState.Landlord.FirstName)
	initLastName := util.UnPtrString(o.InitialState.Landlord.LastName)
	curFirstName := util.UnPtrString(o.Instance.Landlord.FirstName)
	curLastName := util.UnPtrString(o.Instance.Landlord.LastName)

	if initFirstName != curFirstName || initLastName != curLastName {
		fieldsToLog = append(fieldsToLog, &OfferChangelog{
			Action: offer.OfferAction_LANDLORD_NAME_CHANGED,
			Old:    fmt.Sprintf("%s %s", initFirstName, initLastName),
			New:    fmt.Sprintf("%s %s", curFirstName, curLastName),
		})
	}

	return
}

// UnPtrTime safely gets time value from pointer
// TODO: move this to common/util
func UnPtrTime(value *time.Time) time.Time {
	if value == nil {
		return time.Time{}
	}
	return *value
}

// LogDiff creates OfferLog record with changed fields if no error occurred.
// Intended to be called with defer and return error variable address
func (o *OfferLog) LogDiff() {
	if o.Instance == nil {
		return
	}

	newLogs := o.getLogs()

	if len(newLogs) == 0 {
		return
	}

	var docs []interface{}

	for _, v := range newLogs {
		v.Initiator = o.User
		v.Offer = o.Instance.ID.Hex()
		v.CreatedAt = time.Now()
		docs = append(docs, v)
	}

	_, err := OfferLogs.InsertMany(o.ctx, docs)
	if err != nil {
		log.Error(err)
	}
}
