package model

import (
	"context"
	"crypto/tls"
	"fmt"
	"time"

	"gitlab.mashplace.com/mashplace/backend/common/generic/log"
	"gitlab.mashplace.com/mashplace/backend/common/generic/migrations"
	migrationsMongo "gitlab.mashplace.com/mashplace/backend/common/generic/migrations/mongo"
	"go.mongodb.org/mongo-driver/bson"
	"go.mongodb.org/mongo-driver/mongo"
	"go.mongodb.org/mongo-driver/mongo/options"
)

// Mongo collection name
const (
	MongoOffersCollection            = "offer"
	MongoOfferHistoryCollection      = "offer_history"
	MongoOfferLogsCollection         = "offer_logs"
	MongoInvitesCollection           = "invite"
	MongoReceiptCollection           = "rcpt"
	MongoLCGsCollection              = "lcg"
	MongoDocumentsCollection         = "documents"
	MongoScheduledPaymentsCollection = "scheduled_payments"
)

// Mongo collection instance
var (
	Offers            *mongo.Collection
	OffersHistory     *mongo.Collection
	OfferLogs         *mongo.Collection
	Invites           *mongo.Collection
	Receipts          *mongo.Collection
	LCGs              *mongo.Collection
	Documents         *mongo.Collection
	ScheduledPayments *mongo.Collection
)

// Client is Mongo database client instance.
var Client *mongo.Client

// DB is mongo database instance.
var DB *mongo.Database

// LCG - is ...
type LCG struct {
	Name  string `bson:"name"`
	Value uint32 `bson:"value"`
}

// Internal setting
const (
	PaymentLCG = "payment"
	LCGA       = 1249
	LCGC       = 35977
	LCGB       = 1 << 28
	LCGM       = (1 << 30) - LCGB
)

// NextLCG - is ...
func NextLCG(name string) (res uint32, err error) {
	olcg := &LCG{}
	for {
		err := LCGs.FindOne(context.Background(), bson.M{"name": name}).Decode(olcg)
		if err != nil {
			return 0, err
		}
		res = (LCGA*olcg.Value + LCGC) % LCGM
		err = LCGs.FindOneAndUpdate(
			context.TODO(),
			bson.M{"name": name, "value": olcg.Value},
			bson.M{"$set": bson.M{"value": res}},
		).Err()
		if err == nil {
			return res + LCGB, nil
		}
	}
}

// InitRepo initializes repository collections and performs migrations
func InitRepo() {
	clientOptions := options.Client().ApplyURI(fmt.Sprintf("mongodb://%s", MongoURL)).
		SetAuth(options.Credential{
			AuthSource:  "admin",
			Username:    MongoUsername,
			Password:    MongoPassword,
			PasswordSet: true,
		})

	if MongoSSLMode {
		clientOptions.SetTLSConfig(&tls.Config{})
	}

	c, err := mongo.NewClient(clientOptions)
	if err != nil {
		panic(err)
	}

	Client = c

	ctx, cancel := context.WithTimeout(context.Background(), 10*time.Second)
	defer cancel()
	if err := Client.Connect(ctx); err != nil {
		panic(err)
	}

	DB = Client.Database(MongoDB)
	Offers = DB.Collection(MongoOffersCollection)
	OffersHistory = DB.Collection(MongoOfferHistoryCollection)
	OfferLogs = DB.Collection(MongoOfferLogsCollection)
	Invites = DB.Collection(MongoInvitesCollection)
	Receipts = DB.Collection(MongoReceiptCollection)
	LCGs = DB.Collection(MongoLCGsCollection)
	Documents = DB.Collection(MongoDocumentsCollection)
	ScheduledPayments = DB.Collection(MongoScheduledPaymentsCollection)

	err = migrations.MigrateSchema(migrationsMongo.WithInstance(ctx, DB),
		&migrationsMongo.Migration{
			Version: "v0",
			Up: func(ctx context.Context, db *mongo.Database, migrationID string) (int, error) {
				return -1, nil
			},
		},
		&migrationsMongo.Migration{
			Version: "v1",
			Up:      modelMigrationV1,
		},
		&migrationsMongo.Migration{
			Version: "v2",
			Up:      modelMigrationV2,
		},
		&migrationsMongo.Migration{
			Version: "v3",
			Up:      modelMigrationV3,
		},
		&migrationsMongo.Migration{
			Version: "v4",
			Up:      modelMigrationV4,
		},
		&migrationsMongo.Migration{
			Version: "v5",
			Up:      modelMigrationV5,
		},
		&migrationsMongo.Migration{
			Version: "v6",
			Up:      modelMigrationV6,
		},
		&migrationsMongo.Migration{
			Version: "v7",
			Up:      modelMigrationV7,
		},
		&migrationsMongo.Migration{
			Version: "v8",
			Up:      modelMigrationV8,
		},
		&migrationsMongo.Migration{
			Version: "v9",
			Up:      modelMigrationV9,
		},
		&migrationsMongo.Migration{
			Version: "v10",
			Up:      modelMigrationV10,
		},
		&migrationsMongo.Migration{
			Version: "v11",
			Up:      modelMigrationV11,
		},
		&migrationsMongo.Migration{
			Version: "v12",
			Up:      modelMigrationV12,
		},
		&migrationsMongo.Migration{
			Version: "v13",
			Up:      modelMigrationV13,
		},
		&migrationsMongo.Migration{
			Version: "v14",
			Up:      modelMigrationV14,
		},
	)

	if err != nil {
		log.WithContext(ctx).WithContext(ctx).Error(err)
	}

	go func() {
		err = migrateReceiptsToInvoice(context.Background(), DB)
		if err != nil {
			log.WithContext(ctx).Error(err)
		}
	}()

	go func() {
		err = migratePropertyIVR(context.Background(), DB)
		if err != nil {
			log.WithContext(ctx).Error(err)
		}
	}()

	go func() {
		err = migrateContractStates(context.Background(), DB)
		if err != nil {
			log.WithContext(ctx).Error(err)
		}
	}()

	go func() {
		var count int64 = -1

		for count != 0 {
			count, err = migrateOfferManagers(context.Background(), DB)
			if err != nil {
				log.WithContext(ctx).Error(err)
			}

			time.Sleep(time.Minute)
		}
	}()

	go func() {
		var count int64 = -1

		for count != 0 {
			count, err = migrateSchedulePayments(context.Background(), DB)
			if err != nil {
				log.WithContext(ctx).Error(err)
			}

			time.Sleep(time.Minute)
		}
	}()
}
