package model

import (
	"context"
	"fmt"
	"strconv"
	"time"

	expenses "gitlab.mashplace.com/mashplace/backend/expenses/api"
	extras "gitlab.mashplace.com/mashplace/backend/extras/api"
	invoice "gitlab.mashplace.com/mashplace/backend/invoice/api"
	offer "gitlab.mashplace.com/mashplace/backend/offer/api"
	"google.golang.org/protobuf/types/known/timestamppb"
)

// ScheduledPaymentAdvanceModel represents request for payout in advance
type ScheduledPaymentAdvanceModel struct {
	Created         time.Time          `bson:"created"`
	Updated         time.Time          `bson:"updated"`
	LandlordConsent time.Time          `bson:"landlord_consent"`
	Admin           string             `bson:"admin"`
	PaymentID       string             `bson:"payment_id"`
	User            string             `bson:"user"`
	InvoiceID       string             `bson:"invoice_id"`
	Request         string             `bson:"request"`
	PaymentDate     DateModel          `bson:"payment_date"`
	FeePercentage   float64            `bson:"fee_percentage,omitempty"`
	PaymentSerial   uint64             `bson:"payment_serial"`
	State           offer.AdvanceState `bson:"state"`
	Initialized     bool               `bson:"initialized"`
	Eligible        bool               `bson:"eligible"`
}

// ScheduledPaymentPayoutModelLog represents payout status transition
type ScheduledPaymentPayoutModelLog struct {
	Time  time.Time         `bson:"time"`
	Admin string            `bson:"admin"`
	Prev  offer.PayoutState `bson:"prev"`
	Next  offer.PayoutState `bson:"next"`
}

// ScheduledPaymentPayoutModel represents payout state
type ScheduledPaymentPayoutModel struct {
	InvoiceID   string                            `bson:"invoice_id"`
	Transaction string                            `bson:"transaction"`
	Transitions []*ScheduledPaymentPayoutModelLog `bson:"transitions,omitempty"`
	Amount      float64                           `bson:"amount"`
	DueDay      int                               `bson:"due_day"`
	State       offer.PayoutState                 `bson:"state"`
}

// DateModel represents year-month date span
type DateModel struct {
	Year  int `bson:"year"`
	Month int `bson:"month"`
}

// ScheduledPaymentPartsModel represents one of monthly payments
type ScheduledPaymentPartsModel struct {
	Due       time.Time           `bson:"due"`
	InvoiceID string              `bson:"invoice_id"`
	PaymentID string              `bson:"payment_id"`
	Services  []*OfferService     `bson:"services"`
	Serial    uint64              `bson:"serial"`
	Status    offer.PaymentStatus `bson:"status"`
}

// ScheduledPaymentModel represents payment scheduled by a contract
type ScheduledPaymentModel struct {
	ID       string                        `bson:"id"`
	Offer    string                        `bson:"offer"`
	Landlord string                        `bson:"landlord"`
	Tenants  []string                      `bson:"tenants"`
	Parts    []*ScheduledPaymentPartsModel `bson:"parts"`
	Payout   ScheduledPaymentPayoutModel   `bson:"payout,omitempty"`
	Advance  ScheduledPaymentAdvanceModel  `bson:"advance,omitempty"`
	Date     DateModel                     `bson:"date"`
	Migrated bool                          `bson:"migrated"`
}

// ScheduledDepositModel - is ...
type ScheduledDepositModel struct {
	Due      time.Time       `bson:"due"`
	ID       string          `bson:"id"`
	Offer    string          `bson:"offer"`
	Landlord string          `bson:"landlord"`
	Invoice  string          `bson:"invoice"`
	Tenants  []string        `bson:"tenants"`
	Services []*OfferService `bson:"services,omitempty"`
	Months   int             `bson:"months"`
}

// GenerateScheduleDepositResponse - ...
type GenerateScheduleDepositResponse struct {
	Months            int32
	FirstMonthPayment float64
	MonthlyPayment    float64
	DueDay            int
}

// RenderServices - is ...
func RenderServices(offid string, models []*OfferService, mdue time.Time) []*extras.ServiceSnapshot {
	due := timestamppb.New(mdue)

	var svcs []*extras.ServiceSnapshot

	for _, s := range models {
		svcs = append(svcs, &extras.ServiceSnapshot{
			Id:            fmt.Sprintf("%s-%s-%s", offid, mdue.Format("20060102"), s.ID),
			ServiceId:     s.ID,
			Created:       due,
			Updated:       due,
			Code:          s.Code,
			Type:          s.Type,
			Role:          s.Role,
			TaxType:       s.TaxType,
			TaxAmount:     s.TaxAmount,
			TaxPercentage: s.TaxPercentage,
			Price:         s.Price,
			Title:         s.Title,
			Description:   s.Description,
			Primary:       s.Primary,
		})
	}

	return svcs
}

// RenderScheduleAvailable returns true if schedule due date is within bounds of advance request
// (excluding first two and last two months)
func RenderScheduleAvailable(date DateModel, movein time.Time, term, brclause int) bool {
	end := term - 2

	if brclause > 0 {
		end = brclause + 1
	}

	first := movein.AddDate(0, 2, 0)
	last := movein.AddDate(0, end, 0)

	firstp := first.Year()*12 + int(first.Month())
	lastp := last.Year()*12 + int(last.Month())

	datep := date.Year*12 + date.Month

	return datep > firstp && datep < lastp
}

// RenderScheduleFee returns expected fee for given date model and rent payment due day of month
func RenderScheduleFee(now time.Time, date DateModel, dueday int) float64 {
	nowmonths := now.Year()*12 + int(now.Month())

	if now.Day() < dueday {
		nowmonths--
	}

	schedmonths := date.Year*12 + date.Month
	diffmonths := schedmonths - nowmonths

	switch diffmonths {
	case 1:
		return 1
	case 2:
		return 4
	case 3:
		return 9
	default:
		return 0
	}
}

// RenderScheduleParts renders list of payment part models
func RenderScheduleParts(offid string, ms []*ScheduledPaymentPartsModel) (parts []*offer.ScheduledPaymentPart) {
	for _, m := range ms {
		parts = append(parts, RenderSchedulePart(offid, m))
	}

	return
}

// RenderSchedulePart renders payment part model
func RenderSchedulePart(offid string, m *ScheduledPaymentPartsModel) *offer.ScheduledPaymentPart {
	return &offer.ScheduledPaymentPart{
		InvoiceId:     m.InvoiceID,
		PaymentId:     m.PaymentID,
		PaymentSerial: strconv.FormatUint(m.Serial, 10),
		PaymentStatus: m.Status,
		PaymentDue:    timestamppb.New(m.Due),
		Services:      RenderServices(offid, m.Services, m.Due),
	}
}

// RenderSchedule converts model.ScheduledPaymentModel into gRPC API data.
func (model *ScheduledPaymentModel) RenderSchedule(
	ctx context.Context,
	user string,
) *offer.ScheduledPayment {
	var advance *offer.Advance

	var payout *offer.Payout

	if user == model.Landlord || user == "" {
		now := OfferTimeNowString(ctx, model.Offer)

		feeperc := RenderScheduleFee(now, model.Date, model.Payout.DueDay)

		advance = &offer.Advance{
			User:          model.Advance.User,
			State:         model.Advance.State,
			CreatedAt:     timestamppb.New(model.Advance.Created),
			UpdatedAt:     timestamppb.New(model.Advance.Updated),
			FeePercentage: feeperc,
			FeeValue:      (feeperc * model.Payout.Amount) / 100,
			Total:         model.Payout.Amount - ((feeperc * model.Payout.Amount) / 100),
			Available: model.Advance.Eligible &&
				model.Advance.State != offer.AdvanceState_ADVANCE_REQUESTED &&
				model.Advance.State != offer.AdvanceState_ADVANCE_GRANTED &&
				model.Date.Year*12+model.Date.Month > now.Year()*12+int(now.Month()) &&
				feeperc != 0,
			Request: model.Advance.Request,
		}

		var transitions []*offer.PayoutLog

		if user == "" {
			for _, t := range model.Payout.Transitions {
				transitions = append(transitions, &offer.PayoutLog{
					Admin: t.Admin,
					Time:  timestamppb.New(t.Time),
					Prev:  t.Prev,
					Next:  t.Next,
				})
			}
		}

		payout = &offer.Payout{
			Transitions: transitions,
			Transaction: model.Payout.Transaction,
			State:       model.Payout.State,
			InvoiceId:   model.Payout.InvoiceID,
		}
	} else {
		advance = &offer.Advance{}
		payout = &offer.Payout{}
	}

	return &offer.ScheduledPayment{
		Id:      model.ID,
		Offer:   model.Offer,
		Advance: advance,
		Payout:  payout,
		Parts:   RenderScheduleParts(model.Offer, model.Parts),
		Date: &offer.ScheduledPaymentDate{
			Year:  int32(model.Date.Year),
			Month: int32(model.Date.Month),
		},
	}
}

// LegacyTransferOfferPayoutState convert expenses.PayoutState into offer.PayoutState
func LegacyTransferOfferPayoutState(state expenses.PayoutStatus) (offer.AdvanceState, offer.PayoutState) {
	switch state {
	case expenses.PayoutStatus_PAYOUT_STATUS_PENDING:
		return 0, offer.PayoutState_PAYOUT_PENDING
	case expenses.PayoutStatus_PAYOUT_STATUS_PAID:
		return 0, offer.PayoutState_PAYOUT_PAID
	case expenses.PayoutStatus_PAYOUT_STATUS_CONFIRMED:
		return 0, offer.PayoutState_PAYOUT_CONFIRMED
	default:
		return 0, 0
	}
}

// LegacyTransferOfferAdvanceState convert expenses.AdvanceStatus into offer.PayoutState
func LegacyTransferOfferAdvanceState(state expenses.AdvanceStatus) (offer.AdvanceState, offer.PayoutState) {
	switch state {
	case expenses.AdvanceStatus_ADVANCE_STATUS_REQUESTED:
		return offer.AdvanceState_ADVANCE_REQUESTED, offer.PayoutState_PAYOUT_PENDING
	case expenses.AdvanceStatus_ADVANCE_STATUS_CANCELLED:
		return offer.AdvanceState_ADVANCE_CANCELLED, offer.PayoutState_PAYOUT_PENDING
	case expenses.AdvanceStatus_ADVANCE_STATUS_DENIED:
		return offer.AdvanceState_ADVANCE_DENIED, offer.PayoutState_PAYOUT_PENDING
	case expenses.AdvanceStatus_ADVANCE_STATUS_GRANTED:
		return offer.AdvanceState_ADVANCE_GRANTED, offer.PayoutState_PAYOUT_CONFIRMED
	default:
		return 0, 0
	}
}

// LegacyTransferExpensesPayoutState convert offer.PayoutState into expenses.PayoutState
func LegacyTransferExpensesPayoutState(state offer.PayoutState) expenses.PayoutStatus {
	switch state {
	case offer.PayoutState_PAYOUT_PENDING:
		return expenses.PayoutStatus_PAYOUT_STATUS_PENDING
	case offer.PayoutState_PAYOUT_CONFIRMED, offer.PayoutState_PAYOUT_PAID:
		return expenses.PayoutStatus_PAYOUT_STATUS_CONFIRMED
	default:
		return expenses.PayoutStatus_PAYOUT_STATUS_UNKNOWN
	}
}

// RenderScheduleDeposit converts model.ScheduledDepositModel into gRPC API data.
func (model *ScheduledDepositModel) RenderScheduleDeposit() *offer.ScheduledDeposit {
	return &offer.ScheduledDeposit{
		Id:        model.ID,
		Offer:     model.Offer,
		InvoiceId: model.Invoice,
		Due:       timestamppb.New(model.Due),
		Services:  RenderServices(model.Offer, model.Services, model.Due),
	}
}

// LegacyTransferExpensesPaymentStatusToOffer converts invoice's PaymentStatus to offer's PaymentStatus
func LegacyTransferExpensesPaymentStatusToOffer(s expenses.PaymentStatus) offer.PaymentStatus {
	switch s {
	case expenses.PaymentStatus_PAYMENT_STATUS_SUCCESS:
		return offer.PaymentStatus_PAYMENT_SUCCESS
	case expenses.PaymentStatus_PAYMENT_STATUS_PENDING:
		return offer.PaymentStatus_PAYMENT_PENDING
	case expenses.PaymentStatus_PAYMENT_STATUS_FAILED:
		return offer.PaymentStatus_PAYMENT_FAILED
	case expenses.PaymentStatus_PAYMENT_STATUS_CONFIRM:
		return offer.PaymentStatus_PAYMENT_CONFIRM
	default:
		return offer.PaymentStatus_UNKNOWN_PAYMENT_STATUS
	}
}

// LegacyTransferOfferPaymentStatusToExpenses converts offer's PaymentStatus to expenses PaymentStatus
func LegacyTransferOfferPaymentStatusToExpenses(s offer.PaymentStatus) expenses.PaymentStatus {
	switch s {
	case offer.PaymentStatus_PAYMENT_SUCCESS:
		return expenses.PaymentStatus_PAYMENT_STATUS_SUCCESS
	case offer.PaymentStatus_PAYMENT_PENDING:
		return expenses.PaymentStatus_PAYMENT_STATUS_PENDING
	case offer.PaymentStatus_PAYMENT_FAILED:
		return expenses.PaymentStatus_PAYMENT_STATUS_FAILED
	case offer.PaymentStatus_PAYMENT_CONFIRM:
		return expenses.PaymentStatus_PAYMENT_STATUS_CONFIRM
	default:
		return expenses.PaymentStatus_PAYMENT_STATUS_UNKNOWN
	}
}

// LegacyRenderAPISchedulePaymentModel render record into scheduled payment
func LegacyRenderAPISchedulePaymentModel(
	ctx context.Context,
	records []*expenses.Record,
	user, landlord, off string,
) *offer.ScheduledPayment {
	var advance *offer.Advance

	var payout *offer.Payout

	var parts []*offer.ScheduledPaymentPart

	for _, record := range records {
		switch {
		case record.GetIncome() != nil:
			advance, payout = legacyRenderScheduleIncome(ctx, user, landlord, off, record)
		case record.GetExpense() != nil:
			expense := record.GetExpense()

			switch {
			case expense.GetDrp() != nil:
				parts = append(parts, &offer.ScheduledPaymentPart{
					InvoiceId:     expense.GetDrp().GetInvoiceId(),
					PaymentId:     expense.GetDrp().GetPaymentId(),
					PaymentSerial: strconv.FormatUint(expense.GetDrp().GetSerial(), 10),
					PaymentStatus: LegacyTransferExpensesPaymentStatusToOffer(
						expense.GetDrp().GetStatus(),
					),
					PaymentDue: timestamppb.New(record.GetTime().AsTime()),
					Services: []*extras.ServiceSnapshot{
						{
							Id:            record.GetId(),
							Created:       record.GetCreated(),
							Updated:       record.GetUpdated(),
							Code:          invoice.CorrelationTargetDepositReplacement,
							Type:          extras.ServiceType_DEPOSIT_INSURANCE,
							Role:          extras.ServiceRole_TENANT_SERVICES,
							TaxType:       extras.TaxType_TAX_TYPE_VAT,
							Price:         record.GetAmount(),
							TaxAmount:     record.GetTax().GetAmount(),
							TaxPercentage: record.GetTax().GetVat().GetPercent(),
						},
					},
				})
			case expense.GetRent() != nil:
				rent := &extras.ServiceSnapshot{
					Id:      record.GetId(),
					Created: record.GetCreated(),
					Updated: record.GetUpdated(),
					Code:    extras.ProductTenantStandingOrder.String(),
					Type:    extras.ServiceType_RENT,
					Role:    extras.ServiceRole_TENANT_SERVICES,
					Price:   record.GetAmount(),
					TaxType: extras.TaxType_TAX_TYPE_NONE,
				}

				parts = append(parts, &offer.ScheduledPaymentPart{
					InvoiceId:     expense.GetRent().GetInvoiceId(),
					PaymentId:     expense.GetRent().GetPaymentId(),
					PaymentSerial: strconv.FormatUint(expense.GetRent().GetSerial(), 10),
					PaymentStatus: LegacyTransferExpensesPaymentStatusToOffer(
						expense.GetRent().GetStatus(),
					),
					PaymentDue: timestamppb.New(record.GetTime().AsTime()),
					Services:   []*extras.ServiceSnapshot{rent},
				})
			}
		}
	}

	return &offer.ScheduledPayment{
		Id:      records[0].Time.AsTime().Format("200601"),
		Offer:   records[0].GetCorrelationId(),
		Advance: advance,
		Payout:  payout,
		Parts:   parts,
		Date: &offer.ScheduledPaymentDate{
			Year:  int32(records[0].GetTime().AsTime().Year()),
			Month: int32(records[0].GetTime().AsTime().Month()),
		},
	}
}

func legacyRenderScheduleIncome(
	ctx context.Context,
	user, landlord, off string,
	record *expenses.Record,
) (advance *offer.Advance, payout *offer.Payout) {
	var available bool

	var transaction, invoiceid, request string

	var pstate offer.PayoutState

	var astate offer.AdvanceState

	switch {
	case record.GetIncome().GetRent() != nil:
		var pout *expenses.IncomePayout

		for _, p := range record.GetIncome().GetRent().GetPayouts() {
			pout = p

			break
		}

		available = record.GetIncome().GetRent().GetAdvanceAvailable() > 0
		invoiceid = pout.GetInvoiceId()
		astate, pstate = LegacyTransferOfferPayoutState(pout.GetPayoutStatus())
	case record.GetIncome().GetAdvance() != nil:
		var pout *expenses.IncomePayout

		for _, p := range record.GetIncome().GetRent().GetPayouts() {
			pout = p

			break
		}

		available = true
		invoiceid = pout.GetInvoiceId()
		request = record.GetIncome().GetAdvance().GetRequestId()
		astate, pstate = LegacyTransferOfferAdvanceState(record.GetIncome().GetAdvance().GetAdvanceStatus())
	}

	if user == landlord || user == "" {
		now := OfferTimeNowString(ctx, off)

		feeperc := RenderScheduleFee(now, DateModel{
			Year:  record.GetTime().AsTime().Year(),
			Month: int(record.GetTime().AsTime().Month()),
		}, record.GetTime().AsTime().Day())

		advance = &offer.Advance{
			User:          record.GetUserId(),
			State:         astate,
			CreatedAt:     record.GetCreated(),
			UpdatedAt:     record.GetUpdated(),
			FeePercentage: feeperc,
			FeeValue:      (feeperc * record.GetAmount()) / 100,
			Total:         record.GetAmount() - ((feeperc * record.GetAmount()) / 100),
			Available:     available,
			Request:       request,
		}

		payout = &offer.Payout{
			Transaction: transaction,
			State:       pstate,
			InvoiceId:   invoiceid,
		}
	}

	return
}
